#DEBUG_STACK = y

ifndef MAKEOKFLAG

RM = rm -f

ifdef OS_WIN32
ifdef BUILD_ON_WIN32_PC
WINE = 
else
WINE = wine
endif
AR = $(WINE) ar
CC = $(WINE) mingw32-gcc
STRIP = $(WINE) strip
WINDRES = $(WINE) windres --use-temp-file
INSTALL = true
else
AR = ar
ifdef USE_uClibc
 STATIC_ALDLIB = y
 CC = /usr/uClibc/0.9.33.2/i386-static/bin/i386-uclibc-gcc
else
 CC = gcc
endif
STRIP = strip
#ifndef DEBUG
# STRIP_BIN = y
#endif
INSTALL = /usr/bin/install
# Chk binutils-2.21
LDFLAGS_LDL := $(shell if ld --copy-dt-needed-entries -V >/dev/null 2>&1; \
    then echo "-Wl,--copy-dt-needed-entries"; fi)
endif

GCC_VER = $(subst ., ,$(shell $(CC) -dumpversion))
GCC_MAJOR_VER = $(word 1,$(GCC_VER))

LDFLAGS = $(LDFLAGS_LDL)

ifdef DEBUG
 ifndef CFLAGS
 CFLAGS = -O -march=i386 -Wall -std=gnu99
 endif
 CFLAGS += -g
else
 ifndef CFLAGS
  CFLAGS  = -O2 -march=i386 -Wall -std=gnu99
 endif
 ifndef DEBUG_STACK
  LDFLAGS += -s
  CFLAGS += -fomit-frame-pointer
  STRIP_BIN = y
 endif

 CFLAGS += -mpreferred-stack-boundary=2 -falign-functions=1 -falign-jumps=1 -falign-labels=1 -falign-loops=1 -fno-guess-branch-probability
# -finline-limit=0 -fno-builtin-strlen -ffunction-sections -fdata-sections
#CFLAGS += -fgnu89-inline
 ifeq ($(GCC_MAJOR_VER),4)
  LDFLAGS += -Wl,--gc-sections -Wl,--sort-common -Wl,--sort-section=alignment
  CFLAGS += -ffunction-sections -fdata-sections
 endif
endif


ifndef USE_ASM
# USE_ASM = c or i386
USE_ASM = i386
endif

ifdef OS_WIN32
CFLAGS += -D_OS_WIN32
LIBRARY = $(TOP)/lib/libaldccp_Win32.a

ifdef USE_JPEG
 ifdef USE_JPEG_TURBO
  LIBRARY += $(TOP)/lib/Xapp/W32_lib/libjpeg-turbo.a
 else
  LIBRARY += $(TOP)/lib/Xapp/W32_lib/libjpeg.a
 endif
endif

ifdef USE_PNG
LIBRARY += $(TOP)/lib/Xapp/W32_lib/libpng.a
ifndef USE_Z
USE_Z = y
endif
endif

ifdef USE_Z
LIBRARY += $(TOP)/lib/Xapp/W32_lib/zlib-1.1.4/libz.a
endif


else # OS_WIN32

PLIBS = -L$(TOP)/lib
PXLIBS = -L/usr/X11/lib
PATHLIBS = $(PXLIBS) $(PLIBS)

ifdef STATIC_ALDLIB
 ifdef USE_uClibc
  LIBRARY = $(TOP)/lib/libaldccpuC.a
 else
  LIBRARY = $(TOP)/lib/libaldccp.a
 endif
 XLIBRARY = -lX11
 ifdef USE_PTHREAD
  LIBRARY += -lpthread
 endif
else  # STATIC_ALDLIB
 LIBRARY = -laldccp -lpthread -ldl
 XLIBRARY = -laldccpw -lX11
 ifdef RPATH_ALDLIB
  RPATH = -Wl,-rpath,$(CURDIR)/$(TOP)/lib
 endif
endif  # STATIC_ALDLIB

ifdef USE_ACL
# CFLAGS += -DUSE_ACL
 LIBRARY += -lacl
endif

ifdef USE_XPM
 ifdef USE_libPM
  XLIBRARY += -lXpm
 endif
endif

ifdef USE_JPEG
 ifdef USE_JPEG_TURBO
  LIBRARY += $(TOP)/lib/Xapp/i386_lib/libjpeg-turbo.a
 else
  LIBRARY += -ljpeg
 endif
endif

ifdef USE_Z
 ifdef STATIC_Z
  LIBRARY += /usr/lib/libz.a
 else
  LIBRARY += -lz
 endif
endif

ifdef USE_uClibc
 CFLAGS +=  -DUSE_uC
endif

ifdef USE_PTHREAD
 CFLAGS += -DUSE_PTHREAD
endif

endif # OS_WIN32


ifdef USE_EXT_FONT_DLG
 CFLAGS += -DUSE_EXT_FONT_DLG
endif




export PLIBS PXLIBS DEBUG_STACK
export MAKEOKFLAG CC CFLAGS LDFLAGS STRIP AR USE_ASM RM
export DEBUG STATIC_ALDLIB RPATH_ALDLIB
export USE_uClibc
export USE_ACL USE_Z USE_XPM USE_libPM USE_PNG USE_JPEG USE_BMP USE_GIF USE_PNM USE_ONLY_ZICON
export USE_EXT_FONT_DLG

ifdef OS_WIN32
export WINDRES
endif

endif  #MAKEOKFLAG

INCL = -I$(TOP)/include
DEPH += $(addprefix $(TOP)/include/,$(DEPSYSH)) $(TOP)/vernr Makefile

ifdef OS_WIN32
 DEPLIB += $(TOP)/lib/libaldccp_Win32.a
 BUILD_OS = Win32
 USE_uClibc =
 loc_obj_PHAT = obj_Win32
 INCL += -I$(TOP)/include/w32_ext_h
 export USE_uClibc
 prefix = /aldccp
else
 BUILD_OS = Linux
 ifdef USE_uClibc
  DEPLIB += $(TOP)/lib/libaldccpuC.a
  loc_obj_PHAT = obj_uC
 else
  DEPLIB += $(TOP)/lib/libaldccp.a
  loc_obj_PHAT = obj

  ifeq (,$(wildcard /usr/include/X11/Xlib.h))
    ifeq (,$(wildcard /usr/X11/include/X11/Xlib.h))
      $(error no X11 include found !!!)
    else
      INCL += -I/usr/X11/include
    endif
  endif

 endif
 prefix = /opt/aldccp
endif

ifndef DESTDIR
 DESTDIR = $(TOP)
endif

bindir = ${prefix}/bin
datadir = ${prefix}/share
libdir = ${prefix}/lib

lib_obj_PHAT = $(TOP)/lib/$(loc_obj_PHAT)

