/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 This library is free software; you can redistribute it and/or
 modify it under the terms of the GNU Lesser General Public License
 as published by the Free Software Foundation; either
 version 2.1 of the License, or (at your option) any later version. 

 This library is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of 
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 See the GNU Lesser General Public License for more details. 

 You should have received a copy of the GNU Lesser General Public
 License along with this library; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA
*/

#ifndef	__APPBASE_H__
#define	__APPBASE_H__    1

#include "sysbase.h"
#include "sysbaseio.h"
#include <sys/time.h>

#ifndef _OS_WIN32
#include <sys/wait.h>
#endif

//******************** Error *******************************

typedef struct TMSGs {
  void *obj;
  void (*pgMSGs)(void *obj, int ErrType, char *msg, ...);
} TMSGs;
extern void appErrObj(TMSGs *MSGs, int ErrType, char *msg, ...);
typedef void (*TpgMSGs)(void *obj, int ErrType, char *msg, ...);

#define Timer_Enabled		1
#define Timer_Single		(1<<1)
#define Timer_OnTimerAndFree	(1<<2)
#define Timer_NoLostOnTimer	(1<<3)


typedef struct TgTimer {
  AlgTTC
  struct TgTimer *next;
  dword func;
  dword intern_func;
  dword Interval;
  void (*OnTimer)(struct TgTimer*);
  dword Tmp_ms;
  void *obj;
} TgTimer;

extern TgTimer *gTimerX_Add(dword func, dword Interval_ms, void (*OnTimer)(struct TgTimer* gTimer));
extern void gTimer_set_next_ev(TgTimer *gTimer, int next_ms);
extern void gTimer_Init();

typedef struct TPid_Array {
    int pid;
    int tag;
    void *obj;
    void (*On_pid_close)(int pid, int stat_loc, struct TPid_Array*);
    struct TPid_Array *next;
} TPid_Array;

typedef struct Tfd_list {
  void (*Proc)(FileHandle fd, struct Tfd_list*);
  FILE *io;
  void *obj;	//in TExecProc Texec_pipe*
  int m;	//in TExecProc Texec_pipe->pp[m]
} Tfd_list;

typedef void (*TEv_OnPipe)(FileHandle fd, Tfd_list *fdl);

extern void REGPARM3_FUNC fd_set_count(FileHandle fd);

typedef struct TEv_Array {
    int tag;
    void *obj;
    void (*On_appEv)(struct TEv_Array *appEv);
    struct TEv_Array *next;
    bool run_mk, nc1, nc2, nc3;
} TEv_Array;

#define appFu_select_t_NULL	1
#define appFu_select_t0_single	(1<<1)


typedef void (*TOnAHServ)(void *sh);

typedef struct {
  int self_pid;
  char *self_exe;
  char *sHOME;
  char *ALDini_dir;
  char *inifile_global;
  char *inifile_app;
  dword appFu;
  bool gQUIT_Ev, gQUIT_Ev_Mk, gSIGPIPE_Ev, nc3_Ev;
  void *obj;
  TPid_Array *Pid_Array;
  void (*OnPipeError)(int fd, int err);

  TEv_Array *appEv_Array;
  int t_io_us;		//default = 10000us

  void (*TimerDispatch)();
  int x_fd;
  void (*X11DispatchEvent)();
  void (*Wins_PaintOnEvEnd)();

  dword app_us;
  dword app_ms;
  dword tapp_us;
  dword next_zyk_k;
#ifndef _OS_WIN32
  int fd_set_count;
  Tfd_list *fd_list;
  byte AktPipes[FD_SETSIZE];  //(b7=OutPipe+b0=SendAutoDat)
#endif
  TpcIni *globalIni;
  TpcIni *appIni;
  dword ev_count;

#ifndef _OS_WIN32
  struct Tpthd_arr *pthd_arr;
  void (*MSG_Buf_Ev)();
#endif
} TappDat;

extern TappDat IappDat;

#define app_NoIni	1
extern void initapp(int fu, int argc, char **argv);
extern bool AppProcessMessages();
extern bool AppHandleMessage();

extern void setDefaultIniDat(const char *DefaultIniDat);

extern bool fd_in_wait_sigle(int fd, int us);

typedef void (*TOn_pid_close)(int pid, int stat_loc, TPid_Array *Pid_Array);
extern TPid_Array *Add_pid_Handle(int pid, TOn_pid_close On_pid_close);
extern TPid_Array *Get_pid_Handle(int pid);

extern TEv_Array *Add_appEv(void (*On_appEv)(TEv_Array *appEv));
extern void Free_appEv(TEv_Array *appEv);

typedef void (*TOnPipe)(int fd, Tfd_list *fd_list);
#define InPipe		1
#define OutPipe		0x80
extern Tfd_list *AddPipeProc_TmpRes(FileHandle fd, int fu, TOnPipe OnPipe, void *obj);
extern void AddPipeProc(FileHandle fd, int fu, TOnPipe OnPipe, void *obj);
#define clpp_fd		0x01
#define clpp_io		0x02
#define clpp_fdx	0x04	/* force close(fd) */
extern void REGPARM3_FUNC ClosePipeProc(FileHandle fd, int fu);
extern void REGPARM3_FUNC pClosePipe(FileHandle *fd);
extern void REGPARM3_FUNC ClosePipe(FileHandle fd);

#define carfu_cmd	1
extern bool ChkAppRun(char *fn, int fu, char **ppc);

extern TpcIni *get_appIni();
extern void free_appIni(TpcIni *pcIni);

extern TpcIni *get_globalIni();
extern void free_globalIni(TpcIni *pcIni);

//*********************** app/app_ChkIsThread.c **********************
extern bool (*app_ChkIsThread)();
extern bool ChkIsThread();

//*********************** app/cmdtoargs.c ****************************
extern void cmdtoargs(const char *cmd, int *argc, char ***argv);
extern void args_free(char **argv);

//*********************** app/exec_pipe.c ****************************

typedef struct _Texec_pipe Texec_pipe;

#define fd0_outfd		0x00
#define fd1_infd		0x01
#define fd2_errfd		0x02
#define fd3_Doutfd		0x03
#define fd4_Dinfd		0x04

#define evtyp_outfd		0x00
#define evtyp_infd		0x01
#define evtyp_errfd		0x02
#define evtyp_Doutfd		0x03
#define evtyp_Dinfd		0x04
#define evtyp_fd_Close_offset	0x10
#define evtyp_outfd_Close	0x10
#define evtyp_infd_Close	0x11
#define evtyp_errfd_Close	0x12
#define evtyp_Doutfd_Close	0x13
#define evtyp_Dinfd_Close	0x14
#define evtyp_ProcStart		0x20
#define evtyp_ProcClose		0x21

typedef void (*TOnExecProcEv)(Texec_pipe* expip, int evtyp);

typedef struct {
  int fd;
  TLineBuf *lb;
  void (*Proc)(int fd, Tfd_list *fd_list);
} Tprocpipe;

struct _Texec_pipe{
  AlgTTC0  
  char *args;
  void *owner;
  TOnExecProcEv OnExecProcEv;
  char *cd_path;
  TPid_Array *pida;
  void (*On_pid_close)(int pid, int stat_loc, TPid_Array *Pid_Array);
  Tprocpipe pp[5];
  int tag;
  void *p1;	//user
  char *fname;	//user
};

extern Texec_pipe *Texec_pipe_Create();
extern void CloseAll_exec_pipe(Texec_pipe *expip);

#define expipfu_args_free		1
#define expipfu_addPID_H		(1<<1)
// #define expipfu_OpenOut_Fd		(1<<2)

extern int exec_pipe_argv(char **argv, int fu, Texec_pipe *expip);

//*********************** app/exec_pc.c ****************************
extern int exec_pc_co_path(const char *args, TOn_pid_close On_pid_close, char *path);
extern int exec_pc_co(const char *args, TOn_pid_close On_pid_close);
extern int exec_pc(const char *args);

//*********************** app/exec_proc.c ****************************

#define outfd_Open		1
#define infd_Open		(1<<1)
#define errfd_Open		(1<<2)
#define Doutfd_Open		(1<<3)
#define Dinfd_Open		(1<<4)

extern int ExecuteProcApp_argv(char *args, int argc, char **argv, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip, int tag);
extern int ExecuteProcAppTag(char *cmd, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip, int tag);
extern int ExecuteProcApp(char *args, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip);
extern int ExecuteProc(char *args);

//*********************** misc ****************************

#ifdef MEM_CHK
#ifdef MEM_CHKappAZ
extern bool MEM_AZ_Disable;
#endif
#endif

#ifdef TEST_NOTIFY_TIME
extern dword test_notify_us;
extern dword test__us;
#endif


#endif /* __APPBASE_H__ */
