/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include <stdarg.h>
#include <X11/keysym.h>

#ifdef Test_Root_WinId
Window TestRootWinId = 0;
Window OrgRootWinId = 0;
#endif

TappXDat IappXDat;

bool (*On_wmError)(Display *dpy, XErrorEvent *e) = NULL;

#ifndef USE_NANO_X
int wmErrorHandler(Display *dpy, XErrorEvent *e) {
  char  number[16], msg[80], req[80];

	if (AppToAtExit) return 0;
	if (On_wmError && On_wmError(dpy, e)) return 0;

	XGetErrorText(dpy, e->error_code, msg, sizeof(msg));
	inttopc(number, e->request_code);
	XGetErrorDatabaseText(dpy, "XRequest", number, "<unknown>", req, sizeof(req));

//	fprintf(stderr, "win: %s(0x%lx): %s\n", req, e->resourceid, msg);
	printf("ERROR X11: %s: %s(0x%lx): %s\n", 
		GetFileName_pc(IappDat.self_exe), req, e->resourceid, msg);
/*
  printf("%s(%d): wmErrorHandler t=%d id=%x s=%d e=%d r=%d m=%d\n", FL,
          e->type, (int) e->resourceid, (int) e->serial,
          e->error_code, e->request_code, e->minor_code);
*/
  return 0;
}
#endif /* USE_NANO_X */

int CatchFatal(Display *dpy) {
	printf("ERROR: X Server msg. CatchFatal\n");
//	IappDat.gQUIT_Ev = True;
	return 0;
}
static const char *AH_AtomsTab[] = {
  "WM_DELETE_WINDOW",
  "WM_STATE",
  "WM_PROTOCOLS",
  "_MOTIF_WM_HINTS",
  "ALD_selection",
  "NONE",
  "CLIPBOARD",
  "TARGETS",
  "INCR",
  "UTF8_STRING",
  "text/plain",
  "text/uri-list",
  "image/png",
  "application/x-kde-cutselection",
  "application/x-kde-urilist",
  "_NET_WM_STATE",
  "_NET_WM_USER_TIME",
#ifdef FORM_NET_WM_PID
  "_NET_WM_PID",
#endif
  NULL
};

#ifdef APP_INI_DAT
static const char DefaultIniDat[] = {
  "[Global]\n"
  "; Clipbord_Mode 0=X11 1=QT3\n"
  "Clipbord_Mode=1\n"
  "Font=-adobe-helvetica-medium-r-normal-*-12-*-75-75-*-*-*-1\n"
  "FontBold=-adobe-helvetica-bold-r-normal-*-12-*-75-75-*-*-*-1\n"
  "; DateFormat 0=d-m-y 1=m-d-y 2=y-m-d\n"
  "DateFormat=0\n"
  "FileSystemUTF8=0\n"
  "\n"
  "[Icon]\n"
  "IconPaths=\"/opt/aldccp/share/icons/\" \"/usr/share/icons/hicolor/\"\n"
  "\n"
  "[App]\n"
  ";dotdesktop=/usr/share/applications /opt/kde3/share/applications/kde\n"
  ";dotdesktop_mimelnk=/usr/share/mimelnk\n"
  "\n"
  "[Dlg_File]\n"
  "Az0_ItemE_Dbl=True\n"
  "Az0_PathE_Dbl=False\n"
  "Az1_ItemE_Dbl=True\n"
  "Az1_PathE_Dbl=False\n"
};
#endif

static const byte XC_Dat[18] = {
    XC_X_cursor,
    XC_arrow,
    XC_bottom_left_corner,
    XC_bottom_right_corner,
    XC_bottom_side,
    XC_dot,
    XC_fleur,
    XC_left_side,
    XC_pirate,
    XC_right_side,
    XC_sb_h_double_arrow,
    XC_sb_v_double_arrow,
    XC_top_left_arrow,
    XC_top_left_corner,
    XC_top_right_corner,
    XC_top_side,
    XC_watch,
    XC_xterm};

#define g_width 2
#define g_height 2
static const char g_bits[] = {0x02, 0x01};
char *display_name = NULL;

TdpyMonitors dpyMonitors;

void initXapp(int fu, int argc, char **argv) {
  Display *dpy;
  int i;

	initapp(fu, argc, argv);
	Fill0(&IappXDat, sizeof(TappXDat));

#ifdef APP_INI_DAT
  char *pc;

	if (!(fu & app_NoIni)) {
	    IappDat.inifile_global = pcadd(IappDat.ALDini_dir, "/global.ini");
	    if (euidaccess(IappDat.inifile_global, F_OK)) {
		FileHandle fd;
		FileOpen_wr_erl(IappDat.inifile_global, &fd, 0640, EMSG_ERR_EXIT);
		FileWrite(fd, DefaultIniDat, sizeof(DefaultIniDat)-1, NULL);
		FileClose(fd);
	    }
	} 
#endif
	IappDat.X11DispatchEvent = X11DispatchEvent;
	IappDat.Wins_PaintOnEvEnd = Wins_PaintOnEvEnd;

	dpy = XOpenDisplay(display_name);

	if (!dpy) appError(EMSG_ERR_EXIT, "cant open display %s", XDisplayName(display_name));
	IappXDat.dpy = dpy;
	IappDat.x_fd = XConnectionNumber(dpy);
	fd_set_count(IappDat.x_fd);

// printf("%s(%d): %d %s\n", FL, DefaultScreen(dpy), DisplayString(dpy));

	IappXDat.RootW = DefaultRootWindow(dpy);
	if (!IappXDat.RootW) 
	  appError(EMSG_ERR_EXIT, "Screen %d is not a valid screen", IappXDat.RootW);

	IappXDat.default_screen = DefaultScreen(dpy);
#if appChacheSIZE > 0
	apphChache_CLR();
#endif

	MonitorsInit();

// printf("%d %d %d \n", IappXDat.Root_width, IappXDat.Root_height, IappXDat.RootW_depth);
	IappXDat.bixformat = B16_RGB565;
	if (IappXDat.RootW_depth > 16) IappXDat.bixformat = B32_RGB888;

  Atom *at = &IappXDat.XA_WM_DELETE_WINDOW;
  const char **ats = AH_AtomsTab;
	while(*ats) {
	    *at = XInternAtom(dpy, *ats, 0);
	    at++; ats++;
	}
	IappXDat.CLIPBOARD_ATOM = XA_PRIMARY;

	appXeventDefaultInit();
//	SelectionInit();

	XSetIOErrorHandler(CatchFatal);
	XSetErrorHandler(wmErrorHandler);

  long gcm;
  XGCValues gcv;
	Fill0(&gcv, sizeof(XGCValues));
	IappXDat.Default_Color = 0xc0c0c0;
	IappXDat.Paint_Color = 0xc0c0c0;
#ifdef USE_X11GC_plane_mask
	gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth;
	gcv.plane_mask = AllPlanes;
#else
	gcm = GCFunction | GCGraphicsExposures | GCLineWidth;
#endif
	gcv.function = GXcopy;
	gcv.fill_style = FillSolid;
	IappXDat.IconGC = CreateGC(gcm, &gcv);

	IappXDat.Default_GC = CreateGC_RGB(IappXDat.Default_Color);
	IappXDat.bordH_GC = CreateGC_RGB(0xe0e0e0);
	IappXDat.bordL_GC = CreateGC_RGB(0x808080);
	IappXDat.GC_FFFFFF = CreateGC_RGB(0xffffff);
	IappXDat.GC_C0C0C0 = CreateGC_RGB(0xc0c0c0);
	IappXDat.GC_404040 = CreateGC_RGB(0x404040);
	IappXDat.GC_000000 = CreateGC_RGB(0x000000);
	IappXDat.Paint_GC = CreateGC_RGB(IappXDat.Paint_Color);
	IappXDat.Select_GC_Back = CreateGC_RGB(0x00007f);

#ifdef USE_X11GC_plane_mask
	gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth |
	      GCForeground | GCFillStyle;
#else
	gcm = GCFunction | GCGraphicsExposures | GCLineWidth | GCForeground | GCFillStyle;
#endif
	gcv.function  = GXnand;
	gcv.foreground = GetColor_RGB(0xffff80);
//	gcv.function  = GXequiv;
//	gcv.foreground = GetColor_RGB(0x7fffff);
	IappXDat.Select_GC = CreateGC(gcm, &gcv);

	IappXDat.Focus_bitmap = XCreateBitmapFromData(dpy, IappXDat.RootW, g_bits, g_width, g_height);
	gcv.function = GXxor;
	gcv.line_width = 1;
	gcv.foreground = GetColor_RGB(0x000000);
	gcv.background = GetColor_RGB(0xc0c0c0);
	gcv.fill_style = FillOpaqueStippled;
	gcv.stipple = IappXDat.Focus_bitmap;
	gcv.subwindow_mode = IncludeInferiors;
	gcm |= GCStipple | GCSubwindowMode | GCBackground;
	IappXDat.FocusLine_GC = CreateGC(gcm, &gcv);

	gcv.function = GXandInverted;
	IappXDat.GC_000000_Dot = CreateGC(gcm, &gcv);

	gcv.function = GXcopy;
	gcv.foreground = GetColor_RGB(0xffffff);
	IappXDat.SB_GC = CreateGC(gcm, &gcv);

	for (i=0;i<18;i++) {
	  IappXDat.XCursors[XC_Dat[i]] = XCreateFontCursor(dpy, XC_Dat[i]);  //in X11/cursorfont.h
	}

	IappXDat.Menue_Cursor = IappXDat.XCursors[XC_arrow];
	set_ptr_arr(&IappXDat.PaintOnEvEnd_List, 254);

#ifdef APP_INI_DAT
  TpcIni *pcIni;
	if (fu & app_NoIni) {
	    IappXDat.Default_Font = MakeFont("-*-helvetica-medium-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
	    IappXDat.Default_bold_Font = MakeFont("-*-helvetica-bold-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
	} else {
	    pcIni = TpcIni_Create(IappDat.inifile_global);
	    pc = TpcIni_RD_pc(pcIni, "Global", "Font", NULL);
	    IappXDat.Default_Font = MakeFont(pc, "-*-helvetica-medium-r-*-*-12-*-75-75-*-*-*-*", NULL);
	    IappXDat.Default_bold_Font = MakeFont(TpcIni_RD_pc(pcIni, "Global", "FontBold", pc), pc, NULL);

	    i = TpcIni_RD_int(pcIni, "Global", "Clipbord_Mode", 0);
	    if (i == 1) IappXDat.CLIPBOARD_ATOM = IappXDat.XA_CLIPBOARD;

	    IappXDat.UTF8on = TpcIni_RD_boolean(pcIni, "Global", "UTF8", 0);
	    IappXDat.FileSystemUTF8on = TpcIni_RD_boolean(pcIni, "Global", "FileSystemUTF8", 0);

	    GlobDateFormat = TpcIni_RD_int(pcIni, "Global", "DateFormat", 0);
	    Tfree(pcIni);
	}
#else
	IappXDat.Default_Font = MakeFont("-*-helvetica-medium-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
	IappXDat.Default_bold_Font = MakeFont("-*-helvetica-bold-r-*-*-12-*-75-75-*-*-*-*", NULL, NULL);
#endif
	SetFont_Default(IappXDat.Default_Font, True);
	SetFont_Default(IappXDat.Default_bold_Font, True);
}

/* ******************* X11 Event's ******************** */


int Send_MSG(Tobjmsg *msg, Tapph *ap) {
	if (!msg || !ap || !ap->EvProc) return 0;
	if (ap->wtyp != wtyp_Tapph) {
	     printf("%s(%d): ERROR Tapph not wtyp_Tapph\n", FL);
	     return 0;
	}

	msg->ap = ap;
	return ap->EvProc(msg);
}

int Exec_Win_Ev(Tobjmsg *msg, Window win) {
  Tapph *ap;
	ap = Find_Exec_WinsX_Ev(IappXDat.apph, win);
	return Send_MSG(msg, ap);
}


/* *********************************************** */
void X11DispatchEvent() {
  Tobjmsg msg;
  dword t;
  int mk;
  XEvent Event;

	mk = 0;
	t = IappDat.app_us + 16666;
	while (IappXDat.dpy && (XPending(IappXDat.dpy) > 0)) {
	    msg.msg = msgXEv;
	    msg.f = msg.m = 0;
	    msg.p = &Event;
	    XNextEvent(IappXDat.dpy, &Event);

	    Window win = exec_XEventHandle((void*)&msg);
	    if (win) Exec_Win_Ev(&msg, win);
	    if (mk && cpTime(t, GetTime_us())) break;
	    mk++;
	}
}

void handle_wait_X11Event() {
	if (IappDat.X11DispatchEvent)  IappDat.X11DispatchEvent();
	if (IappDat.Wins_PaintOnEvEnd) IappDat.Wins_PaintOnEvEnd();
}


void Wins_PaintOnEvEnd() {
  int i;
  Tobjmsg msg;
  Tapph *ap;

	msg.msg = msgPaint;
	for (i=0; i < IappXDat.PaintOnEvEnd_Count; i++) {
	    ap = (Tapph*)IappXDat.PaintOnEvEnd_List->val[i];
	    IappXDat.PaintOnEvEnd_List->val[i] = NULL;
	    Send_MSG(&msg, ap);
	}
	IappXDat.PaintOnEvEnd_Count = 0;
	if (IappXDat.InfoWin && cpTime(IappXDat.InfoWin_us, IappDat.app_us)) {
	    SetVisible(IappXDat.InfoWin, True);
	}
}

static void InitStartForm() {
  Tapph *ap;
  Tobjmsg msg;
  long XMask;
  XSetWindowAttributes XAttr;

	msg.msg = msgVisibleFormShow;
	msg.f = msg.m = 0;
	msg.p = NULL;

	ap = IappXDat.apph;
	while (ap) {
	    if (ap->aos & aosIsForm) Send_MSG(&msg, ap);
	    ap = ap->next;
	}

	Fill0(&XAttr, sizeof(XAttr));
	XMask = CWCursor;
	XAttr.cursor = IappXDat.XCursors[XC_watch];
	IappXDat.SelHandleWin =  XCreateWindow (IappXDat.dpy, 
			IappXDat.apph->win, 
			0, 0, 1, 1,
			0, CopyFromParent, InputOutput,
			CopyFromParent, XMask, &XAttr);

}

void AppRun() {
	InitStartForm();
//	XSync(IappXDat.dpy, 0);
	AppHandleMessage();
	if (IappXDat.OnAppStart) IappXDat.OnAppStart();
NotClose:
	while (AppHandleMessage());

	if (IappXDat.OnAppClose) IappXDat.OnAppClose();
	if (!IappDat.gQUIT_Ev) goto NotClose;

	appError_ev = NULL;

  Display *dpy = IappXDat.dpy;

	AppToAtExit = True;
	IappXDat.dpy = NULL;
	app_atexit();

	XCloseDisplay(dpy);
}

void Close() {
	IappDat.gQUIT_Ev = True;
}


void MapWinID(Window WinID, int fu) {  //fu 0=UnMap 1=Map 2=MapRaised
  Display *dpy = IappXDat.dpy;

	if (!fu) {
	    XUnmapWindow(dpy, WinID);
	} else if (fu == 1) {
	    XMapWindow(dpy, WinID);
	} else {
	    XMapRaised(dpy, WinID);
	}
}




