/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

bool FontExists(const char *fontname) {
  int L;
  char **FontNameList;
	FontNameList = XListFonts(IappXDat.dpy, fontname, 1, &L);
	XFreeFontNames(FontNameList);
	return (L > 0);
}

char *GetFullFontName(const char *fontname) {
  int L;
  char **FontNameList;
  char *res = NULL;
	if (!fontname || !*fontname) return NULL;
	FontNameList = XListFonts(IappXDat.dpy, fontname, 1, &L);
	if (L > 0) res = pcdup(*FontNameList);
	XFreeFontNames(FontNameList);
	return res;
}

static TFontStructCache *get_fsco(TFont *Font) {
	if (!Font || !Font->xfont) return NULL;
	for (int i=0;i<FSCsitze;i++) {
	    if (IappXDat.FontStructCache[i].xfont == Font->xfont) {
		return &IappXDat.FontStructCache[i];
	    }
	}
	return NULL;
}

void SetFont_Default(TFont *Font, bool def) {
  TFontStructCache *fsco = get_fsco(Font);
	if (!fsco) return;
	fsco->fu &= ~tf_isDefaultFont;
	Font->xfu &= ~tf_isDefaultFont;
	if (def) {
	    fsco->fu |= tf_isDefaultFont;
	    Font->xfu |= tf_isDefaultFont;
	}
}


void FreeFont(TFont *Font) {
  TFontStructCache *fsco = get_fsco(Font);

	if (!fsco) {
	    printf("ERROR -- can t del old_Font font ??????????????\n");
	} else {
	    fsco->count--;
	    if (fsco->count <= 0) {
		if (fsco->fu & tf_isDefaultFont) {
		    fsco->count = 1;
		} else {
		    XFreeFont(IappXDat.dpy, fsco->xfont);
		    xfree(fsco->name);
		    fsco->xfont = NULL;
		}
	    }
	}
}

void TFont_Destroy(TFont *Font) {
	if (Font && !(Font->xfu & tf_isDefaultFont)) {
	    FreeFont(Font);
	    FreeGC(Font->g.FontGC);
	    FreeGC(Font->gtmp.FontGC);
	    xfree(Font);
	}
}

static void MakeTmpFontGC(TFont *Font, TFontGC *g, long ForeColor, long BackColor) {
  long gcm;
  XGCValues gcv;

	if (!Font || !Font->xfont || !g) return;
	Fill0(&gcv, sizeof(XGCValues));
	gcm = 0;
	if (g->FontGC) {
	    if (g->ForeColor != ForeColor) gcm |= GCForeground;
	    if (g->BackColor != BackColor) gcm |= GCBackground;
	    if (!gcm) return;
	} else {
#ifdef USE_X11GC_plane_mask
	    gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth | 
		  GCForeground | GCBackground | GCFillStyle | GCFont;
	    gcv.plane_mask = AllPlanes;
#else
	    gcm = GCFunction | GCGraphicsExposures | GCLineWidth | 
		  GCForeground | GCBackground | GCFillStyle | GCFont;
#endif
	    gcv.function  = GXcopy;
	    gcv.font = Font->xfont->fid;
	    gcv.fill_style = FillSolid;
	}

	g->ForeColor = ForeColor;
	g->BackColor = BackColor;
	if (gcm & GCForeground) gcv.foreground = GetColor_RGB(ForeColor);
	if (gcm & GCBackground) gcv.background = GetColor_RGB(BackColor);
	if (g->FontGC) {
	    ChangeGC(g->FontGC, gcm, &gcv);
	} else {
	    g->FontGC = CreateGC(gcm, &gcv);
	}
}

void ChangeFontGC(TFont *Font, int ForeColor, int BackColor) {
	if (!(Font->xfu & tf_isDefaultFont)) MakeTmpFontGC(Font, &Font->g, ForeColor, BackColor);
}

GC GetTmpFontGC(TFont *Font, int ForeColor, int BackColor) {

	if (!Font || !Font->xfont) return NULL;
	if ((Font->g.ForeColor == ForeColor) && (Font->g.BackColor == BackColor)) {
	    return Font->g.FontGC;
	}
	MakeTmpFontGC(Font, &Font->gtmp, ForeColor, BackColor);
	return Font->gtmp.FontGC;
}

TFont *MakeFont(const char *fontname, const char *defaultfont, TFont *old_Font) {
  char *fontpc = NULL;

	if (!fontname) fontname = "fixed";
	fontpc = GetFullFontName(fontname);
	if (!fontpc) {
	    printf("WARNING -- can t get font \"%s\", trying \"%s\"\n", fontname, defaultfont);
	    fontpc = GetFullFontName(defaultfont);
	}

	if (!fontpc) {
#ifdef _OS_WIN32
	    fontpc = pcdup("-*-system-*-*-*-*-*-*-*-*-*-*-*-*");
#else
	    fontpc = pcdup("fixed");
#endif
	    printf("WARNING -- can t get font \"%s\", trying \"%s\"\n", defaultfont, fontpc);
	}

  XFontStruct *pxfont = NULL;
  TFontStructCache *fsc;
  int i;

// printf("%s(%d) MakeFont: fontname=%s\n", FL, fontname);
// printf("%s(%d) MakeFont:          %s\n", FL, fontpc);

	for (i=0;i<FSCsitze;i++) {
	    fsc = &IappXDat.FontStructCache[i];
	    if (fsc->xfont && pccmp(fsc->name, fontpc)) {
		xfree(fontpc);
		fontpc = fsc->name;
		pxfont = fsc->xfont;
		fsc->count++;
// printf("%s(%d) FontCache: count=%d %s\n", FL, fsc->count, fontpc);
		break;
	    }
	}

	if (!pxfont) {
	  for (i=0;i<FSCsitze;i++) {
	    if (!IappXDat.FontStructCache[i].xfont) {
// printf("%s(%d) MakeFont: Add %d    %s\n", FL, i, fontpc);
	      pxfont = XLoadQueryFont(IappXDat.dpy, fontpc);
// printf("%s(%d) MakeFont:  0x%08x   %s\n", FL, (int)pxfont, fontpc);
	      fsc = &IappXDat.FontStructCache[i];
	      fsc->xfont = pxfont;
	      fsc->name = fontpc;
	      fsc->count = 1;
	      break;
	    }
	  }
	}

	if (!pxfont) {
	  printf("ERROR -- can t get font \"%s\"\n", fontpc);
	  exit(1);
	}

  int ForeColor = 0x000000;
  int BackColor = IappXDat.Default_Color;
	if (old_Font) {
	    ForeColor = old_Font->g.ForeColor;
	    BackColor = old_Font->g.BackColor;
	    TFont_Destroy(old_Font);
	}

  TFont *F = xcalloc(sizeof(TFont));
	F->xfont = pxfont;
	F->height = pxfont->ascent + pxfont->descent;
	F->name = fontpc;
	MakeTmpFontGC(F, &F->g, ForeColor, BackColor);
// printf("%s(%d): 0x%x font=0x%x 0x%x 0x%x\n", FL, (int)F, (int)F->xfont, (int)F->xfont->ext_data, (int)F->xfont->per_char);
	return F;
}

//**************** X11 ****************

#ifdef _OS_WIN32
#define Utf8ToX11Unicode(a,b,c,d) Utf8ToWChars(a,b,c,d, 0)
#else
#define Utf8ToX11Unicode(a,b,c,d) Utf8ToWChars(a,b,c,d, UC_WCharBE)
#endif

int TextnWidth(XFontStruct *xfont, const char *pc, int lg, bool UTF8on) {
  int L, result;
  word  *pwc;
  word  wcs[256];

	if ((lg <= 0) || !pc) return 0;

	if (!UTF8on) {
	    return XTextWidth(xfont, pc, lg);
	} 

	if (lg < 256) {
	    L = 256;
	    pwc =  wcs;
	} else {
	    L = lg + 1;
	    pwc = xmalloc(L*2);
	}

	L = Utf8ToX11Unicode(pwc, L, pc, lg);
	result = 0;
	if (L > 0) result = XTextWidth16(xfont, (XChar2b*)pwc, L);
	if (pwc !=  wcs) xfree(pwc);
	return result;
}

int TextWidth(XFontStruct *xfont, char *pc, bool UTF8on) {
	return TextnWidth(xfont, pc, pclen(pc), UTF8on);
}

static void DrawString(int x, int y, int lg, const char *pc, TVDrawString *vds) {
  int L, xo, yo;
  GC gc;
  XFontStruct *xfont;
  word  *pwc;
  word  wcs[256];

	if ((lg <= 0) || !pc || !vds) return;

	xfont = vds->TmpFont->xfont;
	xo = vds->Xoff + x;
#ifdef _OS_WIN32
	yo = vds->Yoff + y;
#else
	yo = vds->Yoff + xfont->ascent + y;
#endif
	gc = GetTmpFontGC(vds->TmpFont, vds->TmpForeColor, vds->TmpBackColor);

	if (!vds->UTF8on) {
	    if (vds->Image)
		XDrawImageString(IappXDat.dpy, vds->WinID, gc, xo, yo, pc, lg);
	    else
		XDrawString(IappXDat.dpy, vds->WinID, gc, xo, yo, pc, lg);
	    return;
	}

	if (lg < 256) {
	    L = 256;
	    pwc =  wcs;
	} else {
	    L = lg + 1;
	    pwc = xmalloc(L*2);
	}
	L = Utf8ToX11Unicode(pwc, L, pc, lg);
	if (L > 0) {
	  if (vds->Image)
	    XDrawImageString16(IappXDat.dpy, vds->WinID, gc, xo, yo, (XChar2b*)pwc, L);
	  else
	    XDrawString16(IappXDat.dpy, vds->WinID, gc, xo, yo, (XChar2b*)pwc, L);
	}
	if (pwc !=  wcs) xfree(pwc);
}

static int DrawStringSeg(int x, int y, int lgs, const char *pcs, TVDrawString *vds) {
  int i, lg;
  const char *pcsa;
  XFontStruct *xfont;

	if (!pcs) return x;
	xfont = vds->TmpFont->xfont;
	pcsa = pcs;
	if (vds->TabSpaceOn && (vds->TabSpaceWidth > 0)) {
	  for (i=0; i<lgs; i++) {
	    if (*pcs == 0x09) {
	      int w = pcs - pcsa;
	      int x0 = x;
	      if (w > 0) x0 = x + TextnWidth(xfont, pcsa, w, vds->UTF8on);
	      int xn = (x0 / vds->TabSpaceWidth + 1) * vds->TabSpaceWidth;
	      if (!vds->OnlyLg && (xn >= -vds->Xoff)) {
		if ((w > 0) && (x0 >= -vds->Xoff)) {
		  DrawString(x, y, w, pcsa, vds);
		}
		int xt = x0 + vds->Xoff;
		int yt = y + vds->Yoff;
		FillRectRGB(vds->WinID, vds->TmpBackColor, xt, yt, xn - xt, vds->TmpFont->height);
		if (vds->TabSpaceDot) {
		  yt += xfont->ascent;
		  DrawLine(vds->WinID, IappXDat.GC_000000, xt, yt, xt + 1, yt);
		  DrawLine(vds->WinID, IappXDat.GC_000000, xt, yt - 1, xt, yt);
		}
	      }
	      x = xn;
	      pcsa = pcs + 1;
	    }
	    pcs++;
	  }
	  lg = pcs - pcsa;
	} else lg = lgs;

	if (lg > 0) {
	    int x0 = x + TextnWidth(xfont, pcsa, lg, vds->UTF8on);
	    if (!vds->OnlyLg && (x0 >= -vds->Xoff)) {
		DrawString(x, y, lg, pcsa, vds);
	    }
	    x = x0;
	}
	return x;
}

int DrawColorString(int x, int y, TVDrawString *vds) {
  int x0, yo, L;
  char *pc;
  bool UTF8on;

	if (!vds) return x;
	L = vds->lg;
	pc = vds->s;
	x0 = x;

	if (vds->scda && vds->scda->items) {
	    UTF8on = vds->UTF8on;
	    int it = vds->scda->items;
	    TStrColorDat *scd = vds->scda->scd;
	    while ((it > 0) && (L > 0) && scd->pc) {
		int lg = scd->pc - pc;

		if (lg > L) lg = L;
		if (lg > 0) {
		    vds->UTF8on = UTF8on;
		    vds->TmpFont = vds->Font;
		    vds->TmpForeColor = vds->ForeColor;
		    vds->TmpBackColor = vds->BackColor;
		    x = DrawStringSeg(x, y, lg, pc, vds);
		    L -= lg;
		    pc += lg;
		}
		lg = scd->lg;
		if (lg > L) lg = L;
		if (lg > 0) {
		    if (scd->fu & scdfu_override_utf8) {
			vds->UTF8on = (scd->fu & scdfu_UTF8on) != 0;
		    }
		    vds->TmpFont = scd->Font;
		    vds->TmpForeColor = scd->ForeColor;
		    vds->TmpBackColor = scd->BackColor;
		    x = DrawStringSeg(x, y, lg, pc, vds);
		    L -= lg;
		    pc += lg;
		}
		it--; scd++;
	    }
	    vds->UTF8on = UTF8on;
	}

	if (L > 0) {
	    vds->TmpFont = vds->Font;
	    vds->TmpForeColor = vds->ForeColor;
	    vds->TmpBackColor = vds->BackColor;
	    x = DrawStringSeg(x, y, L, pc, vds);
// printf("%dx%d %d %s %x %x %x\n", x, y, L, pc, vds->WinID, vds->TmpForeColor, vds->TmpBackColor);
	}
	if (vds->OnlyLg) return x;

	yo = y + vds->Yoff;
	if (vds->Wins && vds->ClsToEnd) {
	  int yh = vds->TmpFont->height;
	  int xa = x + vds->Xoff;
	  if (xa < 0) xa = 0;
	  int e = vds->Wins->width - xa;
	  if (e > 0) {
	    if (vds->SelToEnd) {
		Fill_Rect(vds->WinID, IappXDat.Select_GC_Back, xa, yo, e, yh);
	    } else {
		FillRectRGB(vds->WinID, vds->BackColor, xa, yo, e, yh);
	    }
	  }
	}

	if ((vds->SelA != vds->SelE) || vds->SelToEnd) {
	    int xe = x + vds->Xoff;
	    int xa = 0;
	    if (vds->SelToEnd != 2) {
		int a = vds->SelA;
		int e = vds->SelE;
		if (a > e) {e = a; a = vds->SelE;}

		xa = GetChar_XPos(a, vds) + x0;
		if (!vds->SelToEnd) xe = GetChar_XPos(e, vds) + x0;
	    }
	    if (xe > xa)
		Fill_Rect(vds->WinID, IappXDat.Select_GC, xa, yo, xe-xa, vds->TmpFont->height);
	} 

	return x;
}

int TextWidthColorString(TVDrawString *vds) {
  int res;
  bool OnlyLg = vds->OnlyLg;

	vds->OnlyLg = True;
	res = DrawColorString(0, 0, vds);
	vds->OnlyLg = OnlyLg;
	return res;
}

void DrawLineString(const char *s, int x, int y, int *max_w, int *max_h, TVDrawString *vds) {
  const char *pc;
  int mx;

	*max_w = 0;
	*max_h = vds->Font->height;
	if (!vds || !s) return;
	vds->TmpFont = vds->Font;
	vds->TmpForeColor = vds->ForeColor;
	vds->TmpBackColor = vds->BackColor;

	mx = x;
	pc = s;
	while (*pc) {
	    if ((*pc == 10) || ((*pc == 13) && (pc[1] == 10))) {
		int lg = pc - s;
		if (lg > 0) {
		    int m = DrawStringSeg(x, y, lg, s, vds);
		    if (m > mx) mx = m;
		}
		if (*pc == 13) pc++;
		s = pc+1;
		y += vds->Font->height;
		*max_h += vds->Font->height;
	    }
	    pc++;
	}

	int lg = pclen(s);
	if (lg > 0) {
	    int m = DrawStringSeg(x, y, lg, s, vds);
	    if (m > mx) mx = m;
	}

	*max_w = mx;
}

