/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

//************************** Colors *****************************************
word REGP1 RGB24To16(dword Col) {
  //Color    rrrrrrrr gggggggg bbbbbbbb
  //pf16bit          rrrrr gggggg bbbbb
  return ((Col >> 8) & 0xf800) | ((Col >> 5) & 0x7e0) | ((Col >> 3) & 0x1f);
}

dword REGP1 RGB16To24(word RGB16) {
  //pf16bit: rrrrr gggggg bbbbb
  return ((RGB16 & 0x1f) << 3) | ((RGB16 & 0x7e0) << 5) | ((RGB16 & 0xf800) << 8);
}

#ifdef _OS_WIN32

dword REGP1 GetColor_RGB(dword RGB) {
	return ((RGB & 0xff0000) >> 16) | (RGB & 0xff00) | ((RGB & 0xff) << 16);
}

#else /* _OS_WIN32 */

dword REGP1 GetColor_RGB(dword RGB) {
  if (IappXDat.RootW_depth > 16) return RGB;
  return ((RGB >> 8) & 0xf800) | ((RGB >> 5) & 0x7e0) | ((RGB >> 3) & 0x1f);
}
#endif /* ! _OS_WIN32 */

dword AddColor(dword RGB, dword RGB_add) {
  int r,g,b;
	r = (RGB & 0xff)+(RGB_add & 0xff);
	RGB >>= 8; RGB_add >>= 8;
	g = (RGB & 0xff)+(RGB_add & 0xff);
	RGB >>= 8; RGB_add >>= 8;
	b = (RGB & 0xff)+(RGB_add & 0xff);
	return Clip255[r] | (Clip255[g] << 8) | (Clip255[b] << 16);
}

dword SubColor(dword RGB, dword RGB_sub) {
  int r,g,b;
	r = (RGB & 0xff)-(RGB_sub & 0xff);
	RGB >>= 8; RGB_sub >>= 8;
	g = (RGB & 0xff)-(RGB_sub & 0xff);
	RGB >>= 8; RGB_sub >>= 8;
	b = (RGB & 0xff)-(RGB_sub & 0xff);
	return Clip255[r] | (Clip255[g] << 8) | (Clip255[b] << 16);
}


GC CreateGC(long mask, XGCValues *dat) {
	return XCreateGC(IappXDat.dpy, IappXDat.RootW, mask, dat);
}

static bool appChkIsAppGC(GC gc) {
	return ((gc == IappXDat.Default_GC) ||
		(gc == IappXDat.bordL_GC) ||
		(gc == IappXDat.bordH_GC) ||
		(gc == IappXDat.GC_FFFFFF) ||
		(gc == IappXDat.GC_C0C0C0) ||
		(gc == IappXDat.GC_404040) ||
		(gc == IappXDat.GC_000000) ||
		(gc == IappXDat.GC_000000_Dot) || 
		(gc == IappXDat.Select_GC) ||
		(gc == IappXDat.Select_GC_Back) ||
		(gc == IappXDat.FocusLine_GC) ||
		(gc == IappXDat.Paint_GC) ||
		(gc == IappXDat.IconGC) ||
		(gc == IappXDat.SB_GC));
}

void FreeGC(GC gc) {
	if (gc && !appChkIsAppGC(gc)) XFreeGC(IappXDat.dpy, gc);
}

GC CreateGC_RGB(long RGB) {
  XGCValues gcv;
  long gcm;
	Fill0(&gcv, sizeof(XGCValues));
#ifdef USE_X11GC_plane_mask
	gcm = GCFunction | GCPlaneMask | GCGraphicsExposures | GCLineWidth |
		GCForeground | GCFillStyle;
	gcv.plane_mask = AllPlanes;
#else
	gcm = GCFunction | GCGraphicsExposures | GCLineWidth |
		GCForeground | GCFillStyle;
#endif
	gcv.function = GXcopy;
	gcv.fill_style = FillSolid;
	gcv.foreground = GetColor_RGB(RGB);
	return CreateGC(gcm, &gcv);
}

void ChangeGC(GC gc, long gcm, XGCValues *gcv) {
	XChangeGC(IappXDat.dpy, gc, gcm, gcv);
}

GC GetPaintGC(int RGB) {
  XGCValues gcv;

	if (IappXDat.Paint_Color != RGB) {
	    IappXDat.Paint_Color = RGB;
	    Fill0(&gcv, sizeof(XGCValues));
	    gcv.foreground = GetColor_RGB(RGB);
	    ChangeGC(IappXDat.Paint_GC, GCForeground, &gcv);
	}
	return IappXDat.Paint_GC;
}

void DrawLine(Window WinID, GC TGC, int X1, int Y1, int X2, int Y2) {
  XDrawLine(IappXDat.dpy, WinID, TGC, X1, Y1, X2, Y2);
}

void DrawLineRGB(Window WinID, int RGB, int X1, int Y1, int X2, int Y2) {
  XDrawLine(IappXDat.dpy, WinID, GetPaintGC(RGB), X1, Y1, X2, Y2);
}

void DrawRect(Window WinID, GC TGC, int x, int y, int w, int h) {
  XDrawRectangle(IappXDat.dpy, WinID, TGC, x, y, w-1, h-1);
}

void DrawRectRGB(Window WinID, int RGB, int x, int y, int w, int h) {
  XDrawRectangle(IappXDat.dpy, WinID, GetPaintGC(RGB), x, y, w-1, h-1);
}

void DrawRectFocus(Window WinID, int x, int y, int w, int h) {
	if (w < 0) {x += w; w = -w;}
	if (h < 0) {y += h; h = -h;}
#ifdef _OS_WIN32
  int isTempDC;
  RECT rc;
  HDC hDC = GetTmpPaintDC(WinID, &isTempDC, IappXDat.FocusLine_GC, gcdc_hpen);
	rc.left = x;
	rc.top = y;
	rc.right = x + w;
	rc.bottom = y + h;
	DrawFocusRect(hDC, &rc);
	ReleaseTmpPaintDC(WinID, hDC, isTempDC);
#else
	DrawRect(WinID, IappXDat.FocusLine_GC, x, y, w, h);
#endif
}

static void StdBorder(Window WinID, GC L_GC, GC H_GC, int x, int y, int w, int h, int BorderFu) {
  if ((w > 1) && (h > 1)) {
	w = x + w - 1;
	h = y + h - 1;
	if (!(BorderFu & mBorderNoTop)) DrawLine(WinID, H_GC, x, y, w, y);
	if (!(BorderFu & mBorderNoLeft)) DrawLine(WinID, H_GC, x, y, x, h);
	if (!(BorderFu & mBorderNoBottom)) DrawLine(WinID, L_GC, x, h, w, h);
	if (!(BorderFu & mBorderNoRigth)) DrawLine(WinID, L_GC, w, y, w, h);
  }
}

void StdBorderPaint(Window WinID, int BorderFu, int x, int y, int w, int h) {
  GC L_GC, H_GC;
  GC L2_GC, H2_GC;

	if (BorderFu & mBorder) {
	    if (BorderFu & mBorder2D) {
		DrawRect(WinID, IappXDat.GC_000000, x, y, w, h);
	    } else {
		if (BorderFu & mBorderL) {
		    H_GC = IappXDat.bordL_GC;
		    L_GC = IappXDat.bordH_GC;
		    L2_GC = IappXDat.GC_FFFFFF; //bordH_GC;
		    H2_GC = IappXDat.GC_000000;

		} else {
		    L_GC = IappXDat.bordL_GC;
		    H_GC = IappXDat.GC_FFFFFF; 
		    L2_GC = IappXDat.GC_000000;
		    H2_GC = IappXDat.Default_GC; 
		    if (BorderFu & mBorderBu) H2_GC = IappXDat.GC_404040;
		}

		if ((BorderFu & mBorderW2) == 0) {
		    StdBorder(WinID, L_GC, H_GC, x, y, w, h, BorderFu);
		} else {
		    StdBorder(WinID, L2_GC, H2_GC, x, y, w, h, BorderFu);
		    StdBorder(WinID, L_GC, H_GC, x+1, y+1, w-2, h-2, BorderFu);
		}
	    }
	}
}

void Fill_Rect(Window WinID, GC gc, int x, int y, int w, int h) {
  XFillRectangle(IappXDat.dpy, WinID, gc, x, y, w, h);
}

void FillRectRGB(Window WinID, int RGB, int x, int y, int w, int h) {
  XFillRectangle(IappXDat.dpy, WinID, GetPaintGC(RGB), x, y, w, h);
}

