/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"
#include <jpeglib.h>
#include <setjmp.h>

/*____________________________________________________________________________*/

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */
  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

/*____________________________________________________________________________*/
void jpeg_wr_error_exit(j_common_ptr cinfo) {
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}


/*____________________________________________________________________________*/
int write_jpeg(TImage *self) {
  struct jpeg_compress_struct cinfo;
  struct my_error_mgr jerr;
  JSAMPROW row_pointer[1];	/* pointer to JSAMPLE row[s] */
  bool gray;

  /* Step 1: allocate and initialize JPEG compression object */
	cinfo.err = jpeg_std_error(&jerr.pub);
	jerr.pub.error_exit = jpeg_wr_error_exit;
	if (setjmp(jerr.setjmp_buffer)) {
	    jpeg_destroy_compress(&cinfo);
	    return IMGWR_ERR_HEAD;
	}

	jpeg_create_compress(&cinfo);
  /* Step 2: specify data destination (eg, a file) */
	gray = (self->out_fu & fuWR_gray) != 0;
	jpeg_stdio_dest(&cinfo, self->out_fp);
  /* Step 3: set parameters for compression */
	cinfo.image_width = self->width; 	/* image width and height, in pixels */
	cinfo.image_height = self->height;
	cinfo.input_components = 3;	/* # of color components per pixel */
	cinfo.in_color_space = JCS_RGB; /* colorspace of input image */
	if (gray) {
	    cinfo.input_components = 1;	/* # of color components per pixel */
	    cinfo.in_color_space = JCS_GRAYSCALE; /* colorspace of input image */
	}

	jpeg_set_defaults(&cinfo);
  int quality = (self->out_fu >> 8) & 0xff;
	if (quality <= 0) quality = 95;
	jpeg_set_quality(&cinfo, quality, TRUE);
//	jpeg_simple_progression(&cinfo);
  /* Step 4: Start compressor */
	jpeg_start_compress(&cinfo, TRUE);
  /* Step 5: while (scan lines remain to be written) */
  /*           jpeg_write_scanlines(...); */
	row_pointer[0] = xmalloc(cinfo.image_width * 3);

  bool rgb565 = (self->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp;
  byte *sb0 = self->data;
	for (int y=self->height;y;y--) {
  byte *sb, *db;
  int r, g, b;

	    sb = sb0;
	    db = row_pointer[0];
	    for (int x=self->width;x;x--) {
		if (rgb565) {
		    dword c = *((word*)sb);
		    b = (c<<3) & 0xf8; g = (c>>3) & 0xfc; r = (c>>8) & 0xf8;
		    sb += 2;
		} else {
		    b = sb[0]; g = sb[1]; r = sb[2];
		    sb += 4;
		}

		if (gray) {
		    *db++ = ((r + g + b) * 0x55a) >> 12;  //div 3
		} else {
		    *db++ = r; *db++ = g; *db++ = b;
		}
	    }
	    jpeg_write_scanlines(&cinfo, row_pointer, 1);
	    sb0 += self->bytes_per_line;
	}

  /* Step 6: Finish compression */
	jpeg_finish_compress(&cinfo);
	xfree(row_pointer[0]);
	jpeg_destroy_compress(&cinfo);

	return IMGWR_OK;
}

static bool ChkType(TglobPath *gp, const char *fn, FILE *F) {
	return TImage_ChkPattern(gp, fn, hImageWR_jpeg.pattern);
}

const ThImageLoader hImageWR_jpeg = {
  "jpg jpeg",
  filetyp_JPEG,
  ChkType,
  write_jpeg
};
