/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under the LGPL v2.1, see the file COPYING.LIB
*/

#include "image.h"

void image_copy32_mask_chk(TImage *src, TImage *dest, int x, int y, int w, int h, 
					int dest_x, int dest_y) {

  byte *sd0, *dd0;

	sd0 = src->data  + (src->bytes_per_line * y) + (x * 4);
	dd0 = dest->data + (dest->bytes_per_line * dest_y) + (dest_x * 4);
	for (;h>0;h--) {
	    dword *sd = (dword*)sd0;
	    dword *dd = (dword*)dd0;
	    for (int i=w;i>0;i--) {
		if (*sd & 0x80000000) *dd = *sd; 
		dd++; sd++;
	    }
	    sd0 += src->bytes_per_line;
	    dd0 += dest->bytes_per_line;
	}
}


void image_copy(TImage *src, TImage *dest, int x, int y, int w, int h, 
					int dest_x, int dest_y, int fu) {
  TImage *nsrc, *ndest;
  Timg_rsz rsz;
  bool alpha;

	if (!src && !dest) return;
	if ((x + w) > src->width) w = src->width - x;
	if ((y + h) > src->height) h = src->height - y;
	if ((dest_x + w) > dest->width) w = dest->width - dest_x;
	if ((dest_y + h) > dest->height) h = dest->height - dest_y;
	if (w <= 0) return;
	if (h <= 0) return;

	nsrc = src;
	ndest = dest;

	alpha = (src->stat & pmxStat_ckh_alpha) || (dest->stat & pmxStat_ckh_alpha);

	if (alpha) {
	    if ((src->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
		nsrc = TImage_rgb565to888(src);
	    }
	    if ((dest->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
		ndest = TImage_rgb565to888(dest);
	    }

	} else if ((dest->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
	    if ((src->pxm_LD & pxmLD_bppMask) != pxm_2b_16bpp) {
		nsrc = TImage_rgb888to565(src);
	    }

	} else if ((src->pxm_LD & pxmLD_bppMask) == pxm_2b_16bpp) {
	    nsrc = TImage_rgb565to888(src);
	}

	if (alpha && (fu & dupfu_copy_mask)) {
	    image_copy32_mask_chk(nsrc, ndest, x, y, w, h, dest_x, dest_y);
	} else {
	    TImage_SetBase_rsz(nsrc, &rsz);
	    TImage_SetDestBase_rsz(ndest, &rsz);
	    rsz.dest_off_x = dest_x;
	    rsz.dest_off_y = dest_y;
	    rsz.dest_lg_x = w;
	    rsz.dest_lg_y = h;
	    ImgDat_Copy(&rsz);
	}

	if (src != nsrc) Tfree(nsrc);

	dest->stat &= ~(pmxStat_XPixAkt|pmxStat_XMaskPixAkt);
	if (dest != ndest) {
	    dest->stat &= ~pmxStat_MaskImgAkt;
	    xfree(dest->mask_data);
	    dest->mask_data = NULL;
	    alpha_bm_stat(ndest);
	    if (!ndest->rgb565_alpha_data && alpha_chk_mask(ndest, 0xff)) {
		dest->rgb565_alpha_data = xmalloc(dest->width * dest->height);
	    }

	    dest->stat &= ~pmxStat_has_alpha_tgl;
	    dest->stat |= (ndest->stat & pmxStat_has_alpha_tgl);
	    TImage_rgb888move565(ndest, dest);
	    Tfree(ndest);
	}
}
