/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <unistd.h>

char *appError_LastBuf = NULL;
int appError_LastType = 0;
int appError_Mk;

int (*appError_ev)(int ErrType, const char *err) = NULL;

static const char *ErrTab[] = {
  NULL,
  "DEBUG ",
  "INFO ",
  "WARN ",
  "ERROR ",
  "ERROR ",
  "ERROR ",
  "ERROR EXIT "
};

static void ppc_xmove(Tlgs *lgs, const char *src) {
  int lg = pcxmove(lgs->s, src, lgs->lg);
	lgs->s += lg;
	lgs->lg -= lg;
}

char *appError_va_Str(char *buf, int size, int ErrType, const char *msg, va_list args) {
  int m;
  char *pc;
  Tlgs lgs;
  char line[2048];

//	if (!app_name) app_name = get_pid_cmd_name(getpid());
	appError_Mk = 1;
	appError_LastType = ErrType;

	if (!buf) {
	    size = 2048;
	    buf = xcalloc(size);
	}

	pc = buf; m = size;
	if (ErrType) { pc = line; m = 2048;}
	vsnprintf(pc, m, msg, args);

	if (ErrType) {
	    lgs.s = buf;
	    lgs.lg = 2046;

	    if (!(ErrType & EMSG_NOH)) {
		m = ErrType & 0x0f;
		if (m > EMSG_ERR_EXIT) m = EMSG_ERR_EXIT;
		ppc_xmove(&lgs, ErrTab[m]);
		ppc_xmove(&lgs, app_name);
		pc = ": ";
		if (ErrType & EMSG_H_LF) pc = ":\n";
		ppc_xmove(&lgs, pc);
	    }
	    ppc_xmove(&lgs, line);
	    if ((ErrType & EMSG_SYS) && !(ErrType & EMSG_NOSYS)) {
		ppc_xmove(&lgs, " [");
		ppc_xmove(&lgs, SysError_pc());
		ppc_xmove(&lgs, "]");
	    }
	    if (!(ErrType & EMSG_NO_LF)) {
		*lgs.s = '\n';
		lgs.s++;
		*lgs.s = 0;
	    }
	}
	if (ErrType & EMSG_ToLastBuf) ppcchange(&appError_LastBuf, buf);
	return buf;
}

void appError(int ErrType, const char *msg, ...) {
  va_list args;
  char *pc;
  char line[2048];
  bool mk;

	if (ErrType & EMSG_None) return;

	if (ErrType & EMSG_FromLastBuf) {
	    pc = appError_LastBuf;
	    if (!pc) return;
	    ErrType = appError_LastType & ~EMSG_ToLastBuf;
	    ErrType |= EMSG_FromLastBuf;
	} else {
	    va_start(args, msg);
	    appError_va_Str(line, 2048, ErrType, msg, args);
	    va_end(args);
	    pc = line;
	}

	mk = (ErrType & EMSG_MASK_Bef) >= EMSG_ERR_EXIT;
	if (mk || (!(ErrType & EMSG_ToLastBuf) && 
		   (!appError_ev || !appError_ev(ErrType, pc)))) {
  FILE *io = stdout;
	    if ((ErrType & EMSG_MASK_Bef) >= EMSG_ERR2) io = stderr;
	    fprintf (io, pc);
	}
	if (mk) exit(EXIT_FAILURE);
}

void appError_LastClr() {
	xfree(appError_LastBuf);
	appError_LastBuf = NULL;
}
