/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include <errno.h>
#include <fcntl.h>

TappDat IappDat;

#ifdef MEM_CHK
#ifdef MEM_CHKappAZ
int LastMEM_CHK_Count;
bool MEM_AZ_Disable = False;
#endif
#endif

#ifdef TEST_NOTIFY_TIME
dword test_notify_us = 0;
static dword az_notify_us;
#endif


void Ev_QUIT(int nonsense) {
  IappDat.gQUIT_Ev = True;
  IappDat.gQUIT_Ev_Mk = True;
}

void Ev_SIGPIPE(int nonsense) {
  IappDat.gSIGPIPE_Ev = True;
  printf("ERROR %s(%d): slave_pipe Ev_SIGPIPE\n", FL);
}

void initapp(int fu, int argc, char **argv) {
  const int SigTab[] = {SIGHUP, SIGQUIT, SIGKILL, SIGTERM};
  int i;
  char tmpf[4096];

#ifdef DEBUG_STACK
	signal(SIGSEGV, abort_SIGSEGV);
#endif
	app_name = GetFileName_pc(*argv);

	Fill0(&IappDat, sizeof(TappDat));

	for (i=0; i < 4; i++) signal(SigTab[i], Ev_QUIT);
	IappDat.gQUIT_Ev = False;
	IappDat.gSIGPIPE_Ev = False;
	fd_set_count(31);

	IappDat.self_pid = getpid();
	i = readlink("/proc/self/exe", tmpf, sizeof(tmpf));
	if (i>0) IappDat.self_exe = pcndup(tmpf, i);

	IappDat.sHOME = getenv("HOME");
#ifdef APP_INI_DAT
	if (!IappDat.sHOME) IappDat.sHOME = "/tmp";
	IappDat.ALDini_dir = pcadd(IappDat.sHOME, "/.ALDccp");
	if ((fu & app_NoIni) == 0) {
	    file_mkdir_erl(IappDat.ALDini_dir, 0750, EMSG_ERR_EXIT);
	}
#endif
//	SetCPU_Flags(NULL);

	IappDat.t_io_us = 30000;
	IappDat.next_zyk_k = IappDat.app_us = GetTime_us();
#ifdef TEST_NOTIFY_TIME
	az_notify_us = GetTime_us();
#endif
}

bool AppProcessMessages() {
	if (IappDat.TimerDispatch) IappDat.TimerDispatch();
	if (IappDat.X11DispatchEvent) IappDat.X11DispatchEvent();
	if (IappDat.MSG_Buf_Ev) IappDat.MSG_Buf_Ev();
	if (IappDat.gQUIT_Ev) return False;
	if (IappDat.Wins_PaintOnEvEnd) IappDat.Wins_PaintOnEvEnd();
	return True;
}

static fd_set in_fdset, out_fdset;

static void fd_wait_sigle_set(int fd) {
	FD_ZERO (&in_fdset);
	FD_ZERO (&out_fdset);
	if (fd) FD_SET (fd, &in_fdset);
}

static int fd_wait_us(int us) {
  struct timeval tv;
	tv.tv_sec = 0;
	tv.tv_usec = us;
	return select(IappDat.fd_set_count, &in_fdset, &out_fdset, 0, &tv);
}

bool fd_in_wait_sigle(int fd, int us) {
	fd_wait_sigle_set(fd);
	if (!fd_wait_us(us)) return False;
	return FD_ISSET(fd, &in_fdset) != 0;
}

bool AppHandleMessage() {
  int i, retval, fd_set_count, us, stat_loc, pid;
  byte *pPipe;
  bool ret;

	fd_wait_sigle_set(IappDat.x_fd);
	fd_set_count = IappDat.fd_set_count;
	pPipe = IappDat.AktPipes;

	for (i=0; i<fd_set_count; i++) {	//max.FD_SETSIZE
	    if (*pPipe) {
		if (*pPipe & OutPipe) {
		    if (*pPipe == OutPipe) FD_SET(i, &out_fdset); else *pPipe -= 1;
		} else {
		    if (*pPipe == InPipe) FD_SET(i, &in_fdset); else *pPipe -= 1;
		}
	    }
	    pPipe++;
	}

  dword ttus = GetTime_us();
	IappDat.tapp_us += ttus - IappDat.app_us;

	us = IappDat.t_io_us;
	if (IappDat.appFu & (appFu_select_t_NULL|appFu_select_t0_single)) us = 0;
	retval = fd_wait_us(us);
	atomic_and(&IappDat.appFu, ~appFu_select_t0_single);
//	IappDat.app_us = GetTime_us();

  struct timeval tv;
  u64 uval;
	gettimeofday(&tv, NULL);
	uval = tv.tv_sec * 1000000;
	uval += tv.tv_usec;
	IappDat.app_us = uval;
	uval = tv.tv_sec * 1000;
	uval += (tv.tv_usec / 1000);
	IappDat.app_ms = uval;
	IappDat.ev_count++;


	if (retval > 0) {
  Tfd_list *fdl = IappDat.fd_list;
	  for (i=0; i<fd_set_count; i++) {
	    if (FD_ISSET(i, &in_fdset) || FD_ISSET (i, &out_fdset)) {
	      if (fdl->Proc) fdl->Proc(i, fdl);
	    }
	    fdl++;
	  }

	} else if (retval == -1) {
	  if (errno == EBADF) {  //9	Bad file number
	    pPipe = IappDat.AktPipes;
	    for (i=0; i<fd_set_count; i++) {
	      if (*pPipe && (fcntl(i, F_GETFD) == -1)) {
		if (IappDat.OnPipeError) IappDat.OnPipeError(i, EBADF);
		*pPipe = 0;
		appError(EMSG_ERR1, "%s(%d) : Pipe %d Bad file number \n", FL, i);
	      }
	      pPipe++;
	    }
	  }
	}

	pid = waitpid(-1, &stat_loc, WNOHANG);
	if ((pid > 0) && IappDat.Pid_Array) {
	    TPid_Array *pida, **ppida;
	    ppida = &IappDat.Pid_Array;
	    while (*ppida && (((*ppida)->pid) != pid)) ppida = &(*ppida)->next;
	    pida = *ppida;
	    if (pida) {
		*ppida = pida->next;
		if (pida->On_pid_close) pida->On_pid_close(pid, stat_loc, pida);
		xfree(pida);
	    }
	}

	ret = AppProcessMessages();

	if (IappDat.appEv_Array) {
  TEv_Array *pev, **ppev;
	    ppev = &IappDat.appEv_Array;
	    while (*ppev) {
		pev = *ppev;
		if (pev->On_appEv) {
		    if (!pev->run_mk) {
			pev->run_mk = True;
			pev->On_appEv(pev);
			pev->run_mk = False;
		    }
		    ppev = &pev->next;
		} else {
		    *ppev = pev->next;
		    xfree(pev);
		}
	    }
	}

#ifdef MEM_CHK
#ifdef MEM_CHKappAZ
	if (!MEM_AZ_Disable) {
  int lMCt, lMsize;
	    MEM_CHK_Count(&lMCt, &lMsize);
	    if (LastMEM_CHK_Count != lMCt) {
		    LastMEM_CHK_Count = lMCt;
    printf("Debug %s: MEM_CHK_Count=%3d  size=%d\n", app_name, lMCt, lMsize);
	    }
	}
#endif
#endif
#ifdef TEST_NOTIFY_TIME
	if (GetCurTimeDiv_us(az_notify_us) > 10000000) {
	    printf("Debug %s(%d): notify_us=%uus\n", FL, test_notify_us);
	    az_notify_us = GetTime_us();
	    test_notify_us = 0;
	}
#endif
	return ret;
}



