/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include <errno.h>

static void TExecProc_DoEv(Texec_pipe *expip, int evtyp) {
	if (expip->OnExecProcEv) expip->OnExecProcEv(expip, evtyp);
}

static void TExecProc_inProc(int fd, Tfd_list* fdl) {
  Texec_pipe *expip = fdl->obj;
  int nr = fdl->m;

	if (TLineBuf_fd_Add(expip->pp[nr].lb, fd) == 0) {
	    ClosePipeProc(fd, clpp_fd);
	    expip->pp[nr].fd = -1;
	    nr += evtyp_fd_Close_offset;
	}
	TExecProc_DoEv(expip, nr);
}


static void inProc_close(Texec_pipe *expip, int nr) {
	if (expip->pp[nr].fd > 0) {
	    if (TLineBuf_fd_Add(expip->pp[nr].lb, expip->pp[nr].fd)) {
		TExecProc_DoEv(expip, nr);
	    }
	    ClosePipeProc(expip->pp[nr].fd, clpp_fd);
	    TExecProc_DoEv(expip, nr + evtyp_fd_Close_offset);
	}
}

static void TExecProc_close(int pid, int stat_loc, TPid_Array *pida) {
  int i;
  Texec_pipe *expip = pida->obj;

	if (expip->pp[0].fd > 0) ClosePipeProc(expip->pp[0].fd, clpp_fdx);
	inProc_close(expip, 1);
	inProc_close(expip, 2);
	if (expip->pp[3].fd > 0) ClosePipeProc(expip->pp[3].fd, clpp_fdx);
	inProc_close(expip, 4);

	TExecProc_DoEv(expip, evtyp_ProcClose);
	for (i=0;i<5;i++) ppTfree(&expip->pp[i].lb);
	xfree(expip->args);
	xfree(expip->cd_path);
	xfree(expip->fname);
	xfree(expip);
}

// fu = outfd_Open | infd_Open | errfd_Open
int ExecuteProcApp_argv(char *cmd, int argc, char **argv, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip, int tag) {
  int pid, expipfu;
  Texec_pipe *expip = NULL;

	if (!argc || !argv) return 0;
	expipfu = expipfu_args_free | expipfu_addPID_H;
	if (OnExecProcEv) {
	    expip = Texec_pipe_Create();
	    expip->owner = owner;
	    expip->tag = tag;
	    expip->args = pcdup(cmd);
	    expip->OnExecProcEv = OnExecProcEv;
	    expip->On_pid_close = TExecProc_close;

	    if (fu & outfd_Open) {	//out_fd
		expip->pp[0].Proc = (void*)1;
	    }

	    if (fu & infd_Open) {	//out_fd
		expip->pp[1].lb = TLineBuf_Create();
		expip->pp[1].Proc = TExecProc_inProc;
	    }

	    if (fu & errfd_Open) {	//err_fd
		expip->pp[2].lb = TLineBuf_Create();
		expip->pp[2].Proc = TExecProc_inProc;
	    }

	    if (fu & Doutfd_Open) {	//Dout_fd
		expip->pp[3].Proc = (void*)1;
	    }

	    if (fu & Dinfd_Open) {	//Din_fd
		expip->pp[4].lb = TLineBuf_Create();
		expip->pp[4].Proc = TExecProc_inProc;
	    }
	}
	pid = exec_pipe_argv(argv, expipfu, expip);
	if (pid > 0) {
	    if (expip) TExecProc_DoEv(expip, evtyp_ProcStart);
	}
	if (ret_expip) *ret_expip = expip;

	if (!pid && expip) {
	    TExecProc_DoEv(expip, evtyp_ProcClose);
	}
	return pid;
}

// fu = outfd_Open | infd_Open | errfd_Open
int ExecuteProcAppTag(char *cmd, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip, int tag) {
  int argc;
  char **argv = NULL;

	if (!cmd || !*cmd) return 0;
	cmdtoargs(cmd, &argc, &argv);
	return ExecuteProcApp_argv(cmd, argc, argv, fu, OnExecProcEv, owner, ret_expip, tag);
}

int ExecuteProcApp(char *cmd, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip) {
	return ExecuteProcAppTag(cmd, fu, OnExecProcEv, owner, ret_expip, 0);
}

int ExecuteProc(char *args) {
	return ExecuteProcApp(args, 0, NULL, NULL, NULL);
}
