/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "pthd_pv.h"

static pthread_t app_pthd_id;

bool pthd_app_ChkIsThread() {
	return app_pthd_id != pthread_self();
}


void Tpthd_PTH_SYNC_SIG(int sig) {
//   printf("%s(%d): Tpthd_PTH_SYNC_SIG pid=%d thd=0x%lx\n", FL, getpid(), pthread_self());
}

static bool chk_pthd_start(Tpthd *pthd) {
	return (pthd && pthd->run && (pthd->wtyp == wtyp_Tpthd));
}

/* void Tpthd_terminates(void *data) {
 printf("%s(%d): Tpthd_terminates 0x%x\n", FL, (int)data);
	return;
} */

static void msg_free(Tpthd *pthd) {
	xfree(pthd->w_msg);
	pthd->w_msg = NULL;
	if (pthd && pthd->ret_mmMSG) {
	    xfree(pthd->ret_mmMSG->msg.ret_msg);
	    xfree(pthd->ret_mmMSG);
	    pthd->ret = 0;
	    pthd->ret_mmMSG = NULL;
	}
}

void *Tpthd_Execute(void *data) {
  Tpthd *pthd = data;

	if (!pthd || (pthd->wtyp != wtyp_Tpthd)) return NULL;
	pthd->run = True;
	FD_ZERO(&pthd->fdset_in);
	if (setjmp(pthd->ptAbort)) goto ende;

	if (pthd->DoExecute) pthd->DoExecute(pthd);

ende:
// printf("%s(%d): Tpthd_Execute Ende\n", FL);
	msg_free(pthd);
	Tpthd_MSG_deregister(pthd->app2pthd);
	pthd->run = False;
	return NULL;
}

bool Tpthd_Start(Tpthd *pthd) {
	if (!pthd || (pthd->wtyp != wtyp_Tpthd)) return False;
	if (pthd->run) return True;
	if (pthread_create(&pthd->pthd_id, NULL, Tpthd_Execute, pthd)) {
	    appError(EMSG_ERR1 | EMSG_SYS, "%s(%d): pthread_create", FL);
	    return False;
	}
	pthd->run = True;
	pthread_detach(pthd->pthd_id);
	Tpthd_MSG_registration(pthd->app2pthd);
	return True;
}

void Tpthd_Stop(Tpthd *pthd) {
	if (!chk_pthd_start(pthd)) return;
	pthd_MSG_to(pthd->app2pthd, thdMSG_CloseProcess);
	pthd->stop = True;
	Tpthd_Sync(pthd);
	pthd->stop = True;
}


void Tpthd_Sync(Tpthd *pthd) {
	if (!chk_pthd_start(pthd)) return;
	atomic_or(&pthd->fu, Tpthd_select_t0_single);
	pthread_kill(pthd->pthd_id, PTH_SYNC_SIG);
}


int Tpthd_Wait_fd(Tpthd *pthd, int fd, dword TimeOut_us) {
  struct timeval totv;

	msg_free(pthd);
	pthd->w_msg = pthd_Get_MSG(pthd->app2pthd);
	if (pthd->w_msg) {
	    atomic_and(&pthd->fu, ~Tpthd_select_t0_single);
	    return pthdWait_msg;
	}

	if (pthd->MsgToApp) {
	    pthd->MsgToApp = False;
	    if (pthd->app_pid) {
		atomic_or(&IappDat.appFu, appFu_select_t0_single);
		kill(pthd->app_pid, PTH_SYNC_SIG);
	    }
	}

	if (pthd->fu & Tpthd_select_t0_single) TimeOut_us = 0;
	totv.tv_sec = 0;
	totv.tv_usec = TimeOut_us;
	FD_ZERO (&pthd->fdset_in);
	if (fd > 0) {
	    FD_SET(fd, &pthd->fdset_in);
	    if (select(FD_SETSIZE, &pthd->fdset_in, NULL, 0, &totv) > 0) {
		if (FD_ISSET(fd, &pthd->fdset_in)) {
		    atomic_and(&pthd->fu, ~Tpthd_select_t0_single);
		    return pthdWait_fd;
		}
	    }
	} else {
	    if (TimeOut_us > 0) {
		select(FD_SETSIZE, &pthd->fdset_in, NULL, 0, &totv);
	    }
	}
	pthd->w_msg = pthd_Get_MSG(pthd->app2pthd);
	atomic_and(&pthd->fu, ~Tpthd_select_t0_single);
	if (pthd->w_msg) return pthdWait_msg;
	return pthdWait_TimeOut;
}

/*_____________________________ Destroy _________________________________*/
static bool Tpthd__Destroy(Tpthd *self) {
  Tpthd_MSG_Obj *MSG_Obj;
  Tpthd **ppthd;
  int i = 1000;

	Tpthd_Stop(self);
	while (i && self->run) {
	    usleep(1000);
	    i--;
	}

	ppthd = &IappDat.pthd_arr->pthd;
	while (*ppthd != self) ppthd = &(*ppthd)->next;
	if (*ppthd) *ppthd = self->next;

	MSG_Obj = self->pthd2app;
	self->pthd2app = NULL;
	Tfree(MSG_Obj);

	if (self->run) {
 printf("%s(%d): Tpthd__Destroy TimeOut mem not free !!!\n", FL);
	    return True;
	}

	MSG_Obj = self->app2pthd;
	self->app2pthd = NULL;
	Tfree(MSG_Obj);

//	pthread_mutex_destroy(&self->mutex);
	return False;
}

/*_____________________________ Create __________________________________*/
Tpthd *Tpthd_Create(int fu, int wtyp_obj, void *obj, Tpthd_OnMSG OnMSG) {
  Tpthd *pthd, **ppthd;
  Tpthd_arr *pthd_arr = IappDat.pthd_arr;


	if (!app_pthd_id) {
	    app_pthd_id = pthread_self();
	    app_ChkIsThread = pthd_app_ChkIsThread;
	}


	if (!pthd_arr) {
	    pthd_arr = xcalloc(sizeof(Tpthd_arr));
	    sigfillset (&pthd_arr->block_mask);
	    pthd_arr->action.sa_handler = Tpthd_PTH_SYNC_SIG;
	    pthd_arr->action.sa_mask = pthd_arr->block_mask;
	    pthd_arr->action.sa_flags = 0;
	    sigaction (PTH_SYNC_SIG, &pthd_arr->action, &pthd_arr->old_action);
	    IappDat.pthd_arr = pthd_arr;
	}

	ppthd = &pthd_arr->pthd;
	while (*ppthd) ppthd = &(*ppthd)->next;
	pthd = ttc_create(sizeof(Tpthd), wtyp_Tpthd);
	pthd->destroy = (TDESTROY)Tpthd__Destroy;

	*ppthd = pthd;
	pthd->fu = fu;
	pthd->wtyp_obj = wtyp_obj;
	pthd->obj = obj;
	pthd->app_pid = IappDat.self_pid;
	if (!(fu & Tpthd_NoMSG)) {
	    Tpthd_MSG_Obj *MSG_Obj = Tpthd_MSG_Buffer_App_Get();
	    pthd->pthd2app = Tpthd_MSG_PthdToApp_Create(MSG_Obj, wtyp_obj, obj);
	    pthd->pthd2app->OnMSG = OnMSG;
	    pthd->pthd2app->pthd = pthd;
	    pthd->app2pthd = Tpthd_MSG_AppToPthd_Create(wtyp_obj, obj);
	}
//	pthread_mutex_init(&pthd->mutex, NULL);

	return pthd;
}
