/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

#ifndef _OS_WIN32
#include <fcntl.h>
#endif

char *tmp_ChkFileName(const char *FileName, int err_level) {
  char *tmp;
	tmp = path_to_local(FileName);
	if (!tmp) {
	    appError(err_level, "\"%s\" is not a filename", FileName);
	}
	return tmp;
}

static bool fd_ChkResult(FileHandle fd, char *tmp, int err_level) {
	if (!ChkFileHandleOK(fd) && err_level) appError(err_level | EMSG_SYS, "\"%s\"", tmp);
	xfree(tmp);
	return ChkFileHandleOK(fd);
}

/* --------------------- RD --------------------------------------------- */
bool FileOpen_rd_erl(const char *FileName, FileHandle *fd, int err_level) {
  char *tmp;

	if (!fd) return False;
	tmp = tmp_ChkFileName(FileName, err_level);
	if (!tmp) return False;
#ifdef _OS_WIN32
  word *wfn = UTF8topwc(tmp);
	*fd = CreateFileW(wfn, GENERIC_READ, FILE_SHARE_READ, NULL, 
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	xfree(wfn);
#else
	*fd = open64(tmp, O_RDONLY);
#endif
	return fd_ChkResult(*fd, tmp, err_level);
}

bool FileOpen_rd(const char *FileName, FileHandle *fd) {
	return FileOpen_rd_erl(FileName, fd, EMSG_ERR1);
}

bool FileOpen_rd_gp(TglobPath *gp, FileHandle *fd) {
  char *fn = gptopc(gp, gptfu_noRootArt);
  bool res = FileOpen_rd(fn, fd);
	xfree(fn);
	return res;
}
/* --------------------- WR --------------------------------------------- */
bool FileOpen_wr_erl(const char *FileName, FileHandle *fd, mode_t mode, int err_level) {
  char *tmp;

	if (!fd) return False;
	tmp = tmp_ChkFileName(FileName, err_level);
	if (!tmp) return False;
#ifdef _OS_WIN32
  word *wfn = UTF8topwc(tmp);
	*fd = CreateFileW(wfn, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, 
			CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	xfree(wfn);
#else
	if (!mode) mode = 0644;
	*fd = open64(tmp, O_CREAT | O_TRUNC | O_WRONLY, mode);
#endif
	return fd_ChkResult(*fd, tmp, err_level);
}

bool FileOpen_wr(const char *FileName, FileHandle *fd, mode_t mode) {
	return FileOpen_wr_erl(FileName, fd, 0, EMSG_ERR1);
}

bool FileOpen_wr_gp(TglobPath *gp, FileHandle *fd) {
  char *fn = gptopc(gp, gptfu_noRootArt);
  bool res = FileOpen_wr(fn, fd, 0);
	xfree(fn);
	return res;
}

/* --------------------- IO --------------------------------------------- */
bool FileRead_erl(FileHandle fd, void *buf, int bufsize, int *retsize, int err_level) {
	if (retsize) *retsize = 0;
#ifdef _OS_WIN32
  DWORD lg;
	if (!ReadFile(fd, buf, bufsize, &lg, NULL)) {
#else
  int lg = read(fd, buf, bufsize);
	if (lg < 0) {
#endif
	    if (err_level) appError(err_level | EMSG_SYS, NULL);
	    return False;
	}
	if (retsize) *retsize = lg;
	return True;
}

bool FileRead(FileHandle fd, void *buf, int bufsize, int *retsize) {
	return FileRead_erl(fd, buf, bufsize, retsize, EMSG_ERR1);
}

bool FileWrite_erl(FileHandle fd, const void *buf, int bufsize, int *retsize, int err_level) {
	if (retsize) *retsize = 0;
#ifdef _OS_WIN32
  DWORD lg;
	if (!WriteFile(fd, buf, bufsize, &lg, NULL)) {
#else
  int lg = write(fd, buf, bufsize);
	if (lg < 0) {
#endif
	    appError(err_level | EMSG_SYS, NULL);
	    return False;
	}
	if (retsize) *retsize = lg;
	return True;
}

bool FileWrite(FileHandle fd, const void *buf, int bufsize, int *retsize) {
	return FileWrite_erl(fd, buf, bufsize, retsize, EMSG_ERR1);
}

void FileClose(FileHandle fd) {
#ifdef _OS_WIN32
	if (fd) CloseHandle(fd);
#else
	if (fd > 0) close(fd);
#endif
}
