/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "../net_xxx.h"

void TNet_SYS_ERROR(TNet *self) {
	self->ende = True;
	self->error = True;
	self->run = False;

	self->fmsg.art = fmArt_AlgErr;
	self->fmsg.typD = fmTyp_char;
	self->fmsg.D.s = get_DisplayFileStr(NULL, &self->gp);
	ppcadd2(&self->fmsg.D.s, "\n      ", SysError_pc());
	TNet_msg((TNet*)self);
	xfree(self->fmsg.D.s);
}

static void TNetLineBuf_OnPipeWrite(FileHandle fd, Tfd_list *fdl) {
  int blg;
  char *pc;
  TNetLineBuf *self = fdl->obj;

	TNetLineBuf_sync_wr(self);
	blg = TLineBuf_Get(self->lb, BUFSIZ, &pc);
	if (blg > 0) {
	    self->error = !FileWrite(fd, pc, blg, NULL);
	    self->ioinfo.DataNowSize += blg;
	    self->ioinfo.Time_us = GetCurTimeDiv_us64(self->ref_us);
	    blg = TLineBuf_Sub(self->lb, blg);
	}

	if ((self->lb->eof && !blg) || self->error) {
#ifdef _OS_WIN32
	    FileClose(fd);
	    self->iofd = NULL;
#else
	    ClosePipeProc(fd, clpp_fd);
	    self->iofd = 0;
#endif
	    self->ioinfo.stat |= fmIOInfo_EOF;
	    self->ende = True;
	    self->run = False;
	}
	TNet_send_ioinfo((TNet*)self);
}

static void TNetLineBuf_OnPipeRead(FileHandle fd, Tfd_list *fdl) {
  TNetLineBuf *self = fdl->obj;

	if (self->OnData) self->OnData((TNetLineBuf*)self, False);
	self->ioinfo.DataNowSize += TLineBuf_fd_Add(self->lb, fd);
	self->ioinfo.Time_us = GetCurTimeDiv_us64(self->ref_us);
	if (self->OnData) self->OnData((TNetLineBuf*)self, True);
	TNetLineBuf_sync_rd(self);
	if (self->lb->eof) {
	    self->run = False;
#ifdef _OS_WIN32
	    FileClose(fd);
	    self->iofd = NULL;
#else
	    ClosePipeProc(fd, clpp_fd);
	    self->iofd = 0;
#endif
	}
}


static bool TNetLineBuf_wait(TNetLineBuf *self) {
	if (!self->omod) return False;
#ifdef _OS_WIN32
  Tfd_list fd_list;
	fd_list.io = NULL;
	fd_list.obj = self;
	fd_list.m = 0;

	if (self->omod == NetOpenMod_Write) {
	    while (!self->lb->eof || self->lb->lg) {
		fd_list.Proc = TNetLineBuf_OnPipeWrite;
		TNetLineBuf_OnPipeWrite(self->iofd, &fd_list);
		AppProcessMessages();
	    }
	} else if (self->omod == NetOpenMod_Read) {
	    while (!self->lb->eof) {
		fd_list.Proc = TNetLineBuf_OnPipeRead;
		TNetLineBuf_OnPipeRead(self->iofd, &fd_list);
		AppProcessMessages();
	    }
	} else {
	    self->ende = True;
	    self->error = True;
	}
#else
	while (AppHandleMessage() && !self->ende);
#endif
	return self->ende && !self->error;
}


static bool TNetLineBuf_open(TNetLineBuf *self, int omod) {

	if (!self->lb || self->run) return False;
	TNet_Reset((TNet*)self);

	if (omod == NetOpenMod_Read) {
	    self->omod = omod;
	    TLineBuf_Clr(self->lb);
	    if (FileOpen_rd_gp(&self->gp, &self->iofd)) {
	        self->run = True;
#ifndef _OS_WIN32
		AddPipeProc(self->iofd, InPipe, TNetLineBuf_OnPipeRead, self);
#endif
		return True;
	    }
	    TNet_SYS_ERROR((TNet*)self);

	} else if (omod == NetOpenMod_Write) {
	    self->omod = omod;
	    if (FileOpen_wr_gp(&self->gp, &self->iofd)) {
	        self->run = True;
#ifndef _OS_WIN32
		AddPipeProc(self->iofd, OutPipe, TNetLineBuf_OnPipeWrite, self);
#endif
		return True;
	    }
	    TNet_SYS_ERROR((TNet*)self);
	}
	self->omod = 0;
	return False;
}

static bool TNetFileLineBuf_set_gp(TNetLineBuf *self, TglobPath *gp) {
	if (self->run || !gp_test_local_drive(gp)) return False;
	gp_pc_free(&self->gp);
	gp_dup_gpdata(&self->gp, gp);
	return True;
}
/*_____________________________ Destroy _________________________________*/
static bool TNetFileLineBuf__Destroy(TNetLineBuf *self) {
#ifdef _OS_WIN32
	FileClose(self->iofd);
#else
	ClosePipeProc(self->iofd, clpp_fd);
#endif
	return TNet__Destroy((TNet*)self);
}


/*_____________________________ Create __________________________________*/
TNetLineBuf *TNetFileLineBuf_Create(void *owner, TglobPath *gp) {
  TNetLineBuf *self;

	self = TNet_Create(sizeof(TNetLineBuf), wtyp_TNetLineBuf, gp);
	self->destroy = (TDESTROY)TNetFileLineBuf__Destroy;
	self->owner = owner;
	self->set_gp = (void*)TNetFileLineBuf_set_gp;
	self->open = TNetLineBuf_open;
	self->wait = TNetLineBuf_wait;

	return self;
}
