/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

bool file_mkdir_erl(const char *dir, mode_t mode, int err_level) {
  char *tmp, *pc;
  Tstat64 *st = NULL;
  bool ok = False;

	tmp = path_to_local(dir);
	if (!tmp) goto err;

	if (!mode) mode = 0755;
	st = file_stat_fn(tmp, fsNoFileName | fsNoLinkName | fsLinkDirekt);
	if (st) {
	    ok = (st->st & st_Dir) != 0;
	    xfree(st);
	    if (!ok) {
err:
		appError(err_level, "\"%s\" is not a directory", dir);
	    }
	    if (ok) ok = True2;

	} else {
	    if (path_sub(tmp, &pc)) {
		if (file_mkdir_erl(pc, mode, err_level)) {
#ifdef _OS_WIN32
		    word *tmpwc = UTF8topwc(tmp);
		    ok = CreateDirectoryW(tmpwc, NULL) != 0;
		    if (!ok) appError(err_level | EMSG_SYS, "\"%s\"", tmp);
		    xfree(tmpwc);
#else
		    ok = mkdir(tmp, mode) == 0;
		    if (!ok) appError(err_level | EMSG_SYS, "\"%s\"", tmp);
#endif
		}
		xfree(pc);
	    }
	}
	xfree(tmp);
	return ok;
}

bool file_mkdir_e(const char *dir, mode_t mode) {
	if (!dir || !*dir) return False;
	return file_mkdir_erl(dir, mode, EMSG_ERR1);
}
