/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include <errno.h>

#ifndef _OS_WIN32
#include <fcntl.h>
#endif

bool file_open_rd(const char *fn, TNetErrRet *ErrRet, FileHandle *fd) {
  bool result;
  char *tmp;

	if (!fd) return False;
	tmp = path_to_local(fn);
	if (!tmp) {
	    if (ErrRet) ErrRet->err = pcdup("not a filename");
	    return False;
	}
#ifdef _OS_WIN32
  word *wfn = UTF8topwc(tmp);
	*fd = CreateFileW(wfn, GENERIC_READ, FILE_SHARE_READ, NULL, 
			OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, 0);
	xfree(wfn);
#else
	*fd = open64(tmp, O_RDONLY);
#endif
	xfree(tmp);
	result = ChkFileHandleOK(*fd);
	if (!result && ErrRet) ErrRet->sys_errno = errno;
	return result;
}

bool file_open_wr(const char *fn, TNetErrRet *ErrRet, FileHandle *fd, mode_t mode) {
  bool result;
  char *tmp;

	if (!fd) return False;
	tmp = path_to_local(fn);
	if (!tmp) {
	    if (ErrRet) ErrRet->err = pcdup("not a filename");
	    return False;
	}
#ifdef _OS_WIN32
  word *wfn = UTF8topwc(tmp);
	*fd = CreateFileW(wfn, GENERIC_WRITE, FILE_SHARE_WRITE, NULL, 
			CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, 0);
	xfree(wfn);
#else
	if (!mode) mode = 0644;
	*fd = open64(tmp, O_CREAT | O_TRUNC | O_WRONLY, mode);
#endif
	xfree(tmp);
	result = ChkFileHandleOK(*fd);
	if (!result && ErrRet) ErrRet->sys_errno = errno;
	return result;
}

bool file_close(FileHandle fd, TNetErrRet *ErrRet) {
#ifdef _OS_WIN32
	if (fd) CloseHandle(fd);
#else
	if (fd > 0) close(fd);
#endif
	return True;
}
