/*
 Copyright (C) 2010 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"

static int TDirList_CpF_Directory(TDirFileList *dfl1, TDirFileList *dfl2) {
  int m1=0, m2=0;

	if (pccmp(dfl1->fn, dfl2->fn)) {
	    return 0;
	} else if (pccmp(dfl1->fn, "..")) {
	    return -1;
	} else if (pccmp(dfl2->fn, "..")) {
	    return 1;
	} else {
	    if ((dfl1->st->mode & S_IFMT) == S_IFDIR) m1++;
	    if ((dfl2->st->mode & S_IFMT) == S_IFDIR) m2++;
	    if (m1 > m2) return -1;
	    if (m1 < m2) return 1;
	}
	return 0;
}

static int TDirList_CpS(char *s1, char *s2, int fu) {
  if (fu & dlsortFu_Case) return pccasecmp_fi(s1, s2, 0); else return pccmp_i(s1, s2);
}

int TDirList_CpF_Name_UP(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1;
	m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (!m1) m1 = TDirList_CpS(dfl1->fn, dfl2->fn, fu);
	return m1;
}
int TDirList_CpF_Name_DOWN(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1;
	m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (!m1) m1 = TDirList_CpS(dfl2->fn, dfl1->fn, fu);
	return m1;
}

int TDirList_CpF_Size_UP(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (m1) return m1;
	if (dfl1->st->size > dfl2->st->size) return 1;
	if (dfl1->st->size < dfl2->st->size) return -1;
	return TDirList_CpS(dfl1->fn, dfl2->fn, fu);
}
int TDirList_CpF_Size_DOWN(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (m1) return m1;
	return TDirList_CpF_Size_UP(dfl2, dfl1, fu);
}

int TDirList_CpF_Type_UP(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  char *s1, *s2;
  int m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (m1) return m1;
        s1 = (char*)GetArtExt_pc(dfl1->st);
	s2 = (char*)GetArtExt_pc(dfl2->st);
	m1 = pccasecmp_fi(s1, s2, 0);
	if (m1) return m1;

	s1 = GetFileExt_pc(dfl1->fn);
	s2 = GetFileExt_pc(dfl2->fn);
	m1 = pccasecmp_fi(s1, s2, 0);
	if (m1) return m1;

	return TDirList_CpS(dfl1->fn, dfl2->fn, fu);
}
int TDirList_CpF_Type_DOWN(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (m1) return m1;
	return TDirList_CpF_Type_UP(dfl2, dfl1, fu);
}

int TDirList_CpF_Date_UP(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (m1) return m1;
	if (dfl1->st->mtime > dfl2->st->mtime) return 1;
	if (dfl1->st->mtime < dfl2->st->mtime) return -1;
	return TDirList_CpS(dfl1->fn, dfl2->fn, fu);
}
int TDirList_CpF_Date_DOWN(TDirFileList *dfl1, TDirFileList *dfl2, int fu) {
  int m1 = TDirList_CpF_Directory(dfl1, dfl2);
	if (m1) return m1;
	return TDirList_CpF_Date_UP(dfl2, dfl1, fu);
}

#define MaxSortProc	20
static void *dl_SortProcArr[MaxSortProc] = {
	TDirList_CpF_Name_UP, TDirList_CpF_Name_DOWN,	//Name
	TDirList_CpF_Size_UP, TDirList_CpF_Size_DOWN,	//Size
	TDirList_CpF_Type_UP, TDirList_CpF_Type_DOWN,	//Type
	TDirList_CpF_Date_UP, TDirList_CpF_Date_DOWN,	//Date
	NULL, NULL,		//Permisson
	NULL, NULL,		//User
	NULL, NULL,		//Group
//	TDirList_CpF_Link_UP, TDirList_CpF_Link_DOWN,
	NULL, NULL,		//Link
	NULL, NULL,
	NULL, NULL
};

bool TDirFileList_ChkSortOk(int fu) {
  int i;
	i = fu & 0xff;
	return ((i < MaxSortProc) && (dl_SortProcArr[i] != NULL));
}

void TDirFileList_Sort(TDirFileList **ppdfl, int fu) {
  TDirFileList *dfl;
  int i;
  int count = 0;
  TDirFileList **pptmpdfl, **ppfln, **ppfls;

	if (!ppdfl) return;
//	self->dfl_sort = NULL;
	count = 0;
	dfl = *ppdfl;
	while (dfl) {
	    count++;
	    dfl = dfl->next;
	}
// printf("%s(%d): count=%d\n", FL, count);
	if (count < 2) return;

	pptmpdfl = xmalloc(sizeof(TDirFileList*) * count);
	ppfln = pptmpdfl;
	dfl = *ppdfl;
	while (dfl) {
	    *ppfln = dfl;
	    ppfln++;
	    dfl = dfl->next;
	}
	i = fu & 0xff;
	if ((i < MaxSortProc) && dl_SortProcArr[i]) {
	    QuickSort_PL((char**)pptmpdfl, 0, count-1, fu, dl_SortProcArr[i]);
	}

	ppfls = ppdfl; //&self->dfl_sort;
	ppfln = pptmpdfl;
	for (;count;count--) {
	    *ppfls = *ppfln;
	    ppfls = &(*ppfls)->next;
	    ppfln++;
	}
	*ppfls = NULL;
	xfree(pptmpdfl);
	return;
}
