/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appbase.h"
#include "tdirlist.h"
#include "xcopy_pv.h"
#include <utime.h>
#include <fcntl.h>
#if USE_MMAP
 #include <sys/mman.h>
#endif

static bool Get_Dat_RD(TXCopy_wd *wd, int *lg, int *fuEOL, TXCopy_ld *dest) {
  TXCopy_ld *src;
  Tlgs lgs;
  bool res;

	src = wd->xcy_src;
	if (!src) {
	   if (fuEOL) *fuEOL = fio_EOF;
	   return False;
	}
	pthread_mutex_lock(&src->io_mutex);
	dest = wd->xcy_dest;
	res = TLineBuf_GetData_lgs(src->lb, &lgs, dest->buf_size, fuEOL);
	if (res) {
	    pnmove(dest->buf, lgs.s, lgs.lg);
	    *lg = lgs.lg;
	}
	pthread_mutex_unlock(&src->io_mutex);
	return res;
}

static int Chk_Dat_RD(TXCopy_wd *wd) {
  TXCopy_ld *src;
  int res = 0;

	src = wd->xcy_src;
	if (!src) return ChkDat_EOF; //0x02;

	pthread_mutex_lock(&src->io_mutex);
	if (src->lb->lg > 0) res = ChkDat_DataOk; //0x01;
	if (src->lb->eof) res |= ChkDat_EOF; //0x02;
	pthread_mutex_unlock(&src->io_mutex);
	return res;
}

/*_____________________________ Thread RD _______________________________*/
static bool dev_open_RD(TXCopy_wd *wd) {
  TXCopy_ld *ld = wd->xcy_src;
	if (!ld) return False;
	ld->fd = open(ld->fname, O_RDONLY);
	if (ld->fd > 0) return True;
	ppcchange(&ld->msg_s, SysError_pc());
	return False;
}

static bool dev_close_RD(TXCopy_wd *wd, TXCopy_ld *src) {
	if (src->fd > 0) close(src->fd);
	src->fd = 0;
	return True;
}

static int send_Mt(void *obj, int RunStat, int MsgSt) {
	if (RunStat == thdMSG_XCopy_RunStat) {
	    pt_send_Mt(obj, RunStat, MsgSt);
	    return 0;
	}
	return pt_send_M(obj, RunStat);
}

static void RD_pth_Init(Tpthd *pthd, TXCopy_wd *wd) {
	step_Init_RD(pthd, wd);
}

static void RD_pth_ChkStart(Tpthd *pthd, TXCopy_wd *wd) {
	step_ChkStart_RD(pthd, wd);
}

static void RD_pth_Start(Tpthd *pthd, TXCopy_wd *wd) {
	step_Start_RD(pthd, wd);
}

static bool RD_pth_Move(Tpthd *pthd, TXCopy_wd *wd, TXCopy_ld *src) {
  bool rdWait = False;

	pthread_mutex_lock(&src->io_mutex);
	TLineBuf_fd_Add(src->lb, src->fd);
	if (src->lb->eof) {
	    pthread_mutex_unlock(&src->io_mutex);
	    close(src->fd);
	    src->fd = 0;
	    src->srt = srt_ok;
	    if (wd->mode & cpwdMode_cut) src->srt = srt_del_wait;
	} else {
	    int lgo = TLineBuf_sync0_mem(src->lb);
	    rdWait = (lgo + src->lb->lg) > wd->MaxBufSize;
	    pthread_mutex_unlock(&src->io_mutex);
	}
	return rdWait;
}

int TXCopy_ld_File_Thread_RD_DoExecute(Tpthd *pthd) {
  int w_ret;
  TXCopy_ld *ld = pthd->obj;
  TXCopy_wd *wd = ld->wd;
  Tpthd_msg *app_msg;
  bool rdWait = True;

loop:
	if (!rdWait && (ld->fd > 0) && ((ld->srt & srt_bef_mask) == srt_move)) {
	    w_ret = Tpthd_Wait_fd(pthd, ld->fd, 4000);
	} else {
	    rdWait = True;
	    w_ret = Tpthd_Wait(pthd, 4000);
	}

	if (w_ret == pthdWait_msg) {
	    app_msg = pthd->w_msg;
	    switch(app_msg->bef & thdMSG_Mask) {
	    case thdMSG_Free:
	    case thdMSG_CloseProcess:
		goto ende;

	    case thdMSG_Init:
		if (!MSG_Init_RD(wd, app_msg->tag, ld->fd > 0)) {
 printf("%s(%d): %7d pth RD Init Internal ERROR !!!\n", FL, GetCurTimeDiv_us(wd->us));
		    pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_err);
		}
		break;

	    case thdMSG_Close:
// if (ld->fd > 0) 
// printf("%s(%d): %7d pth RD Close fd=%d isend=%d\n", FL, GetCurTimeDiv_us(wd->us), ld->fd, ld->init_send);
		MSG_Close_RD(wd, ld);
		break;
	    }

	} else if (w_ret == pthdWait_fd) {
	    rdWait = RD_pth_Move(pthd, wd, ld);
	}

bef_loop:
	if ((ld->srt & srt_bef_mask) == srt_init) {
	    RD_pth_Init(pthd, wd);
	}

	if ((ld->srt & srt_bef_mask) == srt_ChkStart) {
	    RD_pth_ChkStart(pthd, wd);
	}

	if ((ld->srt & srt_bef_mask) == srt_Start) {
	    RD_pth_Start(pthd, wd);
	}

	if ((ld->srt & srt_bef_mask) == srt_move) {
	    if (rdWait && (ld->fd > 0)) {
		pthread_mutex_lock(&ld->io_mutex);
		int lgo = TLineBuf_sync0_mem(ld->lb);
		rdWait = (lgo + ld->lb->lg) > wd->MaxBufSize;
		pthread_mutex_unlock(&ld->io_mutex);
	    }
	}

	if ((ld->srt & srt_bef_mask) == srt_del) {
	    step_Del_RD(wd);
	}

	if ((ld->srt & srt_bef_mask) == srt_ok) {
 if (ld->fd > 0)
 printf("%s(%3d): %7d RD OK\n", FL, GetCurTimeDiv_us(wd->us));
// if (!ld->init_send)
// printf("%s(%3d): %7d RD OK no init_send ???\n", FL, GetCurTimeDiv_us(wd->us));
	    if (ld->init_send) {
		ld->init_send = False;
		pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_close);
	    }
	    ld->srt = srt_0;
	}

	if ((ld->srt & srt_bef_mask) == srt_skip) {
	    ld->srt = srt_0;
	    if (ld->fd > 0) close(ld->fd);
	    ld->fd = 0;
	    pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_skip);
	}

	if (ld->srt & srt_msg_wait) {
	    ld->srt &= ~srt_msg_wait;

// printf("%s(%d): %7d WR srt_msg_wait=0x%05x \n", FL, GetCurTimeDiv_us(wd->us), ld->srt);
	    int ret = pt_send_M(pthd, thdMSG_XCopy | thdMSG_rv_mmMSG_free);
	    if (ret & xcyRetEv_BefMask) {
		if (ret & xcyRetEv_ALL) wd->skip |= ld->msgOk_skip;
		ld->skip |= wd->skip;
		ld->skip |= ld->msgOk_skip;
		ret &= xcyRetEv_BefMask;
		if (ret == xcyRetEv_SKIP) ld->srt = srt_skip;

// printf("%s(%d): %7d RD skip wd=0x%06x ld=0x%06x\n", FL, GetCurTimeDiv_us(wd->us), wd->skip, ld->skip);
		goto bef_loop;

	    } else {
		ld->srt = srt_0;
		pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_err);
	    }
	}

	if (!pthd->stop) goto loop;
ende:
	if (ld->fd > 0) close(ld->fd);
#if 0
 float a;
 float fa = ld->data_ges_lg;
 float b = ld->data_ges_us;
 float c = GetTime_us64() - wd->ges_us;
	fa /= 0x100000;
	b /= 1000000;
	c /= 1000000;
	if (b > 0) a = fa / b; else a = 0;
	if (c > 0) b = fa / c; else b = 0;
 printf("%s(%d): %7d RD Thread Stop %llu %llu %fMB/s %fMB/s\n", FL, GetCurTimeDiv_us(wd->us), ld->data_ges_lg, ld->data_ges_us, a, b);
#endif
	return 0;
}

static bool Init_RD(TXCopy_wd *wd, int funkt) {
  TXCopy_ld *ld = wd->xcy_src;

	if (!ld || !Tpthd_Start(ld->pthd)) return False;
	pt_send_Mt(ld->pthd, thdMSG_Init, funkt);
	Tpthd_Sync(ld->pthd);
	return True;
}

static bool Close_RD(TXCopy_wd *wd, int funkt) {
  TXCopy_ld *ld = wd->xcy_src;

	if (!ld || !Tpthd_Start(ld->pthd)) return False;
	pt_send_Mt(ld->pthd, thdMSG_Close, funkt);
	Tpthd_Sync(ld->pthd);
	return True;
}

static void Clear_RD(TXCopy_wd *wd) {
  TXCopy_ld *src = wd->xcy_src;
	xfree(src->st);
	src->st = NULL;
	xfree(src->fname);
	src->fname = NULL;
}

static bool set_gp_RD(TXCopy_ld *xcy, TglobPath *gp) {
	if (!gp_test_local_drive(gp)) return False;
	ppfree(&xcy->st);
	ppcfree(&xcy->fname);
	ppcchange_cmp(&xcy->gp.path, gp->path);
	ppcchange(&xcy->gp.name, gp->name);
	return True;
}
/*_____________________________ Thread WR _______________________________*/
static void WR_inc_FileCount(TXCopy_wd *wd, dword size, u64 *t64) {
	wd->xcs.FileCount++;
	wd->xcs.GesDataSize += size;
	wd->xcs.GesDataTime += GetTime_us64() - *t64;
}

static void WR_pth_Init(Tpthd *pthd, TXCopy_wd *wd) {
  TXCopy_ld *src, *dest;
  char *pc;
  int xcyChk;
  int src_ifmt;
  u64 t64;

	src = wd->xcy_src;
	dest = wd->xcy_dest;

	if (!step_Init_WR(pthd, wd, &xcyChk, &src_ifmt)) goto err_skip;

	switch (xcyChk & xcyChkDest_BefMask) {

	case xcyChkDest_DelDest:
	    if (!unlink(dest->fname)) {
		xfree(dest->st);
		dest->st = NULL;
	    } else {
		dest->msgOk_skip = xcySkip_errDelFile;
		dest->msg_nr = xpyErr_Alg | xpyErr_DelFile;
		goto err_sys_skip;
	    }
	    break;

	case xcyChkDest_MsgSkip:
	    goto err_skip;

	case xcyChkDest_Skip:
	    dest->srt = srt_skip;
	    return;

	case xcyChkDest_Preserv:
	    dest->srt = srt_preserv;
	    return;
	}

	t64 = GetTime_us64();

	switch (src_ifmt) {

	case S_IFDIR:
	    dest->stmode = wd->mode_dir;
	    if (mkdir(dest->fname, dest->stmode) < 0) {
		dest->msgOk_skip = xcySkip_errMkDir;
		dest->msg_nr = xpyErr_Alg | xpyErr_MkDir;
		wd->xcyChk |= xcyChkDest_NoDirCreate;
		goto err_sys_skip;
	    }
	    pc = pcadd(dest->gp.path, dest->gp.name);
	    add_dest_dir(wd, pc);
	    xfree(pc);
	    WR_inc_FileCount(wd, src->st->size, &t64);
	    dest->srt = srt_preserv;
	    return;

	case S_IFREG:
	    dest->atime = 0;
	    dest->mtime = 0;
	    if (!dest->st) dest->stmode = wd->mode_file;
	    dest->srt = srt_open;
	    return;

	case S_IFLNK:
	    if (!dest->st) {
		pc = Get_ExFileDat_pc(src->st, exfdt_LinkName);
		if (pc && *pc && (symlink(pc, dest->fname) < 0)) {
		    dest->msgOk_skip = xcySkip_errMkLink;
		    dest->msg_nr = xpyErr_Alg | xpyErr_MkLink;
		    goto err_sys_skip;
		}
		WR_inc_FileCount(wd, src->st->LinkSize, &t64);
	    }
	    dest->srt = srt_ok;
	    return;

	case S_IFIFO:
	    if (!dest->st) {
		dest->stmode = wd->mode_blk;
		if (mkfifo(dest->fname, dest->stmode | S_IFIFO) < 0) {
		    dest->msgOk_skip = xcySkip_errMkNod;
		    dest->msg_nr = xpyErr_Alg | xpyErr_MkNod;
		    goto err_sys_skip;
		}
		WR_inc_FileCount(wd, 0, &t64);
	    }
	    dest->srt = srt_preserv;
	    return;

	case S_IFCHR:
	case S_IFBLK:
	case S_IFSOCK:
	    if (!dest->st) {
		dest->stmode = wd->mode_blk;
		if (mknod(dest->fname, dest->stmode | (src->st->mode & S_IFMT), src->st->rdev) < 0) {
		    dest->msgOk_skip = xcySkip_errMkNod;
		    dest->msg_nr = xpyErr_Alg | xpyErr_MkNod;
		    goto err_sys_skip;
		}
		WR_inc_FileCount(wd, 0, &t64);
	    }
	    dest->srt = srt_preserv;
	    return;
	}

	dest->srt = srt_ok | srt_msg_wait;
	dest->msgOk_skip = xcySkip_errType;
	dest->msg_nr = xpyErr_Alg | xpyErr_Type;
	goto err_sys_skip;

err_sys_skip:
	ppcchange(&dest->msg_s, SysError_pc());
	dest->msg_nr |= xpyErr_Msg;
err_skip:
	dest->srt |= srt_msg_wait;
	if (dest->skip & dest->msgOk_skip) dest->srt = srt_skip;
}

static bool dev_open_WR(TXCopy_wd *wd) {
  TXCopy_ld *ld = wd->xcy_dest;
	ld->fd = open(ld->fname, O_WRONLY|O_CREAT|O_TRUNC, ld->stmode);
	if (ld->fd > 0) return True;
	ppcchange(&ld->msg_s, SysError_pc());
	return False;
}

static bool dev_close_WR(TXCopy_wd *wd, TXCopy_ld *dest) {
	if (dest->fd > 0) close(dest->fd);
	dest->fd = 0;
	return True;
}


static bool WR_pth_Move(Tpthd *pthd, TXCopy_wd *wd) {
	return step_Move_WR(pthd, wd);
}

static void WR_pth_Preserv(Tpthd *pthd, TXCopy_wd *wd) {
  TXCopy_ld *src, *dest;  
  u64 t;

	dest = wd->xcy_dest;
	dest->srt = srt_ok;
	src = wd->xcy_src;
	if (!(dest->skip & xcySkip_Chmod) && (src->stmode != dest->stmode)) {
	    t = GetTime_us64();
	    if (chmod(dest->fname, src->stmode) && !(wd->skip & xcySkip_errChmod)) {
 printf("%s(%d): Err chmod=%s\n", FL, SysError_pc());
	    }
	    wd->xcs.GesDataTime += GetTime_us64() - t;
	}

	if (!(dest->skip & xcySkip_Chown)
	    && ((src->uid != dest->uid)||(src->gid != dest->gid))) {
	    t = GetTime_us64();
	    if ((chown(dest->fname, src->st->uid, src->st->gid) < 0) && !(wd->skip & xcySkip_errChown)) {
    printf("%s(%d): Err chown=%s\n", FL, SysError_pc());
	    }
	    wd->xcs.GesDataTime += GetTime_us64() - t;
	}

	if (!(dest->skip & xcySkip_Utime)
	    && ((src->atime != dest->atime)||(src->mtime != dest->mtime))) {
	    t = GetTime_us64();
	    struct utimbuf ut;
	    ut.actime = src->atime;
	    ut.modtime = src->mtime;
	    if ((utime(dest->fname, &ut) < 0) && !(wd->skip & xcySkip_errUtime)) {
    printf("%s(%d): Err utime=%s\n", FL, SysError_pc());
	    }
	    wd->xcs.GesDataTime += GetTime_us64() - t;
	}
}

int TXCopy_ld_File_Thread_WR_DoExecute(Tpthd *pthd) {
  int w_ret;
  TXCopy_ld *ld = pthd->obj;
  TXCopy_wd *wd = ld->wd;
  Tpthd_msg *app_msg;
  bool wrWait = True;

// printf("%s(%d): %7d pth WR Thread Start 0x%x\n", FL, GetCurTimeDiv_us(wd->us), (int)pthd);

loop:
	if (!wrWait && (ld->fd > 0) && ((ld->srt & srt_bef_mask) == srt_move)) {
	    w_ret = Tpthd_Wait_fd(pthd, ld->fd, 4000);
	} else {
	    w_ret = Tpthd_Wait(pthd, 4000);
	}
	if (w_ret == pthdWait_msg) {
	    app_msg = pthd->w_msg;
	    switch(app_msg->bef & thdMSG_Mask) {
	    case thdMSG_Free:
	    case thdMSG_CloseProcess:
		goto ende;

	    case thdMSG_Init:
		if (!MSG_Init_WR(pthd, wd, app_msg->tag, ld->fd > 0)) {
 printf("%s(%d): pth WR Init Internal ERROR !!!\n", FL);
		}
		break;

/*	    case thdMSG_Open:
		break; */

	    case thdMSG_Close:
// printf("%s(%d): %7d pth WR Close\n", FL, GetCurTimeDiv_us(wd->us));
		ld->srt = srt_skip;
		break;
	    }

	} else if (w_ret == pthdWait_fd) {
	    wrWait = WR_pth_Move(pthd, wd);
	}


Bef_loop:
	if ((ld->srt & srt_bef_mask) == srt_init) {
	    WR_pth_Init(pthd, wd);
	}

	if ((ld->srt & srt_bef_mask) == srt_open) {
	    if (!step_Open_WR(pthd, wd)) ld->srt = srt_0;
	}

	if ((ld->srt & srt_bef_mask) == srt_move) {

	    if (wrWait && (ld->fd > 0)) {
		int m = wd->xcy_src->chk_dat(wd);
		if (m & 0x01) {
		    wrWait = False;
		} else if (m & 0x02) {
		    close(ld->fd);
		    ld->fd = 0;
		    ld->srt = srt_preserv;
		    wrWait = True;
		}
	    }
	}

	if ((ld->srt & srt_bef_mask) == srt_preserv) {
	    WR_pth_Preserv(pthd, wd);
	}

	if ((ld->srt & srt_bef_mask) == srt_skip) {
// printf("%s(%3d): %7d WR Skip\n", FL, GetCurTimeDiv_us(wd->us));
	    if (ld->fd > 0) close(ld->fd);
	    ld->fd = 0;
	    ld->srt = srt_0;
	    pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_skip);
	}

	if ((ld->srt & srt_bef_mask) == srt_ok) {
// printf("%s(%d): %7d WR OK\n", FL, GetCurTimeDiv_us(wd->us));
	    ld->srt = srt_0;
	    pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_close);
	}

	if (ld->srt & srt_msg_wait) {
	    ld->srt &= ~srt_msg_wait;

// printf("%s(%d): %7d WR srt_msg_wait=0x%05x \n", FL, GetCurTimeDiv_us(wd->us), ld->srt);
	    int ret = pt_send_M(pthd, thdMSG_XCopy | thdMSG_rv_mmMSG_free);
	    if (ret & xcyRetEv_BefMask) {
		if (ret & xcyRetEv_ALL) wd->skip |= ld->msgOk_skip;
		ld->skip |= wd->skip;
		if (wd->stat & xcywdstat_SameFileMk) {
		    ld->skip = 0;
		}
		ld->skip |= ld->msgOk_skip;
		ret &= xcyRetEv_BefMask;
		if (ret == xcyRetEv_SKIP) ld->srt = srt_skip;

// printf("%s(%d): %7d WR skip=0x%05x\n", FL, GetCurTimeDiv_us(wd->us), ld->skip);
		goto Bef_loop;
	    } else {
// printf("%s(%d): %7d WR ERROR srt=0x%05x\n", FL, GetCurTimeDiv_us(wd->us), ld->srt);
		ld->srt = srt_0;
		pt_send_Mt(pthd, thdMSG_XCopy_RunStat, MsgSt_err);
	    }
	}


	if (!pthd->stop) goto loop;
ende:
	if (ld->fd > 0) close(ld->fd);
#if 0
 float a;
 float fa = ld->data_ges_lg;
 float b = ld->data_ges_us;
 float c = GetTime_us64() - wd->ges_us;
	fa /= 0x100000;
	b /= 1000000;
	c /= 1000000;
	if (b > 0) a = fa / b; else a = 0;
	if (c > 0) b = fa / c; else b = 0;
 printf("%s(%d): %7d WR Thread Stop %llu %llu %fMB/s %fMB/s\n", FL, GetCurTimeDiv_us(wd->us), ld->data_ges_lg, ld->data_ges_us, a, b);
#endif
	return 0;
}


static bool Init_WR(TXCopy_wd *wd, int funkt) {
  TXCopy_ld *ld = wd->xcy_dest;

	if (!ld || !Tpthd_Start(ld->pthd)) return False;
	pt_send_Mt(ld->pthd, thdMSG_Init, funkt);
	Tpthd_Sync(ld->pthd);
	return True;
}

static bool dev_WR(TXCopy_wd *wd, byte *data, int lg, int fuEOL) {
  TXCopy_ld *dest = wd->xcy_dest;

	write(dest->fd, data, lg);
	wd->xcs.DataNowSize += lg;
	wd->xcs.DataNowTime = GetTime_us64() - dest->open_data_us;

	return True;
}

static bool Close_WR(TXCopy_wd *wd, int funkt) {
  TXCopy_ld *ld = wd->xcy_dest;

	if (!ld || !Tpthd_Start(ld->pthd)) return False;
	pt_send_Mt(ld->pthd, thdMSG_Close, funkt);
	Tpthd_Sync(ld->pthd);
	return True;
}

static void Clear_WR(TXCopy_wd *wd) {
  TXCopy_ld *dest = wd->xcy_dest;
	xfree(dest->st);
	dest->st = NULL;
	xfree(dest->fname);
	dest->fname = NULL;
}

/*_____________________________ AppEv ___________________________________*/
int TXCopy_ld_WR_OnMSG(Tpthd_msg *msg) {
  TXCopy_ld *ld = msg->obj;
  int bef = msg->bef & thdMSG_Mask;
  TXCopy_wd *wd;

	if (ld && (ld->wtyp == wtyp_TXcopy_ld) && 
	    ld->wd && (ld->wd->wtyp == wtyp_TXcopy)) {
	    wd = ld->wd;

	    if ((bef == thdMSG_XCopy) && wd->hcyEv) {
// printf("%s(%d): %7d WR MSG \n", FL, GetCurTimeDiv_us(wd->us));
		u64 t64 = GetTime_us64();
		int res = wd->hcyEv(wd, ld->msg_nr | xpyMsg_dest);
		wd->ges_us += GetTime_us64() - t64;
		return res;

	    } else if (bef == thdMSG_XCopy_RunStat) {

		return handle_Mt_WR(wd, msg->tag);
	    } else {
 printf("%s(%d): WR wtyp=0x%x size=%d msg=0x%x wtyp_obj=0x%x obj=0x%x\n", FL, msg->wtyp, msg->size, msg->bef, msg->wtyp_obj, (int)msg->obj);
	    }

	} else {

 printf("%s(%d): WR wtyp=0x%x size=%d msg=0x%x wtyp_obj=0x%x obj=0x%x\n", FL, msg->wtyp, msg->size, msg->bef, msg->wtyp_obj, (int)msg->obj);
	}
	return 0;
}

int TXCopy_ld_RD_OnMSG(Tpthd_msg *msg) {
  TXCopy_ld *ld = msg->obj;
  int bef = msg->bef & thdMSG_Mask;
  TXCopy_wd *wd;

	if (ld && (ld->wtyp == wtyp_TXcopy_ld) && 
	    ld->wd && (ld->wd->wtyp == wtyp_TXcopy)) {
	    wd = ld->wd;

	    if ((bef == thdMSG_XCopy) && wd->hcyEv) {
// printf("%s(%d): %7d RD MSG \n", FL, GetCurTimeDiv_us(wd->us));
		u64 t64 = GetTime_us64();
		int res = wd->hcyEv(wd, ld->msg_nr | xpyMsg_src);
		wd->ges_us += GetTime_us64() - t64;
		return res;

	    } else if (bef == thdMSG_XCopy_RunStat) {
		return handle_Mt_RD(wd, msg->tag);

	    } else {
 printf("%s(%d): RD wtyp=0x%x size=%d msg=0x%x wtyp_obj=0x%x obj=0x%x\n", FL, msg->wtyp, msg->size, msg->bef, msg->wtyp_obj, (int)msg->obj);
	    }
	} else {

 printf("%s(%d): RD wtyp=0x%x size=%d msg=0x%x wtyp_obj=0x%x obj=0x%x\n", FL, msg->wtyp, msg->size, msg->bef, msg->wtyp_obj, (int)msg->obj);
	}
	return 0;
}


/*_____________________________ Destroy _________________________________*/
bool TXCopy_ld_File__Destroy(TXCopy_ld *self) {
	Tfree(self->pthd);
	if (self->fd > 0) close(self->fd);
	gp_pc_free(&self->gp);
	xfree(self->st);
	xfree(self->fname);
	xfree(self->msg_s);
	Tfree(self->lb);
	xfree(self->buf);
	pthread_mutex_destroy(&self->io_mutex);
	return False;
}

/*_____________________________ Create __________________________________*/
TXCopy_ld *TXCopy_ld_File_Create(int fu, TXCopy_wd *wd, TglobPath *gp) {
  TXCopy_ld *self;

	self = ttc_create(sizeof(TXCopy_ld), wtyp_TXcopy_ld);
	self->destroy = (TDESTROY)TXCopy_ld_File__Destroy;
	self->art = ldart_file;
	self->wd = wd;
	gp_dup_gpdata(&self->gp, gp);

	if (fu & xcyFU_OUTPUT) {
	    if (wd->xcy_src && (wd->xcy_src->art == ldart_file)) {
		self->gp.fu |= gpi_dest;
	    }
	    self->clear = Clear_WR;
	    self->init = Init_WR;
	    self->close = Close_WR;

	    self->send_Mt = send_Mt;
	    self->dev_open = dev_open_WR;
	    self->dev_close = dev_close_WR;
	    self->dev_wr = dev_WR;
	    self->stat_gp = file_stat_gp_net;


	    self->pthd = Tpthd_Create(0, self->wtyp, self, TXCopy_ld_WR_OnMSG);
	    self->pthd->DoExecute = TXCopy_ld_File_Thread_WR_DoExecute;
	    self->buf_size = 0x2000; //BUFSIZ;
	    self->buf = xmalloc(self->buf_size);
	} else {
	    gp->fu &= ~gpi_dest;
	    self->clear = Clear_RD;
	    self->init = Init_RD;
	    self->close = Close_RD;
	    self->set_gp = set_gp_RD;
	    self->rmdir_gp = file_rmdir_gp;
	    self->unlink_gp = file_unlink_gp;

	    self->send_Mt = send_Mt;
	    self->dev_open = dev_open_RD;
	    self->dev_close = dev_close_RD;

	    self->stat_gp = file_stat_gp_net;

	    self->pthd = Tpthd_Create(0, self->wtyp, self, TXCopy_ld_RD_OnMSG);
	    self->get_dat = Get_Dat_RD;
	    self->chk_dat = Chk_Dat_RD;
	    self->pthd->DoExecute = TXCopy_ld_File_Thread_RD_DoExecute;
	    self->lb = TLineBuf_Create();
	}
	pthread_mutex_init(&self->io_mutex, NULL);

	return self;
}

bool TXCopy_ld_File_ChkType(TglobPath *gp, bool dest) {
	return gp_test_local_drive(gp);
}

ThXCopy hXCopy_ld_File = {
  TXCopy_ld_File_ChkType,
  TXCopy_ld_File_Create
};
