/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

//BUFSIZ
#define LBSEGSIZE 0x2000

void TLineBuf_Clr(TLineBuf *lb) {
	if (lb->memlg > (LBSEGSIZE<<1)) {
	    xfree(lb->UTF8_buf);
	    lb->UTF8_buf = NULL;
	    lb->UTF8_buf_size = 0;
	    xfree(lb->pc);
	    lb->pc = NULL;
	    lb->memlg = 0;
	}
	lb->pc_rdidx = NULL;
	lb->lg = 0;
	lb->wrlg = 0;
	lb->eof = False;
}

int TLineBuf_sync0_mem(TLineBuf *lb) {
	if (!lb->pc_rdidx) return 0;
	if (lb->lg <= 0) {
	    lb->lg = 0;
	    lb->pc_rdidx = NULL;
	    return 0;
	}

	int lgo = lb->pc_rdidx - lb->pc;
	if ((lgo > (lb->memlg >> 2)) || ((lgo > 256) && (lb->lg < 256))) {
	    pcnmove(lb->pc, lb->pc_rdidx, lb->lg);
	    lb->pc_rdidx = NULL;
	    return 0;
	}
	return lgo;
}

char *TLineBuf_Set_mem(TLineBuf *lb, int lg) {
  int m, lgi, lgo;

	if (!lb->pc) {
	    m = lg / (LBSEGSIZE<<1) + 1;	//New=BUFSIZ * 2
	    lb->memlg = (LBSEGSIZE<<1) * m;
	    lb->pc = xmalloc(lb->memlg);
	    return lb->pc;
	}

	lgo = TLineBuf_sync0_mem(lb);
	lgi = lgo + lb->lg;
	if ((lb->memlg - lgi) < lg) {
	    m = lg - (lb->memlg - lgi);
	    m = m / (LBSEGSIZE<<2) + 1;		//BUFSIZ * 4
	    lb->memlg += (LBSEGSIZE<<2) * m;
	    lb->pc = xrealloc(lb->pc, lb->memlg);
	    if (lb->pc_rdidx) lb->pc_rdidx = &lb->pc[lgo];
	}
	return &lb->pc[lgi];
}

void TLineBuf_Add(TLineBuf *lb, const char *s, int lg) {
  char *pcd;

	if (!s || (lg <= 0)) {
	    lb->eof = True;
	    return;
	}
	pcd = TLineBuf_Set_mem(lb, lg+1);
	ppcnmove(&pcd, s, lg);
	lb->lg += lg;
	lb->wrlg += lg;
}

int TLineBuf_fd_Add(TLineBuf *lb, FileHandle fd) {
  int lg;
  char *pcd;

	if (!ChkFileHandleOK(fd)) {
	    lb->eof = True;
	    return 0;
	}
	pcd = TLineBuf_Set_mem(lb, LBSEGSIZE+1);

	FileRead(fd, pcd, LBSEGSIZE, &lg);
	if (lg > 0) {
	    lb->lg += lg;
	    lb->wrlg += lg;
	    pcd[lg] = 0;
	} else {
	    lb->eof = True; 
	    lg = 0;
	    *pcd = 0;
	}
	return lg;
}

bool TLineBuf_GetData_lgs(TLineBuf *lb, Tlgs *lgs, int max, int *fuEOL) {
  int tmpEOL = 0;

	lgs->s = NULL;
	lgs->lg = 0;
	if (!fuEOL) fuEOL = &tmpEOL;
	if (!lb->lg) goto chk_EOF;

	lgs->s = lb->pc;
	lgs->lg = lb->lg;
	if (lb->pc_rdidx) lgs->s = lb->pc_rdidx;

	if (max && (lgs->lg > max)) {
	    lgs->lg = max;
	    lb->lg -= lgs->lg;
            lb->pc_rdidx = lgs->s + lgs->lg;
	} else {
	    lb->lg = 0;
	    lb->pc_rdidx = NULL;
	}


chk_EOF:
	if (lb->eof && (lb->lg < 1)) *fuEOL |= fio_EOF;
	return (lgs->s != NULL);
}

int TLineBuf_Sub(TLineBuf *lb, int lg) {
  char *pc = lb->pc;

	if (lg <= 0) return 0;
	if (lb->pc_rdidx) pc = lb->pc_rdidx;

	lb->lg -= lg;
	if (lb->lg > 0) {
            lb->pc_rdidx = pc + lg;
	} else {
	    lb->lg = 0;
	    lb->pc_rdidx = NULL;
	}
	return lb->lg;
}

int TLineBuf_Get(TLineBuf *lb, int max, char **ppc) {
  int lg;

	lg = lb->lg;
	if (max && (lg > max)) lg = max;

	if (ppc) {
	    if (lb->pc_rdidx) {
		*ppc = lb->pc_rdidx;
	    } else {
		*ppc = lb->pc;
	    }
	}
	return lg;
}

/*_____________________________ Destroy _________________________________*/
bool TLineBuf__Destroy(TLineBuf *self) {
	xfree(self->UTF8_buf);
	xfree(self->pc);
	return False;
}

/*_____________________________ Create __________________________________*/
TLineBuf *TLineBuf_Create() {
  TLineBuf *self = ttc_create(sizeof(TLineBuf), wtyp_TLineBuf);
	self->destroy = (TDESTROY)TLineBuf__Destroy;
	self->pc = xmalloc(LBSEGSIZE<<1);	//new BUFSIZ * 2
	self->memlg = LBSEGSIZE<<1;
	return self;
}
