/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

//*********************** TLines ************************
void TLines_DoCount(TLines *tls) {
	if (!tls->EvStop && tls->OnCount) tls->OnCount(tls);
}

void TLines_DoLineChange(TLines *tls, TColS *pcs, int idx) {
	if (!tls->EvStop && tls->OnLineChange) {
	    if (idx >= tls->Count) idx = tls->Count - 1;
	    tls->OnLineChange(tls, pcs, idx);
	    tls->Changed = True;
	}
}

void TLines_ResLinec(TLines *tls, TColS * pcs, int idx) {
	if (pcs->c) {
	    if (tls->OnLineDelc) tls->OnLineDelc(tls, pcs, idx);
	    xfree(pcs->c);
	    pcs->c = NULL;
	}
}

void TLines_ResLine(TLines *tls, int idx) {
  TColS *pcs = tls->lines->cs[idx];
	if (!pcs) return;
	if (tls->OnLineDel) tls->OnLineDel(tls, pcs, idx);
	xfree(pcs->s);
	TLines_ResLinec(tls, pcs, idx);
	xfree(pcs);
	tls->lines->cs[idx] = NULL;
}

void TLines_DelN(TLines *tls, int idx, int n) {
  int i;
  TColS **acs, **ecs;

	if ((idx >= tls->Count) || (tls->Count == 0) || (n < 1)) return;
	if ((idx+n) >= tls->Count) n = tls->Count - idx;
	for (i=idx; i<(idx+n); i++) TLines_ResLine(tls, i);
	tls->Count -= n;
	if (tls->Count <= 0) {
	    TLines_Clear(tls);
	    return;
	}
	acs = &tls->lines->cs[idx];
	ecs = &tls->lines->cs[idx+n];
	for (i = idx; i < tls->Count; i++) {
	    *acs = *ecs;
	    *ecs = NULL;
	    acs++; ecs++;
	}
	tls->Changed = True;
	TLines_DoCount(tls);
}

void TLines_Del(TLines *tls, int idx) {
	TLines_DelN(tls, idx, 1);
}

void TLines_Clear(TLines *tls) {
  int i;
	if (tls->lines) {
	    for (i=0; i<tls->Count; i++) TLines_ResLine(tls, i);
	    xfree(tls->lines);
	    tls->lines = NULL;
	}
	tls->Count = 0;
	if (tls->OnCount) tls->OnCount(tls);
//	TLines_DoCount(tls);
}

void TLines_Clear_c(TLines *tls) {
  TColS *pcs;

	if (!tls || !tls->lines) return;
	for (int i=0;i<tls->Count;i++) {
	    pcs = tls->lines->cs[i];
	    if (pcs) TLines_ResLinec(tls, pcs, i);
	}
}

static void *TLines_Get_csMem(TLines *tls) {
  int lg = sizeof(TColS);
	if (tls->csType == csType_TFileS) lg = sizeof(TFileS);
	return xcalloc(lg);
}

void *TLines_Add_(TLines *tls) {
  TColS *pcs;
	if (!tls->lines) set_ptr_arr((void*)&tls->lines, 256);
	if (tls->Count >= tls->lines->items) set_ptr_arr((void*)&tls->lines, tls->lines->items + 256);
	pcs = TLines_Get_csMem(tls);
	tls->lines->cs[tls->Count] = pcs;
	tls->Count++;
	return pcs;
}

TColS *TLines_nAdd(TLines *tls, const char *s, int lg) {
  TColS *pcs = TLines_Add_(tls);
	pcs->lg = lg;
	ppcnchange(&pcs->s, s, lg);
	TLines_DoLineChange(tls, pcs, tls->Count-1);
	TLines_DoCount(tls);
	return pcs;
}

TColS *TLines_Add(TLines *tls, const char *s) {
	return TLines_nAdd(tls, s, pclen(s));
}

TColS *TLines_Add_lgs(TLines *tls, Tlgs *lgs, int eol_fu) {
  TColS *cs = TLines_nAdd(tls, lgs->s, lgs->lg);

	cs->fu = eol_fu & fio_EOL_Mask;
	if (tls->DefaultEOL == -1) {
	    tls->DefaultEOL = eol_fu & fio_EOL_Mask;
	} else if (tls->DefaultEOL != (eol_fu & fio_EOL_Mask)) {
	    if (!(eol_fu & fio_EOF)) tls->DefaultEOL = fio_EOL_mix;
	}
	return cs;
}

static TColS *TLines_Ins_(TLines *tls, int idx) {
  int i;
  TColS *pcs;
	if (idx >= tls->Count) return TLines_Add_(tls);
	if (tls->Count >= tls->lines->items) set_ptr_arr((void*)&tls->lines, tls->lines->items + 256);
	for (i=tls->Count; i>idx; i--) tls->lines->cs[i] = tls->lines->cs[i-1];
	pcs = TLines_Get_csMem(tls);
	tls->lines->cs[idx] = pcs;
	tls->Count++;
	return pcs;
}

TColS *TLines_nIns(TLines *tls, int idx, const char *s, int lg) {
  TColS *pcs = TLines_Ins_(tls, idx);
	pcs->lg = lg;
	ppcnchange(&pcs->s, s, lg);
	TLines_DoLineChange(tls, pcs, idx);
	TLines_DoCount(tls);
	return pcs;
}

TColS *TLines_Ins(TLines *tls, int idx, const char *s) {
	return TLines_nIns(tls, idx, s, pclen(s));
}



char *TLines_Get_pc(TLines *self, int idx) {
	if (!self || (idx >= self->Count) || (idx < 0)) return NULL;
	return self->lines->cs[idx]->s;
}

TColS *TLines_Get_pcs(TLines *self, int idx) {
	if (!self || (idx >= self->Count) || (idx < 0)) return NULL;
	return self->lines->cs[idx];
}

int REGPARM3_FUNC TLines_GetLineCount(TLines *self) {
	if (!self) return 0;
	return self->Count;
}

bool TLines__Destroy(TLines *self) {
	self->OnCount = NULL;
	TLines_Clear(self);
	xfree(self->FileName);
	xfree(self->st);
	return False;
}

TLines *TLines_Create() {
  TLines *tls = ttc_create(sizeof(TLines), wtyp_TLines);

	tls->destroy = (TDESTROY)TLines__Destroy;
#ifdef _OS_WIN32
	tls->DefaultEOL = fio_EOLwindos;
#else
	tls->DefaultEOL = fio_EOLunix;
#endif
	return tls;
}
