/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "sysbase.h"

void *ttc_create(int tcsize, int wtyp) {
  TTC0 *tc;
	if (tcsize < sizeof(TTC0)) {
	    appError(EMSG_ERR1, "%s(%d): ttc_create: tcsize=%d < sizeof(TTC)=%d", FL, tcsize, sizeof(TTC));
	    tcsize = sizeof(TTC0);
	}
	tc = xcalloc(tcsize);
	tc->wtcsize = tcsize;
	if (tcsize > TTC_MAXSIZE) {
	    tc->wtcsize = TTC_MAXSIZE;
	    appError(EMSG_WARN, "%s(%d): ttc_create: wtyp=0x%x tcsize=%d > %d", FL, tcsize, TTC_MAXSIZE, wtyp);
	}
	tc->wtyp = wtyp;
	tc->a55a = 0xa55a;
	return tc;
}

static bool chk_struct(void *tc) {
	if (((TTC*)tc)->a55a != 0xa55a) {
	    appError(EMSG_ERR1, "%s(%d): 0x%x ttc_chk_struct: not a TTC struct ???", FL, (int)tc);
	    return False;
	}
	if (((TTC*)tc)->wtcsize < sizeof(TTC0)) {
	    appError(EMSG_ERR1, "%s(%d): 0x%x ttc_chk_struct: tcsize=%d < sizeof(TTC)=%d", FL, (int)tc, ((TTC*)tc)->wtcsize, sizeof(TTC));
	    return False;
	}
	return True;
}

bool ttc_chk_struct(void *tc) {
	if (!tc) return False;
	return chk_struct(tc);
}

bool REGPARM3_FUNC ttc_chk_is_wtyp(void *tc, int wtyp) {
	return  tc && 
		(((TTC*)tc)->a55a == 0xa55a) &&
		(((TTC*)tc)->wtcsize >= sizeof(TTC0)) &&
		((TTC*)tc)->wtyp == wtyp;
}

static int tc_err_count = 0;

void Tfree(void *tc) {
	if (!tc) return;
	if (!chk_struct(tc)) {
	    tc_err_count++;
	    if (tc_err_count > 10) appError(EMSG_ERR_EXIT, "%s(%d): err_count > 10", FL);
	    return;
	}
	tc_err_count = 0;
	if (((TTC*)tc)->destroy && ((TTC*)tc)->destroy(tc)) return;
	((TTC*)tc)->a55a = 0;	//set old mem free
	xfree(tc);
}

void pp_Tfree(TTC0 **pptc) {
  void *tc;
	if (!pptc) return;
	tc = *pptc;
	if (!tc) return;
	if (!chk_struct(tc)) {
	    tc_err_count++;
	    if (tc_err_count > 10) appError(EMSG_ERR_EXIT, "%s(%d): err_count > 10", FL);
	    return;
	}
	tc_err_count = 0;
	*pptc = NULL;
	if (((TTC*)tc)->destroy && ((TTC*)tc)->destroy(tc)) return;
	((TTC*)tc)->a55a = 0;	//set old mem free
	xfree(tc);
}
