/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include <signal.h>

int TForm_EvProc(Tobjmsg *msg) {
  TForm *self;
  XEvent *ev;

  Window JRoot, JParent, *JChildren, JFrame;
  dword NChildren;
  TGeometry gm;

	self = (TForm *) msg->ap->wins;
	if (!ttc_chk_struct(self)) return 0;

	switch(msg->msg) {
	case msgXEv:
	    ev = (XEvent *) msg->p;

	    switch(ev->type) {

	    case ConfigureNotify: //22
/*
 printf("ConfigureNotify: 0x%x 0x%x 0x%x %d %d %dx%d\n", 
  (int)ev->xconfigure.event, (int)ev->xconfigure.window, (int)ev->xconfigure.above,
  ev->xconfigure.x, ev->xconfigure.y,
  ev->xconfigure.width, ev->xconfigure.height);
*/
		JChildren = NULL;
		JFrame = ev->xconfigure.window;
		while (XQueryTree(IappXDat.dpy, JFrame, &JRoot, &JParent, &JChildren, &NChildren)) {
		    if (JChildren) {XFree(JChildren); JChildren = NULL;}
		    if (JParent == IappXDat.RootW) break;
		    JFrame = JParent;
		}
		if (JChildren) XFree(JChildren);
		if (get_geometry(JFrame, &gm)) {
		    self->left = gm.x;
		    self->top = gm.y;
		} else {
		    self->left = ev->xconfigure.x;
		    self->top = ev->xconfigure.y;
		}
		self->width = ev->xconfigure.width;
		self->height = ev->xconfigure.height;
		if ((self->xwc.x != self->left) || (self->xwc.y != self->top)) {
		    if (self->OnMove) self->OnMove(self);
		    self->xwc.x = self->left;
		    self->xwc.y = self->top;
		}
		if ((self->xwc.width != self->width) || (self->xwc.height != self->height)) {
		    self->xwc.width = self->width;
		    self->xwc.height = self->height;
		    if (self->ResizeStartMk && self->OnResize && self->Visible) 
			self->OnResize((TWins *)self);
		    TWins_Resize((TWins *)self);
		}
		break;

	    case ClientMessage:
		if ((ev->xclient.format == 32) && 
		    (ev->xclient.data.l[0] == IappXDat.XA_WM_DELETE_WINDOW)) {
			if (self->Close) self->Close(self);
			return 1;
		}
		break;
//	    default:
//		break;
	    }
	    break;

	case msgVisibleFormShow:
		if (self->Visible && self->Show) self->Show(self);
		return 1;
		break;
#ifdef DRAG_AND_DROP
/*	case msg_dnd:
		switch(msg->f) {
		case msgf_dndEnter:
			if (IappXDat.dnd) return 1;
			break;
		}
		break; */
#endif
	} 

	return TWins_EvProc(msg);
}


void TForm_Close(TForm *self) {
  int CanClose;

	CanClose = True;
	if (self->OnCanClose) self->OnCanClose(self, &CanClose);
	if (CanClose) {
	    if (IappXDat.StartForm == self) {
		IappDat.gQUIT_Ev = True;
	    } else {
		if (CanClose == FormCloseAndFree) {
		    Tfree(self);
		} else {
		    TWins_SetVisible((TWins*)self, False);
		}
	    }
	}
}

void  TForm_Show(TForm *self) {
  bool send_WMProp = False;
  int x, y, xm, ym;
  XID ParentWinID;
  XSizeHints SizeHints;
  XSizeHints *pSizeHints;
  XTextProperty xname, xIconName;
  XTextProperty *pxname, *pxIconName;
  XClassHint *FormClass;
  XWMHints *pWMHints;

	ParentWinID = IappXDat.RootW;
	if (self->Parent) ParentWinID = self->Parent->WinID;
	pSizeHints = NULL;
	pWMHints = NULL;
	pxname = NULL;
	pxIconName = NULL;
	FormClass = NULL;
	Fill0(&SizeHints, sizeof(XSizeHints));
	if (self->Visible && self->IsParent) {
	    XSetInputFocus (IappXDat.dpy, self->WinID, RevertToParent, CurrentTime);
	} else {

	    SizeHints.flags = 0;
	    if (self->FormPosition != poDefault) {
  TWRect rc;
		if (GetXWinAbs_rc(ParentWinID, &rc)) {
		    x = rc.left + rc.width;
		    y = rc.top + rc.height;
		    xm = rc.left + ((rc.width - self->width) >> 1);
		    ym = rc.top + ((rc.height - self->height) >> 1);
		} else {
		    x = IappXDat.Root_width;
		    y = IappXDat.Root_height;
		    xm = (x - self->width) >> 1;
		    ym = (y - self->height) >> 1;
		}
		if (xm < 0) xm = 0;
		if (ym < 0) ym = 0;

		if (self->FormPosition == poCenter) {
		    self->left = xm;
		    self->top = ym;
		} else if (self->Gravity) {
		    SizeHints.flags |= PWinGravity;
		    SizeHints.win_gravity = self->Gravity;
		    switch(self->Gravity) {
		    case NorthGravity:
			self->left = xm + self->left;
			break;
		    case NorthEastGravity:
			self->left = x - self->left - self->width;
			break;
		    case WestGravity:
			self->top = ym + self->top;
			break;
		    case CenterGravity:
			self->left = xm + self->left;
			self->top = ym + self->top;
			break;
		    case EastGravity:
			self->left = x - self->left - self->width;
			self->top = ym + self->top;
			break;
		    case SouthWestGravity:
			self->top = y - self->top - self->height;
			break;
		    case SouthGravity:
			self->left = xm + self->left;
			self->top = y - self->top - self->height;
			break;
		    case SouthEastGravity:
			self->left = x - self->left - self->width;
			self->top = y - self->top - self->height;
			break;
		    }
		    self->Gravity = 0;
		}
		SizeHints.x = self->left;
		SizeHints.y = self->top;
		SizeHints.flags |= USPosition;
	    }

	    if (self->FormZStyle & fszNoResize) {
		self->min_width = self->width;
		self->max_width = self->width;
		self->min_height = self->height;
		self->max_height = self->height;
	    }

	    if ((self->min_width > 0) || (self->min_height > 0)) {
		SizeHints.min_width = self->min_width;
		SizeHints.min_height = self->min_height;
		SizeHints.flags |= PMinSize;
	    }
	    if ((self->max_width > 0) || (self->max_height > 0)) {
		SizeHints.max_width = self->max_width;
		SizeHints.max_height = self->max_height;
		SizeHints.flags |= PMaxSize;
	    }

	    if (SizeHints.flags) {
		pSizeHints = &SizeHints;
		send_WMProp = True;
	    }

	    if (self->Icon && self->Icon->pic) {
		pWMHints = xcalloc(sizeof(XSizeHints));
		pWMHints->flags |= IconPixmapHint;
		pWMHints->icon_pixmap = self->Icon->pic;
		if (self->Icon->mask) {
		    pWMHints->flags |= IconMaskHint;
		    pWMHints->icon_mask = self->Icon->mask;
		}
		send_WMProp = True;
	    }

	    if (!self->IsParent) {
		self->IsParent = True;
#ifndef USE_NANO_X
		if ((self->FormZStyle & fszTransient) && self->Parent) {
		    TForm *fm = GetParentForm(self->Parent);
		    if (fm) XSetTransientForHint(IappXDat.dpy, self->WinID, fm->WinID);
		}
#endif /* USE_NANO_X */

		TWins_AdjustSize((TWins*)self);

		if (self->vds.s) {
		    xname.encoding = XA_STRING;
		    xname.format = 8;
		    xname.nitems = self->vds.lg;
		    xname.value = (unsigned char*)self->vds.s;
		    pxname = &xname;
		}
		if (self->IconName) {
		    xIconName.encoding = XA_STRING;
		    xIconName.format = 8;
		    xIconName.nitems = pclen(self->IconName);
		    xIconName.value = (unsigned char*)self->IconName;
		    pxIconName = &xIconName;
		}
		if (self->XFormClass.res_name || self->XFormClass.res_class)
			FormClass = &self->XFormClass;
		send_WMProp = True;
	    }
	    if (send_WMProp)
		XSetWMProperties(IappXDat.dpy, self->WinID, pxname, pxIconName,
				 NULL, 0, pSizeHints, pWMHints, FormClass);

	    xfree(pWMHints);
	    self->Visible = True;
	    MapWinID(self->WinID, 2);
	    Resize((TWins *)self);
	}
	if (self->OnShow) self->OnShow(self);

}

int TForm_DoKeyDown(TWins *self, TKeyEvent *KeyEv) {
  TForm *sw = (TForm*)self;

	if (sw->Root_MenuButton && sw->Root_MenuButton->mbuStat) return 1;
	return TWins_DoKeyDown(self, KeyEv);
}

int TForm_DoKeyUp(TWins *self, TKeyEvent *KeyEv) {
  TForm *sw = (TForm*)self;
	if (!sw->Root_MenuButton || !sw->Root_MenuButton->mbuStat) {
	    TWins_DoKeyUp(self, KeyEv);
	}

	if (!(KeyEv->state & 0x4) && 
            ((KeyEv->keysym == XK_Meta_L) || (KeyEv->keysym == XK_Alt_L))) {
	    if (sw->DoMenuButtonExec) sw->DoMenuButtonExec(sw);
	}
	return 1;
}


void TForm_SetCaption(TWins *self, const char *s) {
  XTextProperty xtp;

	if (pccmp(s, self->vds.s)) return;
	ppcchange(&self->vds.s, s);
	self->vds.lg = pclen(s);
	if (self->IsParent) {
	    xtp.encoding = XA_STRING;
	    xtp.format = 8;
	    xtp.nitems = self->vds.lg;
	    xtp.value = (unsigned char*)self->vds.s;
	    XSetWMProperties(IappXDat.dpy, self->WinID, &xtp, NULL, NULL, 0, NULL, NULL, NULL);
	}
}

void TForm_SetVisible(TWins *self, bool Value) {
	if (self->Visible != Value) {
	    if (self->Visible) {
		if (((TForm*)self)->Close) ((TForm*)self)->Close((TForm*)self);
	    } else {
		if (((TForm*)self)->Show) ((TForm*)self)->Show((TForm*)self);
	    }
	}
}

/*_____________________________ Destroy _________________________________*/
bool TForm__Destroy(TForm *self) {
	return TWins__Destroy((TWins*)self);
}

/*____________________________ Create ___________________________________*/
void TForm_CreateParams(TWins_Create_struct *wcs) {
  TForm *self;
  TForm *parentForm;

	self = wcs->wins;
	parentForm = GetParentForm((TWins*)wcs->SubP);
	if (parentForm) {
	    self->Icon = parentForm->Icon;
	    self->IconName = parentForm->IconName;
	}

//	self->destroy = (TDESTROY)TForm__Destroy;
	self->dnd_isEnable = False;

	self->width = 320;
	self->height = 240;

	self->apph->EvProc = TForm_EvProc;
	self->SetCaption = TForm_SetCaption;
	self->SetVisible = TForm_SetVisible;

	self->DoKeyDown = TForm_DoKeyDown;
	self->DoKeyUp = TForm_DoKeyUp;
	self->Show = TForm_Show;
	self->Close = TForm_Close;

	if (!IappXDat.StartForm) IappXDat.StartForm = self;

	self->apph->aos |= aosIsForm;
	wcs->XAttr.event_mask |= FocusChangeMask;
}


void TForm_CreateXWin(TWins_Create_struct *wcs) {
  TForm *self = wcs->wins;
	TWins_CreateXWin(wcs);
#ifndef USE_NANO_X
	XSetWMProtocols(IappXDat.dpy, self->WinID, &IappXDat.XA_WM_DELETE_WINDOW, 1);
#endif /* USE_NANO_X */

#ifdef FORM_NET_WM_PID
	XChangeProperty(IappXDat.dpy, self->WinID, 
			IappXDat.XA_NET_WM_PID, XA_CARDINAL, 32, PropModeReplace,
			(unsigned char *) &IappDat.self_pid, 1);
#endif /* FORM_NET_WM_PID */
}

void *TForm_CreatParamAlg_WCSfkt(TWins *Parent, int FormMemSize, dword WCSfkt) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, FormMemSize, wtyp_Form);
	wcs.CreateParams = TForm_CreateParams;
	wcs.CreateXWin = TForm_CreateXWin;
	wcs.SubP = Parent;
	wcs.WCSfkt = WCSfkt;
	TWins_DefaultConf(&wcs);
	((TForm*)(wcs.wins))->Parent = Parent;
	return wcs.wins;
}

void *TForm_CreatParamAlg(TWins *Parent, int FormMemSize) {
	return TForm_CreatParamAlg_WCSfkt(Parent, FormMemSize, 0);
}

TForm *TForm_Create(TWins *Parent) {
	return TForm_CreatParamAlg_WCSfkt(Parent, sizeof(TForm), 0);
}
