/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

void TWins_Refresh(TWins *self) {
  int i;
	if (self->PaintOnEvEndMk || !self->IsParent || !self->Visible) return;
	self->PaintOnEvEndMk = True;
	i = IappXDat.PaintOnEvEnd_Count;
	if (i >= IappXDat.PaintOnEvEnd_List->items) 
		set_ptr_arr(&IappXDat.PaintOnEvEnd_List, IappXDat.PaintOnEvEnd_List->items+256);
	IappXDat.PaintOnEvEnd_List->val[i] = self->apph;
	IappXDat.PaintOnEvEnd_Count++;
}

static void TWins_AppMouseDown(TWins *self, XButtonEvent *ev) {
	TMouseDown *md = &IappXDat.MouseDown;
	md->WinIsEnabled = TestWinIsEnabled(self);
	if (md->WinIsEnabled && (ev->button == 1)) Set_Focus(self);
	md->oldapph = md->apph;
	md->apph = self->apph;
	md->x = ev->x;
	md->y = ev->y;
	md->x_root = ev->x_root;
	md->y_root = ev->y_root;
	md->obj_left = self->left;
	md->obj_top = self->top;
	md->obj_width = self->width;
	md->obj_height = self->height;
	if (self->DoMouseDown) self->DoMouseDown(self, ev);
}

static void TWins_ChkInfoWin(TWins *self, bool Leave) {
  TWins *InfoWin = IappXDat.InfoWin;
	if (!Leave && self->InfoWin) {
	    IappXDat.InfoWin = self->InfoWin;
	    IappXDat.InfoWin_us = IappDat.app_us + 400000;
	    if (InfoWin && (InfoWin != self->InfoWin)) {
		SetVisible(InfoWin, False);
	    }
	} else if (InfoWin) {
	    IappXDat.InfoWin = NULL;
	    SetVisible(InfoWin, False);
	}
}

int TWins_EvProc(Tobjmsg *msg) {
  TWins *self;
  THdnd *Hdnd;

	self = msg->ap->wins;
	if (!ttc_chk_struct(self)) return 0;

	if (msg->msg == msgXEv) {
  XEvent *ev;
	    ev = (XEvent *) msg->p;
	    switch(ev->type) {
	    case Expose:
		if ((self->KeyFu & mHasFocus) && (self->CaptionFu & mFocusRecIsPaint)) {
		    XClearWindow(IappXDat.dpy, self->WinID);
#ifdef _OS_WIN32
		    self->CaptionFu &= ~mFocusRecIsPaint;
		    return 1;
#endif /* ! _OS_WIN32 */
		}
		self->CaptionFu &= ~mFocusRecIsPaint;
		Refresh(self);
		return 1;

	    case ButtonPress:
		TWins_AppMouseDown(self, &ev->xbutton);
		return 1;

	    case ButtonRelease:
		if (self->DoMouseUp) self->DoMouseUp(self, &ev->xbutton);
		return 1;

	    case MotionNotify:
		TWins_ChkInfoWin(self, False);
		if (self->DoMouseMove) self->DoMouseMove(self, &ev->xmotion);
		return 1;

	    case EnterNotify:
	    case LeaveNotify:
		if (self->DoMouseOver) {
		    bool Leave = ev->type == LeaveNotify;
		    TWins_ChkInfoWin(self, Leave);
		    self->DoMouseOver(self, &ev->xcrossing, Leave);
		}
		return 1;

	    default:
		break;
	    }
	} else {
  TKeyEvent *KeyEv;
	    switch(msg->msg) {

	    case msgKey:
		KeyEv = (TKeyEvent*)msg->p;
		if (!TestWinIsAktiv(self)) {
		    if (!(self->KeyFu & mK_Ev_OnDisWins)) return -1;
		    KeyEv->keyst |= keyst_WinsIsDis;
		}
		if (KeyEv->keyst & keyst_Press) {
		    if (self->DoKeyDown) self->DoKeyDown(self, KeyEv);
		} else if (KeyEv->keyst & keyst_Release) {
		    if (self->DoKeyUp) self->DoKeyUp(self, KeyEv);
		} else return 0;
		return 1;

	    case msgPaint:
		Paint(self);
		return 1;

	    case msgFocus:
		if (self->DoFocus) self->DoFocus(self, msg->m);
		return 1;

	    case msgDestroy:
		Tfree(self);
		return 1;
#ifdef DRAG_AND_DROP
	    case msg_dnd:
// printf("%s(%d): Wins=0x%x f=%d\n", FL, (int)self->WinID, msg->f);
// fflush(stdout);
	      Hdnd = self->Hdnd;
	      if (Hdnd) {
		switch(msg->f) {
		case msgf_dndEnter: return 1;
		case msgf_dndPosition:
		    if (Hdnd->OndndMouseMove)
		      return Hdnd->OndndMouseMove(self, (XMotionEvent*)msg->p);
// printf("%s(%d): Wins=0x%x f=%d 0x%x\n", FL, (int)self, msg->f, self->DodndMouseMove);
		    break;
		case msgf_dndDrop: 
		    if (Hdnd->OndndDrop) 
		      return Hdnd->OndndDrop(self);
		    break;
		case msgf_dndStatus: 
		    if (Hdnd->OndndStatus) 
		      return Hdnd->OndndStatus(self, (XClientMessageEvent*)msg->p);
		    break;
		}
	      }
#endif
	    }
	}
	return 0;
}


bool WinsCtl(TWins *self, TWinsCtl *wctl) {
  bool res = False;
	if (self && wctl) {
	    if (self->WinsCtl) res = self->WinsCtl(self, wctl);
	    if (wctl->flag & wctlf_ptrFree) xfree(wctl->ptr);
	    if (wctl->flag & wctlf_SelfFree) xfree(wctl);
	}
	return res;
}

#ifdef USE_DECOR
bool TWins_Decor(TWins *self, int bef, int mk, void *p) {
  bool res = False;

	if (self->Decor && self->Decor->DoDecor) {
		res = self->Decor->DoDecor(self, bef, mk, p);
	}
	return res;
}
#endif

static void set_Align(TWins *wins, TWRect *rc) {
  int Align;
	if (wins->Visible) {
	  Align = wins->Align;
	  if (Align & alTop) {
	    rc->top += wins->height;
	    rc->height -= wins->height;
	  } else if (Align & alLeft) {
	    rc->left += wins->width; 
	    rc->width -= wins->width; 
	  } else if (Align & alRight) {
	    rc->width -= wins->width; 
	  } else if (Align & alBottom) {
	    rc->height -= wins->height;
	  }
	}
}

void TWins_GetAlignClientRect(TWins *self, TWRect *rc) {
  Tapph *ap;

	Get_ClientRect(self->Parent, rc);

	if (self->Align & (alTop | alLeft)) {
	    ap = self->apph;
	    while (ap->prev) ap = ap->prev;

	    while (ap != self->apph) {
		set_Align(ap->wins, rc);
		ap = ap->next;
	    }
	    if (self->Align & alLeft) {
		ap = self->apph->next;
		while (ap) {
		    if (ap->wins->Visible && (ap->wins->Align & alBottom)) {
			rc->height -= ap->wins->height;
		    }
		    ap = ap->next;
		}
	    }

	} else if (self->Align & (alRight | alBottom)) {
	    ap = self->apph;
	    while (ap->next) ap = ap->next;

	    while (ap != self->apph) {
		set_Align(ap->wins, rc);
		ap = ap->prev;
	    }
	} else if (self->Align & alClient) {
	    ap = self->apph;
	    while (ap->prev) ap = ap->prev;

	    while (ap != self->apph) {
		set_Align(ap->wins, rc);
		ap = ap->next;
	    }
	    ap = self->apph;
	    while (ap->next) ap = ap->next;

	    while (ap != self->apph) {
		set_Align(ap->wins, rc);
		ap = ap->prev;
	    }
	}
}

void TWins_GetClientRect(TWins *self, TWRect *rc) {
	rc->left   = self->BorderOffLeft;
	rc->width  = self->width - (self->BorderOffLeft + self->BorderOffRigth);
	rc->top    = self->BorderOffTop;
	rc->height = self->height - (self->BorderOffTop + self->BorderOffBottom);
}

void TWins_SetSize(TWins *self) {
  int mkPos, mkSize;

	if (!self->IsParent) return;
	TWins_ChkWinSizeMinMax(self, &self->width, &self->height);
	mkPos = mkSize = 0;
	if (self->left != self->xwc.x) mkPos = CWX;
	if (self->top != self->xwc.y) mkPos |= CWY;
	if (self->width != self->xwc.width) mkSize = CWWidth;
	if (self->height != self->xwc.height) mkSize |= CWHeight;
	if (!mkPos && !mkSize) return;

	self->xwc.x = self->left;
	self->xwc.y = self->top;
	self->xwc.width = self->width;
	self->xwc.height = self->height;

	if (mkSize && mkPos) {
	    XMoveResizeWindow(IappXDat.dpy, self->WinID,
			self->left, self->top, self->width, self->height);
	    if (self->ResizeStartMk && self->OnResize) self->OnResize(self);

	} else if (mkSize) {
	    XResizeWindow(IappXDat.dpy, self->WinID, self->width, self->height);
	    if (self->ResizeStartMk && self->OnResize) self->OnResize(self);

	} else {
	    XMoveWindow(IappXDat.dpy, self->WinID, self->left, self->top);
	}
#ifdef USE_DECOR
	TWins_Decor(self, decor_NewSize, mkPos | mkSize, NULL);
#endif
} 

void TWins_AdjustSize(TWins *self) {
  TWRect rc;

	if (self->IsParent && self->Parent && self->Align &&
	    (self->Parent->WinID != IappXDat.RootW)) {

	    GetAlignClientRect(self, &rc);

	    self->left = rc.left;
	    self->top = rc.top;
	    if (self->Align & alTop) {
		self->width = rc.width;
	    } else if (self->Align & alLeft) {
		self->height = rc.height;
	    } else if (self->Align & alBottom) {
		self->top += rc.height - self->height; self->width = rc.width;
	    } else if (self->Align & alRight) {
		self->left += rc.width - self->width; self->height = rc.height;
	    } else if (self->Align & alClient) {
		self->width = rc.width; self->height = rc.height;
	    }
	}
	TWins_SetSize(self);
}

void TWins_DoActivate(TWins *self) {
	if (self->RunStatus & runst_Activate) return;
	if (self->IsParent && TestWinIsAktiv(self)) {
	    self->RunStatus |= runst_Activate;
	    if (self->OnActivate) self->OnActivate(self);
	    self->RunStatus |= runst_Activate1;
	}
}

void TWins_Resize(TWins *self) {
  Tapph *ap;

	if (self->IsParent) {
	    self->AdjustSize(self);
	    if (!self->ResizeStartMk && self->OnResize && self->Visible) {
		self->OnResize(self);
		self->ResizeStartMk = True;
	    }
	    ap = self->apph->next_child;
	    while (ap) {
		Resize(ap->wins);
		ap = ap->next;
	    }
	}
	if (self->DoActivate) self->DoActivate(self);
}

void TWins_PaintEnCaption(TWins *self, bool En) {
  TVDrawString *vds = &self->vds;
  int ForeColor = vds->ForeColor;

	vds->Image = ((self->CaptionFu & mCaptionSolid) != 0);
	if (!En) {
	    vds->ForeColor = AddColor(self->Color, 0x202020);
	    DrawColorString(1, 1, vds);
	    vds->ForeColor = SubColor(self->Color, 0x404040);
	}
	DrawColorString(0, 0, vds);
	vds->ForeColor = ForeColor;
}

void TWins_DoPaint(TWins *self) {
  int x, y, tw, m;
  TWRect rc;
  TVDrawString *vds = &self->vds;

	if (!vds->lg && !(self->CaptionFu & mCaptionRePaint)) return;
	if ((self->CaptionFu & mCaptionPaint) && vds->Font) {
	    if (!(self->CaptionFu & mCaptionNoAutoXY)) {
		Get_ClientRect(self, &rc);
		x = rc.left;
		y = rc.top;
		m = rc.height - vds->Font->height;
		if (self->CaptionFu & mCaptionBottom) {
		    y = rc.top + m;
		} else if ((self->CaptionFu & mCaptionTop) == 0) {
		    y = rc.top + (m >> 1);
		}
		vds->Yoff = self->CaptionOffY + y;

		tw = TextWidthColorString(vds);
		if (self->CaptionFu & mCaptionRigth) {
		    x = rc.left + (rc.width - tw);
		} else if ((self->CaptionFu & mCaptionLeft) == 0) {
		    x = rc.left + ((rc.width - tw) >> 1);
		}
		if (self->CaptionFu & mCaptionRigth) {
		    x -= self->CaptionOffX;
		} else {
		    x += self->CaptionOffX;
		}
		vds->Xoff = x;
		if (self->CaptionFu & mCaptionRePaint) {
		    self->CaptionFu &= ~mCaptionRePaint;
		    FillRectRGB(self->WinID, self->Color, rc.left, vds->Yoff, rc.width, vds->Font->height);
		}
		TWins_PaintEnCaption(self, TestWinIsEnabled(self));

	    } else {
		TWins_PaintEnCaption(self, TestWinIsEnabled(self));
	    }

	}
}

void TWins_Paint(TWins *self) {
	self->LastPaintTime_us = GetTime_us();
	self->PaintOnEvEndMk = False;
	if (!self->IsParent || !self->Visible) return;
#ifdef _OS_WIN32
	SetTmp_hGC(self->WinID);
#endif /* _OS_WIN32 */
	if (!self->ResizeStartMk) {
	    Resize(self);
	    self->ResizeStartMk = True;
	} else {
	    if (self->DoActivate) self->DoActivate(self);
	}
	if (self->DoPaint) self->DoPaint(self);
	if (self->OnPaint) self->OnPaint(self);
#ifdef USE_DECOR
	if (!TWins_Decor(self, decor_Border, 0, NULL))
#endif
		StdBorderPaint(self->WinID, self->BorderFu, 0, 0, self->width, self->height);
	self->RunStatus |= runst_Paint1;
#ifdef _OS_WIN32
	ReleaseTmp_hGC(self->WinID);
#endif /* _OS_WIN32 */
}

void TWins_RePaint(TWins *self) {
	if (!TestWinIsVisible(self) || !self->IsParent) return;
	self->CaptionFu &= ~mFocusRecIsPaint;
	XClearWindow(IappXDat.dpy, self->WinID);
	Refresh(self);
}


void TWins_RePaintAllNext(Tapph *ap) {
	while (ap) {
	    if (ap->next_child) TWins_RePaintAllNext(ap->next_child);
	    RePaint(ap->wins);
	    ap = ap->next;
	}
}

void TWins_SetEnabled(TWins *self, bool Value) {
	if (self->Enabled != Value) {
	    self->Enabled = Value;
	    RePaint(self);
	    if (self->apph->next_child) {
		TWins_RePaintAllNext(self->apph->next_child);
	    }
	}
}

static void Clr_Activate(Tapph *ap) {
	while (ap) {
	    if (ap->next_child) Clr_Activate(ap->next_child);
	    ap->wins->RunStatus &= ~runst_Activate;
	    ap = ap->next;
	}
}

void TWins_SetVisible(TWins *self, bool Value) {
	if (self->Visible != Value) {
	    self->Visible = Value;
	    if (!self->IsParent) return;
	    if (self->Align && self->Parent) {
		Resize(self->Parent);
	    } else if (Value) {
		Resize(self);
	    }
	    if (!Value) {
		Clr_Activate(self->apph->next_child);
		self->RunStatus &= ~runst_Activate;
	    }
	    MapWinID(self->WinID, Value ? 1 : 0);
	    if (self->OnVisible) self->OnVisible(self, Value);
	}
}

void TWins_ChkWinSizeMinMax(TWins *self, int *w, int *h) {
  int m;
	if (w) {
	    m = *w;
	    if (m < self->min_width) m = self->min_width;
	    if (self->max_width && (m > self->max_width)) m = self->max_width;
	    if (m < 1) m = 1;
	    if (m > IappXDat.Root_width) m = IappXDat.Root_width;
	    *w = m;
	}
	if (h) {
	    m = *h;
	    if (m < self->min_height) m = self->min_height;
	    if (self->max_height && (m > self->max_height)) m = self->max_height;
	    if (m < 1) m = 1;
	    if (m > IappXDat.Root_height) m = IappXDat.Root_height;
	    *h = m;
	}
}

void TWins_SetRC(TWins *self, int left, int top, int width, int height) {
	TWins_ChkWinSizeMinMax(self, &width, &height);

	if ((self->width != width) || (self->height != height)) {
	    self->left = left;
	    self->top = top;
	    self->width = width;
	    self->height = height;
	    Resize(self);
	    if (self->IsParent && self->Align && self->Parent) Resize(self->Parent);

	} else if ((self->left != left) || (self->top != top)) {
	    self->left = left;
	    self->top = top;
	    AdjustSize(self);
	}
}


void TWins_DoClick(TWins *self, long KeyState, bool DblClick) {
  if (self->OnClick && TestWinIsEnabled(self)) self->OnClick(self, KeyState, DblClick);
}

void TWins_DoMouseDown(TWins *self, XButtonEvent *ev) {
  int m;

	self->MouseClickStop = False;
	if (self->OnMouseDown) self->OnMouseDown(self, ev);

	if (TestWinIsEnabled(self)) {
	    if (ev->button == 1) {
		m = ev->time - IappXDat.MouseDown.time_ev;
		IappXDat.MouseDown.DblClickAkt = ((m < 350) && 
			(IappXDat.MouseDown.oldapph == self->apph));
	    }
	}
	IappXDat.MouseDown.time_ev = ev->time;
}

void TWins_DoMouseUp(TWins *self, XButtonEvent *ev) {
  dword ahID;
  bool UpDownOK;

	ahID = self->apph->ahID;
#ifdef DRAG_AND_DROP
  Tdnd *dnd = IappXDat.dnd;
	if (dnd && (dnd->dndM_SrcWin != 0) && dnd->MoveSet) {
	    dnd->MoveSet(dndM_Move_Drop, 0, 0, 0);
	    return;
	}
#endif

#ifdef _OS_WIN32
	UpDownOK = (IappXDat.MouseDown.apph == self->apph);
#else
  int jx, jy;
  Window UpWin, Jwin;

	UpWin = Jwin = IappXDat.RootW;
	while (Jwin && XTranslateCoordinates(IappXDat.dpy, IappXDat.RootW, Jwin, 
	    ev->x_root, ev->y_root, &jx, &jy, &Jwin)) {
	    if (Jwin) UpWin = Jwin;
// printf("%s(%d): DoMouseUp %dx%d %dx%d %dx%d 0x%08x -> 0x%08x\n", FL, ev->x, ev->y, ev->x_root, ev->y_root, jx, jy, (int)self->WinID, (int)Jwin);
	}

	UpDownOK = (IappXDat.MouseDown.apph == self->apph) && 
		   (self->WinID == UpWin);
#endif

	if (TestWinIsEnabled(self) && !self->MouseClickStop) {
	    if (ev->button == 1) {

		if (UpDownOK && self->DoClick) {
		    bool DblClick = IappXDat.MouseDown.DblClickAkt && 
			      ((ev->time - IappXDat.MouseDown.time_ev) < 500);
		    self->DoClick(self, ev->state, DblClick);
		}

		IappXDat.MouseDown.DblClickAkt = False;
		/* chk self is Destroy */
		if (!Find_ahID(IappXDat.apph, ahID)) return;

	    } else if (ev->button == 3) {
		if (UpDownOK && self->PopupMenu) {
		    self->PopupMenu->owner = self;
#ifdef Test_Root_WinId
	      int jx, jy;
	      XID Jwin;
		    if (TestRootWinId) {
			XTranslateCoordinates(IappXDat.dpy, TestRootWinId, OrgRootWinId, 0, 0,
			    &jx, &jy, &Jwin);
			ev->x_root -= jx;
			ev->y_root -= jy;
		    }
#endif
		    TMenu_Show(self->PopupMenu, self, ev->x_root, ev->y_root);
		}
	    }
	}

	if (self->OnMouseUp) self->OnMouseUp(self, ev);
}

void TWins_DoMouseMove(TWins *self, XMotionEvent *ev) {
	if (self->OnMouseMove) self->OnMouseMove(self, ev);
}

void TWins_DoMouseOver(TWins *self, XCrossingEvent *ev, bool Leave) {
	if (self->OnMouseOver) self->OnMouseOver(self, ev, Leave);
}

int TWins_DoKeyDown(TWins *self, TKeyEvent *KeyEv) {
  TWins *ws;
  int ks;

	if (self->OnKeyDown) self->OnKeyDown(self, KeyEv);
// printf("TWins_DoKeyDown: win=%d xkey=%d(0x%x) c=%d(0x%x)\n", (int)self->WinID, ev->keycode, ev->keycode, (int)*keysym, (int)*keysym);

	if ((KeyEv->state & (ShiftMask | ControlMask))) return 0;

	ks = KeyEv->keysym;
	if (((ks == XK_Right) && (self->KeyFu & mK_Right)) ||
	    ((ks == XK_Down) && (self->KeyFu & mK_Down)) ||
	    ((ks == XK_Return) && (self->KeyFu & mK_ReturnTab)) ||
	    ((ks == XK_Tab) && (self->KeyFu & mK_Tab))) {
		ws = GetNextFocEl(self->apph, False, ks == XK_Tab);
		if (ws) Set_Focus(ws);
		return 1;

	} else if (((ks == XK_Left) && (self->KeyFu & mK_Left)) ||
		   ((ks == XK_Up) && (self->KeyFu & mK_Up))) {
		ws = GetNextFocEl(self->apph, True, False);
		if (ws) Set_Focus(ws);
		return 1;
	}
	return 0;
}

int TWins_DoKeyUp(TWins *self, TKeyEvent *KeyEv) {

	if (self->OnKeyUp) self->OnKeyUp(self, KeyEv);
	return 1;
}

void TWins_DoFocus(TWins *self, int Change) {

	if (Change & WinsFocusIN) {
	    TForm *tmpWin = GetParentForm(self);
	    IappXDat.KeyInWin = self;
	    if (tmpWin) {
		tmpWin->FocusWin = self;
		int WinsBase = self->wtyp & wtyp_WinsBaseMask;
		if ((WinsBase != wtyp_MenuButton) && (WinsBase != wtyp_Menu)) {
		    tmpWin->LastNoMenuFocusWin = self;
		}
	    }
	    self->KeyFu |= mHasFocus;
	} else {

	    self->KeyFu &= ~mHasFocus; 
	}
	if (self->OnFocus) self->OnFocus(self, Change);
}

void Set_Focus(TWins *self) {
	if (!self || !TestWinIsEnabled(self)) return;
	if ((self->KeyFu & (mCanFocus | mHasFocus)) != mCanFocus) return;

  TForm *tmpWin = GetParentForm(self);
  Tobjmsg msg;

	msg.msg = msgFocus;
	msg.m = WinsFocusChange;
	if (tmpWin && tmpWin->FocusWin) {
	    Exec_Win_Ev(&msg, tmpWin->FocusWin->WinID);
	}
	msg.m = WinsFocusIN|WinsFocusChange;
	Exec_Win_Ev(&msg, self->WinID);
}

void vds_SetStr(TVDrawString *vds, const char *s) {
	ppcchange(&vds->s, s);
	vds->lg = pclen(s);
	vds->scda = NULL;
}

void TWins_SetCaption(TWins *self, const char *s) {
	if (!pccmp(s, self->vds.s)) {
	    self->CaptionFu |= mCaptionRePaint;
	    vds_SetStr(&self->vds, s);
	    if (self->OnSetCaption) self->OnSetCaption(self, s);
	    if (self->vds.lg) Refresh(self); else RePaint(self);
	}
}

void TWins_SetUTF8(TWins *self, bool UTF8) {
	if (self->vds.UTF8on == UTF8) return;
	self->vds.UTF8on = UTF8;
	RePaint(self);
}

void TWins_SetColor(TWins *self, dword Color) {
  XSetWindowAttributes XAttr;
	if (self->Color != Color) {
	    self->Color = Color;
	    Fill0(&XAttr, sizeof(XSetWindowAttributes));
	    XAttr.background_pixel = GetColor_RGB(Color);
	    XChangeWindowAttributes(IappXDat.dpy, self->WinID, CWBackPixel, &XAttr);
	    RePaint(self);
	}
	self->vds.BackColor = Color;
}

void TWins_SetBorder(TWins *self, int BorderFu) {
	if (self->BorderFu != BorderFu) {
	    self->BorderFu = BorderFu;
	    TWins_AdjustSize(self);
//	    Refresh(self);
	}
}

bool TWins_Ctl(TWins *self, TWinsCtl *wctl) {
	if (wctl->bef == wctlb_SetFont) {
	    self->vds.Font = MakeFont(wctl->ptr, IappXDat.Default_Font->name, self->vds.Font);
	    ChangeFontGC(self->vds.Font, self->vds.ForeColor, self->Color);
	    return True;
	}
	return False;
}

void Set_Parent(TWins *self, TWins *Parent) {
	if (self->SetParent) self->SetParent(self, Parent);
}

void SetCaption(TWins *self, const char *s) {
	if (self->SetCaption) self->SetCaption(self, s);
}

char *GetCaption_pc(TWins *self) {
	return self->vds.s;
}

void SetColor(TWins *self, dword col) {
	if (self->SetColor) self->SetColor(self, col);
}

void SetBorder(TWins *self, int BorderFu) {
	if (self->SetBorder) self->SetBorder(self, BorderFu);
}

void SetFont(TWins *self, const char *name) {
  TWinsCtl wctl;
	wctl.bef = wctlb_SetFont;
	wctl.flag = 0;
	wctl.ptr = (char*)name;
	WinsCtl(self, &wctl);
}

void SetEnabled(TWins *self, bool Value) {
	if (self->SetEnabled) self->SetEnabled(self, Value);
}

void SetVisible(TWins *self, bool Value) {
	if (self->SetVisible) self->SetVisible(self, Value);
}

void Set_Cursor(TWins *self, int Value) {
  XSetWindowAttributes XAttr;

	if (self->cursor_nr != Value) {
	    self->cursor_nr = Value;
	    Fill0(&XAttr, sizeof(XSetWindowAttributes));
	    XAttr.cursor = IappXDat.XCursors[Value];
	    XChangeWindowAttributes(IappXDat.dpy, self->WinID, CWCursor, &XAttr);
	}
}

void SetPosition(TWins *self, int left, int top) {
	if (self->SetRC) self->SetRC(self, left, top, self->width, self->height);
}
void SetLeft(TWins *self, int left) {
	SetPosition(self, left, self->top);
}
void SetTop(TWins *self, int top) {
	SetPosition(self, self->left, top);
}
void SetWinSize(TWins *self, int Width, int Height) {
	if (self->SetRC) self->SetRC(self, self->left, self->top, Width, Height);
}
void SetWidth(TWins *self, int Width) {
	SetWinSize(self, Width, self->height);
}
void SetHeight(TWins *self, int Height) {
	SetWinSize(self, self->width, Height);
}
void SetRC(TWins *self, int left, int top, int width, int height) {
	if (self->SetRC) self->SetRC(self, left, top, width, height);
}

void Paint(TWins *self) {
	if (self->Paint) self->Paint(self);
}

void RePaint(TWins *self) {
	if (self->RePaint) self->RePaint(self);
}

void Refresh(TWins *self) {
	if (self->Refresh) self->Refresh(self);
}

void Resize(TWins *self) {
	if (self->Resize) self->Resize(self);
}

void Get_ClientRect(TWins *self, TWRect *rc) {
	if (self->GetClientRect) self->GetClientRect(self, rc);
}

void GetAlignClientRect(TWins *self, TWRect *rc) {
	if (self->GetAlignClientRect) self->GetAlignClientRect(self, rc);
}

void AdjustSize(TWins *self) {
	if (self->AdjustSize) self->AdjustSize(self);
}

void *GetParentForm(TWins *sw) {
	if (sw && !(sw->apph->aos & aosIsForm)) sw = sw->ParentForm;
	if (sw && !(sw->apph->aos & aosIsForm)) sw = NULL;
	return sw;
}

bool TestWinIsEnabled(TWins *self) {
  TWins *tmpWin = self;
	if (!tmpWin) return False;
	while (tmpWin->Enabled && tmpWin->Parent && 
		!(tmpWin->apph->aos & aosIsForm)) tmpWin = tmpWin->Parent;
	return tmpWin->Enabled;
}

bool TestWinIsVisible(TWins *self) {
  TWins *tmpWin = self;
	if (!tmpWin) return False;
	while (tmpWin->Visible && tmpWin->Parent && 
	       !(tmpWin->apph->aos & aosIsForm)) tmpWin = tmpWin->Parent;
	return tmpWin->Visible;
}

bool TestWinIsAktiv(TWins *self) {
  TWins *tmpWin = self;
	if (!tmpWin) return False;
	while (tmpWin->Enabled && tmpWin->Visible && tmpWin->Parent && 
	       !(tmpWin->apph->aos & aosIsForm)) tmpWin = tmpWin->Parent;
	return (tmpWin->Enabled && tmpWin->Visible);
}

bool TestWinCanFocus(TWins *self) {
	return (TestWinIsAktiv(self) && (self->KeyFu & mCanFocus));
}

void TWins_SetBorderWidth(TWins *self, int BorderWidth) {
  int fu = self->BorderFu;
	if (!(fu & mBorderNoLeft))   self->BorderOffLeft = BorderWidth;
	if (!(fu & mBorderNoRigth))  self->BorderOffRigth = BorderWidth;
	if (!(fu & mBorderNoTop))    self->BorderOffTop = BorderWidth;
	if (!(fu & mBorderNoBottom)) self->BorderOffBottom = BorderWidth;
}

void TWins_SetDefaultBorderWidth(TWins *self) {
  int m;

	m = self->BorderOffLeft | self->BorderOffRigth | 
	    self->BorderOffTop | self->BorderOffBottom;
	if (m) return;

	if (self->BorderFu & mBorder) {
	    m = 1;
	    if (self->BorderFu & mBorderW2) m++;
	}
	TWins_SetBorderWidth(self, m);
}
/*____________________________ Destroy __________________________________*/
bool TWins__Destroy(TWins *self) {
  Window win;
  TForm *tmpW = GetParentForm(self);

  if (self->RunStatus & runst_GoDestroy) {
     printf("TWins__Destroy: win=0x%x wtyp=0x%x tmpW=0x%x\n", (int)self->WinID, self->wtyp, (int)tmpW);  
  }

	self->RunStatus |= runst_GoDestroy;
	DelAppWin(self->apph);
	if (IappXDat.GrabWin == self) Std_UngrabPointer();
	if (self->OnDestroy) self->OnDestroy(self);
	if (IappXDat.InfoWin == self) IappXDat.InfoWin = NULL;
	if (tmpW && (tmpW->FocusWin == self)) tmpW->FocusWin = NULL;
	win = self->WinID;
	self->vds.lg = 0;
	xfree(self->vds.s);
	TFont_Destroy(self->vds.Font);
	if (IappXDat.dpy) XDestroyWindow(IappXDat.dpy, win);
	Tfree(self->InfoWin);
#ifdef USE_DECOR
	TWins_Decor(self, decor_Destroy, 0, NULL);
#endif
	return False;
}

/*____________________________ Create ___________________________________*/
void TWins_SetParent(TWins *self, TWins *Parent) {
	self->IsParent = True;
	TWins_SetDefaultBorderWidth(self);
	AdjustSize(self);
	if (self->Visible) MapWinID(self->WinID, 2);
//	TWins_Decor(self, decor_SetParent, 0, NULL);
}

void TWins_CreateXWin(TWins_Create_struct *wcs) {
  TWins *self = wcs->wins;
	wcs->XAttr.background_pixel = GetColor_RGB(self->Color);
	wcs->XAttr.cursor = IappXDat.XCursors[self->cursor_nr];

	self->xwc.x = self->left;
	self->xwc.y = self->top;
	self->xwc.width = self->width;
	self->xwc.height = self->height;

	self->WinID = XCreateWindow (IappXDat.dpy, wcs->ParentWinID, 
			self->left, self->top, self->width, self->height,
			0, CopyFromParent, InputOutput,
			CopyFromParent, wcs->XMask, &wcs->XAttr);
	self->apph->win = self->WinID;
	self->vds.WinID = self->WinID;
#ifdef _OS_WIN32
	self->OS_var.background_pixmap = wcs->XAttr.background_pixmap;
	self->OS_var.OverrideRedirect = ((wcs->XMask & CWOverrideRedirect) && 
					  wcs->XAttr.override_redirect);
#endif
}

void TWins_DefaultConf(TWins_Create_struct *wcs) {
  TWins *self;
  TWins *Parent;
  Tapph *ap;

	self = wcs->wins;
	self->destroy = (TDESTROY)TWins__Destroy;

	Parent = wcs->Parent;
	if (Parent) ap = AddAppWin(Parent->apph, self); else ap = AddAppWin(NULL, self);
	ap->EvProc = TWins_EvProc;

	self->Parent = Parent;
	if (Parent) {
	    wcs->ParentWinID = Parent->WinID;
	    self->Color = Parent->Color;
	    self->vds.Font = Parent->vds.Font;
	    self->vds.UTF8on = Parent->vds.UTF8on;
	    self->subClass = Parent->subClass;
	} else {
	    wcs->ParentWinID = IappXDat.RootW;
	    self->Color = IappXDat.Default_Color;
	    self->vds.Font = IappXDat.Default_Font;
#ifdef _OS_WIN32
	    self->vds.UTF8on = True;
#endif
	}
	if (self->width <= 0) self->width = 32;
	if (self->height <= 0) self->height = 16;
	self->Align = alNone;
	self->Enabled = True;
	self->Visible = True;
	self->vds.Wins = self;

	self->SetParent = TWins_SetParent;
	self->SetCaption = TWins_SetCaption;
	self->SetColor = TWins_SetColor;
	self->SetBorder = TWins_SetBorder;
	self->SetEnabled = TWins_SetEnabled;
	self->SetVisible = TWins_SetVisible;


	self->SetRC = TWins_SetRC;
	self->Refresh = TWins_Refresh;
	self->WinsCtl = TWins_Ctl;

	self->GetClientRect = TWins_GetClientRect;
	self->GetAlignClientRect = TWins_GetAlignClientRect;
	self->AdjustSize = TWins_AdjustSize;
	self->Resize = TWins_Resize;
	self->Paint = TWins_Paint;
	self->RePaint = TWins_RePaint;

	self->DoPaint = TWins_DoPaint;
	self->DoClick = TWins_DoClick;
	self->DoMouseDown = TWins_DoMouseDown;
	self->DoMouseUp = TWins_DoMouseUp;
	self->DoMouseMove = TWins_DoMouseMove;
	self->DoMouseOver = TWins_DoMouseOver;
	self->DoKeyDown = TWins_DoKeyDown;
	self->DoKeyUp = TWins_DoKeyUp;
	self->DoFocus = TWins_DoFocus;
	self->DoActivate = TWins_DoActivate;

	self->cursor_nr = XC_top_left_arrow;
#ifdef DRAG_AND_DROP
//	self->DodndMouseMove = TWins_DodndMouseMove;
//	self->DodndDrop = TWins_DodndDrop;
//	self->DodndStatus = TWins_DodndStatus;
#endif

	wcs->XMask = CWBackPixel | CWEventMask | CWCursor; // or CWSaveUnder;

	wcs->XAttr.event_mask = ButtonPressMask | ButtonMotionMask |
                      ButtonReleaseMask | PointerMotionMask | EnterWindowMask |
                      LeaveWindowMask | ExposureMask | KeyPressMask | KeyReleaseMask |
                      StructureNotifyMask;

	if (wcs->WCSfkt & WCSfkt_OverrideRedirect) {
	    wcs->XMask |= CWOverrideRedirect;
	    wcs->XAttr.override_redirect = 1;
	}

	if (wcs->CreateParams) wcs->CreateParams(wcs);

	TWins_SetDefaultBorderWidth(self);
	
	if (wcs->CreateXWin) wcs->CreateXWin(wcs);
}

void TWins_Create_struct_init(TWins_Create_struct *wcs, int tcsize, int wtyp) {
	Fill0(wcs, sizeof(TWins_Create_struct));
	wcs->CreateXWin = TWins_CreateXWin;
	wcs->wins = ttc_create(tcsize, wtyp);
}

TWins *TWins_Create(TWins *Parent) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TWins), wtyp_Wins);
	wcs.Parent = Parent;
	TWins_DefaultConf(&wcs);
	return wcs.wins;
}


TWins *TWins_CreateB(TWins *Parent, int Left, int Top, int Width, int Height, int BorderFu) {
  TWins *self;
	self = TWins_Create(Parent);
	self->BorderFu = BorderFu;
	SelfCreateS(self, Left, Top, Width, Height, NULL);
	return self;
}

TWins *TWins_CreateBA(TWins *Parent, int Left, int Top, int Width, int Height, int BorderFu, int Align) {
  TWins *self;
	self = TWins_Create(Parent);
	self->BorderFu = BorderFu;
	self->Align = Align;
	SelfCreateS(self, Left, Top, Width, Height, NULL);
	return self;
}

void REGP3 SelfCreateS(TWins *self, int Left, int Top, int Width, int Height, const char *s) {
	self->top = Top;
	self->left = Left;
	self->width = Width;
	self->height = Height;
	if (s) {
	    self->CaptionFu |= mCaptionPaint;
	    SetCaption(self, s);
	}
	Set_Parent(self, self->Parent);
}
