/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

TPage *GetParentPage(TWins* ws) {
	while (ws) {
	    if (ws->wtyp == wtyp_TPage) break;
	    ws = ws->Parent;
	}
	return (TPage*)ws;
}

void TPage_SetCaption(TWins *self, const char *s) {
  TPageControl *pagc;

	if (pccmp(self->vds.s, s)) return;
	vds_SetStr(&self->vds, s);
	if (self->OnSetCaption) self->OnSetCaption(self, s);
	pagc = ((TPage*)self)->PageControl;
	if (pagc && pagc->ResizeStartMk) {
	    pagc->ResizeStartMk = False;
	    Refresh((TWins*)pagc);
	}
}

void TPage_SetPageAktiv(TPage *self) {
  int fu = 0;
  TPage *pag;
  TPageControl *pagc;

	if (!self || !self->IsParent) return;
	pagc = self->PageControl;
	if (pagc && (pagc->AktPage != self)) {
	    if (pagc->OnChangeTab) pagc->OnChangeTab(pagc, pagc->AktPage, self, &fu);
	    if ((fu & pgNotChange) == 0) {
		pagc->AktPage = self;
		pag = pagc->Page0;
		while (pag) {
		    if (pag != self) TWins_SetVisible((TWins*)pag, False);
		    pag = pag->next;
		}
		TWins_SetVisible((TWins*)self, True);
		pagc->ResizeStartMk = False;
		Refresh((TWins*)pagc);
		if (self->FocusWin) Set_Focus(self->FocusWin);
		if (pagc->OnActivate_pg) pagc->OnActivate_pg(pagc, pagc->AktPage);
	    }
	}
}

void TPage_SetVisible(TPage *self, bool Value) {
// printf("%s(%d): 0x%x\n", FL, self->PageControl->RunStatus);
	if (self->PageVisible != Value) {
	    self->PageVisible = Value;
	    if (self->PageControl->RunStatus & runst_Paint1) {
		Refresh((TWins*)self->PageControl);
	    }
	}
}

/* ____________________ TPageControl ____________________________________ */
void TPageControl_GetClientRect(TWins *self, TWRect *rc) {
  TPageControl *pagc = (TPageControl*)self;

	rc->left = 2;
	rc->top = pagc->TabsGesHeight + 2;
	rc->width = pagc->width - 4;
	rc->height = pagc->height - pagc->TabsGesHeight - 4;
	if (rc->height < 4) rc->height = 4;
}

void TPageControl_SetPagesSize(TPageControl *pagc) {
  int x, xa, xe, w, mt0, h;
  TPage *TmpPage;

	for (x=0; x<16; x++) pagc->TabMaxX[x] = -1;

	TmpPage = pagc->Page0;
	pagc->xv1 = 0;
	pagc->xv2 = 0;
	pagc->Ynr = 0;
	x = 2;
	mt0 = 0;
	while (TmpPage) {
	    if (TmpPage->PageVisible) {
		TmpPage->vds.WinID = pagc->WinID;
		w = 0;
		if (TmpPage->vds.s) {
		    if (*TmpPage->vds.s == ' ') {
			*TmpPage->vds.s = '*';
			w = TextWidthColorString(&TmpPage->vds);
			*TmpPage->vds.s = ' ';
		    } else {
			w = TextWidthColorString(&TmpPage->vds);
		    }
		}
		TmpPage->Wtxt = w;
		w = w + pagc->TabTxtOff;
		if (w < pagc->MinTabWidth) w = pagc->MinTabWidth;
		if (w < 16) w = 16;

		if ((mt0 > 0) && ((x + w + 4) > pagc->width)) {
		    pagc->Ynr++;
		    x = 2;
		    mt0 = 0;
		} else mt0++;

		xa = x;
		xe = x + w;
		if (TmpPage->Visible) {
		    xa = x - 2;
		    xe = xe + 2;
		    pagc->AktYnr = pagc->Ynr;
		    pagc->xv1 = xa;
		    pagc->xv2 = xe+1;
		}

		TmpPage->TabXa = xa;
		TmpPage->TabXe = xe;
		TmpPage->TabYnr = pagc->Ynr;
		x = x + w + 1;
		if (pagc->Ynr < 16) pagc->TabMaxX[pagc->Ynr] = x;
	    } //if PageVisible
	    TmpPage = TmpPage->next;
	}

	h = pagc->TabsHeight * (pagc->Ynr + 1) - pagc->Ynr * 2;
	if (pagc->Ynr > 0) h = h + 2;

	if (pagc->TabsGesHeight != h) {
	    pagc->TabsGesHeight = h;
	}
}

void TPageControl_Resize(TPageControl *self) {
  TPage *TmpPage;
  TWRect rc;

	if (self->IsParent) {
	    TWins_AdjustSize((TWins*)self);
	    if (!self->ResizeStartMk && self->OnResize && self->Visible) {
		self->OnResize((TWins*)self);
		self->ResizeStartMk = True;
	    }
	    TPageControl_SetPagesSize(self);
	    Get_ClientRect((TWins*)self, &rc);
	    TmpPage = self->Page0;
	    while (TmpPage) {
		TmpPage->left = rc.left;
		TmpPage->top = rc.top;
		TmpPage->width = rc.width;
		TmpPage->height = rc.height;
		Resize((TWins*)TmpPage);
		TmpPage = TmpPage->next;
	    }
	}
	if (self->DoActivate) self->DoActivate((TWins*)self);
}

void TPageControl_PaintTabs(TPageControl *pagc, int y0, int Ynr) {
  int x, x2, y, y2, xt;
  TPage *TmpPage;

	TmpPage = pagc->Page0;
	y2 = pagc->TabsGesHeight;
	if (Ynr < 16) Fill_Rect(pagc->WinID, IappXDat.Default_GC, 0, y0+3, pagc->TabMaxX[Ynr], y2-y0-3);
	x2 = 0;
	while (TmpPage) {
	    if (TmpPage->PageVisible && (TmpPage->TabYnr == Ynr)) {
		x = TmpPage->TabXa;
		x2 = TmpPage->TabXe;
		y = y0;
		if (TmpPage->Visible) {
		    Fill_Rect(pagc->WinID, IappXDat.Default_GC, x, y, x2-x, y2-y+1);
		} else {
		    y += 2;
		}
		TmpPage->TabYo = y;
		TmpPage->TabYu = y0 + pagc->TabsHeight;

		xt = ((x2 - x) - TmpPage->Wtxt) >> 1;
		TmpPage->vds.Xoff = x+xt;
		TmpPage->vds.Yoff = y+2;
		TmpPage->vds.WinID = pagc->WinID;
		TWins_PaintEnCaption((TWins*)TmpPage, TestWinIsEnabled((TWins*)TmpPage));

		if (TmpPage->prev && TmpPage->prev->Visible) {
		} else {
		    DrawLine(pagc->WinID, IappXDat.GC_FFFFFF, x, y+2, x, y2);
		    DrawLine(pagc->WinID, IappXDat.GC_FFFFFF, x, y+2, x+2, y);
		}

		DrawLine(pagc->WinID, IappXDat.GC_FFFFFF, x+2, y, x2-2, y);
		DrawLine(pagc->WinID, IappXDat.bordL_GC, x2-2, y, x2, y+2);
		DrawLine(pagc->WinID, IappXDat.GC_000000, x2, y+2, x2, y2);

	    } //if FPageVisible 
	    TmpPage = TmpPage->next;
	} 
}

void TPageControl_DrawFocus(TPageControl *pagc) {
  TPage *Page = pagc->AktPage;
  if (Page) DrawRectFocus(pagc->WinID, Page->TabXa+2, Page->TabYo+2,
                                       Page->TabXe-Page->TabXa-4,
                                       Page->TabYu-Page->TabYo-4);
}

void TPageControl_DoPaint(TWins *self) {
  TPageControl *pagc = (TPageControl*)self;
  int i, y, w, h;

	Fill_Rect(pagc->WinID, IappXDat.Default_GC, 0, 0, pagc->width, pagc->TabsGesHeight+2);

	y = 0;
	for (i=pagc->Ynr; i>=0 ; i--) {
	    if (pagc->AktYnr != i) {
		TPageControl_PaintTabs(pagc, y, i);
		y = y + pagc->TabsHeight - 2;
	    }
	}
	if (pagc->Ynr > 0) y = y + 2;
	TPageControl_PaintTabs(pagc, y, pagc->AktYnr);

	if (pagc->AktPage && (pagc->KeyFu & mHasFocus)) {
	    TPageControl_DrawFocus(pagc);
	}
	w = pagc->width-1;
	h = pagc->height-1;
	y = pagc->TabsGesHeight;

	if (pagc->xv1 != pagc->xv2) DrawLine(pagc->WinID, IappXDat.GC_FFFFFF, 0, y, pagc->xv1, y);
	DrawLine(pagc->WinID, IappXDat.GC_FFFFFF, pagc->xv2, y, w, y);

	DrawLine(pagc->WinID, IappXDat.GC_FFFFFF, 0, y, 0, h);
	DrawLine(pagc->WinID, IappXDat.GC_404040, 0, h, w, h);
	DrawLine(pagc->WinID, IappXDat.GC_404040, w, y, w, h);
}

/******************************************************************************
 * insto_pag = NULL; then ins to pagc->Page0
 ******************************************************************************/
void TPageControl_InsPageToList(TPageControl *pagc, TPage *pag, TPage *insto_pag) {
	if (!pagc->Page0) {
	    pagc->Page0 = pag;
	    pag->prev = NULL;
	    pag->next = NULL;
	} else if (!insto_pag) {
	    insto_pag = pagc->Page0;
	    insto_pag->prev = pag;
	    pag->next = insto_pag;
	    pag->prev = NULL;
	    pagc->Page0 = pag;
	} else {
	    pag->prev = insto_pag;
	    pag->next = insto_pag->next;
	    insto_pag->next = pag;
	    if (pag->next) pag->next->prev = pag;
	}
}

TPage *TPageControl_AddNewUserPage(TPageControl *pagc, char *Caption, struct TPage* (*DoTPageCreate)(struct TPageControl *Parent)) {

  TPage *pag, *TmpPage;
	pag = NULL;
	if (DoTPageCreate) pag = DoTPageCreate(pagc);
	if (!pag) return NULL;
	pag->SetVisible = (TSetVisibleEv)TPage_SetVisible;
	SetCaption((TWins*)pag, Caption);

	if (!pagc->Page0) {
	    pagc->Page0 = pag;
	    pagc->AktPage = pag;
	    pag->Visible = True;
	} else {
	    TmpPage = pagc->Page0;
	    while (TmpPage->next) TmpPage = TmpPage->next;
	    TmpPage->next = pag;
	    pag->prev = TmpPage;
	}
	Set_Parent((TWins*)pag, (TWins*)pagc);
	if (pagc->RunStatus & runst_Paint1) Refresh((TWins*)pagc);
	return pag;
}

bool TPageControl_ClrPageFromList(TPageControl *pagc, TPage *pag) {
  TPage **ppag;
  bool res = False;

	if (pagc && pag && pagc->Page0) {
	    if (pagc->AktPage == pag) pagc->AktPage = NULL;
	    ppag = &pagc->Page0;
	    while (*ppag && (*ppag != pag)) ppag = &(*ppag)->next;
	    if (*ppag) {
		*ppag = pag->next;
		if (pag->prev) pag->prev->next = pag->next;
		if (pag->next) pag->next->prev = pag->prev;
		res = True;
	    }
	}
	return res;
}

void TPageControl_FreePage(TPageControl *pagc, TPage *pag) {
  TPage *npag = NULL;
	if (TPageControl_ClrPageFromList(pagc, pag)) {
	    if (pag->prev) {
		npag = pag->prev;
		while (npag && !npag->PageVisible) npag = npag->prev;
	    }
	    if (!npag && pag->next) {
		npag = pag->next;
		while (npag && !npag->PageVisible) npag = npag->next;
	    }

	    Tfree(pag);
	    pagc->ResizeStartMk = False;
	    Refresh((TWins*)pagc);
	    TPage_SetPageAktiv(npag);
	}
}

void TPageControl_DoMouseDown(TWins *self, XButtonEvent *ev) {
  TPageControl *pagc = (TPageControl*)self;
  TPage  *TmpP;

	TWins_DoMouseDown(self, ev);
	if (ev->button != 1) return;

	TmpP = pagc->Page0;
	while (TmpP) {
	    if (TmpP->PageVisible && TestWinIsEnabled((TWins*)TmpP)) {
		if ((ev->y >= TmpP->TabYo) && (ev->y <= TmpP->TabYu)) {
		    if ((ev->x >= TmpP->TabXa) && (ev->x <= TmpP->TabXe)) {
		    TPage_SetPageAktiv(TmpP);
		    break;
		    }
		}
	    }
	    TmpP = TmpP->next;
	}
}

int TPageControl_DoKeyDown(TWins *self, TKeyEvent *KeyEv) {
  TPageControl *pagc = (TPageControl*)self;
  TPage  *TmpP;

	if (self->OnKeyDown) self->OnKeyDown(self, KeyEv);

	if (pagc->AktPage) {
	    if ((KeyEv->keysym == XK_Right) && (self->KeyFu & mK_Right)) {
		TmpP = pagc->AktPage->next;
		while (TmpP && !TmpP->PageVisible) TmpP = TmpP->next;
		if (!TmpP) {
		  TmpP = pagc->Page0;
		  while (TmpP && !TmpP->PageVisible) TmpP = TmpP->next;
		}
		TPage_SetPageAktiv(TmpP);
		return 1;
	    } else if ((KeyEv->keysym == XK_Left) && (self->KeyFu & mK_Left)) {
		TmpP = pagc->AktPage;
		if (TmpP == pagc->Page0) {
		  while (TmpP->next) TmpP = TmpP->next;
		} else if (TmpP->prev) TmpP = TmpP->prev;
		while (TmpP && !TmpP->PageVisible) TmpP = TmpP->prev;
		TPage_SetPageAktiv(TmpP);
		return 1;
	    } 
	}
	return 0;
}

void TPageControl_DoFocus(TWins *self, int Change) {
	TWins_DoFocus(self, Change);
	TPageControl_DrawFocus((TPageControl*)self);
}

/*________________________ Create TPage __________________________________*/
void TPage_CreateParams(TWins_Create_struct *wcs) {
  TPage *self = wcs->wins;
	self->Visible = False;
	self->PageVisible = True;
	self->SetCaption = TPage_SetCaption;
	self->PageControl = (TPageControl*)wcs->Parent;
}

void *TPage_CreateParamAlg(TPageControl *Parent, int PageMemSize) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, PageMemSize, wtyp_TPage);
	wcs.Parent = (TWins*)Parent;
	wcs.CreateParams = TPage_CreateParams;
	TWins_DefaultConf(&wcs);
	return wcs.wins;
}

TPage *TPage_Create(TPageControl *Parent) {
	return TPage_CreateParamAlg(Parent, sizeof(TPage));
}

TPage *TPageControl_AddNewPage(TPageControl *pagc, char *Caption) {
	return TPageControl_AddNewUserPage(pagc, Caption, TPage_Create);
}

/*___________________ Create TPageControl ______________________________*/
void TPageControl_CreateParams(TWins_Create_struct *wcs) {
  TPageControl *self = wcs->wins;
	self->width = 240;
	self->height = 160;
	self->TabsHeight = 18;
	self->MinTabWidth = 20;
	self->TabTxtOff = 6;

	self->DoPaint = TPageControl_DoPaint;
	self->GetClientRect = TPageControl_GetClientRect;
//	self->AdjustSize = TPageControl_AdjustSize;
	self->Resize = (TWinsEv)TPageControl_Resize;
	self->DoMouseDown = TPageControl_DoMouseDown;
	self->DoKeyDown = TPageControl_DoKeyDown;
	self->DoFocus = TPageControl_DoFocus;
}

TPageControl *TPageControl_Create(TWins *Parent) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TPageControl), wtyp_TPageControl);
	wcs.Parent = Parent;
	wcs.CreateParams = TPageControl_CreateParams;
	TWins_DefaultConf(&wcs);
	return wcs.wins;
}
