/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "cicons.h"

//*********************** TScrollBar ************************
void TScrollBar_AdjustSize(TWins *self);
void TScrollBar_DoPaint(TWins *self);
void TScrollBar_DoMouseDown(TWins *self, XButtonEvent *ev);
void TScrollBar_DoMouseUp(TWins *self, XButtonEvent *ev);
void TScrollBar_DoMouseMove(TWins *self, XMotionEvent *ev);
static void TScrollBar_scPosSet(TScrollBar *sb);
void TScrollBar_Refresh(TWins *self);


static void sb_GetClientRect(TScrollBar *self, TWRect *rc) {
  int m;
	TWins_GetClientRect((TWins*)self, rc);
	if (!self->Hsc) {
	    m = rc->left; rc->left = rc->top; rc->top = m;
	    m = rc->width; rc->width = rc->height; rc->height = m;
	}
}

static void sb_BorderPaint(TScrollBar *self, int fu, int x, int y, int w, int h) {
	if (self->Hsc) {
	    StdBorderPaint(self->WinID, fu,  x, y, w, h);
	} else {
	    StdBorderPaint(self->WinID, fu, y, x, h, w);
	}
}

static void sb_Fill_Rect(TScrollBar *self, GC gc, int x, int y, int w, int h) {
	if (w <= 0) return;
	if (self->Hsc) {
	   Fill_Rect(self->WinID, gc, x, y, w, h);
	} else {
	   Fill_Rect(self->WinID, gc, y, x, h, w);
	}
}

void TScrollBar_AdjustSize(TWins *self) {
  TScrollBar *sb = (TScrollBar*)self;
  int l;
  TWRect rc;
  int w = sb->width;
  int h = sb->height;
  bool rsz = (w != sb->xwc.width) || (h != sb->xwc.height);

	TWins_AdjustSize((TWins*)sb); //+Chk Align
	if (!rsz) rsz = (w != sb->width) || (h != sb->height);
	sb_GetClientRect(sb, &rc);

	l = 15;
	if (rc.width < 48) {
	    l = (rc.width-3) / 3;
	    if (l < 7) l = 7;
	}
	if (rsz || (sb->bulg != l)) {
	    sb->bulg = l;
	    TScrollBar_scPosSet(sb);
	}
}

void TScrollBar_ButtonPaint(TScrollBar *sb, bool pos, bool down) {
  TWRect rc;
  int ix, iy;
  int fu = mBorder | mBorderW2 | mBorderBu;
  TPic *pic;

	if (down) fu = mBorder | mBorderW2 | mBorderBu | mBorderL;

	if (sb->Hsc) {
	    if (pos) {
		pic = get_pic9_hsb_plus(False);
	    } else {
		pic = get_pic9_hsb_minus(False);
	    }
	} else {
	    if (pos) {
		pic = get_pic9_vsb_plus(False);
	    } else {
		pic = get_pic9_vsb_minus(False);
	    }
	}

	sb_GetClientRect(sb, &rc);
	if (pos) rc.left += rc.width - sb->bulg;

	iy = rc.top + ((rc.height-9) >> 1);
	ix = rc.left + ((sb->bulg-9) >> 1);

	DrawPic(sb->WinID, pic, sb->Hsc ? ix : iy, sb->Hsc ? iy : ix);
	sb_BorderPaint(sb, fu, rc.left, rc.top, sb->bulg, rc.height);
}

void TScrollBar_DoPaint(TWins *self) {
  TScrollBar *sb = (TScrollBar*)self;
  TWRect rc;
  int l1;

	TScrollBar_ButtonPaint(sb, False, (sb->MouseDownSeg == 1));

	sb_GetClientRect(sb, &rc);
	if (sb->scPaint) {
	    l1 = sb->scPos - sb->bulg - rc.left;
	    sb_Fill_Rect(sb, IappXDat.SB_GC, sb->bulg, rc.top, l1, rc.height);
	    l1 = rc.left + rc.width - sb->scPos - sb->scLg - sb->bulg;
	    sb_Fill_Rect(sb, IappXDat.SB_GC, sb->scPos+sb->scLg, rc.top, l1, rc.height);
	    sb_Fill_Rect(sb, IappXDat.Default_GC, sb->scPos, rc.top, sb->scLg, rc.height);
	    sb_BorderPaint(sb, mBorder | mBorderW2 | mBorderBu, sb->scPos, rc.top, sb->scLg, rc.height);
	} else {
	    rc.left += sb->bulg;
	    rc.width -= sb->bulg + sb->bulg;
	    sb_Fill_Rect(sb, IappXDat.SB_GC, rc.left, rc.top, rc.width, rc.height);
	}

	TScrollBar_ButtonPaint(sb, True, (sb->MouseDownSeg == 5));
}


static void SetCount(TScrollBar *sb, int Count) {
	if (Count > (sb->MaxCount - sb->ButtonCountSize)) Count = sb->MaxCount - sb->ButtonCountSize;
	if (Count < 0) Count = 0;
	TScrollBar_DoCount(sb, Count);
}

void TScrollBar_OnTimer(TgTimer *ti) {
  TScrollBar *sw = (TScrollBar*)ti->obj;
  int m;
	if (sw->RepTimer) sw->RepTimer->Interval = 50;
	m = sw->Count;
	switch (sw->MouseDownSeg) {
	case 1: m -= sw->SmallChange; break;
	case 2: m -= sw->LargeChange; break;
	case 4: m += sw->LargeChange; break;
	case 5: m += sw->SmallChange; break;
	default: return;
	}
	SetCount(sw, m);
}

static int TScrollBar_MouseSeg(TScrollBar *self, int evx) {
  TWRect rc;

	sb_GetClientRect(self, &rc);
	if (evx <= self->bulg) return 1;
	if (evx >= (rc.width - self->bulg)) return 5;
	if (!self->scPaint) return 0;
	if (evx < self->scPos) return 2;
	if (evx > (self->scPos + self->scLg)) return 4;
	return 3;
}

void TScrollBar_DoMouseDown(TWins *self, XButtonEvent *ev) {
  int count, evx;
  TScrollBar *sb = (TScrollBar*)self;
  bool ref;
	if ((ev->button == 1) && TestWinIsEnabled(self)) {

	    if (sb->Hsc) evx = ev->x; else evx = ev->y;
	    sb->MouseDownSeg = TScrollBar_MouseSeg(sb, evx);
	    ref = True;
	    count = sb->Count;
	    switch (sb->MouseDownSeg) {
	    case 1: count -= sb->SmallChange; break;
	    case 2: count -= sb->LargeChange; break;
	    case 3:
		sb->MouseDownPos = evx;
		sb->MDscPos = sb->scPos;
		ref = False;
		break;
	    case 4: count += sb->LargeChange; break;
	    case 5: count += sb->SmallChange; break;
	    default: ref = False;
	    }

	    if (ref) {
		if (!sb->RepTimer) {
		    sb->RepTimer = gTimerX_Add(Timer_Enabled, 300, TScrollBar_OnTimer);
		    sb->RepTimer->obj = self;
		}
		SetCount(sb, count);
		TWins_Refresh(self);
	    }
	}
	TWins_DoMouseDown(self, ev);
}

void TScrollBar_DoMouseUp(TWins *self, XButtonEvent *ev) {
  TScrollBar *sw = (TScrollBar*)self;

	sw->MouseDownSeg = 0;
	if (sw->RepTimer) {
	    Tfree(sw->RepTimer);
	    sw->RepTimer = NULL;
	}
#ifdef _OS_WIN32
	Std_UngrabPointer();
#endif /* ! _OS_WIN32 */
	TWins_DoMouseUp(self, ev);

	if ((ev->button == 1) && TestWinIsEnabled(self)) {
	    TWins_Refresh(self);
	}
}

static void TScrollBar_scPosSet(TScrollBar *self) {
  TWRect rc;
  int l;

	sb_GetClientRect(self, &rc);
	self->scPaint = ((self->MaxCount > 0) && (self->MaxCount > self->ButtonCountSize));
	if (self->scPaint) {
	    rc.width -= self->bulg * 2;
	    if (self->ButtonCountSize > 0) {
		l = rc.width * self->ButtonCountSize / self->MaxCount;
		if (l < 10) l = 10;
	    } else l = self->bulg-1;
	    self->scLg = l;
	    rc.width -= l;
	    rc.left += self->bulg;
	    self->scPos = (rc.width * self->Count / (self->MaxCount - self->ButtonCountSize)) + rc.left;
	}
	TWins_Refresh((TWins*)self);
}

void TScrollBar_DoMouseMove(TWins *self, XMotionEvent *ev) {
  TScrollBar *sb = (TScrollBar*)self;
  TWRect rc;
  int m, evx;

	sb_GetClientRect(sb, &rc);
	if ((ev->state & Button1MotionMask) && (sb->MouseDownSeg == 3)) {
	    
	    if (sb->Hsc) evx = ev->x; else evx = ev->y;
	    m = evx - sb->MouseDownPos + sb->MDscPos;
	    rc.width -= sb->scLg;
	    if (m > (rc.width - sb->bulg)) m = rc.width - sb->bulg;
	    if (m < sb->bulg) m = sb->bulg;
	    sb->scPos = m + rc.left;
	    m -= sb->bulg;
	    rc.width -= sb->bulg << 1;
	    if (rc.width > 0) {
		m = m * (sb->MaxCount - sb->ButtonCountSize) / rc.width;
		SetCount(sb, m);
	    }
	}
	TWins_DoMouseMove((TWins*)sb, ev);
}

void TScrollBar_DoCount(TScrollBar *sb, int Count) {
	if (sb->Count != Count) {
	    sb->Count = Count;
	    if (sb->OnCount) sb->OnCount(sb, TScrollBar_GetCount(sb));
	    TScrollBar_scPosSet(sb);
	}
}

static void TScrollBar_ChkCount(TScrollBar *sb, int *Count) {
  int res = *Count;
  int max = sb->MaxCount - sb->ButtonCountSize;

	res -= sb->OffCount;
	if (sb->sbfu & sbfu_inv) res = max - res;
	if (res > max) res = max;
	if (res < 0) res = 0;
	*Count = res;
}

void TScrollBar_SetCount(TScrollBar *sb, int Count) {
	TScrollBar_ChkCount(sb, &Count);
	if (sb->Count != Count) {
	    sb->Count = Count;
	    TScrollBar_scPosSet(sb);
	}
}

void TScrollBar_SetCount_ev(TScrollBar *sb, int Count) {
	TScrollBar_ChkCount(sb, &Count);
	TScrollBar_DoCount(sb, Count);
}

void TScrollBar_SetButtonCountSize(TScrollBar *sb, int ButtonCountSize) {
	if (sb->ButtonCountSize != ButtonCountSize) {
	    sb->ButtonCountSize = ButtonCountSize;
	    TScrollBar_scPosSet(sb);
	}
}

void TScrollBar_SetMaxCount(TScrollBar *sb, int MaxCount) {
	if (sb->MaxCount != MaxCount) {
	    sb->MaxCount = MaxCount;
	    TScrollBar_scPosSet(sb);
	}
}

void TScrollBar_SetMaxCount_ButtonSize(TScrollBar *sb, int MaxCount, int ButtonCountSize) {
	sb->MaxCount = MaxCount;
	sb->ButtonCountSize = ButtonCountSize;
	TScrollBar_Refresh((TWins*)sb);
}

void TScrollBar_Refresh(TWins *self) {
  TScrollBar *sb = (TScrollBar*)self;

	if (sb->ButtonCountSize > sb->MaxCount) sb->ButtonCountSize = sb->MaxCount;
	if ((sb->ButtonCountSize + sb->Count) > sb->MaxCount) {
	    SetCount(sb, sb->MaxCount - sb->ButtonCountSize);
	}
	TScrollBar_scPosSet(sb); //+TWins_Refresh(self);
}

int TScrollBar_GetCount(TScrollBar *sb) {
  int res = sb->Count;
  int max = sb->MaxCount - sb->ButtonCountSize;
	if (res > max) res = max;
	if (res < 0) res = 0;
	if (sb->sbfu & sbfu_inv) res = max - res;
	res += sb->OffCount;
	return res;
}

void TScrollBar_SetCountParam(TScrollBar *sb, int Count, int buCount, 
			      int MaxCount, int OffCount, int sbfu) {
	sb->ButtonCountSize = buCount;
	sb->MaxCount = MaxCount;
	sb->OffCount = OffCount;
	sb->sbfu = sbfu;
	TScrollBar_SetCount(sb, Count);
}

/*_____________________________ Create __________________________________*/
void TScrollBar_CreateParams(TWins_Create_struct *wcs) {
  TScrollBar *self = wcs->wins;

	self->AdjustSize = TScrollBar_AdjustSize;

	self->Refresh = TScrollBar_Refresh;

	self->DoPaint = TScrollBar_DoPaint;
	self->DoMouseDown = TScrollBar_DoMouseDown;
	self->DoMouseUp = TScrollBar_DoMouseUp;
	self->DoMouseMove = TScrollBar_DoMouseMove;

	self->MaxCount = 100;
	self->SmallChange = 1;
	self->LargeChange = 10;
	self->bulg = 15;
}

TScrollBar *TScrollBar_Create(TWins *Parent) {
  TWins_Create_struct wcs;
	TWins_Create_struct_init(&wcs, sizeof(TScrollBar), wtyp_ScrollBar);
	wcs.Parent = Parent;
	wcs.CreateParams = TScrollBar_CreateParams;
	TWins_DefaultConf(&wcs);
	TScrollBar_scPosSet((TScrollBar*)wcs.wins);
	return wcs.wins;
}


TScrollBar *TScrollBarHorz_CreateP(TWins *Parent) {
  TScrollBar *sb = TScrollBar_Create(Parent);
	sb->Hsc = True;
	sb->width = 60;
	sb->height = 15;
	Set_Parent((TWins*)sb, Parent);
	return sb;
}

TScrollBar *TScrollBarVert_CreateP(TWins *Parent) {
  TScrollBar *sb = TScrollBar_Create(Parent);
	sb->width = 15;
	sb->height = 60;
	Set_Parent((TWins*)sb, Parent);
	return sb;
}
