/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "highl.h"

//*********************** highlight - bash ******************************
void highlight_Bash(Thighlight *hl, TLines *tls, int edAnf, int edEnd, int fu);

const char *sDat_Bash = 
"bg cd do fc fg fi if in for let pwd set \
bind case dirs done echo elif else esac exit eval exec hash jobs kill popd read test then time trap true type \
alias break false local pushd shift shopt times umask unset until while \
disown enable export logout printf return select source ulimit \
command declare getopts suspend typeset unalias \
continue function ";

static const char C_zBash[256] = {
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x22,0x23 ,0x00,0x00,0x00,0x27 ,0x00,0x00,0x00,0x00 ,0x00,0x01,0x00,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x00,0x00 ,0x00,0x00,0x00,0x00,

  0x00,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x00 ,0x5c,0x00,0x00,0x01,
  0x60,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01,
  0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x01 ,0x01,0x01,0x01,0x00 ,0x00,0x00,0x00,0x00,

  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,

  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00,
  0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00 ,0x00,0x00,0x00,0x00};

Thighlight *Thighlight_Bash_Create(TWins *wins, TLines *tls) {
  Thighlight *hl;

	hl = Thighlight_Create(wins, tls);
	if (!hl) return NULL;

	hl_Set_keyword(hl, &hl->fc0_Dat, sDat_Bash, FONT_Bold, 0x000000, hl->Font->g.BackColor);
//	hl_Set_keyword(&hl->fc1_Dat, sDat_C2, hl->Font, 0x800000, hl->Font->BackColor);

	hl->ChangeLine = highlight_Bash;
//	hl->SetCaret = highlight_SetCaret_C;
	return hl;
}

bool Get_Bash_STok(Thldat *hid, Tstoken *t) {
  unsigned char *s = t->s;
  int i;
  char m;
	if (!s || !*s) return False;
	while (*s && !C_zBash[*s]) s++;
	t->s = s;
	if (!*s) return False;
	i = 0;
	m = C_zBash[*s];
	if (*s) {s++; i++;}
	if (m != '"') while (*s && (m == (C_zBash[*s]))) {s++; i++;}
	t->next = s;
	t->lg = i;
	t->c = m;
	return True;
}

void hlSetLine_Bash(Thldat *hid) {
  Tstoken t;
  int i, m;

	xfree(hid->cs->c);
	hid->cs->c = NULL;
	hid->t = &t;
	t.s = (byte*)hid->cs->s;
	t.next = (byte*)hid->cs->s;
	if (hid->mk & hl_strline) {
	    add_scd_str(hid, &t);
	    t.s = t.next;
	} else if (hid->mk & hl_defline) {
	    add_scd_str_fc(hid, &t, hl_defline, '`', &hid->hl->fc_def);
	    t.s = t.next;
	}

	while (Get_Bash_STok(hid, &t)) {
	    if (t.c == 0x01) {
		m = 0;
		if ((t.s[0] == '0') && (t.lg > 2) && (t.s[1] == 'x') && (zC_HEX[t.s[2]])) {
		    m = 1;
		    for (i=3; i<t.lg; i++) {
			if (!zC_HEX[t.s[i]]) {m = 0; break;}
		    }
		    if (m) add_scd_nr(hid, (char*)t.s, t.lg);
		} else if (zC_DIGIT[t.s[0]]) {
		    m = 1;
		    for (i=1; i<t.lg; i++) {
			if (!zC_DIGIT[t.s[i]]) {m = 0; break;}
		    }
		    if (m) add_scd_nr(hid, (char*)t.s, t.lg);
		}
		if (!m) {
		    if (t.lg > 1) {
			    hl_keyword_c0(hid, hid->hl->fc0_Dat);
		    }
		}

	    } else if (t.c == '"') {
		add_scd_str(hid, &t);

	    } else if (t.c == '`') {
		add_scd_str_fc(hid, &t, hl_defline, '`', &hid->hl->fc_def);

	    } else if (t.c == '#') {
		t.lg = pclen((char*)t.s);
		add_scd_comment(hid, (char*)t.s, t.lg);
		break;
	    }
	    t.s = t.next;
	}
}

void highlight_Bash(Thighlight *hl, TLines *tls, int edAnf, int edEnd, int fu) {
  int i;
  Thldat hld;

	hld.hl = hl;
	hld.mk = 0;

	if (fu & clfu_use_yaye) {
	    if (edAnf > 0) {
		hld.cs = tls->lines->cs[edAnf-1];
		hld.mk = hld.cs->hl_mk;
	    }
	    if (edEnd >= tls->Count) edEnd = tls->Count-1;
	    for (i=edAnf; i<=edEnd; i++) {
		hld.cs = tls->lines->cs[i];
		hlSetLine_Bash(&hld);
		if (hld.cs->c) hld.cs->hl_mk = hld.mk; else hld.cs->hl_mk = 0;
	    }

	    for (; i<tls->Count; i++) {
		hld.cs = tls->lines->cs[i];
		if (hld.mk == hld.cs->hl_mk) break;
		hld.cs = tls->lines->cs[i];
		hlSetLine_Bash(&hld);
		if (hld.cs->c) hld.cs->hl_mk = hld.mk; else hld.cs->hl_mk = 0;
	    }
	    if (i < tls->Count) {
		hld.cs = tls->lines->cs[i];
		hlSetLine_Bash(&hld);
		if (hld.cs->c) hld.cs->hl_mk = hld.mk; else hld.cs->hl_mk = 0;
	    }
	} else {
	    for (i=0; i<tls->Count; i++) {
		hld.cs = tls->lines->cs[i];
		hlSetLine_Bash(&hld);
		if (hld.cs->c) hld.cs->hl_mk = hld.mk; else hld.cs->hl_mk = 0;
	    }
	}
}
