/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"

//*********************** TLineEdit ************************
TPlugIn_KeyDown *TLineEdit_keyplugin_Global = NULL;

static void set_SB(TScrollBar *SB, int max, int lew) {
	if (SB->Visible && ((SB->MaxCount != max) || 
			    (SB->ButtonCountSize != lew))) {
	    SB->MaxCount = max;
	    SB->ButtonCountSize = lew;
	    SB->LargeChange = lew - 1;
	    Refresh((TWins*)SB);
	}
}

int TLineEdit_GetLineFontHeight(TLineEdit *self) {
	return self->vds.Font->height + self->LineOffset;
}

static void TLineEdit_Adjust_VHSB(TLineEdit *self) {
  int m, wu, lew;
  TVDrawString *vds = &self->vds;

	vds->s = "        ";
	vds->lg = 8;
	vds->TabSpaceWidth = TextWidthColorString(vds);
	vds->s = "u";
	vds->lg = 1;
	wu = TextWidthColorString(vds);
	vds->s = NULL;
	vds->lg = 0;

	if (wu < 1) wu = 1;
	self->HSB->SmallChange = wu;
	self->VisibleLines = self->height / TLineEdit_GetLineFontHeight(self);
	m = self->tls->Count;
	if (!(self->LEfu & LEfu_VSBnoVisibleLines)) m += self->VisibleLines - 1;
	set_SB(self->VSB, m, self->VisibleLines);

	lew = self->width + 4;
	if (lew < 2) lew = 2;
	m = self->width;
	if (self->MaxStrX > lew) m = self->MaxStrX + wu;

	set_SB(self->HSB, m, lew);
}

void TLineEdit_AdjustSize(TWins *self) {
	TWins_AdjustSize(self);
	TLineEdit_Adjust_VHSB((TLineEdit*)self);
}

static TColS *TLineEdit_Set_vds(TLineEdit *le, int idy) {
  TColS *cs = TLines_Get_pcs(le->tls, idy);
	if (cs) {
	    le->vds.s = cs->s;
	    le->vds.lg = cs->lg;
	    le->vds.scda = (TStrColorDatArr*)cs->c;
	} else {
	    le->vds.s = NULL;
	    le->vds.lg = 0;
	    le->vds.scda = NULL;
	}
	return cs;
}

static int TLineEdit_PaintLine(TLineEdit *self, int idy, int y) {
  int xa, xe, ya, ye;
  TVDrawString *vds;
  TColS *cs = TLineEdit_Set_vds(self, idy);

	if (!cs) return y;
	vds = &self->vds;
	vds->SelE = vds->SelA;
	vds->SelToEnd = False;

	xa = self->SelAx;
	xe = self->CaretCharIdx;
	ya = self->SelAy;
	ye = self->CaretCharIdy;

	if ((xa != xe) || (ya != ye)) {
	    if (ya == ye) {
		if (idy == ya) {
		    vds->SelA = xa;
		    vds->SelE = xe;
		}
	    } else {
		if (ya > ye) {
		    xa = xe; 
		    xe = self->SelAx;
		    ya = ye; 
		    ye = self->SelAy;
		}
		if ((idy >= ya) && (idy <= ye)) {
		    if (idy == ya) {
			vds->SelA = xa;
			vds->SelE = cs->lg;
			vds->SelToEnd = True;
		    } else if (idy == ye) {
			vds->SelA = 0;
			vds->SelE = xe;
		    } else {
			vds->SelA = 0;
			vds->SelE = cs->lg;
			vds->SelToEnd = 2;
		    }
		}
	    }
	}

	vds->Xoff = -self->HSB->Count;
	vds->BackColor = self->Color;
	if ((cs->fu & fio_EOL_Mask) == fio_EOLnone) {
	    vds->BackColor = 0xffe0e0;
	} else {
	    TColS *csn = TLines_Get_pcs(self->tls, idy-1);
	    if (csn && ((csn->fu & fio_EOL_Mask) == fio_EOLnone)) {
		vds->BackColor = 0xffe8f0;
	    }
	}
	if (self->LineOffset > 0) {
	    int yo = y + self->vds.Font->height;
	    if (self->LineOffset == 1) {
		DrawLineRGB(self->WinID, self->Color, 0, yo, self->width, yo);
	    } else {
		FillRectRGB(self->WinID, self->Color, 0, yo, self->width, yo + self->LineOffset);
	    }
	}
	DrawColorString(0, y, vds);

	return y + TLineEdit_GetLineFontHeight(self);
}

static void TLineEdit_Caret_Draw(TLineEdit *le, int fu) {
	Caret_Draw(fu);
	if (le->highlight && le->highlight->SetCaret) {
	    le->highlight->SetCaret(le->highlight, le->tls, le->CaretCharIdx, le->CaretCharIdy, fu);
	}
}

void TLineEdit_DoPaint(TLineEdit *self) {
  int i, y, m;
	if (self->KeyFu & mHasFocus) TLineEdit_Caret_Draw(self, 0);
	self->vds.ClsToEnd = True;

	self->vds.Image = True;
	self->vds.BackColor = self->Color;
	m = TLineEdit_GetLineCount(self);
	y = 0;
	for (i=self->VSB->Count; i < m; i++) {
	    y = TLineEdit_PaintLine(self, i, y);
	    if (y > self->height) break;
	}
	if (y < self->height)
	    FillRectRGB(self->WinID, self->Color, 0, y, self->width, self->height - y);

	if (self->KeyFu & mHasFocus) TLineEdit_Caret_Draw(self, 1);
}

void TLineEdit_GetChar_xy(TLineEdit *le, int idx, int idy, int *ret_x, int *ret_y, int *ret_h) {
  int i, y, x;
  TColS *cs;

	*ret_x = -1; //*ret_y = -1;
	if (idy < le->VSB->Count) return;
	y = 0;
	for (i=le->VSB->Count; i<idy; i++) {
	    if (!TLineEdit_Set_vds(le, i)) return;
	    y += TLineEdit_GetLineFontHeight(le);
	    if (y > le->height) return;
	}
	cs = TLineEdit_Set_vds(le, idy);
	if (!cs) return;
	le->vds.Xoff = -le->HSB->Count;
	x = GetChar_XPos(idx, &le->vds);
	if (x < le->width) {
	    *ret_x = x;
	    *ret_y = y;
	    *ret_h = TLineEdit_GetLineFontHeight(le);
	}
}

void TLineEdit_GetCaretIdxy(TLineEdit *le, int x, int y, int *idx, int *idy) {
  TColS *cs;
	*idx = 0; *idy = 0;
	if (le->tls->Count <= 0) return;
  int iy = (y / TLineEdit_GetLineFontHeight(le)) + le->VSB->Count;
	if (iy >= le->tls->Count) iy = le->tls->Count;
	le->vds.Xoff = -le->HSB->Count;
	cs = TLineEdit_Set_vds(le, iy);
	if (cs) *idx = GetMouseX_CharPos(x, &le->vds);
	*idy = iy;
}

static void ed_Caret_SetNewPos(TLineEdit *le, int x) {
  int y;
	y = (le->CaretCharIdy - le->VSB->Count) * TLineEdit_GetLineFontHeight(le);
	if ((x > le->width) || (y < 0) || (y > le->height)) x = -1;
	TWins_Caret_SetNewPos((TWins*)le, x, y);
	if (x >= 0) Caret_Draw(1);
}

void TLineEdit_SetCaret(TLineEdit *le) {
  TColS *cs;
  int x;

	x = 0;
	Caret_Draw(0);
	cs = TLineEdit_Set_vds(le, le->CaretCharIdy);
	if (cs) {
	    le->vds.Xoff = -le->HSB->Count;
	    x = GetChar_XPos(le->CaretCharIdx, &le->vds); // - le->HSB->Count;
	    ed_Caret_SetNewPos(le, x);
	} else if (le->CaretCharIdy >= le->tls->Count) {
	    le->CaretCharIdy = le->tls->Count;
	    x = -le->HSB->Count;
	    ed_Caret_SetNewPos(le, x);
	}
	if (!le->OldX) le->OldCaretX = x;
	if (le->OnSetCaret) le->OnSetCaret(le, cs);
}

static void TLineEdit_SetCaret_if_Akt(TLineEdit *le) {
	if (TWins_Caret_Chk((TWins*)le)) TLineEdit_SetCaret(le);
}

void TLineEdit_ChkCaret(TLineEdit *le, int fu) {
  TColS *cs;
  int x, xa, m, ma, lh, ym, lx, la;

	lh = le->height / TLineEdit_GetLineFontHeight(le);
	ym = 0;
	if (fu & LE_ChkCaret_Ymiddle) ym = lh >> 1;
	if (le->CaretCharIdy < le->VSB->Count) {
	    TScrollBar_SetCount_ev(le->VSB, le->CaretCharIdy - ym);
	} else if (le->CaretCharIdy >= (le->VSB->Count + lh)) {
	    lh -= ym;
	    TScrollBar_SetCount_ev(le->VSB, le->CaretCharIdy - lh + 1);
	}

	cs = TLineEdit_Set_vds(le, le->CaretCharIdy);
	if (cs) {
	    if (le->CaretCharIdx > cs->lg) le->CaretCharIdx = cs->lg;
	    le->vds.Xoff = 0;
	    m = GetChar_XPos(le->CaretCharIdx, &le->vds);
	    x = m - le->HSB->Count;

	    if (TLineEdit_ChkIsSel(le)) {
		ma = GetChar_XPos(le->SelAx, &le->vds);
		xa = ma - le->HSB->Count;
		lx = x - xa;
		la = lx;
		if (la < 0) la = -la;

		if ((IappXDat.last_KeyState & Button1Mask) &&
		    (IappXDat.MouseDown.apph == le->apph)) {
  // MouseMove and MouseDown
		} else {

		  if (la < le->width) {
		    if (lx > 0) {
		      if (xa < 0) {x = xa; m = ma;}
		    } else {
		      if (xa > le->width) {x = xa; m = ma;}
		    }
		  } else {
		    if (lx > 0) {
		      if (x <= le->width) {x -= le->width; m -= le->width;}
		    }
		  }
		}
	    }

	    if (x < 0) {
		TScrollBar_SetCount_ev(le->HSB, m);
	    } else if (x > (le->width - 4)) {
		TScrollBar_SetCount_ev(le->HSB, m - le->width + 4);
	    }
	}
}

void TLineEdit_DoMouseDown(TLineEdit *self, XButtonEvent *ev) {
	TWins_DoMouseDown((TWins*)self, ev);
	if ((ev->button == Button1) && TestWinIsEnabled((TWins*)self)) {
	    self->OldX = False;
	    TLineEdit_GetCaretIdxy(self, ev->x, ev->y, &self->CaretCharIdx, &self->CaretCharIdy);
	    if (TLineEdit_ChkIsSel(self)) Refresh((TWins*)self);
	    self->SelAx = self->CaretCharIdx;
	    self->SelAy = self->CaretCharIdy;
	    TLineEdit_SetCaret(self);
	}
}


void TLineEdit_DoMouseUp(TLineEdit *self, XButtonEvent *ev) {
  int VSBc;

	TWins_DoMouseUp((TWins*)self, ev);
	if ((ev->button == Button4) || (ev->button == Button5)) {
	    VSBc = self->VSB->Count;
	    if (ev->button == Button4) VSBc -= self->MouseZ_Wheel; else VSBc += self->MouseZ_Wheel;
	    TScrollBar_SetCount_ev(self->VSB, VSBc);
	}
}


void TLineEdit_DoMouseMove(TLineEdit *self, XMotionEvent *ev) {
  int x, y, yo;

	TWins_DoMouseMove((TWins*)self, ev);
	if ((ev->state & Button1Mask) && (IappXDat.MouseDown.apph == self->apph)) {
	    yo = IappXDat.MouseDown.y - ev->y;
	    if (yo > 0) {
		if (yo > 4) yo = 4;
	    } else {
		if (yo < -4) yo = -4;
	    }
	    x = self->CaretCharIdx;
	    y = self->CaretCharIdy;
	    TLineEdit_GetCaretIdxy(self, ev->x, ev->y+yo, &self->CaretCharIdx, &self->CaretCharIdy);
	    if ((x != self->CaretCharIdx) || (y != self->CaretCharIdy)) {
		TLineEdit_ChkCaret(self, 0);
		TLineEdit_SetCaret(self);
		Refresh((TWins*)self);
	    }
	}
}

void TLineEdit_DoClick(TLineEdit *self, long KeyState, bool DblClick) {
  TColS *cs;
  int rx, ry, xa, xe;

	TWins_DoClick((TWins*)self, KeyState, DblClick);
	if (!DblClick) return;

	TLineEdit_GetCaretIdxy(self, IappXDat.MouseDown.x, IappXDat.MouseDown.y, &rx, &ry);
	cs = TLineEdit_Set_vds(self, ry);
	if (!cs || (self->vds.lg == 0)) return;
	xa = xe = rx;
	while ((xa > 0) && (GetTzC(self->TzC, GetVdsPrev_WChar(&rx, &self->vds)) == 1)) {
	    xa = rx;
	}
	if (xa < xe) xe--;
	while (GetTzC(self->TzC, GetVdsNext_WChar(&xe, &self->vds)) == 1);
	if (xa != xe) TLineEdit_SetSel(self, xa, ry, xe, ry);
}

void TLineEdit_SetSel(TLineEdit *le, int xa, int ya, int xe, int ye) {
	le->SelAx = xa;
	le->SelAy = ya;
	le->CaretCharIdx = xe;
	le->CaretCharIdy = ye;
	TLineEdit_SetCaret(le);
	Refresh((TWins*)le);
}

bool TLineEdit_ChkIsSel(TLineEdit *le) {
	return ((le->SelAx != le->CaretCharIdx) || 
		(le->SelAy != le->CaretCharIdy));
}

bool TLineEdit_GetSel_lgs(TLineEdit *le, Tlgs *lgs, int fu) {
  bool res = TLineEdit_ChkIsSel(le);

	lgs->s = NULL;
	lgs->lg = 0;
	if (res) {
	    TLines_GetSel_Tlgs(le->tls, lgs, le->SelAx, le->SelAy, le->CaretCharIdx, le->CaretCharIdy, fu);
	}
	return res;
}


void TLineEdit_Strg_C(TLineEdit *le) {
  Tlgs lgs;
  int fio_fu = 0;
  bool utf8;
	if (le->fio_Strg_C_on) fio_fu = fio_Strg_C;
	if (TLineEdit_GetSel_lgs(le, &lgs, fio_fu)) {
	    Tlgs_Clr(&IappXDat.SelTxt);

	    IappXDat.SelTxt.s = lgs.s;
	    IappXDat.SelTxt.lg = lgs.lg;

	    utf8 = le->vds.UTF8on;
	    if (utf8) {
		int fuEOL = 0;
		Chk_fio_UTF8_lgs(&lgs, &fuEOL);
	        if ((fuEOL & fio_has_0x80_char) && 
		    !(fuEOL & fio_has_UTF8_char)) utf8 = False;
	    }

	    if (utf8) {
#ifdef _OS_WIN32
		SetSelection(2, CF_UNICODETEXT, XA_STRING);
#else
		SetSelection(2, IappXDat.XA_UTF8_STRING, XA_STRING);
#endif
	    } else {
		SetSelection(1, XA_STRING);
	    }
	}
}

TPlugIn_KeyDown *TLineEdit_GetKeyDownPlugIn(TLineEdit *le) {
  TPlugIn_KeyDown *plug = NULL;
	if (le) plug = le->PlugIn_KeyDown;
	if (!plug && !le->Disable_keyplugin_Global) plug = TLineEdit_keyplugin_Global;
	return plug;
}

int TLineEdit_DoKeyDown(TLineEdit *self, TKeyEvent *KeyEv) {
  int m;

	TWins_DoKeyDown((TWins*)self, KeyEv);

	if (IappXDat.SelWait) return 0;
	TLineEdit_Set_vds(self, self->CaretCharIdy);

	if ((KeyEv->keycode == 54) && (KeyEv->state & 0x4)) { //Strg_C
	    TLineEdit_Strg_C(self);
	    return 1;
	}

  TPlugIn_KeyDown *plug = TLineEdit_GetKeyDownPlugIn(self);
	m = 0;
	while (plug && !plug->KeyDownPlugIn((TWins*)self, KeyEv, &m)) {    
	    plug = plug->next;
	}
	return m;
}

void TLineEdit_DoFocus(TWins *self, int Change) {
	TWins_DoFocus(self, Change);
	if (Change & WinsFocusIN) {
	    TLineEdit_SetCaret((TLineEdit*)self);
	} else {
	    Caret_Draw(0);
	}
}

void TLineEdit_VSB_OnCount(TScrollBar* sb, int Count) {
  TLineEdit *le = (TLineEdit*)sb->owner;

	TLineEdit_SetCaret_if_Akt(le);
	if (le->VSB_DoCount) le->VSB_DoCount((TWins*)sb, Count);
	Refresh((TWins*)le);
}

void TLineEdit_HSB_OnCount(TScrollBar* sb, int Count) {
  TLineEdit *le = (TLineEdit*)sb->owner;

	TLineEdit_SetCaret_if_Akt(le);
	if (le->HSB_DoCount) le->HSB_DoCount((TWins*)sb, Count);
	Refresh((TWins*)le);
}

static void TLineEdit_MaxStrLen(TLineEdit *le, int ya, int ye) {
  int i, x, mx;

	mx = 0;
	if (ye > le->tls->Count) ye = le->tls->Count;
	for (i = ya; i < ye; i++) {
	    TLineEdit_Set_vds(le, i);
	    x = TextWidthColorString(&le->vds);
	    if (x > mx) mx = x;
	}
	mx += 10;
	if (mx > le->MaxStrX) le->MaxStrX = mx;
}

void TLineEdit_tls_OnLineChange(TLines *tls, TColS *pcs, int idx) {
  TLineEdit *le = (TLineEdit*)tls->Owner;
	TLineEdit_MaxStrLen(le, idx, idx+1);
	TLineEdit_Adjust_VHSB(le);
}

void TLineEdit_tls_OnCount(TLines *tls) {
  TLineEdit *le = (TLineEdit*)tls->Owner;

	TLineEdit_Adjust_VHSB(le);
}

void TLineEdit_Clr(TLineEdit *le) {
	if (!le) return;
	le->CaretCharIdx = 0;
	le->CaretCharIdy = 0;
	le->SelAx = 0;
	le->SelAy = 0;
	le->OldCaretX = 0;
	le->MaxStrX = 0;
	TLines_Clear(le->tls);
	TLineEdit_SetCaret_if_Akt(le);
	Refresh((TWins*)le);
}

void TLineEdit_Clr_c(TLineEdit *le) {
	if (!le) return;
	TLines_Clear_c(le->tls);
}

void TLineEdit_AllAtrNew(TLineEdit *le) {
	le->MaxStrX = 0;
	le->tls->EvStop = False;
	TLineEdit_MaxStrLen(le, 0, le->tls->Count);
	TLineEdit_Adjust_VHSB(le);

	Refresh((TWins*)le);

	if (le->highlight && le->highlight->ChangeLine) {
	    AppHandleMessage();
	    le->highlight->ChangeLine(le->highlight, le->tls, 0, 0, clfu_all);
	    Refresh((TWins*)le);
	}
}

void TLineEdit_SetCaretXY(TLineEdit *le, int x, int y) {
	le->CaretCharIdx = x;
	le->SelAx = x;
	le->CaretCharIdy = y;
	le->SelAy = y;
	TLineEdit_ChkCaret(le, 0);
	TLineEdit_SetCaret(le);
}

void TLineEdit_AddLine_lgs(TLineEdit *le, Tlgs *lgs) {
	if (le->wtyp != wtyp_LineEdit) return;
	TLines_Add_lgs(le->tls, lgs, 0);
	Refresh((TWins*)le);
}

void TLineEdit_AddLine(TLineEdit *le, const char *s) {
  Tlgs lgs;
	lgs.s = (char*)s;
	lgs.lg = pclen(s);
	TLineEdit_AddLine_lgs(le, &lgs);
}


bool TLineEdit_Ctl(TLineEdit *self, TWinsCtl *wctl) {
	if (wctl->bef == wctlb_SetFont) TLineEdit_Clr_c(self);

	TWins_Ctl((TWins*)self, wctl);
	if (self->highlight && self->highlight->hlCtl) {
	    self->highlight->Font = self->vds.Font;
	    self->highlight->hlCtl(self->highlight, wctl);
	}

	if (wctl->bef == wctlb_SetFont) {
	    TLineEdit_AllAtrNew(self);
	    return True;
	}
	return False;
}

int TLineEdit_GetLineCount(TLineEdit *self) {
	if (!self || !self->tls) return 0;
	return self->tls->Count;
}

/*____________________________ Destroy __________________________________*/
bool TLineEdit__Destroy(TLineEdit *self) {

	TWins_Caret_Clr((TWins*)self);
	Tfree(self->highlight);
	self->highlight = NULL;

	Tfree(self->UnRedo);
	Tfree(self->tls);

	while (self->PlugIn_KeyDown) {
	    TPlugIn_KeyDown *plug = self->PlugIn_KeyDown;
	    self->PlugIn_KeyDown = plug->next;
	    Tfree(plug);
	}

	self->tls = NULL;
	self->vds.s = NULL;
	self->vds.lg = 0;
	self->vds.scda = NULL;
	return TWins__Destroy((TWins*)self);
}

/*_____________________________ Create __________________________________*/
void TLineEdit_CreateParams(TWins_Create_struct *wcs) {
  TLineEdit *self = wcs->wins;

	Caret_Init();

	TScrollWin_CreateParams(wcs);

	self->TzC = (Tss*)&Default_TzC;

	self->tls = TLines_Create();
	self->tls->Owner = (int*) self;
	self->tls->OnLineChange = TLineEdit_tls_OnLineChange;
	self->tls->OnCount = TLineEdit_tls_OnCount;
	self->vds.TabSpaceOn = True;
	self->vds.TabSpaceDot = True;

	self->destroy = (TDESTROY)TLineEdit__Destroy;
	self->AdjustSize = TLineEdit_AdjustSize;
	self->DoPaint = (TWinsEv)TLineEdit_DoPaint;
	self->WinsCtl = (WinsCtlEv)TLineEdit_Ctl;

	self->DoMouseDown = (TMupdownEv)TLineEdit_DoMouseDown;
	self->DoMouseUp = (TMupdownEv)TLineEdit_DoMouseUp;
	self->DoMouseMove = (TMmoveEv)TLineEdit_DoMouseMove;
	self->DoClick = (TClickEv)TLineEdit_DoClick;

	self->DoKeyDown = (TKeyEv)TLineEdit_DoKeyDown;

	self->DoFocus = TLineEdit_DoFocus;

	self->VSB->OnCount = TLineEdit_VSB_OnCount;
	self->HSB->OnCount = TLineEdit_HSB_OnCount;

	self->KeyFu = mCanFocus;
	self->fio_Strg_C_on = True;

	self->vds.Font = MakeFont(IappXDat.Default_Font->name, NULL, NULL);
	ChangeFontGC(self->vds.Font, self->vds.ForeColor, self->Color);

	self->cursor_nr = XC_xterm;
}

TLineEdit *TLineEdit_Create(TWins *Parent) {
  TWins_Create_struct wcs;
	TWins_Create_struct_init(&wcs, sizeof(TLineEdit), wtyp_LineEdit);
	wcs.Parent = Parent;
	wcs.CreateParams = TLineEdit_CreateParams;
	TScrollWin_DefaultConf(&wcs);
	return wcs.wins;
}

TLineEdit *TLineEdit_CreateA(TWins *Parent, int Left, int Top, int Width, int Height, int Align) {
  TLineEdit *self;
	self = TLineEdit_Create(Parent);
	self->pw->top = Top;
	self->pw->left = Left;
	self->pw->width = Width;
	self->pw->height = Height;
	self->Align = alClient;
	self->pw->Align = Align;
	Set_Parent((TWins*)self, Parent);
	return self;
}
