/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under LGPL v2.1, see file COPYING.LIB in this tarball
*/

#include "appXbase.h"
#include "tdirlist.h"

void TMenuFile_SubMenuFree(TMenu *self) {
  TMenuItem *mi;

	if (self->wtyp != wtyp_MenuFile) return;
	mi = self->MenuItem;
	while (mi) {
	    if (mi->next_menu) TMenuFile_SubMenuFree(mi->next_menu);
	    mi->next_menu = NULL;
	    mi = mi->next;
	}
	Tfree(self);
}

void TMenuFile_DataFree(TMenuFile *self) {
  TMenuItem *mi = self->MenuItem;

	while (mi) {
	    if (mi->next_menu) TMenuFile_SubMenuFree(mi->next_menu);
	    mi->next_menu = NULL;
	    mi = mi->next;
	}
	TMenu_Cls((TMenu*)self);
	self->mfk &= ~mfk_init_ok;
}


int TMenuFile_EvProc(Tobjmsg *msg) {
  XEvent *ev;
  TMenuFile *self = (TMenuFile*)msg->ap->wins;

	switch(msg->msg) {
	case msgXEv:
	    ev = (XEvent *) msg->p;
	    switch(ev->type) {

	    case UnmapNotify:
		if (self->OnUnmap) self->OnUnmap(self);
		if (self->mfk & mfk_AutoFree) TMenuFile_DataFree(self);
		return 1;
	    default:
		break;
	    }
	}
	return TWins_EvProc(msg);
}

void TMenuFile_DoPaint(TWins *self) {
  TMenu *mf = (TMenu*)self;
  int y, h;

	TMenu_DoPaint(mf);
	if (mf->side_pic) {
	    h = mf->side_pic->height - mf->height + 4;
	    y = 2;
	    if (h < 0) {
		y -=h;
		h = 0;
	    }
	    DrawPic_wh(mf->WinID, mf->side_pic, 0, h,
		    mf->SideXOff, mf->height-4, 2, y);
	}
}

bool TMenuFileItem__Destroy(TMenuFileItem *self) {
	if (self->wtyp == wtyp_MenuFileItem) {
	    xfree(self->filename);
	    xfree(self->itemname);
	}
	return TMenuItem__Destroy((TMenuItem*)self);
}

TMenuItem *TMenuFile_DoMenuItemCreate(TMenu *self) {
  TMenuFileItem *mi;
	mi = ttc_create(sizeof(TMenuFileItem), wtyp_MenuFileItem);
	mi->destroy = (TDESTROY)TMenuFileItem__Destroy;
	return (TMenuItem*)mi;
}

void TMenuFile_ChkSubMenu_OnFileMsg(TFileMsg *fmsg) {
  TDirListObjNy *self;
  TDirFileList *dfl;
  int bef;

	if (!file_msg_chk_wtyp(fmsg, wtyp_TDirListObjNy, FL)) return;
	self = fmsg->obj;

	if ((fmsg->art != fmArt_DirList) ||
	    (fmsg->typD != fmTyp_int) ||
	    (fmsg->typF != fmTyp_dfl) || !fmsg->F.dfl) {
	    return;
	}

	bef = fmsg->D.m;
	dfl = fmsg->F.dfl;
	if (!dfl || !(bef & dlbef_ScanRun)) return;

	switch (bef & dlbef_BefMask) {
	case dlbef_Add:
	    if ((dfl->st->mode & S_IFMT) == S_IFDIR) {
		self->tag |= 0x01;
	    } else if ((dfl->st->mode & S_IFMT) == S_IFREG) {
		if (dfl->st->FileType == stft_desktop) self->tag |= 0x02;
	    }
	    break;
	}
}

int TMenuFile_ChkSubMenu(char *s) {
  int res = 0;
  TDirListObjNy *ObjNy;

	ObjNy = TDirListObjNy_Create(NULL);
	ObjNy->OnFileMsg = TMenuFile_ChkSubMenu_OnFileMsg;
	ObjNy->findAtr = findREG | findDIR;
	if (TDirListObjNy_Open(ObjNy, s)) {
	    TDirListObjNy_Open_Wait(ObjNy);
	    res = ObjNy->tag;
	}
	Tfree(ObjNy);
	return res;
}

void TMenuFile_Sort(TMenuFile *self, char *sort) {
  TMenuFileItem *mi = (TMenuFileItem*)self->MenuItem;
  int lg, i;
  int count = 0;
  char *indata, *token;
  TMenuFileItem **miarr, **ppmi, *miprev;
  bool inv = (self->mfk & mfk_sort_inv);
  bool mkHead = False;

	while (mi) {
	    mi->mfik &= ~mfik_sort_ok;
	    if (mi->fu & miHead && !mkHead) {
		mkHead = True;
		count++;
	    }
	    count++;
	    mi = (TMenuFileItem*)mi->next;
	}
	if (count < 2) return;

	indata = sort;
	while (indata && *indata) {
	    indata = pcGetToken(indata, &token, &lg);
	    if (token && (*token == '-')) {
		token++;
		if (*token == 't') count++;
	    }
	}
	miarr = xcalloc((count+1) * sizeof(void*));
	
	indata = sort;
	ppmi = miarr;

	if (mkHead) {
	    mi = (TMenuFileItem*)self->MenuItem;
	    while (mi) {
		if (mi->fu & miHead) {
		    mi->mfik |= mfik_sort_ok;
		    *ppmi = mi;
		    ppmi++;
		}
		mi = (TMenuFileItem*)mi->next;
	    }
	    mi = (TMenuFileItem*)TMenu_AddItem((TMenu*)self, "-", 0, NULL);
	    mi->mfik |= mfik_sort_ok;
	    *ppmi = mi;
	    ppmi++;
	}

	if (inv) ppmi = &miarr[count-1];

	while (indata && *indata) {
	    indata = pcGetToken(indata, &token, &lg);
	    if (token && (*token == '-')) {
		token++;
		if (*token == 't') {
		    mi = (TMenuFileItem*)TMenu_AddItem((TMenu*)self, "-", 0, NULL);
		    mi->mfik |= mfik_sort_ok;
		    *ppmi = mi;
		    if (inv) ppmi--; else ppmi++;
		}
	    } else {
		mi = (TMenuFileItem*)self->MenuItem;
		while (mi) {
		    if ((pclen(mi->itemname) == lg) && pcncmp(mi->itemname, token, lg)) {
			mi->mfik |= mfik_sort_ok;
			*ppmi = mi;
			if (inv) ppmi--; else ppmi++;
			mi = NULL;
		    } else mi = (TMenuFileItem*)mi->next;
		}
	    }
	}
	mi = (TMenuFileItem*)self->MenuItem;
	while (mi) {
	    if ((mi->mfik & mfik_sort_ok) == 0) {
		*ppmi = mi;
		if (inv) ppmi--; else ppmi++;
	    }
	    mi = (TMenuFileItem*)mi->next;
	}
	ppmi = miarr;
	miprev = NULL;
	for (i=0; i<count; i++) {
	    mi = *ppmi;
	    mi->prev = (TMenuItem*)miprev;
	    miprev = mi;
	    ppmi++;
	    mi->next = (TMenuItem*)*ppmi;
	}
	mi->next = NULL;
	self->MenuItem = (TMenuItem*)*miarr;
	xfree(miarr);
}

void TMenuFile_InitDat_dfl(TDirListObjNy *self, TDirFileList *dfl) {
  int chk, mifu;
  char *rpe;
  TMenuFileItem *mi;
  TMenuFile *mf, *tmp_mf;
  TpcIni *ini;
  TDesktopHead ddkh;
  bool mkDirHead;
  char line[2048];

	mf = self->owner;
	rpe = line;
	ppcmove(&rpe, self->dl->gp.path);

	if ((dfl->st->mode & S_IFMT) == S_IFDIR) {
	    ppcmove(&rpe, dfl->fn);
	    *rpe = '/'; rpe++;
	    *rpe = 0;
	    tmp_mf = NULL;
	    mifu = 0;
	    chk = TMenuFile_ChkSubMenu(line);
	    if (chk) {
		tmp_mf = TMenuFile_Create(mf->Parent);
		tmp_mf->OnClick = mf->OnClick;
		ppcchange(&tmp_mf->path, line);
	    } else mifu = miNotEnabled;

	    ppcmove(&rpe, ".directory");
	    ini = TpcIni_Create(line);
	    ddkh.name = dfl->fn;
	    TPic *pic = NULL;
	    if (dotdesktop_get_head(ini, &ddkh)) {
		pic = GetDeskIcon(ddkh.ddkt, mf->icon_size, ddkh.icon, 0, NULL);
	    }
	    mi = (TMenuFileItem*)TMenu_AddItem((TMenu*)mf, ddkh.name, mifu, NULL);
	    Tfree(ini);
	    mi->pic = pic;
	    ppcchange(&mi->itemname, dfl->fn);
	    mi->next_menu = (TMenu*)tmp_mf;

	} else if ((dfl->st->mode & S_IFMT) == S_IFREG) {

	    mkDirHead = ((mf->mfk & mfk_AddDirHead) && pccmp(dfl->fn, ".directory"));
	    if ((dfl->st->FileType == stft_desktop) || mkDirHead) {
		ppcmove(&rpe, dfl->fn);
		ini = TpcIni_Create(line);
		ddkh.name = dfl->fn;
		TPic *pic = NULL;
		if (dotdesktop_get_head(ini, &ddkh)) {
		    pic = GetDeskIcon(ddkh.ddkt, mf->icon_size, ddkh.icon, 0, NULL);
		}

		mi = (TMenuFileItem*)TMenu_AddItem((TMenu*)mf, ddkh.name, 0, NULL);
		Tfree(ini);
		mi->pic = pic;
		mi->itemname = ChangeFileExt(dfl->fn, NULL);
		ppcchange(&mi->filename, dfl->fn);
		if (mkDirHead) mi->fu |= miHead | miCanNotSelect;
	    }
	}
}

void TMenuFile_InitDat(TMenuFile *self) {
  char *rpe;
  TDirListObjNy *ObjNy;
  char *pc, *icon, *sort;
  TDirFileList *dfl;
  TpcIni *ini;
  char line[2048];

	if (self->OnInitDat) self->OnInitDat(self, False);
	if (self->mfk & mfk_init_ok) return;
	self->mfk |= mfk_init_ok;
	if (pclen(self->path) < 3) return;

	rpe = line;
	ppcmove(&rpe, self->path);
	rpe--;
	if (*rpe != '/') {
	    rpe++;
	    *rpe = '/';
	}
	rpe++;
	*rpe = 0;

	ObjNy = TDirListObjNy_Create(self);
	ObjNy->findAtr = findDIR | findREG | findIsRootPath;
	if ((self->mfk & mfk_AddDirHead) == 0) ObjNy->findAtr |= findNoHidden;
	if (TDirListObjNy_Open(ObjNy, line)) {
	    TDirListObjNy_Open_Wait(ObjNy);
	    TDirListObjNy_Sort(ObjNy, dlsortFu_Name);
	    dfl = ObjNy->dfl;
	    while (dfl) {
		TMenuFile_InitDat_dfl(ObjNy, dfl);
		dfl = dfl->next;
	    }
	}
	Tfree(ObjNy);

	pc = rpe;
	ppcmove(&pc, ".directory");
	if (TestIsFile(line)) {
	    ini = TpcIni_Create(line);
	    sort = TpcIni_RD(ini, Desktop_Entry, "ALDsort");
	    if (sort) TMenuFile_Sort(self, sort);
	    icon = TpcIni_RD(ini, Desktop_Entry, "ALDSideIcon");
	    if (icon) {
		self->side_pic = GetDeskIcon(0, 0, icon, 0, NULL);
	    }
	    Tfree(ini);
	}
	if (self->OnInitDat) self->OnInitDat(self, True);

}


void TMenuFile_DoShow(TMenu *me, int Left, int Top) {
	TMenuFile_InitDat((TMenuFile*)me);
	TMenu_DoShow(me, Left, Top);
}

/*____________________________ Destroy __________________________________*/
bool TMenuFile__Destroy(TMenuFile *self) {
	if (self->wtyp == wtyp_MenuFile) {
	    xfree(self->path);
	} else {
 printf("ERROR: %s(%d) TMenuFile_Destroy: win=%d wtyp=%d ???\n", FL, (int)self->WinID, self->wtyp);  
	}
	return TMenu__Destroy((TMenu*)self);
}

/*_____________________________ Create __________________________________*/
void TMenuFile_CreateParams(TWins_Create_struct *wcs) {
  TMenuFile *self = wcs->wins;

	TMenu_CreateParams(wcs);

	self->icon_size = 16;
	self->apph->EvProc = TMenuFile_EvProc;
	self->destroy = (TDESTROY)TMenuFile__Destroy;
	self->DoMenuItemCreate = TMenuFile_DoMenuItemCreate;

	self->DoPaint = TMenuFile_DoPaint;
	self->DoShow = TMenuFile_DoShow;
}

TMenuFile *TMenuFile_Create(TWins *Form) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TMenuFile), wtyp_MenuFile);
	wcs.Parent = Form;
	wcs.CreateParams = TMenuFile_CreateParams;
	TWins_DefaultConf(&wcs);
	return wcs.wins;
}
