/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "appXbase.h"
#ifdef USE_DECOR
#include "image.h"
#endif
#include "alsa_mixer.h"
#include <math.h>
#include <fcntl.h>

#define APPNAME "aldalsamixer"

typedef struct {
  AlgTWins
  TScrollBar *sbVol;
  TScrollBar *sbBal;
  int headpicno;
  Tmix_elem *elem;
  bool ev_stop;
} TpnlRegler;


typedef struct {
  AlgTWins
  AlgTForm
  TSound_MIXER *alsa_mix;
  TPageControl *pgc;
  TPage *pgBase;
  TpnlRegler *prMaster;
  TpnlRegler *prPCM;
  TpnlRegler *prSpeaker;
  bool Bal_LR;
} TMainForm;

TPic *get_MixerIcon(int idx);


/* ___________________________ Paint ____________________________________*/
static void pnlPaint_pc(TWins* pnl, char *pc, int x, int y, int fu) {
	vds_SetStr(&pnl->vds, pc);
	if (fu & 0x01) {
	    x -= (TextWidthColorString(&pnl->vds) >> 1);
	}
	if (x < 1) x = 1;
	if (fu & 0x02) FillRectRGB(pnl->WinID, pnl->Color, 1, y, pnl->width-2, pnl->vds.Font->height);
	DrawColorString(x, y, &pnl->vds);
}


void TpnlRegler_DoPaint(TpnlRegler *self) {
  int x, xm, y, m;
//  Tmix_elem *elem;
  char *name;
  char db[16];

	SetFont((TWins*)self, "-adobe-helvetica-medium-r-normal-*-10-*-75-75-*-*-*-1");

	x = self->sbVol->left;
	xm = x + ((self->sbBal->left + self->sbBal->width - x) >> 1);
	y = self->sbVol->top + self->sbVol->height + 2;
	DrawPic(self->WinID, get_MixerIcon(self->headpicno), x + 2, 11);

	if (self->elem) {
	    if (self->elem->mixer_type & MIXER_ELEM_MUTE_SWITCH) {
		m = 8 + (self->elem->mute & 0x03);
		DrawPic(self->WinID, get_MixerIcon(m), xm - 8, 2);
	    }
	}

	name = "-----";
	pcmove(db, "--/--dB");
	if (self->elem) {
	    name = self->elem->item_name;
	    snprintf(db,15,"%.1f/%.1fdB", self->elem->dB_left, self->elem->dB_right);
	}

	pnlPaint_pc((TWins*)self, db, xm, y, 0x03);
	y += self->vds.Font->height;
	SetFont((TWins*)self, IappXDat.Default_bold_Font->name);
	pnlPaint_pc((TWins*)self, name, xm, y, 0x03);
}

void TpnlRegler_OnClick(TpnlRegler *self, long KeyState, bool DblClick) {
  int x, xm, mute;
  TMainForm *mf = IappDat.obj;
  Tmix_elem *elem = self->elem;
  TMouseDown *md = &IappXDat.MouseDown;


	if (!DblClick) return;


	if (!elem || !(elem->mixer_type & MIXER_ELEM_MUTE_SWITCH)) return;
	if ((md->y > 16) || (md->y < 2)) return;

	x = self->sbVol->left;
	xm = x + ((self->sbBal->left + self->sbBal->width - x) >> 1);

	mute = -1;
	if ((md->x > (xm-8)) && (md->x < (xm+8))) {
	    if (elem->mute & 0x03) mute = 0; else mute = 3;
	} else if (md->x <= (xm-8)) {
	    mute = elem->mute ^ 0x01;
	} else if (md->x >= (xm+8)) {
	    mute = elem->mute ^ 0x02;
	}
	if (mute >= 0) {
	    TSound_AlsaMIX_SetMute(mf->alsa_mix, elem->idx, mute);
	    Refresh((TWins*)self);
	}
}

void TpnlRegler_SetUp(TpnlRegler *rg, Tmix_elem *elem) {
  int m;
  float vf;
  float vol, bal;
  TMainForm *mf = IappDat.obj;

	rg->elem = elem;

	m = elem->vmax;
	m >>= 3;
	if (m < 1) m = 1;
	rg->sbVol->LargeChange = m;
	rg->sbBal->LargeChange = m;

	if (!rg->ev_stop) {
	    bal = 0;
	    if (elem->vol_left < elem->vol_right) {
		vol = elem->vol_right;
		bal = vol - elem->vol_left;
		if (mf->Bal_LR) bal = elem->vol_left - vol;
	    } else {
		vol = elem->vol_left;
		bal = elem->vol_right - vol;
		if (mf->Bal_LR) bal = vol - elem->vol_right;
	    }
	    vol *= elem->vmax;
	    TScrollBar_SetCountParam(rg->sbVol, round(vol), 0, elem->vmax, 0, sbfu_inv);

	    vf = elem->dB_vmax - elem->dB_vmin;
	    m = elem->vmax;
	    if (vf > 24) m = round((float)m * 24 / vf);

	    bal *= elem->vmax;
	    TScrollBar_SetCountParam(rg->sbBal, round(bal), 0, m*2, -m, sbfu_inv);
	}

	rg->ev_stop = False;

	Refresh((TWins*)rg);
}

void TpnlRegler_sbOnCount(TScrollBar* sb, int Count) {
  TMainForm *mf = IappDat.obj;
  TpnlRegler *rg = sb->subClass;
  Tmix_elem *elem = rg->elem;
  float vol_left, vol_right;
  float bal;

	if (!elem || !mf->alsa_mix) return;
	vol_left = TScrollBar_GetCount(rg->sbVol) / (float)elem->vmax;
	vol_right = vol_left;
	bal = 0;
	if (elem->channels > 1) {
	    bal = TScrollBar_GetCount(rg->sbBal) / (float)elem->vmax;
	    if (bal < 0) {
		if (mf->Bal_LR) vol_left += bal; else vol_right += bal;
	    } else {
		if (mf->Bal_LR) vol_right -= bal; else vol_left -= bal;
	    }
	}
	rg->ev_stop = True;
	TSound_AlsaMIX_SetVol(mf->alsa_mix, elem->idx, vol_left, vol_right);
}

void TpnlRegler_DefaultConf(TWins_Create_struct *wcs) {
  TpnlRegler *self = wcs->wins;

	TWins_DefaultConf(wcs);
	self->subClass = self;
	self->width = 70;
	self->Align = alLeft;
	self->BorderFu = mBorder | mBorderL;
	self->DoPaint = (TWinsEv)TpnlRegler_DoPaint;
	self->OnClick = (TClickEv)TpnlRegler_OnClick;

	self->sbVol = TScrollBarVert_CreateP((TWins*)self);
	self->sbVol->BorderFu = mBorder | mBorderL | mBorderNoTop | mBorderNoBottom;
	TWins_SetBorderWidth((TWins*)self->sbVol, 1);
	SetRC((TWins*)self->sbVol, 17, 27, 21, 120);
	self->sbVol->OnCount = TpnlRegler_sbOnCount;
	self->sbBal = TScrollBarVert_CreateP((TWins*)self);
	self->sbBal->BorderFu = mBorder | mBorderL | mBorderNoTop | mBorderNoBottom;
	TWins_SetBorderWidth((TWins*)self->sbBal, 1);
	SetRC((TWins*)self->sbBal, 40, 27, 13, 120);
	self->sbBal->OnCount = TpnlRegler_sbOnCount;
}

TpnlRegler *TpnlRegler_Create(TWins *Parent) {
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TpnlRegler), wtyp_USER_Wins);
	wcs.Parent = Parent;
	TpnlRegler_DefaultConf(&wcs);
	Set_Parent(wcs.wins, Parent);
	return wcs.wins;
}

/* ___________________________ Event ____________________________________*/

void mixer_OnElemAdd(TSound_MIXER *mix, Tmix_elem *mix_elem) {
  int type;
  TMainForm *mf = IappDat.obj;

	type = mix_elem->mixer_type & MIXER_ELEM_TYPE_MASK;
	if ((type == MIXER_ELEM_ENUM) || (type == MIXER_ELEM_CAPTURE_ENUM)) {

	} else if (type == MIXER_ELEM_CAPTURE) {

	} else {
	    if (mix_elem->mixer_type & MIXER_ELEM_HAS_VOLUME) {
//	printf("%s(%d): %s\n", FL, mix_elem->item_name);
		if (str_match("*master*", mix_elem->item_name, FU_CASE_ASCII)) {
//	printf("%s(%d): %s\n", FL, mix_elem->item_name);
		    TpnlRegler_SetUp(mf->prMaster, mix_elem);

		} else if (str_match("*pcm*", mix_elem->item_name, FU_CASE_ASCII)) {
//	printf("%s(%d): %s\n", FL, mix_elem->item_name);
		    TpnlRegler_SetUp(mf->prPCM, mix_elem);

		} else if (str_match("*speaker*", mix_elem->item_name, FU_CASE_ASCII) ||
			   str_match("*beep*", mix_elem->item_name, FU_CASE_ASCII)) {
//	printf("%s(%d): %s\n", FL, mix_elem->item_name);
		    TpnlRegler_SetUp(mf->prSpeaker, mix_elem);
		}
	    }
	}
}

void mixer_OnElemDel(TSound_MIXER *mix, Tmix_elem *mix_elem) {
}

void mixer_OnElemChange(TSound_MIXER *mix, Tmix_elem *mix_elem) {
  TMainForm *mf = IappDat.obj;

	if (mf->prMaster->elem == mix_elem) {
//	    Setup_sb(mf->sbMaster, mf->sbMasterB, mix_elem);
	    TpnlRegler_SetUp(mf->prMaster, mix_elem);
	} else if (mf->prPCM->elem == mix_elem) {
	    TpnlRegler_SetUp(mf->prPCM, mix_elem);
	} else if (mf->prSpeaker->elem == mix_elem) {
	    TpnlRegler_SetUp(mf->prSpeaker, mix_elem);
	}
}

/* ___________________________ app Taskbar ______________________________*/
void FormTaskBar_OnPaint(TWins *self) {
	DrawPic(self->WinID, get_MixerIcon(7), 0, 0);
}

void FormTaskBar_OnClick(TWins *self, long KeyState, bool DblClick) {
  TMainForm *mf = IappDat.obj;
	if (!mf->Visible) {
	    TForm_Show((TForm*)mf);
	} else {
	    TForm_Close((TForm*)mf);
	}
}

void m_Close(TMenu *menu, long KeyState, TMenuItem *smi) {
  Close();
}

/* ___________________________ app Event ________________________________*/
void Main_PipeProc(int fd, Tfd_list *fdl) {
  int lg;
  TMainForm *mf = IappDat.obj;
  char line[BUFSIZ];

	lg = read(fd, line, BUFSIZ);
	if (lg > 0) {
	    line[lg] = 0;
	    if (!mf) return;
	    if (!mf->Visible) TForm_Show((TForm*)mf);
// printf("%s(%d): AHbatmon_PipeProc fd=%d\"%s\"\n", FL, fd, line);
	}
}

void OnAppStart() {
  TMainForm *mf = IappDat.obj;
	mf->alsa_mix->init(mf->alsa_mix);
}

void OnAppClose() {
  TMainForm *mf = IappDat.obj;
	mf->alsa_mix->stat = 0;		//Stop Events
	TSound_AlsaMIX_Close(mf->alsa_mix);
}

static const char speaker16xf_AHico[] = {
 0x04,0x0E,0x10,0x00,0x10,0x00,0x00,0x00,0x9A,0x00,0x04,0x21,0xA6,0x31,0xE8,0x41,
 0x49,0x4A,0x8A,0x52,0xCB,0x5A,0x2C,0x63,0x6E,0x73,0xCF,0x7B,0x10,0x84,0x51,0x8C,
 0xB2,0x94,0xF4,0xA4,0x18,0xC6,0xDB,0x0C,0x21,0x26,0xBB,0x3D,0x72,0x55,0x78,0x80,
 0x6C,0x37,0x42,0x94,0x89,0x79,0xDA,0x37,0x32,0x92,0x9A,0x78,0x55,0x77,0xD9,0x7C,
 0x4F,0xC0,0x31,0x27,0x5E,0x09,0x21,0x77,0x03,0x88,0x32,0x5E,0x66,0x4D,0x98,0x76,
 0x0A,0x0C,0x2B,0x59,0xEE,0x4D,0x54,0x36,0x16,0x14,0x1F,0x59,0x32,0x4A,0x10,0x36,
 0xA2,0x14,0xDF,0x58,0x32,0x4A,0x10,0x76,0x22,0x0D,0x2B,0x19,0xE6,0x4D,0x10,0x76,
 0x4F,0x09,0x33,0x1E,0x62,0x4D,0x58,0x62,0xF3,0x54,0x1C,0x73,0xE2,0x95,0x10,0x32,
 0xF6,0x8D,0xCC,0xA4,0x26,0x5E,0xD5,0x89,0x76,0x23,0x44,0x99,0x98,0xC7,0xF6,0x13,
 0x55,0x87,0x78,0xE7,0x76,0x33,0x89,0x87,0xA9,0x0D};

const char *Help_Msg = {
  "\n"
  "Usage: aldalsamixer [options]\n"
  "\n"
  "  -t                   add to taskbar\n"
  "  -h, --help           Prints out this screen\n"
  "  ALDccp: aldalsamixer version "VERSION"  (build %d  %s)\n"
};

int main (int argc, char **argv) {
  TMainForm *mf;

  int i, m, fd;
  char *val;
  char *tline;
  int iofd[2] = {-1,-1};
  TSysTrayWin *FormTaskBar = NULL;
  bool help = False;
  bool err = False;
  bool TaskBar = False;
  char line[1024];

	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'h':
		    help = True; break;
		case 't':
		    TaskBar = True; break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
#ifdef Test_Root_WinId
		    } else if (pcncmp(val, "TestRootWinID", 13)) {
			i++;
			val = argv[i];
			if (val && *val != '-') {
				TestRootWinId = pctoint(val);
			} else err = True;
#endif
		    } else err = True;
		    break;
		default: 
			err = True;
		}  // switch end
	    } else {
		err = True;
	    }
	    i++;
	}

	if (help || err) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
	    exit(err);
	}

	if (!TestIsPath("/proc/asound")) {
	    printf("ERROR: aldalsamixer \"ALSA driver not running\"\n");
	    return 0;
	}

	initXapp(0, argc, argv);

	if (ChkAppRun(*argv, carfu_cmd, &tline)) {
	    m = 0;
	    fd = open("/tmp/.ALDccp/aldalsamixer", O_RDONLY);
	    if (fd > 0) {
		m = read(fd, line, 16);
		close(fd);
		if (m > 0) m = pctoint(line);
	    }
	    if (m <= 0) return 0;

	    snprintf(line, sizeof(line), "%sfd/%d", tline, m);
	    fd = open(line, O_WRONLY);
	    if (fd > 0) {
		write(fd, "show", 4);
		close(fd);
	    } else appError(EMSG_ERR1 | EMSG_SYS, "\"%s\"", line);
	    return 0;
	}

	pipe(iofd);
	AddPipeProc(iofd[0], InPipe, Main_PipeProc, NULL);
// printf("%s(%d): %d %d\n", FL, iofd[0], iofd[1]);

	FileHandle fh;
	FileOpen_wr_mkdir_erl("/tmp/.ALDccp/aldalsamixer", &fh, 0777, EMSG_ERR_EXIT);
	inttopc(line, iofd[0]);
	FileWrite(fh, line, pclen(line), NULL);
	FileClose(fh);

	IappXDat.OnAppStart = OnAppStart;
	IappXDat.OnAppClose = OnAppClose;
#ifdef USE_DECOR
	IappXDat.DefaultButtonDecor = TDecor_button_w3b3_Create();
#endif
	if (TaskBar) {
	    FormTaskBar = TSysTrayWin_Create();
  TMenuItem *mi;
	    FormTaskBar->OnPaint = FormTaskBar_OnPaint;
	    FormTaskBar->OnClick = FormTaskBar_OnClick;
	    FormTaskBar->PopupMenu = TMenu_Create(NULL);
	    mi = TMenu_AddItem(FormTaskBar->PopupMenu, "ALSA Mixer", miCanNotSelect, NULL);
	    mi->pic = get_MixerIcon(6);
	    TMenu_AddItem(FormTaskBar->PopupMenu, "-", 0, NULL);
	    TMenu_AddItem(FormTaskBar->PopupMenu, "Close", 0, m_Close);
	}

	mf = TForm_CreatParamAlg(NULL, sizeof(TMainForm));
	IappDat.obj = mf;
	mf->height = 200;
	mf->width = 214;
	mf->min_width = 214;
	mf->min_height = 200;
	mf->max_height = 200;
	mf->FormPosition = poDesigned;
	mf->Bal_LR = True;
	mf->Visible = FormTaskBar == NULL;

	if (FormTaskBar && FormTaskBar->SysTrayWin) {
	    mf->Gravity = SouthEastGravity;
	}

	
	mf->Icon = GetZIcon(speaker16xf_AHico);
//	mf->Icon = GetDeskIcon(ddkt_Application, 16, APPNAME, 
//				    gdifu_noZoom, speaker16xf_AHico);

	mf->pgc = TPageControl_Create((TWins*)mf);
	mf->pgc->Align = alClient;
	Set_Parent((TWins*)mf->pgc, (TWins*)mf);

	mf->pgBase = TPageControl_AddNewPage(mf->pgc, "Base");

	mf->prMaster = TpnlRegler_Create((TWins*)mf->pgBase);
	mf->prMaster->headpicno = 2;

	mf->prPCM = TpnlRegler_Create((TWins*)mf->pgBase);
	mf->prPCM->headpicno = 1;

	mf->prSpeaker = TpnlRegler_Create((TWins*)mf->pgBase);
	mf->prSpeaker->headpicno = 1;


	mf->alsa_mix = TSound_AlsaMIX_Create(0);
	mf->alsa_mix->OnElemAdd = mixer_OnElemAdd;
	mf->alsa_mix->OnElemDel = mixer_OnElemDel;
	mf->alsa_mix->OnElemChange = mixer_OnElemChange;

//	if (SetFontSXTypeIdx(11, IappXDat.Default_Font->name, "10", &val)) {
// printf("%s\n", val);
//	}

	AppRun();

	unlink("/tmp/.ALDccp/aldalsamixer");
}

