/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

static bool di_ChkDeskPos(TDesktopIcon *self, TwmRoot *wmr, int cx, int cy) {
  bool ok = True;
  Tapph *app = wmr->apph->next_child;
	while (ok && app) {
	    TWins *wins = app->wins;
	    if ((wins != (TWins*)self) && (wins->wtyp == wtyp_TwmDeskIcon)) {
		if (((wins->left > (cx-29)) && 
		    (wins->left < (cx+29))) &&
		    ((wins->top  > (cy-44)) && 
		    (wins->top  < (cy+44)))) ok = False;
	    }
	    app = app->next;
	}
	return ok;
}

static int di_GetMaxRight() {
	return dpyMonitors.mon[0].right - 32;
}
static int di_GetMaxBottom() {
	return dpyMonitors.mon[0].bottom - 72;
}

static void di_GetNewDeskPos(TDesktopIcon *self, TwmRoot *wmr, int *pcx, int *pcy) {
  int cx = *pcx;
  int cy = *pcy;
  int max_right = di_GetMaxRight();
  int max_bottom = di_GetMaxBottom();
  bool ok = False;
	cx = 15; cy = 5;
	while (!ok && (cy < max_bottom)) {
	    if (cx > max_right) {cx = 15; cy += 55;}
	    ok = di_ChkDeskPos(self, wmr, cx, cy);
	    if (ok) {
		*pcx = cx;
		*pcy = cy;
		return;
	    }
	    cx += 55;
	}
}

void TDesktopIcon_DoPaint(TDesktopIcon *self) {

	if (!self->pic) return;
	if (self->pic->mask != self->LastIcoMask) {
	    self->LastIcoMask = self->pic->mask;
	    CombineMask(self->WinID, self->pic->mask);
	}
	DrawPic(self->WinID, self->pic, 0, 0);
}

void TDesktopIconTw_DoPaint(TWins *self) {
  TDesktopIcon *di = (TDesktopIcon*)self->owner;

	if (di->glob->mod & (wmDImod_HighLight | wmDImod_Transparent)) {
	    di->WinCaption->vds.ForeColor = 0x404040;
	    DrawColorString(2, 0, &self->vds);
	    DrawColorString(2, 2, &self->vds);
	    DrawColorString(1, 1, &self->vds);
	    DrawColorString(3, 1, &self->vds);
	    di->WinCaption->vds.ForeColor = 0xffffff;
	    DrawColorString(2, 1, &self->vds);
	} else {
	    DrawColorString(2, 1, &self->vds);
	}
}

TDesktopIcon INTERNAL_FUNC *get_owner_DesktopIcon(TWins *self) {
  TWins *owner;
	if (self->wtyp == wtyp_TwmDeskIcon) return (TDesktopIcon*) self;
	owner = self->owner;
	if (owner && (owner->wtyp == wtyp_Wins)) owner = owner->owner;
	if (!owner || (owner->wtyp != wtyp_TwmDeskIcon)) return NULL;
	return (TDesktopIcon*)owner;
}

void TDesktopIcon_OnTimer(TgTimer *gti) {
  TWins *ws = gti->obj;
  TDesktopIcon *di = get_owner_DesktopIcon(ws);

	if (!di) return;
	di->IcoButtonUpTimer = NULL;
	Set_Cursor(ws, XC_fleur);
}

void TDesktopIcon_DoMouseDown(TWins *self, XButtonEvent *ev) {
  TDesktopIcon *di = get_owner_DesktopIcon(self);

	if (!di) return;
	if (di->IcoButtonUpTimer != NULL) Tfree(di->IcoButtonUpTimer); //gTimerX_Del(di->IcoButtonUpTimer);
	di->IcoButtonUpTimer = NULL;
	IappXDat.MouseDown.obj_left = di->left;
	IappXDat.MouseDown.obj_top = di->top;
	if (ev->button == Button1) {
	    di->IcoButtonUpTimer = gTimerX_Add(Timer_Enabled | Timer_OnTimerAndFree, 400, TDesktopIcon_OnTimer);
	    di->IcoButtonUpTimer->obj = self;
	}
	TWins_DoMouseDown(self, ev);
}

void TwmDeskIcon_App_ExecEv(Texec_pipe* expip, int evtyp);




void TDesktopIcon_DoMouseUp(TWins *self, XButtonEvent *ev) {
  TDesktopIcon *di = get_owner_DesktopIcon(self);

	if (!di) return;
	if (di->IcoButtonUpTimer != NULL) Tfree(di->IcoButtonUpTimer); //gTimerX_Del(di->IcoButtonUpTimer);
	di->IcoButtonUpTimer = NULL;
	TWins_DoMouseUp(self, ev);

  TpcIni *ini = NULL;
	if (self->cursor_nr == XC_fleur) {
  char *fn;
  char Dgr[32], Dxy[32];

	    if (!di_ChkDeskPos(di, (TwmRoot*)di->Parent, di->left, di->top)) {
		int x = di->left;
		int y = di->top;
		di_GetNewDeskPos(di, (TwmRoot*)di->Parent, &x, &y);
		SetPosition((TWins*)di, x, y);
		SetPosition(di->WinCaption, x + di->xo, y + di->yo);
	    }

	    fn = GetFileName_pc(di->fname);
	    ini = TpcIni_Create(((TwmRoot*)di->Parent)->IconPosFile);
	    snprintf(Dgr, 32, "%dx%d", dpyMonitors.mon[0].right, dpyMonitors.mon[0].bottom);
	    snprintf(Dxy, 32, "%d %d", di->left, di->top);
	    TpcIni_WRstr(ini, fn, Dgr, Dxy);
	    TpcIni_Save(ini);
		
	} else if (IappXDat.GrabWin) {
//	    if (ev->button != Button1) Std_UngrabPointer();

	} else if (ev->button == Button1) {
	    switch(di->ddkt) {
	    case ddkt_Application:
		ExecDesktopFile(di->fname, NULL);
		break;

	    case ddkt_FSDevice:
		ini = TpcIni_Create(di->fname);
		if (di->pic == di->pic3) {
		    meDeskIcon_mount_Click(ini);
		} else if (di->pic == di->pic2) {
		    meDeskIcon_umount_Click(ini);
		}
		break;
	    } //switch
	}
	Tfree(ini);
	Set_Cursor(self, XC_top_left_arrow);
}

void TDesktopIcon_DoMouseMove(TWins *self, XMotionEvent *ev) {
  int x, y;
  TDesktopIcon *di = get_owner_DesktopIcon(self);

	if (!di) return;
	if ((self->cursor_nr == XC_fleur) &&
	    (ev->state & Button1Mask) && (IappXDat.MouseDown.apph == self->apph)) {

	    x = IappXDat.MouseDown.obj_left + (ev->x_root - IappXDat.MouseDown.x_root);
	    y = IappXDat.MouseDown.obj_top + (ev->y_root - IappXDat.MouseDown.y_root);

	    if (x > di_GetMaxRight()) x = di_GetMaxRight();
	    if (y > di_GetMaxBottom()) y = di_GetMaxBottom();
	    if (x < 0) x = 0;
	    if (y < 0) y = 0;

	    x = (x / 5) * 5;
	    y = (y / 5) * 5;
	    SetPosition((TWins*)di, x, y);
	    SetPosition(di->WinCaption, x + di->xo, y + di->yo);
	}
	TWins_DoMouseMove(self, ev);
}

static bool TDesktopIcon_ChkMouseOverIcoCaption(TDesktopIcon *di, int x, int y) {
  TWins *tw = di->WinCaption;
	return ((x >= tw->left) && (x <= (tw->left + tw->width)) &&
		(y >= tw->top) && (y <= (tw->top + tw->height)));
}

void TDesktopIcon_DoMouseOver(TWins *self, XCrossingEvent *ev, bool Leave) {
  TDesktopIcon *di = get_owner_DesktopIcon(self);

	if (!di) return;
	if (Leave) {
	    if (!(di->WinCaption->KeyFu & mHasFocus) && 
		!TDesktopIcon_ChkMouseOverIcoCaption(di, ev->x_root, ev->y_root))
		    TDesktopIcon_SetFunkt(di, DIFu_Default);
	} else {
	    if (di->WinCaption->KeyFu & mHasFocus) {
		TDesktopIcon_SetFunkt(di, DIFu_Focus);
	    } else {
		TDesktopIcon_SetFunkt(di, DIFu_HighLight);
	    }
	}
	TWins_DoMouseOver(self, ev, Leave);
}

void TDesktopIconTw_DoFocus(TWins *self, int Change) {
	TWins_DoFocus(self, Change);
	int m = DIFu_Default;
	if (Change & WinsFocusIN) m = DIFu_Focus;
	TDesktopIcon_SetFunkt((TDesktopIcon*)self->owner, m);
}

static void TDesktopIcon_Restack(TDesktopIcon *di) {
  Window win[4];

	if (!di) return;
	if (di->glob->top_di && (di->glob->top_di != di)) {

	    win[0] = di->glob->top_di->WinID;
	    win[1] = di->glob->top_di->WinCaption->WinID;
	    win[2] = di->WinID;
	    win[3] = di->WinCaption->WinID;
	    XRestackWindows(IappXDat.dpy, win, 4);

	    win[0] = di->WinID;
	    win[1] = di->WinCaption->WinID;
	    win[2] = di->glob->top_di->WinID;
	    win[3] = di->glob->top_di->WinCaption->WinID;
	    XRestackWindows(IappXDat.dpy, win, 4);

	}
	di->glob->top_di = di;
}

static void TDesktopIcon_ClrMask(TDesktopIcon *di) {
	if (di->stat & wmDIst_MaskAkt) {
	    di->stat &= ~wmDIst_MaskAkt;
	    CombineMask(di->WinCaption->WinID, 0);
	}
}

void TDesktopIcon_SetFunkt(TDesktopIcon *di, int fu) {
  TPixMap *pxm;

	switch(fu & 0xffff) {

	case DIFu_Focus:
	    TDesktopIcon_ClrMask(di);
	    di->stat &= ~wmDIst_HlAkt;
	    di->WinCaption->vds.ForeColor = di->glob->TwFocFontColor;
	    SetColor(di->WinCaption, di->glob->TwFocColor);
	    break;

	case DIFu_HighLight:
	    TDesktopIcon_ClrMask(di);
	    TDesktopIcon_Restack(di);
	    if (di->glob->mod & wmDImod_HighLight) {
		di->stat |= wmDIst_HlAkt;
		di->WinCaption->vds.ForeColor = di->glob->TwHlFontColor;
		SetColor(di->WinCaption, di->glob->TwHlColor);
	    } else {
		di->WinCaption->vds.ForeColor = di->glob->TwFontColor;
		SetColor(di->WinCaption, di->glob->TwColor);
	    }
	    break;

	default:
	    if (di->glob->mod & wmDImod_Transparent) {
	      if (!(di->stat & wmDIst_MaskAkt) || (fu & DIFu_Force)) {
		di->stat |= wmDIst_MaskAkt;
		pxm = TPixMap_Create_depth(di->WinCaption->width, di->WinCaption->height, IappXDat.RootW_depth);
		TPixMap_CreateXimg(pxm);
		TPixMap_SetFont(pxm, di->glob->DesktopIcoFont);
		TPixMap_ClsRGB(pxm, 0xffffff);
		di->WinCaption->vds.ForeColor = 0x000000;
		TPixMap_DrawString(pxm, di->WinCaption->vds.s, 2, 0);
		TPixMap_DrawString(pxm, di->WinCaption->vds.s, 2, 2);
		TPixMap_DrawString(pxm, di->WinCaption->vds.s, 1, 1);
		TPixMap_DrawString(pxm, di->WinCaption->vds.s, 3, 1);
		TPixMap_DrawString(pxm, di->WinCaption->vds.s, 2, 1);
		TPixMap_XGetImage(pxm);

		TPixMap_SetMaskRGB(pxm, 0xffffff);
		TPixMap_WinCombineMask(pxm, di->WinCaption->WinID);
		Tfree(pxm);
		di->WinCaption->vds.ForeColor = 0xffffff;
		SetColor(di->WinCaption, 0x000000);
	      }
	    } else {
		TDesktopIcon_ClrMask(di);
		di->WinCaption->vds.ForeColor = di->glob->TwFontColor;
		SetColor(di->WinCaption, di->glob->TwColor);
	    }
	    di->stat &= ~wmDIst_HlAkt;
	    break;
	}

}

void TDesktopIcon_ScanMousePos(TWins *self, XMotionEvent *ev, Tapph *app) {
  TDesktopIcon *di;
  TWins *tw;

	while (app) {
	    if (app->wins->wtyp == wtyp_TwmDeskIcon) {
		di = (TDesktopIcon*)app->wins;
		tw = di->WinCaption;
		if (TDesktopIcon_ChkMouseOverIcoCaption(di, ev->x, ev->y)) {
		    if (tw->KeyFu & mHasFocus) {
			TDesktopIcon_SetFunkt(di, DIFu_Focus);
		    } else {
			TDesktopIcon_SetFunkt(di, DIFu_HighLight);
		    }
		    return;
		} else {
		    if (!(tw->KeyFu & mHasFocus))
			    TDesktopIcon_SetFunkt(di, DIFu_Default);
		}
	    }
	    app = app->next;
	}
}

void TDesktopIcon_SetSize(TDesktopIcon *di) {
  TWins *tw = di->WinCaption;
  int w;

	w = TextWidthColorString(&tw->vds) + 4;
	if (w < 32) w = 32;
	di->xo = (di->width - w) / 2;
	di->yo = di->height + 2;
	SetRC(tw, di->left + di->xo, di->top + di->yo, w, tw->vds.Font->height + 2);
}
/*_____________________________ Destroy _________________________________*/

bool TDesktopIcon__Destroy(TDesktopIcon *self) {
	if (self->WinCaption) Tfree(self->WinCaption);
	self->WinCaption = NULL;
	xfree(self->fname);
	xfree(self->s2);
	if (self->glob->top_di == self) self->glob->top_di = NULL;
	return TWins__Destroy((TWins*)self);
}


void TDesktopIcon_Destroy_fname(char *fn) {
  Tapph *ap = wmRoot->apph->next_child;
  TDesktopIcon *wmdi;

	while (ap) {
	  if (ap->wins->wtyp == wtyp_TwmDeskIcon) {
	    wmdi = (TDesktopIcon*)ap->wins;
	    if (pccmp(wmdi->fname, fn)) {
		Tfree(wmdi);
		return;
	    }
	  }
	  ap = ap->next;
	}
}

/*_____________________________ Create __________________________________*/
TDesktopIcon *TDesktopIcon_Create(TwmRoot *wmr, char *fname) {
  TDesktopIcon *self;
  TWins_Create_struct wcs;
  TWins *tw;
  int ddkt;
  char *pc, *fn, *name, *icon;
  TpcIni *inid = NULL;
  TPic *pic;

	ddkt = ddkt_desktop(fname);
	if (ddkt < 1) return NULL;
	self = NULL;

	fn = GetFileName_pc(fname);
	inid = TpcIni_Create(fname);
	name = TpcIni_RD_pc(inid, Desktop_Entry, "Name", fn);
	icon = TpcIni_RD(inid, Desktop_Entry, "Icon");
	if (!name) goto Ende;

	pic = GetDeskIcon(ddkt, 32, icon, gdifu_getico0, NULL);
	if (!pic) goto Ende;

	TWins_Create_struct_init(&wcs, sizeof(TDesktopIcon), wtyp_TwmDeskIcon);
	wcs.Parent = (TWins*)wmr;
	wcs.CreateXWin = NULL;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;
	self->fname = pcdup(fname);

	self->glob = &wmr->di_glob;
	if (!self->glob->DesktopIcoFont) self->glob->DesktopIcoFont = IappXDat.Default_Font;
	self->destroy = (TDESTROY)TDesktopIcon__Destroy;
	self->DoPaint = (TWinsEv)TDesktopIcon_DoPaint;
	self->DoMouseDown = TDesktopIcon_DoMouseDown;
	self->DoMouseUp = TDesktopIcon_DoMouseUp;
	self->DoMouseMove = TDesktopIcon_DoMouseMove;
	self->DoMouseOver = TDesktopIcon_DoMouseOver;

	tw = TWins_Create((TWins*)wmr);
	self->WinCaption = tw;
	tw->owner = (TWins*)self;
	tw->vds.Font = self->glob->DesktopIcoFont;
	tw->DoPaint = TDesktopIconTw_DoPaint;
	tw->DoMouseDown = TDesktopIcon_DoMouseDown;
	tw->DoMouseUp = TDesktopIcon_DoMouseUp;
	tw->DoMouseOver = TDesktopIcon_DoMouseOver;
	tw->DoFocus = TDesktopIconTw_DoFocus;
	tw->DoMouseMove = TDesktopIcon_DoMouseMove;
	tw->PopupMenu = wmr->menu_Default_DeskIcon;


	self->ddkt = ddkt;
	self->PopupMenu = wmr->menu_Default_DeskIcon;
	if (self->ddkt == ddkt_FSDevice) {
	    self->pic3 = pic; //hdd_unmount
	    self->pic2 = GetDeskIcon(ddkt, 32, icon, gdifu_getico1, NULL); //hdd_mount
	    self->pic4 = GetDeskIcon(ddkt, 32, icon, gdifu_getico2, NULL); //hdd_iseject
	    self->pic = self->pic4; //hdd_iseject

	    self->s2 = TpcIni_RDstrdup(inid, Desktop_Entry, "Dev", NULL);
	    char *tpc = self->s2;
	    while (*tpc && (*tpc != '/')  && (*tpc != ':')) tpc++;
	    if ((*tpc == ':') && (tpc[1] == '/')) {
		add_nfs_info(self->s2);
	    } else {
		add_fs_info(GetFileName_pc(self->s2), NULL);
	    }

	} else {
	    self->pic = pic;
	}

  bool ok;
  char line[200];
  TpcIni *inisz = TpcIni_Create(wmr->IconPosFile);
	snprintf(line, sizeof(line), "%dx%d", dpyMonitors.mon[0].right, dpyMonitors.mon[0].bottom);
	ok = False;
	pc = TpcIni_RD(inisz, fn, line);
	if (pc) {
	    self->left = ppctoint(&pc);
	    self->top = ppctoint(&pc);
	    ok = di_ChkDeskPos(self, wmr, self->left, self->top);
	}
	if (!ok) di_GetNewDeskPos(self, wmr, &self->left, &self->top);
	Tfree(inisz);

	if (self->left > di_GetMaxRight()) self->left = di_GetMaxRight();
	if (self->top > di_GetMaxBottom()) self->top = di_GetMaxBottom();
	if (self->left < 0) self->left = 0;
	if (self->top < 0) self->top = 0;

	self->width = 32;
	self->height = 32;
	TWins_CreateXWin(&wcs);

	SetCaption(tw, name);

	Set_Parent((TWins*)self, (TWins*)wmr);
	Set_Parent(tw, (TWins*)wmr);

	TDesktopIcon_SetSize(self);
	TDesktopIcon_SetFunkt(self, DIFu_Default);
	XLowerWindow(IappXDat.dpy, self->WinID);
	XLowerWindow(IappXDat.dpy, self->WinCaption->WinID);

Ende:
	Tfree(inid);
	return self;
}

void DeskIcons_OnFileMsg(TFileMsg *fmsg) {
  TwmRoot *wmr;
  char *rpe;
  int bef;
  TDirFileList *dfl;
  char line[1024];

	bef = get_bef_dflEv(fmsg);
	if (!bef) return;
	dfl = fmsg->F.dfl;
	wmr = ((TDirListObjNy*)fmsg->obj)->owner;

	rpe = line;
	ppcmove(&rpe, wmr->DesktopPath);
	ppcmove(&rpe, dfl->fn);

	switch (bef) {
	case dlbef_Add:
	    TDesktopIcon_Create(wmr, line);
	    break;
	case dlbef_WrClose:
	    TDesktopIcon_Destroy_fname(line);
	    TDesktopIcon_Create(wmr, line);
	    break;
	case dlbef_Del:
	    TDesktopIcon_Destroy_fname(line);
	    break;
	}
}


void TwmRoot_InitDeskIcons(TwmRoot *wmr) {
  TDirListObjNy *ObjNy;

	ObjNy = TDirListObjNy_Create(wmr);
	ObjNy->OnFileMsg = DeskIcons_OnFileMsg;
	ObjNy->findAtr = findREG;
	if (!TDirListObjNy_Open(ObjNy, wmr->DesktopPath)) {
	    Tfree(ObjNy);
	} else {
	    wmr->ObjNy_DeskIcons = ObjNy;
	}
}
