/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

static bool ChkIsFrame(TwmFrame *wmf) {
	return wmf && (wmf->wtyp == wtyp_TwmFrame);
}

static bool ChkIsFrameCurrentDesk(TwmFrame *wmf) {
	return wmf && (wmf->wtyp == wtyp_TwmFrame)
		   && wmf->Visible
		   && (wmRoot->CurrentDesk == wmf->wmWin->Desk);
}

static void TwmFrame_SetNextFocusWin(TwmFrame *self) {
  Tapph *ap = IappXDat.apph;
  Display *dpy = IappXDat.dpy;
  XID wmWinID;
  XEvent tmp_ev;

	GrabXServer();
	while (ap) {
	    if (ChkIsFrameCurrentDesk((TwmFrame*)ap->wins)) {
		wmWinID = ((TwmFrame*)ap->wins)->wmWinID;
		if ((XCheckTypedWindowEvent(dpy, wmWinID, DestroyNotify, &tmp_ev)) || 
		    (XCheckTypedWindowEvent(dpy, wmWinID, UnmapNotify, &tmp_ev))) {
		  XPutBackEvent(dpy, &tmp_ev);
		} else {
		    goto Ende;
		}
	    }
	    ap = ap->next;
	}
	wmWinID = wmRoot->WinID;
	wmRoot->Focus_wmWin = NULL;
Ende:
// Debug_msg(0, "%s(%d) XSetInputFocus w=0x%x\n", FL, wmWinID);
	SetXInputFocus(wmWinID);
	UngrabXServer();
}

void TwmFrame_GetClientRect(TWins *self, TWRect *rc) {
  int m = ((TwmFrame*)self)->Border_width;
	rc->left = m;
	rc->top = m;
	m += m;		//Border_width * 2
	rc->width = self->width - m;
	rc->height = self->height - m;
}

static void TwmFrame_VisibleStat(TwmFrame *self, bool Visible) {
  int State = NormalState;
	if (!Visible) State = IconicState;
	TWins_SetVisible((TWins*)self, Visible);
	TwmWin_SetMapState(self->wmWin, State);
}

static void TwmFrame_transientfor_SetVisible(Window win, bool Visible) {
  Tapph *ap;
  TwmFrame *wmf;

	ap = IappXDat.apph;
	while (ap) {
	    wmf = (TwmFrame*)ap->wins;
	    if ((wmf->wtyp == wtyp_TwmFrame) && wmf->transientfor) {
		if (wmf->transientfor == win) {
		    TwmFrame_transientfor_SetVisible(wmf->wmWinID, Visible);
		    TwmFrame_VisibleStat(wmf, Visible);
		}
	    }
	    ap = ap->next;
	}
}

void TwmFrame_SetVisible(TwmFrame *self, bool Value) {
  TwmWin *wmw = self->wmWin;

	if (self->Visible == Value) return;
// Debug_msg(0, "%s(%d) Set Visible 0x%x %d\n", FL, (int)wmw->WinID, Value);
	if (wmw->Desk == wmRoot->CurrentDesk) {
	    TwmFrame_transientfor_SetVisible(wmw->WinID, Value);
	    if (self->Visible) {
		TwmWin_GrabFocusWin(wmw, True);
		self->Visible = False;	//skip self for SetNextFocusWin
		TwmFrame_SetNextFocusWin(self);
		self->Visible = True;
		TTaskbar_buWin_SetUpDown(self->buTaskbarWin, False);
	    }
	    TwmFrame_VisibleStat(self, Value);
	} else {
	    if (Value) {
		self->Visible = True;
	    } else {
		TWins_SetVisible((TWins*)self, Value);
	    }
	}
}

void TwmFrame_SetCaption(TwmFrame *self, char *s) {
	if ((self->wmWin->tflags & tfl_TITLE) && self->wmTitle) {
	    vds_SetStr(&self->vds, s);
	    SetCaption((TWins*)self->wmTitle, s);
	    if (self->buTaskbarWin) SetCaption((TWins*)self->buTaskbarWin, s);
	}
}

static void TwmFrame_Cursor(TwmFrame *self, int x, int y) {
  int lr;
  int curs;

  if (self->min_width  && (self->min_width == self->max_width) &&
      self->min_height && (self->min_height == self->max_height)) return;

	curs = XC_top_left_arrow;
	lr = 16;
	if (x <= self->Border_width) {
	  curs = XC_left_side;
	  if (y < lr) curs = XC_top_left_corner;
	  if (y >= self->height - lr) curs = XC_bottom_left_corner;

	} else if (y <= self->Border_width) {
	  curs = XC_top_side;
	  if (x < lr) curs = XC_top_left_corner;
	  if (x >= self->width - lr) curs = XC_top_right_corner;

	} else if (x >= (self->width - self->Border_width)) {
	  curs = XC_right_side;
	  if (y < lr) curs = XC_top_right_corner;
	  if (y >= (self->height - lr)) curs = XC_bottom_right_corner;

	} else if (y >= (self->height - self->Border_width)) {
	  curs = XC_bottom_side;
	  if (x < lr) curs = XC_bottom_left_corner;
	  if (x >= (self->width - lr)) curs = XC_bottom_right_corner;
	}

	Set_Cursor((TWins*)self, curs);
} 

static bool TwmFrame_ChkSizeCursor(TwmFrame *self) {
  int cur = self->cursor_nr;
	return ((cur == XC_left_side) || 
		(cur == XC_top_left_corner) ||
		(cur == XC_top_side) || 
		(cur == XC_top_right_corner) ||
		(cur == XC_right_side) || 
		(cur == XC_bottom_right_corner) ||
		(cur == XC_bottom_side) || 
		(cur == XC_bottom_left_corner));
}

void TwmFrame_DoMouseOver(TwmFrame *self, XCrossingEvent *ev, bool Leave) {
	if (!Leave) TwmFrame_Cursor(self, ev->x, ev->y);
}

void TwmFrame_DoMouseMove(TwmFrame *self, XMotionEvent *ev) {
  int x, y, mx, my, mw, mh;

	if ((ev->state & (Button1Mask | Button2Mask | Button3Mask | Button4Mask | Button5Mask)) == 0) {
	    TwmFrame_Cursor(self, ev->x, ev->y);
	} else if ((ev->state & Button1Mask) && 
		   (IappXDat.MouseDown.apph == self->apph)) {
	    x = ev->x_root - IappXDat.MouseDown.x_root;
	    y = ev->y_root - IappXDat.MouseDown.y_root;
	    mx = IappXDat.MouseDown.obj_left;
	    my = IappXDat.MouseDown.obj_top;
	    mw = IappXDat.MouseDown.obj_width;
	    mh = IappXDat.MouseDown.obj_height;

	    if (self->cursor_nr == XC_left_side) {
		mx += x; mw -= x;
	    } else if (self->cursor_nr == XC_top_left_corner) {
		mx += x; mw -= x;
		my += y; mh -= y;
	    } else if (self->cursor_nr == XC_top_side) {
		my += y; mh -= y;
	    } else if (self->cursor_nr == XC_top_right_corner) {
		my += y; mh -= y;
		mw += x;
	    } else if (self->cursor_nr == XC_right_side) {
		mw += x;
	    } else if (self->cursor_nr == XC_bottom_right_corner) {
		mw += x;
		mh += y;
	    } else if (self->cursor_nr == XC_bottom_side) {
		mh += y;
	    } else if (self->cursor_nr == XC_bottom_left_corner) {
		mx += x; mw -= x;
		mh += y;
	    } else {
		SetRC((TWins*)self, IappXDat.MouseDown.obj_left + x, IappXDat.MouseDown.obj_top + y, self->width, self->height);
		return;
	    }
	    TwmWin_ChkSize(self->wmWin, &mw, &mh);
	    self->final_x = mx;
	    self->final_y = my;
	    self->final_w = mw;
	    self->final_h = mh;
	    MoveOutline(IappXDat.RootW, mx, my, mw, mh);
	}
}

void TwmFrame_DoMouseDown(TwmFrame *self, XButtonEvent *ev) {
  Display *dpy = IappXDat.dpy;
  dword ahID;

	if (wmRoot->Focus_wmWin != self->wmWin) {
// Debug_msg(0, "%s(%d) XSetInputFocus w=0x%x\n", FL, self->wmWinID);
	    ahID = self->apph->ahID;
	    SyncX();
	    GrabXServer();
	    TwmFrame_SetOnTop(self, 0x00);
	    SyncX();
	    AppProcessMessages();
	    UngrabXServer();
//	    SyncX();
	    if (!Find_ahID(IappXDat.apph, ahID)) {
  printf("%s(%d): TwmFrame_DoMouseDown: wmFrame Destroy ???\n", FL);
		return;
	    }
	    SetXInputFocus(self->wmWinID);
	    XAllowEvents(dpy, ReplayPointer, 0);
	    return;
	}
	TWins_DoMouseDown((TWins*)self, ev);

	if (ev->button != Button1) return;
	if (TwmFrame_ChkSizeCursor(self)) {
	    self->final_x = IappXDat.MouseDown.obj_left;
	    self->final_y = IappXDat.MouseDown.obj_top;
	    self->final_w = IappXDat.MouseDown.obj_width;
	    self->final_h = IappXDat.MouseDown.obj_height;
	    MoveOutline(IappXDat.RootW, self->final_x, self->final_y, self->final_w, self->final_h);
	    XChangeActivePointerGrab(dpy, ButtonReleaseMask | PointerMotionMask, 
			IappXDat.XCursors[self->cursor_nr], 0);
	    IappXDat.GrabWin = (TWins*)self;
	}
}

void TwmFrame_DoMouseUp(TwmFrame *self, XButtonEvent *ev) {

	TWins_DoMouseUp((TWins*)self, ev);
	if (ev->button != Button1) return;

	if (TwmFrame_ChkSizeCursor(self)) {
	    MoveOutline(IappXDat.RootW, 0, 0, 0, 0);
	    TwmWin_SetupSize(self->wmWin, self->final_x, self->final_y, 
		    self->final_w, self->final_h, wmwPos_SendPos);
	} else if (self->cursor_nr == XC_fleur) {
	    TwmWin_SetupSize(self->wmWin, self->left, self->top, 
		    self->width, self->height, wmwPos_SendPos);
	}
	Std_UngrabPointer();
	Set_Cursor((TWins*)self, XC_top_left_arrow);
}


void TwmFrame_ClrAll_Hilite(TwmFrame *wmf) {
  Tapph *ap;
  TwmFrame *twmf;

	ap = IappXDat.apph;
	while (ap) {
	    twmf = (TwmFrame*)ap->wins;
	    if ((twmf->wtyp == wtyp_TwmFrame) && (twmf != wmf) && twmf->wmTitle) {
		if (twmf->wmTitle->flags & wmtfl_hilite) {
		    TwmWin_GrabFocusWin(twmf->wmWin, True);
		    TwmTitle_SetHilite(twmf->wmTitle, False);
		}
		TTaskbar_buWin_SetUpDown(twmf->buTaskbarWin, False);
	    }
	    ap = ap->next;
	}
}

TwmFrame *GetTop_DeskWin(TwmFrame *wmf) {
  Tapph *ap;

	if (ChkIsFrameCurrentDesk(wmRoot->Top_wmf)) return wmRoot->Top_wmf;

	ap = IappXDat.apph->next;
	while (ap) {
	    if (!(ap->aos & aosOnTop)) {
		TwmFrame *twmf = (TwmFrame*)ap->wins;
		if ((twmf != wmf) && ChkIsFrameCurrentDesk(twmf)) return twmf;
	    }
	    ap = ap->next;
	}

	ap = IappXDat.apph->next;
	while (ap) {
	    TwmFrame *twmf = (TwmFrame*)ap->wins;
	    if ((twmf != wmf) && ChkIsFrameCurrentDesk(twmf)) return twmf;
	    ap = ap->next;
	}
	return NULL;
}

typedef struct {
  Tapph **ppapi;
  Window *pwini;
  int wcount;
} TAddToRestack;

static Tapph *TwmFrame_Getapp(Window win) {
  Tapph *ap = IappXDat.apph->next;
	while (ap) {
	    if ((ap->win == win) || 
		((ap->wins->wtyp == wtyp_TwmFrame) && 
		 (((TwmFrame*)ap->wins)->wmWinID == win))) break;
	    ap = ap->next; 
	}
	return ap;
}

static void AddToRestack(Tapph *ap, TAddToRestack *atrck) {
	if (!ap || (ap->aos & aosRestackMk)) return;
	ap->aos |= aosRestackMk;
	*atrck->ppapi = ap;
	atrck->ppapi++;
	*atrck->pwini = ap->win;
	atrck->pwini++;
	atrck->wcount++;
// Debug_msg(0,"onTop %d 0x%x\n", atrck->wcount, ap->win);
}

static void AddTransientFor(Window win, TAddToRestack *atrck) {
  Tapph *ap;

	if (!win) return;
	ap = IappXDat.apph->next;
	while (ap) {
	    if (!(ap->aos & aosRestackMk) && 
		(ChkIsFrameCurrentDesk((TwmFrame*)ap->wins)) &&
		(((TwmFrame*)ap->wins)->transientfor == win)) {
		AddTransientFor(((TwmFrame*)ap->wins)->wmWinID, atrck);
		AddToRestack(ap, atrck);
	    }
	    ap = ap->next;
	}
}

static void AddTopFrame(Tapph *ap, TAddToRestack *atrck, int fu) {

	if (!ap) return;
	if (!(fu & 0x01)) {
apt_loop:
	    if (ChkIsFrame((TwmFrame*)ap->wins) &&
		((TwmFrame*)ap->wins)->transientfor) {
		Tapph *apt = TwmFrame_Getapp(((TwmFrame*)ap->wins)->transientfor);
		if (apt) {
		    ap = apt;
		    goto apt_loop;
		}
	    }

	    if (ChkIsFrame((TwmFrame*)ap->wins)) {
		AddTransientFor(((TwmFrame*)ap->wins)->wmWinID, atrck);
	    }
	}
	AddToRestack(ap, atrck);
}


void TwmFrame_SetOnTop(TwmFrame *wmf, int fu) {
  Tapph *ap, *apTop, *apt, **ppap;
  Window *pwin;
  TAddToRestack atrck;
  int wmaxcount;

	wmRoot->Top_wmf = wmf;
	wmaxcount = 0;
	ap = IappXDat.apph->next;
	while (ap) {ap->aos &= ~aosRestackMk; ap = ap->next; wmaxcount++;}
	if (wmaxcount < 2) return;

	ppap = xmalloc(sizeof(void*) * (wmaxcount+2));
	pwin = xmalloc(sizeof(void*) * (wmaxcount+2));
	atrck.ppapi = ppap;
	atrck.pwini = pwin;
	atrck.wcount = 0;

	IappXDat.apph->aos |= aosRestackMk;
	*atrck.ppapi = IappXDat.apph;         //add wmRoot
	atrck.ppapi++;

	if (wmRoot->Taskbar && wmRoot->Taskbar->Visible) {
	    ap = TwmFrame_Getapp(wmRoot->Taskbar->WinID);
	    if (ap) AddToRestack(ap, &atrck);
	}

	apTop = NULL;
	if (wmf) apTop = TwmFrame_Getapp(wmf->WinID);
	if (apTop && (apTop->aos & aosOnTop)) {
	    AddTopFrame(apTop, &atrck, fu);
	}

	ap = IappXDat.apph->next;
	while (ap) {
	    if ((ap->aos & aosOnTop) && !(ap->aos & aosRestackMk)) {
		if (ChkIsFrameCurrentDesk((TwmFrame*)ap->wins)) {
		    AddToRestack(ap, &atrck);
		}
	    }
	    ap = ap->next;
	}

	AddTopFrame(apTop, &atrck, fu);

	*atrck.pwini = 0;
	XRaiseWindow(IappXDat.dpy, *pwin);
	XRestackWindows(IappXDat.dpy, pwin, atrck.wcount);

	ap = IappXDat.apph->next;
	while (ap) {
	    if (!(ap->aos & aosRestackMk)) {
		*atrck.ppapi = ap;
		atrck.ppapi++;
// atrck.wcount++;
	    }
	    ap = ap->next;
	}
	*atrck.ppapi = NULL;

	Tapph **ppapn = ppap;
	apt = NULL;
	ap = *ppapn;
	IappXDat.apph = ap;
	while (ap) {
	    ap->prev = apt;
	    apt = ap;
	    ppapn++;
	    ap = *ppapn;
	    apt->next = ap;
	}

// Debug_msg(0,"SetOnTop %d %d\n", wmaxcount, atrck.wcount);

/*
#ifdef	WTestWin
 int i;
 char bufint[32];
 char *s;
 Window *nwin = pwin;
 
  WrInteger(atrck.wcount, &bufint[0]);
  s = pcdup(&bufint[0]);
  ppcadd(&s, " ");
  for (i=0;i<atrck.wcount; i++) {
    WrInteger((int)*nwin, &bufint[0]);
    ppcadd(&s, &bufint[0]);
    ppcadd(&s, " ");
    nwin++;
  }
 WTest_msg_Add("RS %s", s);
 xfree(s);
#endif
*/
	xfree(pwin);
	xfree(ppap);
}


/*____________________________ Destroy __________________________________*/
bool TwmFrame__Destroy(TwmFrame *self) {
	if (wmRoot->Top_wmf == self) wmRoot->Top_wmf = NULL;
	TTaskbar_DelWin(wmRoot->Taskbar, self);
	self->Visible = False;
	TwmFrame_SetNextFocusWin(self);
	self->wmWin = NULL;
	TWins__Destroy((TWins*)self);
// Debug_msg(0, "%s(%d) Destroy TwmFrame\n", FL);
	return False;
}

/*____________________________ Create ___________________________________*/
TwmFrame *TwmFrame_Create(TwmWin *wmw, XWindowAttributes *attr) {
  TwmFrame *self;
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TwmFrame), wtyp_TwmFrame);
	wcs.CreateXWin = NULL;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;

	self->owner = (TWins*)wmw;
	self->wmWin = wmw;
	self->wmWinID = wmw->WinID;
	self->top = attr->y;
	self->left = attr->x;
	self->width = attr->width;
	self->height = attr->height;

	if (wmw->tflags & tfl_BORDER) {
	    self->Border_width = glob_Border_width;
	    self->width += glob_Border_width << 1;
	    self->height += glob_Border_width << 1;
	    self->BorderFu = mBorder | mBorderW2;
	}
	if (wmw->tflags & tfl_TITLE) {
	    self->Caption_height = glob_Caption_height;
	    self->height += glob_Caption_height;
	}

	self->apph->aos |= aosIsForm;
	if (!(wmw->tflags & tfl_BORDER)) self->Paint = NULL;

	if (XGetTransientForHint(IappXDat.dpy, wmw->WinID,  &self->transientfor) != 0) {
	    if ((self->transientfor == IappXDat.RootW) || 
		(self->transientfor == wmw->WinID)) self->transientfor = 0;
	    if (self->transientfor) wmw->tflags &= ~tfl_MINIMIZE;
	}

	self->destroy = (TDESTROY)TwmFrame__Destroy;
	self->GetClientRect = TwmFrame_GetClientRect;
	self->SetVisible = (TSetVisibleEv)TwmFrame_SetVisible;
	self->SetCaption = (TSetCaptionEv)TwmFrame_SetCaption;

	self->DoMouseMove = (TMmoveEv)TwmFrame_DoMouseMove;
	self->DoMouseOver = (TMoverEv)TwmFrame_DoMouseOver;
	self->DoMouseDown = (TMupdownEv)TwmFrame_DoMouseDown;
	self->DoMouseUp = (TMupdownEv)TwmFrame_DoMouseUp;


	wcs.XAttr.event_mask = (SubstructureRedirectMask | ButtonPressMask |
			   ButtonReleaseMask | EnterWindowMask |
			   LeaveWindowMask | ExposureMask | PointerMotionMask);

	if (attr->save_under != 0) {
	    wcs.XAttr.save_under = 1;
	    wcs.XMask |= CWSaveUnder;
	}
#ifdef USE_DECOR
	if (wmw->tflags & tfl_BORDER) self->Decor = TDecor_dup(&wmRoot->Default_Decor);
#endif
	self->Visible = False;
	TWins_CreateXWin(&wcs);

	Set_Parent((TWins*)self, NULL);

	if (wmw->tflags & tfl_TITLE) {
	    self->wmTitle = TwmTitle_Create(self);
	    if (wmw->tflags & tfl_MINIMIZE) TTaskbar_AddWin(wmRoot->Taskbar, self);
	    if (wmw->tcreate) SetCaption((TWins*)self, wmw->tcreate->WMName);
	}

	return self;
}
