/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

#ifdef	USE_DEBUG_WIN
extern int GrabXServer_count;
#endif

Tapph *TwmWin_Find_WinsX_Ev(Tapph *ap, Window win) {
  Tapph *apn;

	while (ap) {
	    apn = ap->next_child;
	    if (apn && ap->wins && (ap->wins->wtyp == wtyp_TwmFrame)) {
		while (apn) {
		    if ((apn->win == win) && apn->wins && 
			(apn->wins->wtyp == wtyp_TwmWin)) return apn;
		    apn = apn->next;
		}
	    }
	    ap = ap->next;
	}
	return NULL;
}

void INTERNAL_FUNC TwmWin_SetMapState(TwmWin *wmWin, int State) {
  int state_data[2];
	    state_data[0] = State;
	    state_data[1] = 0;
	    XChangeProperty(IappXDat.dpy, wmWin->WinID, 
			IappXDat.XA_WM_STATE, IappXDat.XA_WM_STATE, 32,
			PropModeReplace, (unsigned char *)state_data, 2);
}


void TwmWin_Close(TwmWin *wmWin) {
  XClientMessageEvent ev;
	if (!wmWin) return;
	Fill0(&ev, sizeof(ev));
	ev.type = ClientMessage;
	ev.window = wmWin->WinID;
	ev.message_type = IappXDat.XA_WM_PROTOCOLS;
	ev.format = 32;
	ev.data.l[0] = IappXDat.XA_WM_DELETE_WINDOW;
	XSendEvent(IappXDat.dpy, ev.window, 0, 0, (void*)&ev);
}

void TwmWin_GrabFocusWin(TwmWin *wmWin, bool grab) {
	if (grab) {
	    if (!(wmWin->flags & fl_GrabButton)) {
		XGrabButton(IappXDat.dpy, AnyButton, AnyModifier, 
			wmWin->Frame->WinID, 0,
			ButtonPressMask, GrabModeSync, GrabModeAsync, None, None);
		wmWin->flags |= fl_GrabButton;
	    }
	} else {
	    wmRoot->FocusWin = NULL;
	    if (wmWin->flags & fl_GrabButton) {
		XUngrabButton(IappXDat.dpy, AnyButton, AnyModifier, 
			wmWin->Frame->WinID);
		wmWin->flags &= ~fl_GrabButton;
	    }
	}
}

void TwmWin_DoFocus(TwmWin *self, int Change) {
  TwmFrame *wmf;

	if (!(Change & WinsFocusIN)) return;

	wmf = self->Frame;
	TwmWin_GrabFocusWin(self, False);
	if (wmRoot->Focus_wmWin != self) {
	    TwmFrame_ClrAll_Hilite(wmf);
	    if (wmRoot->Top_wmf != wmf) {
		 TwmFrame_SetOnTop(wmf, 0x00);
	    }
	}
	wmRoot->Focus_wmWin = self;
	if (wmf->wmTitle) TwmTitle_SetHilite(wmf->wmTitle, True);
	TTaskbar_buWin_SetUpDown(wmf->buTaskbarWin, True);
}

void INTERNAL_FUNC TwmWin_CalcGravity(TwmWin *self, int *x, int *y) {
  int bw2, m;
  TwmFrame *wmf;

	if (!self || !self->Frame || !(self->tflags & tfl_NOAUTOPOS)) return;
 Debug_msg(0, "WinGravity grav=%d x=%d y=%d --> ", self->WinGravity, *x, *y);
	wmf = self->Frame;
	bw2 = (wmf->Border_width - self->border_width) * 2;
	m= bw2 + wmf->Caption_height;
	if (wmRoot->Taskbar) m += wmRoot->Taskbar->height;
	switch(self->WinGravity) {
	case NorthEastGravity:
	case EastGravity:
	    *x -= bw2;
	    break;
	case SouthWestGravity:
	case SouthGravity:
	    *y -= m;
	    break;
	case SouthEastGravity:
	    *x -= bw2;
	    *y -= m;
	    break;
//	case StaticGravity:
//		*x -= bw2;
//		*y -= chbw1;
//		break;
	}
 Debug_msg(0, "x=%d y=%d\n", *x, *y);
}

void TwmWin_ChkSize(TwmWin *wmw, int *width, int *height) {
  int bw2, w, h;
	bw2 = wmw->Frame->Border_width << 1;
	w = *width - bw2;
	h = *height - bw2 - wmw->Frame->Caption_height;
	TWins_ChkWinSizeMinMax((TWins*)wmw, &w, &h);

	if (w < 84) w = 84;
	w += bw2;
	h += bw2 + wmw->Frame->Caption_height;
	if (w > IappXDat.Root_width) w = IappXDat.Root_width;
	if (h > IappXDat.Root_height) h = IappXDat.Root_height;
	*width = w;
	*height = h;
}

static void TwmWin_SendPos(TwmWin *wmw) {
  XConfigureEvent cEv;

	Fill0(&cEv, sizeof(cEv));
	cEv.type = ConfigureNotify;
	cEv.send_event = True;
	cEv.display = IappXDat.dpy;
	cEv.event   = wmw->WinID;
	cEv.window  = wmw->WinID;

	cEv.x = wmw->left + wmw->Frame->Border_width;
	cEv.y = wmw->top + wmw->Frame->Caption_height + wmw->Frame->Border_width;

	cEv.width  = wmw->width;
	cEv.height = wmw->height;
	XSendEvent(IappXDat.dpy, wmw->WinID, 0, StructureNotifyMask, (XEvent*)&cEv);
 Debug_msg(0, "sendEvent win=0x%x x=%d y=%d w=%d h=%d\n", wmw->WinID, cEv.x, cEv.y, cEv.width, cEv.height);
}

void TwmWin_SetupSize(TwmWin *wmw, int left, int top, int width, int height, int sendEvent) {
  int ww, wh, bw2;
  TwmFrame *wmf = wmw->Frame;
  bool rc_change;

	TwmWin_ChkSize(wmw, &width, &height);
	if (!wmf) return;

	bw2 = (wmf->Border_width << 1);
	ww = wmw->width;
	wh = wmw->height;
	wmw->width = width - bw2;
	if (wmw->width < 1) wmw->width = 1;
	wmw->height = height - wmf->Caption_height - bw2;
	if (wmw->height < 1) wmw->height = 1;

	if (top > (IappXDat.Root_height - 40)) top = IappXDat.Root_height - 40;
	if (top < -16) top = -16;
	if (left > (IappXDat.Root_width - 16)) left = IappXDat.Root_width - 16;
	if ((left + width - 16) < 0) left = 16 - width;

	rc_change = ((ww != wmw->width)  || (wh != wmw->height));
	if (rc_change) {
 Debug_msg(0, "SetupSize change 0x%x x=%d y=%d w=%d h=%d\n", wmw->WinID, left, top, width, height);
#ifdef SAVE_PARENT_WIN
	    XResizeWindow(IappXDat.dpy, wmw->SaveParentWinID, wmw->width, wmw->height);
	    XResizeWindow(IappXDat.dpy, wmw->WinID, wmw->width, wmw->height);
#else
	    XMoveResizeWindow(IappXDat.dpy, wmw->WinID, cx, cy, wmw->width, wmw->height);
#endif
	}

	if (sendEvent && ((wmw->left != left) || (wmw->top != top))) {
	    wmw->left = left;
	    wmw->top = top;
	    rc_change = True;
	}
	if (!rc_change) sendEvent &= ~wmwPos_SendPos;
	SetRC((TWins*)wmf, left, top, width, height);
	if (sendEvent) TwmWin_SendPos(wmw);
}

static __attribute__ ((noinline)) void TwmWin_GetWindowSizeHints(TwmWin *wmw) {
  long supplied = 0;
  TwmFrame *wmf = wmw->Frame;
  XSizeHints SizeHints;

	if (!XGetWMNormalHints(IappXDat.dpy, wmw->WinID, 
				&SizeHints, &supplied)) return;
	if (!SizeHints.flags) return;

	if (SizeHints.flags & PMinSize) {
 Debug_msg(0, "PMinSize w=%d h=%d\n", SizeHints.min_width, SizeHints.min_height);
	    wmw->min_width = SizeHints.min_width;
	    wmw->min_height = SizeHints.min_height;
	}
	if (SizeHints.flags & PMaxSize) {
 Debug_msg(0, "PMaxSize w=%d h=%d\n", SizeHints.max_width, SizeHints.max_height);
	    wmw->max_width = SizeHints.max_width;
	    wmw->max_height = SizeHints.max_height;
	}
//	if (SizeHints.flags & (PMinSize | PMaxSize)) {
//	    ChkWinSizeMinMaxVal(&wmw->max_width, &wmw->max_height);
//	}

	if (SizeHints.flags & (USPosition | PPosition)) {
 Debug_msg(0, "USPosition w=0x%x Flag=0x%x %dx%d wmf=0x%x\n", wmw->WinID, SizeHints.flags, SizeHints.x, SizeHints.y, wmf);
	    if (!wmf) wmw->tflags |= tfl_NOAUTOPOS;
	}

#ifdef	USE_DEBUG_WIN
	if (SizeHints.flags & (USSize | PSize)) {
 Debug_msg(0, "USSize w=0x%x Flag=0x%x %dx%d wmf=0x%x\n", wmw->WinID, SizeHints.flags, SizeHints.width, SizeHints.height, wmf);
	    if (!wmf) wmw->tflags |= tfl_NOAUTOPOS;
	}
#endif

	if (SizeHints.flags & PWinGravity) {
	    wmw->WinGravity = SizeHints.win_gravity;
 Debug_msg(0, "PWinGravity %d\n", wmw->WinGravity);
	}

	if (SizeHints.flags & (PMinSize | PMaxSize)) {
	  if (wmf) {
 Debug_msg(0, "(PMinSize | PMaxSize) w=0x%x Flag=0x%x %dx%d wmf=0x%x\n", wmw->WinID, SizeHints.flags, SizeHints.x, SizeHints.y, wmf);
	    TwmWin_SetupSize(wmw, wmf->left, wmf->top, wmf->width, wmf->height, wmwPos_SendPos);
	  }
	}
}

// ------------------------- _NET_WM_STATE ------------------------
void TwmWin_NetWMState(TwmWin *wmw, XClientMessageEvent *ev) {
  TwmFrame *wmf;
  bool above, below;

	above = below = False;
	for (int i=1;i<=2;i++) {
	    TNetWMnum NetWMnum = GetNetWMnum(ev->data.l[i]);
	    switch (NetWMnum) {
	    case NET_WM_NONE: 
#ifdef	USE_DEBUG_WIN
 if (ev->data.l[i] == NetWMAtom.XA_NET_WM_STATE_MAXIMIZED_VERT) Debug_msg(0, "NetWMState XA_NET_WM_STATE_MAXIMIZED_VERT\n");
 if (ev->data.l[i] == NetWMAtom.XA_NET_WM_STATE_MAXIMIZED_HORZ) Debug_msg(0, "NetWMState XA_NET_WM_STATE_MAXIMIZED_HORZ\n");
 if (ev->data.l[i] == NetWMAtom.XA_NET_WM_STATE_FULLSCREEN) Debug_msg(0, "NetWMState XA_NET_WM_STATE_FULLSCREEN\n");
#endif
		break;
	    case NET_WM_STATE_ABOVE: above = True; break;
	    case NET_WM_STATE_BELOW: below = True; break;
	    }
	}

 Debug_msg(0,"HandleClientMessage 0x%x 0x%x 0x%x -> %d %d\n", ev->data.l[0], ev->data.l[1], ev->data.l[2], above, below);

	wmf = wmw->Frame;
	if (!wmf) return;

	if (ev->data.l[0]) {
	    if (above && !(wmf->apph->aos & aosOnTop)) {
		wmf->apph->aos |= aosOnTop;
 Debug_msg(0,"Always on Top On  0x%x\n", wmw->WinID);
	    }

	} else {
	    if (above) {
		wmf->apph->aos &= ~aosOnTop;
 Debug_msg(0,"Always on Top Off 0x%x\n", wmw->WinID);
	    }
	}
}

static __attribute__ ((noinline)) void TwmWin_GetNetWMState(TwmWin *self) {
  int actual_format;
  Atom actual_type;
  unsigned long nitems, bytesafter;
  Atom *prop_ret;

	prop_ret = NULL;
	if (XGetWindowProperty(IappXDat.dpy, self->WinID, 
			IappXDat.XA_NET_WM_STATE, 0, 32, 0, XA_ATOM, 
			&actual_type, &actual_format, &nitems,
			&bytesafter, (void*)&prop_ret) != Success) return;

	for (int i=0;i<nitems;i++) {
	    TNetWMnum NetWMnum = GetNetWMnum(prop_ret[i]);

	    switch (NetWMnum) {
	    case NET_WM_NONE: 
#ifdef	USE_DEBUG_WIN
 if (prop_ret[i] == NetWMAtom.XA_NET_WM_STATE_MAXIMIZED_VERT) Debug_msg(0, "GetNetWMState XA_NET_WM_STATE_MAXIMIZED_VERT\n");
 if (prop_ret[i] == NetWMAtom.XA_NET_WM_STATE_MAXIMIZED_HORZ) Debug_msg(0, "GetNetWMState XA_NET_WM_STATE_MAXIMIZED_HORZ\n");
 if (prop_ret[i] == NetWMAtom.XA_NET_WM_STATE_FULLSCREEN) Debug_msg(0, "GetNetWMState XA_NET_WM_STATE_FULLSCREEN\n");
#endif
		break;
	    case NET_WM_STATE_ABOVE: self->tflags |= tfl_ABOVE; break;
	    case NET_WM_STATE_BELOW: break;
	    }
	}

 Debug_msg(0, "NET_WM_STATE items=%d 0x%06x\n", nitems, self->tflags);

	if (prop_ret) XFree(prop_ret);

}

// ------------------------- MOTIF_WM_HINTS ------------------------
static __attribute__ ((noinline)) void TwmWin_GetMotifWMHints(TwmWin *self) {
  int actual_format;
  Atom actual_type;
  unsigned long nitems, bytesafter;
  TMwmHints *MwmH;

	MwmH = NULL;
	if (XGetWindowProperty(IappXDat.dpy, self->WinID, IappXDat.XA_MOTIF_WM_HINTS, 
			0, 20, 0, IappXDat.XA_MOTIF_WM_HINTS, 
			&actual_type, &actual_format, &nitems,
			&bytesafter, (void*)&MwmH) != Success) return;

	if (MwmH && (nitems >= 5)) {
	    if ((MwmH->flags & MWM_HINTS_DECORATIONS) &&
		((MwmH->decorations & MWM_DECOR_ALL) == 0)) {
		if ((MwmH->decorations & MWM_DECOR_BORDER) == 0) {
		    self->tflags &= ~tfl_BORDER;
		}
		if ((MwmH->decorations & MWM_DECOR_TITLE) == 0) {
		    self->tflags &= ~tfl_TITLE;
		}
	    }
	}
	if (MwmH) XFree(MwmH);
}

// ------------------------- NET_WM_PID ------------------------
#ifdef FORM_NET_WM_PID
static __attribute__ ((noinline)) void TwmWin_GetNetWMPid(TwmWin *self) {
  int actual_format;
  Atom actual_type;
  unsigned long nitems, bytesafter;
  int *RetPid;
  TPid_Array *pida;
  Texec_pipe *expipe;

	RetPid = NULL;
	if (XGetWindowProperty(IappXDat.dpy, self->WinID, IappXDat.XA_NET_WM_PID, 
			0, 20, 0, XA_CARDINAL, 
			&actual_type, &actual_format, &nitems,
			&bytesafter, (void*)&RetPid) != Success) return;

	if (RetPid && nitems) self->pid_NET_WM = *RetPid;
	if (RetPid) XFree(RetPid);
	if (!self->pid_NET_WM) return;
 Debug_msg(0, "%s(%d): RetPid=%d\n", FL, *RetPid);

	pida = Get_pid_Handle(self->pid_NET_WM);
	if (!pida) return;
	expipe = pida->obj;
	if (expipe && (expipe->wtyp == wtyp_Texec_pipe) && expipe->fname) {
	    TpcIni *ini = TpcIni_Create(expipe->fname);
	    self->tcreate->icon_name = TpcIni_RDstrdup(ini, "Desktop Entry", "icon", NULL);
 Debug_msg(0, "icon_name=\"%s\"\n", self->tcreate->icon_name);
	    Tfree(ini);
	    ppcfree(&expipe->fname);
	}
}
#else
#define TwmWin_GetNetWMPid(a)
#endif /* FORM_NET_WM_PID */

// ------------------------- PropertyNotify ------------------------
void TwmWin_PropertyNotify(TwmWin *wmw, XPropertyEvent *ev) {
  XTextProperty text_prop;

#ifdef	USE_DEBUG_WIN
 char *pcx = XGetAtomName(IappXDat.dpy, ev->atom);
 Debug_msg(1, "%s(%d) PropertyNotify %x flags=%x %d %s\n", FL, wmw->WinID, wmw->flags, ev->atom, pcx);
 XFree(pcx);
#endif

	switch(ev->atom) {
	case XA_WM_HINTS:		//35
		break;

	case XA_WM_CLIENT_MACHINE:	//36
/*		if (XGetWMClientMachine(IappXDat.dpy, wmw->WinID, &text_prop) != 0) {
#ifdef	WTestWin
 WTest_msg_Add("%s(%d) XA_WM_CLIENT_MACHINE: %x %s", FL, wmw->WinID, text_prop.value);
#endif
			XFree(text_prop.value);
		} */
		break;

	case XA_WM_ICON_NAME:		//37
#ifdef	USE_DEBUG_WIN
		if (XGetWMIconName(IappXDat.dpy, wmw->WinID, &text_prop) != 0) {
 Debug_msg(1, "%s(%d) XA_WM_ICON_NAME: 0x%x %s\n", FL, wmw->WinID, text_prop.value);
			XFree(text_prop.value);
		}
#endif
		break;

	case XA_WM_NAME:		//39 
	    if (XGetWMName(IappXDat.dpy, wmw->WinID, &text_prop) != 0) {
		if (wmw->Frame) SetCaption((TWins*)wmw->Frame, (char*)text_prop.value);
/* #ifdef	USE_DEBUG_WIN
 Debug_msg(1, "%s(%d) XA_WM_NAME: %s\n", FL, text_prop.value);
#endif */
 Debug_msg(1, "%s(%d) XA_WM_NAME:      0x%x %s\n", FL, wmw->WinID, text_prop.value);
		XFree(text_prop.value);
	    }
	    break;

	case XA_WM_NORMAL_HINTS:	//40
#ifdef ALL_XwmError
		XwmError_Stop = True;
#endif
		TwmWin_GetWindowSizeHints(wmw);
#ifdef ALL_XwmError
		XwmError_Stop = False;
#endif
		break;
	default:
	    if (ev->atom == IappXDat.XA_NET_WM_USER_TIME) {
		if ((!wmw->Frame || !wmw->Frame->wmTitle) &&
		    (wmRoot->Focus_wmWin != wmw)) SetXInputFocus(wmw->WinID);
// Debug_msg(1, "%s(%d) XA_NET_WM_USER_TIME time=%x\n", FL, ev->time);
	    }


/* #ifdef	USE_DEBUG_WIN
 pcx = XGetAtomName(IappXDat.dpy, ev->atom);
 Debug_msg(1, "%s(%d) PropertyNotify %x flags=%x %d %s\n", FL, wmw->WinID, wmw->flags, ev->atom, pcx);
 XFree(pcx);
#endif */
		break;
	}
}

int TwmWin_EvProc(Tobjmsg *msg) {
  TWins *self;
  XEvent *ev;

	self = msg->ap->wins;
	if (msg->msg == msgXEv) {
	    ev = (XEvent*) msg->p;
	    switch(ev->type) {
	    case FocusIn:
		if (!ttc_chk_struct(self)) return 0;
		TwmWin_DoFocus((TwmWin*)self, WinsFocusIN);
		return 1;
//	    case FocusOut:
//		TwmWin_DoFocusOut(self);
//		return 1;
//	    case PropertyNotify:
//		TwmWin_PropertyNotify((TwmWin*)self, (XPropertyEvent*)ev);
//		return 1;
//	    default:
//		break;
	    }
	}

	return TWins_EvProc(msg);
}

/*____________________________ Destroy __________________________________*/
bool TwmWin__Destroy(TwmWin *self) {

	DelAppWin(self->apph);
	self->vds.lg = 0;
	xfree(self->vds.s);
#ifdef SAVE_PARENT_WIN
	if (self->SaveParentWinID) XDestroyWindow(IappXDat.dpy, self->SaveParentWinID);
#endif
	return False;
}

/*____________________________ Create __________________________________*/
void tmp_create_int(TwmWin *self) {
  Display *dpy = IappXDat.dpy;
  TwmWin_tcreate *tcreate = xcalloc(sizeof(TwmWin_tcreate));
  XTextProperty text_prop;

	self->tcreate = tcreate;
	if (XGetClassHint(dpy, self->WinID, &tcreate->classhint)) {
 Debug_msg(0, "xclass.res_name=\"%s\" xclass.res_class=\"%s\"\n", tcreate->classhint.res_name, tcreate->classhint.res_class);
	}
	tcreate->WMHints = XGetWMHints(dpy, self->WinID);
	if (tcreate->WMHints) {
	    if (tcreate->WMHints->flags & WindowGroupHint) self->window_group = tcreate->WMHints->window_group;
#ifdef	USE_DEBUG_WIN
  if (!(tcreate->WMHints->flags & StateHint)) tcreate->WMHints->initial_state = 0;
  if (!(tcreate->WMHints->flags & IconPixmapHint)) tcreate->WMHints->icon_pixmap = 0;
  if (!(tcreate->WMHints->flags & IconWindowHint)) tcreate->WMHints->icon_window = 0;
  if (!(tcreate->WMHints->flags & IconMaskHint)) tcreate->WMHints->icon_mask = 0;
  Debug_msg(0, "WMHints flg=0x%x st=%d Pixmap=0x%x iWin=0x%x Mask=0x%x group=0x%x\n", tcreate->WMHints->flags, tcreate->WMHints->initial_state, tcreate->WMHints->icon_pixmap, tcreate->WMHints->icon_window, tcreate->WMHints->icon_mask, self->window_group);
#endif
	}

	if (XGetWMName(dpy, self->WinID, &text_prop)) {
 Debug_msg(0, "XGetWMName=\"%s\"\n", text_prop.value);
	    tcreate->WMName = (char*)text_prop.value;
	}
	if (XGetWMIconName(dpy, self->WinID, &text_prop)) {
 Debug_msg(0, "XGetWMIconName=\"%s\"\n", text_prop.value);
	    tcreate->WMIconName = (char*)text_prop.value;
	}
}

void tmp_create_free(TwmWin *self) {
  TwmWin_tcreate *tcreate = self->tcreate;
	if (!tcreate) return;
	self->tcreate = NULL;
	xfree(tcreate->icon_name);
	if (tcreate->WMIconName) XFree(tcreate->WMIconName);
	if (tcreate->WMName) XFree(tcreate->WMName);
	if (tcreate->WMHints) XFree(tcreate->WMHints);
	if (tcreate->classhint.res_name) XFree(tcreate->classhint.res_name);
	if (tcreate->classhint.res_class) XFree(tcreate->classhint.res_class);

	xfree(tcreate);
}

TwmWin *TwmWin_Create_base(Window win, XWindowAttributes *attr) {
  int m;
  TwmWin *self;

 Debug_msg(0, "          ********* ADD NEW WIN = 0x%x grab_count=%d *********\n", (int)win, GrabXServer_count);
	self = ttc_create(sizeof(TwmWin), wtyp_TwmWin);
	self->WinID = win;
	self->Color = IappXDat.Default_Color;
	self->vds.Font = IappXDat.Default_Font;

	self->cursor_nr = XC_top_left_arrow;
	self->destroy = (TDESTROY)TwmWin__Destroy;
#ifdef ALL_XwmError
	XwmError_Stop = True;
#endif
	GrabXServer();
	m = XGetWindowAttributes(IappXDat.dpy, self->WinID, attr);
#ifdef ALL_XwmError
	XwmError_Stop = False;
#endif
	if (!m) {
	    UngrabXServer();
	    xfree(self);
	    return NULL;
	}
	XAddToSaveSet(IappXDat.dpy, self->WinID);

	self->left = attr->x;
	self->top = attr->y;
	self->width = attr->width;
	self->height = attr->height;
	self->border_width = attr->border_width;

 Debug_msg(0, "x=%d y=%d w=%d h=%d bw=%d amk=0x%x ymk=0x%x nmk=0x%x\n",
		attr->x, attr->y, attr->width, attr->height, attr->border_width,
		attr->all_event_masks, attr->your_event_mask, attr->do_not_propagate_mask);
 Debug_msg(0, "class=0x%x bgvy=0x%x wgvy=0x%x sto=0x%x svu=0x%x map=0x%x ov=0x%x\n",
	attr->class, attr->bit_gravity, attr->win_gravity, 
	attr->backing_store, attr->save_under, attr->map_state,
	attr->override_redirect);

	self->Desk = wmRoot->CurrentDesk;
	return self;
}

TwmWin *TwmWin_Create(Window win) {
  int pid, x, y, m;
  TwmWin *self;
  Display *dpy = IappXDat.dpy;
  TwmFrame *wmf;
  XWindowAttributes attr;

 Debug_msg(0, "          ********* ADD NEW WIN = 0x%x grab_count=%d *********\n", (int)win, GrabXServer_count);
	pid = TwmRoot_ClearStartNotify();
	self = TwmWin_Create_base(win, &attr);
	if (!self) return NULL;
	self->DoMouseDown = TWins_DoMouseDown;
	self->DoMouseUp = TWins_DoMouseUp;
	self->DoFocus = (TFocusEv)TwmWin_DoFocus;
	self->left = -4000;
	self->top = -4000;

	tmp_create_int(self);

	self->tflags = tfl_BORDER | tfl_TITLE | tfl_MINIMIZE | tfl_MINMAXIMIZE;

	TwmWin_GetMotifWMHints(self);

	TwmWin_GetNetWMPid(self);

	TwmWin_GetWindowSizeHints(self);
	if (self->min_width && 
	    (self->min_width == self->max_width) &&
	    self->min_height && 
	    (self->min_height == self->max_height)) self->tflags &= ~tfl_MINMAXIMIZE;


	if (self->tflags & tfl_NOAUTOPOS) {
	    if ((attr.x < 0) || (attr.y < 0)) {
		if (attr.x < 0) attr.x = 0;
		if (attr.y < 0) attr.y = 0;
		if ((attr.width < 23) && (attr.height < 23)) {
		    TTaskbar_Add_AppletWin(wmRoot->Taskbar, self);
		    UngrabXServer();
		    tmp_create_free(self);
		    return NULL;
		}
	    }
	}

	TwmWin_GetNetWMState(self);

	wmf = TwmFrame_Create(self, &attr);
	self->Frame = wmf;
  Tapph *ap = AddAppWin(wmf->apph, (TWins*)self);
	ap->EvProc = TwmWin_EvProc;
	ap->win = self->WinID;

  XSetWindowAttributes XAttr;
#ifdef SAVE_PARENT_WIN
	Fill0(&XAttr, sizeof(XSetWindowAttributes));
	XAttr.cursor = IappXDat.XCursors[self->cursor_nr];

	XAttr.event_mask = (SubstructureNotifyMask | SubstructureRedirectMask | ButtonPressMask |
			   ButtonReleaseMask | EnterWindowMask |
			   LeaveWindowMask | ExposureMask | PointerMotionMask);
  int cx = wmf->Border_width;
  int cy = wmf->Caption_height + wmf->Border_width;
	self->SaveParentWinID = XCreateWindow(dpy, wmf->WinID, 
			cx, cy, self->width, self->height,
			0, CopyFromParent, InputOutput, CopyFromParent,
			CWEventMask | CWCursor | CWSaveUnder, &XAttr);

	XReparentWindow(dpy, self->WinID, self->SaveParentWinID, -attr.border_width, -attr.border_width);
 Debug_msg(0, "FrameWinID=0x%x SaveParentWinID=0x%x ap=0x%x\n", (int)self->Frame->WinID, (int)self->SaveParentWinID, (int)ap);
#else
	XReparentWindow(dpy, self->WinID, wmf->WinID, 0, 0);
#endif

	Fill0(&XAttr, sizeof(XSetWindowAttributes));
	XAttr.event_mask = (StructureNotifyMask | PropertyChangeMask |
			   VisibilityChangeMask |  EnterWindowMask |
			   LeaveWindowMask | FocusChangeMask); // or ColormapChangeMask);
	XAttr.do_not_propagate_mask = ButtonPressMask | ButtonReleaseMask;
	XChangeWindowAttributes(dpy, self->WinID, 
			CWEventMask | CWDontPropagate, &XAttr);

	TwmWin_GrabFocusWin(self, True);

#if 1
 Debug_msg(0, "-> fr=%x Trans=%x x=%d y=%d w=%d h=%d\n", 
	wmf->WinID, wmf->transientfor, attr.x, attr.y, attr.width, attr.height);
#endif

	if (self->tflags & tfl_ABOVE) wmf->apph->aos |= aosOnTop;

	if ((self->tflags & tfl_NOAUTOPOS) == 0) {
  TwmFrame *twmf;
	    twmf = GetTop_DeskWin(wmf);
	    if (twmf) {
		x = twmf->left + 22;
		y = twmf->top + 32;
		m = IappXDat.Root_height;
		if (wmRoot->Taskbar) m -= wmRoot->Taskbar->height;
		if (x + wmf->width > dpyMonitors.mon[0].right) {
			x = y = 0;
		}
		if (y + wmf->height > m) y = 0;
		wmf->left = x;
		wmf->top = y;
	    } else {
		wmf->left = 0;
		wmf->top = 0;
	    }
	} else if (self->WinGravity > 0) {
	    TwmWin_CalcGravity(self, &wmf->left, &wmf->top);
	}

	if (self->WinGravity == StaticGravity) {
	    wmf->top -= (wmf->Border_width + wmf->Caption_height);
	    wmf->left -= wmf->Border_width;
	}

	TwmWin_SetupSize(self, wmf->left, wmf->top, wmf->width, wmf->height, wmwPos_SendPosAlways);

#ifdef SAVE_PARENT_WIN
	XMapWindow(dpy, self->SaveParentWinID);
#endif
	MapWinID(self->WinID, 1);
	UngrabXServer();
	SyncX();

  XEvent tmp_ev;
	if (XCheckTypedWindowEvent(dpy, self->WinID, DestroyNotify, &tmp_ev)) {
	    XPutBackEvent(dpy, &tmp_ev);
	} else {
	    TwmFrame_SetOnTop(wmf, 0x01);
	    SetVisible((TWins*)wmf, True);
	}

	tmp_create_free(self);
	return self;
}
