/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

TwmRoot *wmRoot;
int ExtQuitBef = 0;

#ifdef	USE_DEBUG_WIN
Window last_event_window;
int last_event_type;
#endif

bool Chk_another_WM;
bool XwmError_Stop;
bool NoAutoStart;

static bool XwmError(Display *dpy, XErrorEvent *e) {
  bool res;

	if (Chk_another_WM) {
	    printf("ERROR: another WM is running\n");
	    exit(1);
	}
	if (XwmError_Stop) return True;

#ifdef ALL_XwmError
	res = False;
#else
	res = (e->error_code == BadWindow);
#endif

#ifdef	USE_DEBUG_WIN
 char msg[80], req[80], number[80];
    if (!res) {
	XGetErrorText(dpy, e->error_code, msg, sizeof(msg));
	sprintf(number, "%d", e->request_code);
	XGetErrorDatabaseText(dpy, "XRequest", number, "<unknown>", req, sizeof(req));
	Debug_msg(0, "ALDesk XwmError: %s(0x%lx): %s\n", req, e->resourceid, msg);
	Debug_msg(0, "  --------->>: last_ev_w=0x%x last_ev_typ=%d\n", (int)last_event_window, last_event_type);
    }
#endif
	return res;
}

void AutoStart_OnFileMsg(TFileMsg *fmsg) {
  char *pc;

	if (get_bef_dflEv(fmsg)) {
	    if (fmsg->F.dfl->st->FileType == stft_desktop) {
		pc = pcadd(fmsg->R.gp->path, fmsg->F.dfl->fn);
		ExecDesktopFile(pc, NULL);
		xfree(pc);
	    }
	}
}

void AutoStart() {
  TDirListObjNy *ObjNy;

	ObjNy = TDirListObjNy_Create(NULL);
	ObjNy->OnFileMsg = AutoStart_OnFileMsg;
	ObjNy->findAtr = findREG;
	if (TDirListObjNy_Open(ObjNy, wmRoot->AutostartPath)) {
	    TDirListObjNy_Open_Wait(ObjNy);
	}
	Tfree(ObjNy);
}

/*_____________________________ ErrorMsg _____________________________*/
int Err_ev_count = 0;

void MsgBox_Error_ev(Texec_pipe* expip, int evtyp) {
	if (evtyp == evtyp_ProcClose) Err_ev_count--;
}

int Error_ev(int ErrType, const char *err) {

	if (Err_ev_count > 4) return 0;

	if (((ErrType & EMSG_MASK_Bef) == EMSG_ERR1) ||
	    ((ErrType & EMSG_MASK_Bef) == EMSG_ERR2)) {

	    if ((ErrType & EMSG_MASK_Type) == EMSG_IMAGE_RD) return 0;

	    XBell(IappXDat.dpy, 0);
	    if (IappDat.gQUIT_Ev) {
		printf("ERROR:%s(%d): %s", FL, err);
	    } else  {
		Err_ev_count++;
		ShowExtMsgBoxEv(err, MsgBox_Error_ev);
	    }
	    return 1;
	}
	return 0;
}

/*_____________________________ Monitors ________________________________*/
void MonitorsReScane(TgTimer *gti) {
	MonitorsInit();
}

/*_____________________________ Start ___________________________________*/
static const char *path_tab[] = {
  "/Menu/OpenWith/",
  "/Menu/Tools/",
  NULL
};

void OnAppStart() {
  char *pc;
  const char **ppc;

	ppc = path_tab;
	while (*ppc) {
	    pc = pcadd(IappDat.ALDini_dir, *ppc);
	    file_mkdir_e(pc, 0750);
	    xfree(pc);
	    ppc++;
	}

	AppProcessMessages();
	SyncX();
	AppHandleMessage();

	TwmRoot_global(wmRoot);
	TwmRoot_InitDeskIcons(wmRoot);

	gTimerX_Add(Timer_Enabled, 2000, TwmRoot_AzMount);

	if (wmRoot->Taskbar && wmRoot->Taskbar->Visible) {
	    XRaiseWindow(IappXDat.dpy, wmRoot->Taskbar->WinID);
	}

	SyncX();
	AppProcessMessages();

#if USE_SYS_BLOCK
	if (wmRoot->ObjNy_DeskIcons) {
	    while (wmRoot->ObjNy_DeskIcons->dl->runstat < dlrunstat_NotifyRun) {
		AppHandleMessage();
	    }
	}
	InitIcons_AutoMount(wmRoot);
#endif
	AppHandleMessage();
	chk_AutoFsDev();


	MonitorsInit();
	if ((dpyMonitors.stat & dpyM_Active) && !(dpyMonitors.stat & dpyM_AutoReScane)) {
	    gTimerX_Add(Timer_Enabled, 2000, MonitorsReScane);
	}
	AppHandleMessage();
	if (!NoAutoStart) AutoStart();
}

bool exit_msg_run = False;
bool force_exit = False;

void MsgBox_ExitEv(Texec_pipe* expip, int evtyp) {
	if (evtyp == evtyp_ProcClose) {
	    if (pcncmp(expip->pp[evtyp_infd].lb->pc, "YES", 3)) {
		force_exit = True;
		Close();
	    }
	    exit_msg_run = False;
	}
}

bool ChkAnyWinOpen() {
  Tapph *ap;
	ap = IappXDat.apph;
	while (ap) {
	    if (ap->wins && (ap->wins->wtyp == wtyp_TwmFrame)) return True;
	    ap = ap->next;
	}
	return False;
}

void OnAppClose() {
  int t;
  Tapph *ap;
  char *pc;
  char line[1024];
  bool OpenMk;
  bool SysShutDown;

	SysShutDown = ((ExtQuitBef & 3) == 3);
	exec_Buildin(NULL, "close_all_win", NULL, 0);
	if (force_exit) goto CloseWmClose;

	IappDat.gQUIT_Ev = False;

	t = GetTime_ms() + 1000;
	if (SysShutDown) t += 10000;
loop_t:
	OpenMk = ChkAnyWinOpen();
	if (OpenMk) {
//		XSync(IappXDat.dpy, 0);
	    AppHandleMessage();
	    OpenMk = ChkAnyWinOpen();
	    if (!cpTime(t, GetTime_ms())) goto loop_t;
	}

	if (OpenMk) {
	    if (SysShutDown) {
		exec_Buildin(NULL, "close_all_win", NULL, 0);
		AppHandleMessage();
		goto CloseWmClose;
	    }
	    if (!exit_msg_run) {
		pc = line;
#ifdef USE_BULDIN_APP
		pc += sprintf(pc, "%s aldmessagebox ", IappDat.self_exe);
#else
		ppcmove(&pc, "aldmessagebox ");
#endif
#ifdef Test_Root_WinId
		if (TestRootWinId) {
			pc += sprintf(pc, "--TestRootWinID 0x%x ", (int)TestRootWinId);
		}
#endif
		ppcmove(&pc, " -fYES,NO -dNO \"Not all Windows closed\\nForce Exit ???\"");
		ExecuteProcApp(line, infd_Open, MsgBox_ExitEv, NULL, NULL);
	    }
	    return;
	}

CloseWmClose:
	IappDat.gQUIT_Ev = True;
// ******** Close Taskbar AppletWin ********
	if (wmRoot->Taskbar) {
	    ap = wmRoot->Taskbar->tbSysTray->apph->next_child;
	    while (ap) {
  Tapph *apnx = ap->next;
		if (ap->wins->wtyp == wtyp_TTaskbarApplet) {
			TwmWin_Close((TwmWin*)ap->wins);
		}
		ap = apnx;
	    }
	}

	if (SysShutDown) {
	    FileHandle fd;
	    FileOpen_wr("/var/run/ALDesk/aldesk", &fd, 0640);
	    FileWrite(fd, "ShutDown", 8, NULL);
	    FileClose(fd);
	}
}

void menu_desktop_info_OnClick(TMenu *menu, long KeyState, TMenuItem* mi) {
  char *pc;
  char line[1024]; 

	pc = line;
#ifdef USE_BULDIN_APP
	pc += sprintf(pc, "%s Copyright_MsgBox \"%s   (build %d  %s)\"", 
			IappDat.self_exe, VERSION, ALD_BUILD_V, __DATE__);
#else
	pc += sprintf(pc, "Copyright_MsgBox \"%s %s   (build %d  %s)\"", 
			IappDat.self_exe, VERSION, ALD_BUILD_V, __DATE__);
#endif
#ifdef Test_Root_WinId
	if (TestRootWinId) {
	    sprintf(pc, "--TestRootWinID 0x%x ", (int)TestRootWinId);
	}
#endif
	ExecuteProc(line);
}

static const char *Help_Msg = {
  "Usage: aldesk [options]\n"
  "\n"
  "  -f File              Ini-File (default ~/.ALDccp/global.ini)\n"
  "  -d display_name      :1\n"
  "  --NoAutoStart        not exec ~/.ALDccp/Autostart/*.deskdop\n"
#ifdef Test_Root_WinId
  "  --TestRootWinID      wm run in owner window\n"
#endif
#ifdef	USE_DEBUG_WIN
  "  --debug              show debug window\n"
#endif
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: aldesk version "VERSION"  (build %d  %s)\n"
};

int main (int argc, char **argv) {
  int i;
  char *val;
  bool err = False;
  bool help = False;

	cmdline_slash_korr(argc, argv);
#ifdef USE_BULDIN_APP
  TBuildin_app *bi_app;
	bi_app = Get_Buildin_pc(argv[0]);
	if (bi_app) return bi_app(argc, argv);
	if ((argc > 1) && (*argv[1] != '-')) {
	    cmdline_del_argv0(&argc, argv);
	    bi_app = Get_Buildin_pc(argv[0]);
	    if (bi_app) return bi_app(argc, argv);
	    BuildinHelp();
	    return EXIT_FAILURE;
	}
#endif

	NoAutoStart = False;
	XwmError_Stop = False;
	i = 1;
	while (!err && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'd': 
		    i++;
		    err = (i >= argc);
		    if (!err) {
			    display_name = argv[i];
		    }
		    break;
		case 'h': 
		    help = True; break;
		case '-':
		    val++;
		    if (pcncmp(val, "NoAutoStart", 11)) {
			NoAutoStart = True;
#ifdef Test_Root_WinId
		    } else if (pcncmp(val, "TestRootWinID", 13)) {
			i++;
			val = argv[i];
			if (val && *val != '-') {
				TestRootWinId = pctoint(val);
			} else err = True;
#endif
		    } else if (pcncmp(val, "debug", 5)) {
#ifdef	USE_DEBUG_WIN
			DebugOn = True;
#endif
		    } else err = True;
		    break;
		default:
		    err = True;
		}
	    } else err = True;
	    i++;
	}
	if (err || help) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
#ifdef USE_BULDIN_APP
	    printf("\nor:\n");
	    BuildinHelp();
#endif
	    if (err) exit(1); else exit(0);
	}

	while (waitpid(-1, NULL, WNOHANG) > 0);

	initXapp(0, argc, argv);
	ThDirList_Reg(1, &hdirlist_ld_File);
//	ThDirList_ny_Reg(1, &hdirlist_ny_File);
	ThDirList_ny_Reg(2, &hdirlist_nyi_File, &hdirlist_ny_File);

	hGetDeskIcon = &hGetDeskIcon_file;
	ThImageRD_Reg(4, &hImageRD_xpm, &hImageRD_png, &hImageRD_jpeg, 
			 &hImageRD_bmp);

	hGetFileHeadInfo = &hGetFileHeadInfo_buildin;

	IappDat.inifile_app = pcadd(IappDat.ALDini_dir, "/aldesk.ini");
	ChkSetDefault();

	Chk_another_WM = True;

	On_wmError = XwmError;

//	XSetErrorHandler(CatchRedirectError);
	InitEventHandlerJumpTable();

	Create_globGC();

	wmRoot = TwmRoot_Create(NULL);
// return 0;
	Chk_another_WM = False;

//	chk_AutoFsDev();

/*
//create NoFocusWin
	attributes.event_mask = KeyPressMask | FocusChangeMask;
	attributes.override_redirect = 1;
	NoFocusWin = XCreateWindow(IappXDat.dpy, IappXDat.RootW, -10, -10, 10, 10, 0, 0,
                               InputOnly, CopyFromParent,
                               CWEventMask | CWOverrideRedirect, &attributes);
	XMapWindow(IappXDat.dpy, NoFocusWin);
// printf("%s(%d) NoFocusWin=0x%x\n", FL, (int)NoFocusWin);
*/

//	setNumlock(True);

	wmRoot->Taskbar = TTaskbar_Create();

	IappXDat.OnAppStart = OnAppStart;
	IappXDat.OnAppClose = OnAppClose;

	appError_ev = Error_ev;

#ifdef USE_DEBUG_WIN
	if (DebugOn) Chk_DebugWinRun();
#endif

	AppRun();
	return 0;
}
