/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/fs.h>

char *no_AutoFsDev = NULL;

int boot_XBell_sleep = 10;

static void boot_XBell() {
	if (boot_XBell_sleep) return;
	boot_XBell_sleep++;
	XBell(IappXDat.dpy, 0);
}


bool ppcskip(char **ppc, int w) {
  char *pc;
	if (!ppc) return False;
	pc = *ppc;
loop:
	while (*pc && ((byte)*pc <= 0x20)) pc++;
	if (w > 0) {
	    w--;
	    while ((byte)*pc > 0x20) pc++;
	    goto loop;
	}
	*ppc = pc;
	return w == 0;
}


typedef struct TnxExecProc {
  struct TnxExecProc *next;
  char *args;
  int fu;
  TOnExecProcEv OnExecProcEv;
  void *owner;
  Texec_pipe **ret_expip;
} TnxExecProc;

static TnxExecProc *nxExecProc = NULL;

/* ___________________ ExecuteProcApp ________________________________ */
static void Add_ExecProc(char *args, int fu, TOnExecProcEv OnExecProcEv, void *owner, Texec_pipe **ret_expip) {
  TnxExecProc *proc, **ppproc;
	proc = xcalloc(sizeof(TnxExecProc));
	if (!nxExecProc) {
	    nxExecProc = proc;
	    ExecuteProcApp(args, fu, OnExecProcEv, owner, ret_expip);
	    return;
	}
	proc->args = pcdup(args);
	proc->fu = fu;
	proc->OnExecProcEv = OnExecProcEv;
	proc->owner = owner;
	proc->ret_expip = ret_expip;
	ppproc = &nxExecProc;
	while (*ppproc) ppproc = &(*ppproc)->next;
	*ppproc = proc;
}

static void ExecProc_next() {
  TnxExecProc *proc = nxExecProc;
	if (!proc) return;
	nxExecProc = proc->next;
	xfree(proc->args);
	xfree(proc);
	if (!nxExecProc) return;
	ExecuteProcApp(nxExecProc->args, nxExecProc->fu, 
		nxExecProc->OnExecProcEv, nxExecProc->owner, nxExecProc->ret_expip);
}

/* ___________________ get_part_table ________________________________ */
static TFSDevice_part_table *get_part_table(char *dev) {
  TFSDevice_part_table *table;

	if (!dev || !*dev) return NULL;
	table = wmRoot->part_table;
	while (table) {
	    if (pccmp(table->dev, dev)) return table;
	    table = table->next;
	}
	return NULL;
}

static int get_part_table_type(char *dev, int dev_nr) {
  TFSDevice_part_table *table;

	if ((dev_nr < 0)||(dev_nr >= PART_TABLE_MAX)) return -1;
	table = get_part_table(dev);
	if (!table) return -1;
	return table->type[dev_nr];
}

/* ___________________ fs_info _______________________________________ */
static TFSDevice_info *new_fs_info(char *dev, char *dev0) {
  TFSDevice_info **ppfs_info, *fs_info;

	ppfs_info = &wmRoot->fs_info;
	while (*ppfs_info) ppfs_info = &(*ppfs_info)->next;
	fs_info = xcalloc(sizeof(TFSDevice_info));
	*ppfs_info = fs_info;

	fs_info->part_type = -1;
	pcmove(fs_info->dev0, dev0);
	pcmove(fs_info->dev[0], dev);
	return fs_info;
}

TFSDevice_info *add_fs_info(char *dev, char *dev0) {
  int m, major, minor;
  char *pc;
  TFSDevice_info *fs_info;
  Tstat64 *st;
  char buf[32];

	m = pclen(dev);
	if ((m >= sizeof(fs_info->dev)) || (m < 3)) return NULL;
	m--;
	if ((dev[m] >= '0') || (dev[m] <= '9')) {
	    if ((dev[m-1] >= '0') && (dev[m-1] <= '9')) m--;
	    m = pctoint(&dev[m]);
	    if ((m < 0) || (m > 15)) return NULL;
	} else {
	    m = 0;
	}

	pc = buf;
	ppcmove(&pc, "/dev/");
	ppcmove(&pc, dev);
	st = file_stat_fn(buf, fsFollowLink);
	if (!st) {
	    fs_info = wmRoot->fs_info;
	    while (fs_info &&
		   ((dev0 && !pccmp(fs_info->dev0, dev0)) ||
		    (dev && !pccmp(fs_info->dev[0], dev)))) fs_info = fs_info->next;
	    if (fs_info) return NULL;

	    fs_info = new_fs_info(dev, dev0);
	    fs_info->stat = FSINFO_HIDE;

	    printf("%s(%d): /dev/%s not found\n", FL, dev);
	    return NULL;
	} 
	major = (int)((st->rdev >> 8) & 0xff);
	minor = (int)(st->rdev & 0xff);
	xfree(st);

	if (major == 11) {  //CDROM
	    if (!dev0) dev0 = dev;

	} else {
	    if (!dev0) return NULL;
	}

	fs_info = wmRoot->fs_info;
	while (fs_info &&
	       ((fs_info->major != major) ||
	        (fs_info->minor != minor))) fs_info = fs_info->next;

	if (fs_info) {
	    int i = 0;
	    while ((i < 4) && *fs_info->dev[i] && 
			      !pccmp(fs_info->dev[i], dev)) i++;
	    if (i >= 4) return NULL;
	    if (!*fs_info->dev[i]) pcmove(fs_info->dev[i], dev);
	    return fs_info;
	}

	fs_info = new_fs_info(dev, dev0);
	if (major == 11) {
	    fs_info->stat = FSINFO_CDROM;
	}
	fs_info->nr = m;
	fs_info->major = major;
	fs_info->minor = minor;
// printf("%s(%d): add_fs_info (%3d:%-3d) %s %s\n", FL, major, minor, fs_info->dev0, fs_info->dev[0]);
	return fs_info;
}

TNFSDevice_info *add_nfs_info(char *dev) {
  int m;
  TFSDevice_info **ppfs_info;
  TNFSDevice_info *res;

	m = pclen(dev);
	if ((m >= sizeof(res->dev)) || (m < 8)) return NULL;
	res = (TNFSDevice_info*)get_fs_info(dev);
	if (res) return res;

	ppfs_info = &wmRoot->fs_info;
	while (*ppfs_info) ppfs_info = &(*ppfs_info)->next;
	res = xcalloc(sizeof(TNFSDevice_info));
	*ppfs_info = (TFSDevice_info*)res;

	res->stat = FSINFO_NFS | FSINFO_DEVLOCK;
	res->part_type = -1;
	pcmove(res->dev, dev);
// printf("%s(%d): add_fs_info %s %s\n", FL, res->dev0, res->dev);

	return res;
}

TFSDevice_info *get_fs_info(char *dev) {
  TFSDevice_info *fs_info;

	if (!dev) return NULL;
	if (pcncmp(dev, "/dev/", 5)) dev += 5;
	if (!*dev) return NULL;

	fs_info = wmRoot->fs_info;
	while (fs_info) {
	    int i = 0;
	    while ((i < 4) && *fs_info->dev[i]) {
		if (pccmp(fs_info->dev[i], dev)) return fs_info;
		i++;
	    }
	    fs_info = fs_info->next;
	}
	return NULL;
}


void chk_add_mountpoint(TFSDevice_info *fs_info) {
  char *pc;
  TpcIni *Ini;

	if (!(fs_info->stat & FSINFO_AZ_OK) ||
	    (fs_info->stat & FSINFO_HIDE)) return;

	Ini = TpcIni_Create(IappDat.inifile_global);

	pc = pcadd(no_AutoFsDev, TpcIni_RD(Ini, "AutoFsDev", "no_AutoFsDev"));
	if (token_find(pc, fs_info->dev[0])) fs_info->stat |= FSINFO_HIDE;
	xfree(pc);

	pc = TpcIni_RD_pc(Ini, "AutoFsDev", "no_AutoFsTyp", "swap");
	if (token_find(pc, fs_info->type)) fs_info->stat |= FSINFO_HIDE;

	Tfree(Ini);

	if (fs_info->stat & FSINFO_HIDE) return;

	pc = pcadd_n(3, wmRoot->DesktopPath, fs_info->dev, ".desktop");
	if (!TestIsFile(pc)) {
	    char *Dev = pcadd("/dev/", fs_info->dev[0]);
	    char *MountPoint = pcadd_n(3, wmRoot->AutoMountPath, "/", fs_info->dev);
	    char *Icon = "harddisk";
	    char *ReadOnly = "False";
	    if (fs_info->stat & FSINFO_CDROM) {
		Icon = "cdrom";
		ReadOnly = "True";
	    }
	    dotdesktop_new(pc, fs_info->dev[0], Icon, Dev, MountPoint, ReadOnly, ddkt_FSDevice);
	    xfree(MountPoint);
	    xfree(Dev);
	}
	xfree(pc);
}

void fs_info_Ev(Texec_pipe* expip, int evtyp) {
  int lg, m;
  char *pc, *token;
  TFSDevice_info *fs_info;

	if ((evtyp != evtyp_ProcClose) || !expip->owner) return;

	fs_info = get_fs_info(expip->owner);
	xfree(expip->owner);
	if (!fs_info) goto next_proc;
	fs_info->stat &= ~FSINFO_AZ_OK;
	*fs_info->size = 0;
	*fs_info->type = 0;
	m = get_part_table_type(fs_info->dev0, fs_info->nr);
	if ((m == 0)||(m == 5)) fs_info->stat |= FSINFO_HIDE;
	fs_info->part_type = m;

	pc = expip->pp[evtyp_infd].lb->pc;
	if (!pc || (*pc != '"')) goto next_proc;
	pc = pcGetToken(pc, &token, &lg);
	pc = pcGetToken(pc, &token, &lg);
	if ((lg != 4) || !pcncmp(token, "SIZE", 4)) goto next_proc;
	pc = pcGetToken(pc, &token, &lg);
	if (lg >= sizeof(fs_info->size)) goto next_proc;
	pcnmove(fs_info->size, token, lg);
	pc = pcGetToken(pc, &token, &lg);
	if ((lg != 3) || !pcncmp(token, "TYP", 3)) goto next_proc;
	pc = pcGetToken(pc, &token, &lg);
	if (lg >= sizeof(fs_info->type)) goto next_proc;
	pcnmove(fs_info->type, token, lg);
	fs_info->stat |= FSINFO_OK;

	if ((lg == 4) && pcncmp(token, "part", 4)) fs_info->stat |= FSINFO_HIDE;

	chk_add_mountpoint(fs_info);

next_proc:
	ExecProc_next();
}

void read_fs_info(char *dev) {
  char *pc;
  char buf[200];

	pc = buf;
	wmExcCMD_head(&pc, "gethddevtyp --noerr");
	ppcmove(&pc, "/dev/");
	ppcmove(&pc, dev);
	pc = pcdup(dev);
	Add_ExecProc(buf, infd_Open, fs_info_Ev, pc, NULL);
}
/* ___________________ partition table _______________________________ */
static void part_table_del(char *dev) {
  TFSDevice_part_table *table, **pptable;

	pptable = &wmRoot->part_table;
	while (*pptable) {
	    if (pccmp((*pptable)->dev, dev)) {
		table = *pptable;
		*pptable = table->next;
		Tfree(table->dl);
		xfree(table);
		return;
	    }
	    pptable = &(*pptable)->next;
	}
}

static void part_table_add(TFSDevice_part_table *table) {
  TFSDevice_part_table **pptable;

	part_table_del(table->dev);
	pptable = &wmRoot->part_table;
	while (*pptable) pptable = &(*pptable)->next;
	*pptable = table;
}



static void part_table_Ev(Texec_pipe* expip, int evtyp) {
  char *pc;
  TFSDevice_part_table *table;

	if ((evtyp != evtyp_ProcClose) || !expip->owner) return;
	table = get_part_table(expip->owner);
	xfree(expip->owner);
	if (!table) goto err;

	pc = expip->pp[evtyp_infd].lb->pc;
	if (!pc || (*pc != '"')) goto err;

	ppcskip(&pc, 1);
	if (!pcncmp(pc, "TABLE", 5)) goto err;
	ppcskip(&pc, 1);
	for (int i=1;i<5;i++) table->type[i] = ppctoint(&pc);

/* printf("%s(%d): %s 0x%x 0x%x 0x%x 0x%x\n", FL, table->dev, 
		table->type[0], table->type[1],
		table->type[2], table->type[3]);
*/
	ExecProc_next();
	boot_XBell();
	return;
err:
	if (table) part_table_del(table->dev);
	ExecProc_next();
}

void read_part_table(char *dev) {
  char *pc;
  char buf[200];

	if (((dev[0] == 's') && (dev[1] == 'r')) ||
	    ((dev[0] == 's') && (dev[1] == 'c'))) return;

	pc = buf;
	wmExcCMD_head(&pc, "gethddevtyp --noerr -p ");
	ppcmove(&pc, "/dev/");
	ppcmove(&pc, dev);
	pc = pcdup(dev);
	Add_ExecProc(buf, infd_Open, part_table_Ev, pc, NULL);
}

/*      SCAN_ALL_DEV     only kernel 2.6.xx */
#define SCAN_ALL_DEV 0

#if SCAN_ALL_DEV
static bool use_proc_diskstats;
#endif

void scan_diskstats() {
  int lg, lg0;
#if SCAN_ALL_DEV
  int skip;
#endif
  FILE *f;
  char *tline, *pc;
  TFSDevice_part_table *part_table;
  TFSDevice_info *fs_info;
  char dev0[16];
  char dev[16];
  char line[256];

#if SCAN_ALL_DEV
	skip = 2;
	pc = "/proc/diskstats";
	if (!use_proc_diskstats) {
	    skip = 3;
	    pc = "/proc/partitions";
	}
	f = fopen(pc, "r");
#else
	f = fopen("/proc/partitions", "r");
#endif
	if (!f) return;

	lg0 = 0;
LoopUSBHD:
	tline = fgets(line, sizeof(line)-1, f);
	if (tline) {
	    if ((line[0] == ' ') && line[1]) {
		pc = line;
#if SCAN_ALL_DEV
		ppcskip(&pc, skip);
#else
		ppcskip(&pc, 3);
#endif
		char *tpc = pc;
		while ((byte)*tpc > ' ') tpc++;
		lg = tpc - pc;
		if (!lg || (lg>=16)) goto LoopUSBHD;

#if SCAN_ALL_DEV
		if (pcncmp(pc, "ram", 3) || 
		    pcncmp(pc, "loop", 4)) goto LoopUSBHD;
		if ((pc[0] == 's') && (pc[1] == 'r') && 
		    (pc[2] >= '0') && (pc[2] <= '9')) {

		    pcnmove(dev0, pc, lg);
		    pcnmove(dev, pc, lg);
		    lg0 = lg;
		    fs_info = get_fs_info(dev);
		    if (fs_info) {
			fs_info->stat &= ~FSSTAT_EXIST_MK;
		    } else {
			fs_info = add_fs_info(dev, dev0);
			if (fs_info) {
			    fs_info->stat |= FSINFO_CDROM;
			    chk_add_mountpoint(fs_info);
			}
		    }
		    goto LoopUSBHD;
		}
#endif
		if (!lg0 || !pcncmp(dev0, pc, lg0)) {

		    pcnmove(dev0, pc, lg);
		    lg0 = lg;
		    part_table = get_part_table(dev0);
		    if (part_table) {
			part_table->stat &= ~FSSTAT_EXIST_MK;
		    } else {
// printf("%s(%d): %s\n", FL, dev0);
			part_table = xcalloc(sizeof(TFSDevice_part_table));
			for (int i=0;i<PART_TABLE_MAX;i++) part_table->type[i] = -1;
			pcnmove(part_table->dev, dev0, lg);
			part_table_add(part_table);
			read_part_table(dev0);
		    }
		} else {

		    pcnmove(dev, pc, lg);
		    fs_info = get_fs_info(dev);
		    if (fs_info) {
			fs_info->stat &= ~FSSTAT_EXIST_MK;
		    } else {
			fs_info = add_fs_info(dev, dev0);
			if (fs_info) read_fs_info(dev);
		    }
		}
	    }
	    goto LoopUSBHD;
	}
	fclose(f);
}


void Timer_AutoFsDev_proc_part() {
  TFSDevice_part_table *part_table, **pptable;
  TFSDevice_info *fs_info, **ppfs_info;

	part_table = wmRoot->part_table;
	while (part_table) {
	    part_table->stat |= FSSTAT_EXIST_MK;
	    part_table = part_table->next;
	}
	fs_info = wmRoot->fs_info;
	while (fs_info) {
	    if (!(fs_info->stat & FSINFO_DEVLOCK)) fs_info->stat |= FSSTAT_EXIST_MK;
	    fs_info = fs_info->next;
	}
	scan_diskstats();

	pptable = &wmRoot->part_table;
	while (*pptable) {
	    if ((*pptable)->stat & FSSTAT_EXIST_MK) {
		part_table = *pptable;
		*pptable = part_table->next;
		xfree(part_table);
		boot_XBell();
	    } else {
		pptable = &(*pptable)->next;
	    }
	}
	ppfs_info = &wmRoot->fs_info;
	while (*ppfs_info) {
	    if ((*ppfs_info)->stat & FSSTAT_EXIST_MK) {
		fs_info = *ppfs_info;
		*ppfs_info = fs_info->next;
		xfree(fs_info);
	    } else {
		ppfs_info = &(*ppfs_info)->next;
	    }
	}
	if (boot_XBell_sleep > 0) boot_XBell_sleep--;
}

static char *chk_access(int pid, char *buf) {
  char *pc = buf;

	ppcmove(&pc, "/proc/");
	inttoppc(&pc, pid);
	ppcmove(&pc, "/mountinfo");

	if (!access(buf, F_OK) && !access(buf, R_OK)) return buf;
	return NULL;
}

void chk_AutoFsDev() {
  char *pc;
  FILE *f;
  int major, minor;
  char buf[256];

#if SCAN_ALL_DEV
	use_proc_diskstats = !access("/proc/diskstats", R_OK);
#endif

//Find root dev   --> /proc/1/mountinfo
	pc = chk_access(1, buf);
	if (!pc) pc = chk_access(IappDat.self_pid, buf);
	if (!pc) goto find_CD;

	f = fopen(pc, "r");
	if (!f) goto find_CD;
loop_rdv:
	major = minor = 0;
	pc = fgets(buf, sizeof(buf)-1, f);
	if (pc) {
	    ppcskip(&pc, 2);
	    major = ppctoint(&pc);
	    if (*pc != ':') goto loop_rdv;
	    pc++;
	    minor = ppctoint(&pc);
	    while (*pc == ' ') pc++;
	    if ((*pc != '/') || (pc[1] != ' ')) goto loop_rdv;
	    pc++;
	    while (*pc == ' ') pc++;
	    if ((*pc != '/') && (pc[1] != ' ')) goto loop_rdv;
	}
	fclose(f);

	pc = NULL;
	buf[0] = 0;
	if ((major == 3) && (minor & 0x3f)) {
	    buf[0] = 'h';
	    buf[1] = 'd';
	    buf[2] = 0x61 + ((minor & 0xc0) >> 6);
	    pc = &buf[3];
	    inttoppc(&pc, minor & 0x3f);
	} else if ((major == 8) && (minor & 0x0f)) {
	    buf[0] = 's';
	    buf[1] = 'd';
	    buf[2] = 0x61 + ((minor & 0xf0) >> 4);
	    pc = &buf[3];
	    inttoppc(&pc, minor & 0x0f);
	}

	if (pc) {	//add skip root dev
	    *pc = ' '; pc[1] = 0;
	    ppcadd(&no_AutoFsDev, buf);
	}
// printf("%s(%d): major=%d minor=%d %s %s\n", FL, major, minor, buf, no_AutoFsDev);

find_CD:
#if SCAN_ALL_DEV == 0
	f = fopen("/proc/diskstats", "r");
	if (!f) return;
loop_cd:
	pc = fgets(buf, sizeof(buf)-1, f);
	if (pc) {
	    ppcskip(&pc, 2);
	    if ((pc[0] == 's') && 
		(((pc[1] == 'r')&&(pc[2] >= '0')&&(pc[2] <= '9')) ||
		 ((pc[1] == 'c')&&(pc[2] == 'd')&&(pc[3] >= '0')&&(pc[3] <= '9')))) {

		char *pce = pc;
		while ((byte)*pce > 0x20) pce++;
		*pce = 0;
		TFSDevice_info *fs_info = get_fs_info(pc);
		if (!fs_info) {
		    fs_info = add_fs_info(pc, pc);
		    if (fs_info) {
			fs_info->stat |= FSINFO_CDROM | FSINFO_DEVLOCK;
			chk_add_mountpoint(fs_info);
		    }
		} else {
		    fs_info->stat |= FSINFO_CDROM | FSINFO_DEVLOCK;
		}
	    }
	    goto loop_cd;
	}
	fclose(f);
#endif /* SCAN_ALL_DEV == 0 */

	return;
}
