/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"

static Window REGPARM3_FUNC HandleMapRequest(TXev_msg *msg) {
  XMapRequestEvent *ev = (XMapRequestEvent*)msg->ev;
  Tapph *ap;
	if (ev->window == IappXDat.RootW) return 0;

	ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, ev->window);
	if (!ap) {
	    TwmWin_Create(ev->window);
	} else {
	    if (ap->wins->wtyp == wtyp_TwmWin) SetVisible(ap->wins, True);
	}
	return 0;
}

static Window REGPARM3_FUNC HandleMapNotify(TXev_msg *msg) {
  XMapEvent *ev = (XMapEvent*)msg->ev;
  Tapph *ap;
  TwmWin *wmw;
  TwmFrame *wmf;

	if (ev->event != ev->window) return 0;
	if (ev->window == IappXDat.RootW) return 0;

	ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, ev->window);
	if (!ap) return 0;

	wmw = (TwmWin*)ap->wins;
	wmf = wmw->Frame;
	if (wmw->Desk == wmRoot->CurrentDesk) {
	    if (wmf) {
		SetVisible((TWins*)wmf, True);
		TwmFrame_SetOnTop(wmf, 0);
 Debug_msg(0, "%s(%d) XSetInputFocus w=0x%x\n", FL, wmw->WinID);
		SetXInputFocus(wmw->WinID);
	    }
	}
	return 0;
}

static Window REGPARM3_FUNC HandleUnmapNotify(TXev_msg *msg) {
  XUnmapEvent *ev = (XUnmapEvent*)msg->ev;
  Window win = ev->window;
  Tapph *ap;
  TwmWin *wmw;
  TwmFrame *wmf;
  Display *dpy;
  int dx, dy;
  Window dwin;
  XEvent tmp_ev;

	ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, win);
	if (!ap) return win;
// Debug_msg(0, "HandleUnmapNotify 0x%x\n", ev->window);

	wmw = (TwmWin*)ap->wins;
	wmf = ((TwmWin*)ap->wins)->Frame;
	if (!wmf) return 0;
// printf("HandleUnmapNotify: A wtyp=%x\n", ap->wins->wtyp);
	dpy = IappXDat.dpy;
	GrabXServer();
//	SyncX();

	if (XCheckTypedWindowEvent(dpy, win, DestroyNotify, &tmp_ev)) {
 Debug_msg(0, "%s(%d) Destroy CheckType %x\n", FL, wmw->WinID);
	    Tfree(wmf);

	} else { 
#ifdef ALL_XwmError
	  XwmError_Stop = True;
#endif
	  if (XTranslateCoordinates(dpy, win, IappXDat.RootW, 0, 0, &dx, &dy, &dwin)) {
#ifdef ALL_XwmError
	    XwmError_Stop = False;
#endif 
	    XReparentWindow(dpy, win, IappXDat.RootW, wmf->left, wmf->top);

	    MapWinID(wmf->WinID, 0);
	    XRemoveFromSaveSet(dpy, win);
	  }
#ifdef ALL_XwmError
	  XwmError_Stop = False;
#endif 
 Debug_msg(0, "%s(%d) Reparent + to Destroy wmf\n", FL);
	    Tfree(wmf);

	}

// printf("HandleUnmapNotify: F wtyp=%x\n", ap->wins->wtyp);
	UngrabXServer();
	return 0;
}

static Window REGPARM3_FUNC HandleDestroyNotify(TXev_msg *msg) {
  Tapph *ap;
  XUnmapEvent *ev = (XUnmapEvent*)msg->ev;

	if (ev->event != ev->window) return 0;
	if (ev->window == IappXDat.RootW) return 0;

	ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, ev->window);
	if (!ap) return 0;
 Debug_msg(0, "%s(%d) Destroy Notify %x\n", FL, ap->wins->WinID);
	Tfree(((TwmWin*)ap->wins)->Frame);
	return 0;
}

static Window REGPARM3_FUNC HandleConfigureRequest(TXev_msg *msg) {
  XConfigureRequestEvent *ev = (XConfigureRequestEvent*)msg->ev;
  Tapph *ap;
  unsigned int xwcm;
  XWindowChanges xwc;
  int x, y, width, height;
  TwmWin *wmw;
  TwmFrame *wmf;

	if (ev->window == IappXDat.RootW) return 0;

	ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, ev->window);
	if (!ap) {
	    xwcm = ev->value_mask & (CWX | CWY | CWWidth | CWHeight);
	    if (xwcm) {
		xwc.x = ev->x;
		xwc.y = ev->y;
		xwc.width = ev->width;
		xwc.height = ev->height;
		XConfigureWindow(IappXDat.dpy, ev->window, xwcm, &xwc);
	    }
	    return 0;
	}

	wmw = (TwmWin*)ap->wins;
	wmf = wmw->Frame;
	if (ev->value_mask & CWStackMode) {
	    xwc.stack_mode = ev->detail;
 Debug_msg(0, "HandleConfigureRequest w=0x%x stack_mode=%04x\n", (int)ev->window, xwc.stack_mode);
	    XConfigureWindow(IappXDat.dpy, wmf->WinID, CWStackMode, &xwc);
	    TwmFrame_SetOnTop(wmf, 0);
	}

	if (!(ev->value_mask & (CWX | CWY | CWWidth | CWHeight))) return 0;
 Debug_msg(0, "HandleConfigureRequest L2 w=0x%x ap=0x%x m=0x%x x=%d y=%d w=%d h=%d\n", (int)ev->window, ap, ev->value_mask, ev->x, ev->y, ev->width, ev->height);
 Debug_msg(0, "                       L3 wmf x=%d y=%d w=%d h=%d\n", wmf->left, wmf->top, wmf->width, wmf->height);

	x = wmf->left;
	y = wmf->top;
	width = wmf->width;
	height = wmf->height;
	if (ev->value_mask & CWX) x = ev->x;
	if (ev->value_mask & CWY) y = ev->y;
	if (ev->value_mask & CWWidth) width = ev->width + (wmf->Border_width << 1);
	if (ev->value_mask & CWHeight) height = ev->height + wmf->Caption_height + (wmf->Border_width << 1);

	if (wmw->WinGravity == StaticGravity) {
	    y -= (wmf->Border_width + wmf->Caption_height);
	    x -= wmf->Border_width;
	}
 Debug_msg(0, "                       L4   x=%d y=%d w=%d h=%d grav=%d\n", x, y, width, height, wmw->WinGravity);

	TwmWin_SetupSize(wmw, x, y, width, height, wmwPos_SendPos);
	return 0;
}

static Tapph *SetXKey_TKey(XKeyEvent *XKeyEv, Tobjmsg *msg, TKeyEvent *KeyEv, int keyst) {
  int m;

 Debug_msg(0, "%s(%d):      0x%x  0x%x 0x%x 0x%x\n", FL, IappXDat.GrabWin, (int)XKeyEv->window, (int)XKeyEv->root, (int)XKeyEv->subwindow);
	if (IappXDat.GrabWin && chk_AnyMenuOpen()) {
	    XKeyEv->window = IappXDat.GrabWin->WinID;
	} else if ((XKeyEv->window == IappXDat.RootW)
		   && (XKeyEv->root == IappXDat.RootW)) {
	    TwmFrame *wmf = GetTop_DeskWin(NULL);
	    if (wmf) {
		XKeyEv->window = wmf->wmWinID;
		XKeyEv->subwindow = 0;
	    }
	}
	msg->msg = msgKey;
	msg->f = msg->m = 0;
	msg->p = KeyEv;
	IappXDat.last_KeyState = XKeyEv->state;

	KeyEv->keyst = keyst;
	KeyEv->state = XKeyEv->state;
	KeyEv->keycode = XKeyEv->keycode;
	XLookupString(XKeyEv, (char*)&m, 4, &KeyEv->keysym, NULL);
 Debug_msg(0, "%s(%d):      0x%x  code=%d sym=0x%04x\n", FL, KeyEv->state, KeyEv->keycode, KeyEv->keysym);
	NumLook_korr(KeyEv);
	return TwmWin_Find_WinsX_Ev(IappXDat.apph, XKeyEv->window);
}

static bool BKy_cmp(TBuildinKeys *BKy, TKeyEvent *KeyEv) {
	if ((BKy->mod & 0xff) != (KeyEv->state & BKy->modmask)) return False;
	if (BKy->sym) return (BKy->sym == KeyEv->keysym);
	return (BKy->code == KeyEv->keycode);
}

static Window REGPARM3_FUNC HandleKeyPress(TXev_msg *msg) {
  XKeyEvent *ev = (XKeyEvent*)msg->ev;
  Tapph *ap;
  TBuildinKeys *BKy;
  Tobjmsg omsg;
  TKeyEvent KeyEv;

	ap = SetXKey_TKey(ev, &omsg, &KeyEv, keyst_Press);
	wmRoot->ap_KeyDown = ap;
	BKy = wmRoot->BuildinKeys;
	while (BKy) {
	    if (BKy_cmp(BKy, &KeyEv)) break;
	    BKy = BKy->next;
	}
	wmRoot->BKy_key_up = BKy;
	if (BKy) {
 Debug_msg(0, "code=%d sym=0x%04x state=0x%02x %s\n", ev->keycode, (int)KeyEv.keysym, ev->state, BKy->Action);
	    if (exec_Buildin(ap, BKy->Action, BKy->cmd, cond_ChkOnly)) return 0;
	}

	if (!ap) {
	    Exec_Win_Ev(&omsg, ev->window);
	    return 0;
	}
	XSendEvent(IappXDat.dpy, ap->win, 0, KeyPressMask, (XEvent*)ev);
	SyncX();
	return 0;
}

static Window REGPARM3_FUNC HandleKeyRelease(TXev_msg *msg) {
  XKeyEvent *ev = (XKeyEvent*)msg->ev;
  Tapph *ap;
  TBuildinKeys *BKy;
  Tobjmsg omsg;
  TKeyEvent KeyEv;

	ap = SetXKey_TKey(ev, &omsg, &KeyEv, keyst_Release);

	BKy = wmRoot->BKy_key_up;
	wmRoot->BKy_key_up = NULL;
	if (ap == wmRoot->ap_KeyDown) {
// if (BKy) Debug_msg(0, "%s(%d): 0x%x 0x%x  0x%x 0x%x\n", FL, BKy->sym, KeyEv.keysym, BKy->mod & 0xff, ev->state & BKy->modmask);
	    if (BKy && !BKy_cmp(BKy, &KeyEv)) BKy = NULL;
	    if (BKy) {
 Debug_msg(0, "code=%d sym=0x%04x state=0x%02x %s\n", ev->keycode, (int)KeyEv.keysym, ev->state, BKy->Action);
		if (exec_Buildin(ap, BKy->Action, BKy->cmd, cond_0)) return 0;
	    }
	}

	if (!ap) {
	    Exec_Win_Ev(&omsg, ev->window);
	    return 0;
	}
	XSendEvent(IappXDat.dpy, ap->win, 0, KeyReleaseMask, msg->ev);
	SyncX();
	return 0;
}

static Window REGPARM3_FUNC HandleFocusIn(TXev_msg *msg) {
  Tobjmsg objmsg;
	if (wmRoot->FocusWin) {
	    objmsg.msg = msgFocus;
	    objmsg.m   = 0; //WinsFocusOut;
	    Exec_Win_Ev(&objmsg, wmRoot->FocusWin->WinID);	
	}
	return msg->ev->xany.window;
}

static Window REGPARM3_FUNC HandleFocusOut(TXev_msg *msg) {
	return msg->ev->xany.window;
}

static Window REGPARM3_FUNC HandlePropertyNotify(TXev_msg *msg) {	//28
  XPropertyEvent *ev = (XPropertyEvent*)msg->ev;
  Tapph *ap;

	for (int i=0;i<evFilter_max;i++) {
	  if (IappXDat.PropEvFilter[i].Filter_PropEv &&
	      IappXDat.PropEvFilter[i].FilterAtom && 
	      (IappXDat.PropEvFilter[i].FilterAtom == ev->atom)) {

	    IappXDat.PropEvFilter[i].Filter_PropEv(ev);
	    return 0;
	  }
	}

	if (ev->window == IappXDat.RootW) return 0;
	ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, ev->window);
	if (ap) {
	    TwmWin_PropertyNotify((TwmWin*)ap->wins, ev);
	} else TTaskbar_Chk_AppletWin_EvProc((Tobjmsg*)msg, ev->window);
	return 0;
}

static Window REGPARM3_FUNC HandleClientMessage(TXev_msg *msg) {	//33
  XClientMessageEvent *ev = (XClientMessageEvent*)msg->ev;
#ifdef	USE_DEBUG_WIN
  char *pc;
  pc = XGetAtomName(IappXDat.dpy, msg->ev->xclient.message_type);
 Debug_msg(0,"HandleClientMessage w=0x%x Atom=%d %s\n", 
		(int)msg->ev->xany.window, (int)msg->ev->xclient.message_type, pc);
  XFree(pc);
#endif

	if (ev->message_type == wmRoot->XA_ALDesk_DESTROY) {
	    if ((ev->format == 32) && ((ev->data.l[0] & 1) == 1)) {
		ExtQuitBef = ev->data.l[0];
		IappDat.gQUIT_Ev = True;
	    }
	    return 0;

	} else if (ev->message_type == IappXDat.XA_NET_WM_STATE) {
	    Tapph *ap = TwmWin_Find_WinsX_Ev(IappXDat.apph, ev->window);
	    if (ap) {
		TwmWin_NetWMState((TwmWin*)ap->wins, ev);
		return 0;
	    }

	} else if (wmRoot->Taskbar && 
		   (ev->message_type == wmRoot->Taskbar->XA_NET_SYSTEM_TRAY_OPCODE)) {
	    TTaskbar_Handle_SysTray_ev(wmRoot->Taskbar, ev);
	}

#ifdef DRAG_AND_DROP
	if (!IappXDat.dnd || !IappXDat.dnd->dnd_Message
			  || !IappXDat.dnd->dnd_Message((XClientMessageEvent*)msg->ev)) {
#if 0
  char *pc;
  pc = XGetAtomName(IappXDat.dpy, msg->ev->xclient.message_type);

 printf("%s(%d): HandleClientMessage w=0x%x Atom=%d %s\n", FL, (int)msg->ev->xany.window, (int)msg->ev->xclient.message_type, pc);
 fflush(stdout);
#endif
		return msg->ev->xany.window;
	}
#endif
	return 0;
}

// #define USE_ev_Not_uses 1

#ifdef	USE_ev_Not_uses
static Window REGPARM3_FUNC HandleCreateNotify(TXev_msg *msg) {	//16
 printf("%s(%d): HandleCreateNotify\n", FL);
	return 0;
}

static Window REGPARM3_FUNC HandleConfigureNotify(TXev_msg *msg) {	//22
  XConfigureEvent *cEv = (XConfigureEvent*)msg->ev;

 Debug_msg(0,"%s(%d): HandleConfigureNotify above=0x%x\n", FL, (int)cEv->above);

	return 0;
}

static Window REGPARM3_FUNC wmHandleSelectionClear(TXev_msg *msg) {	//29
 printf("%s(%d): HandleSelectionClear\n", FL);
	return 0;
} 

#endif

void InitEventHandlerJumpTable() {

	set_XEventHandle(KeyPress,         HandleKeyPress);		//2
	set_XEventHandle(KeyRelease,       HandleKeyRelease);		//3
//	set_XEventHandle(ButtonPress,      HandleButtonPressRelease);	//4
//	set_XEventHandle(ButtonRelease,    HandleButtonPressRelease);	//5
//	set_XEventHandle(MotionNotify,     HandleMotionNotify);		//6
//	set_XEventHandle(EnterNotify,      HandleEnterNotify);		//7
//	set_XEventHandle(LeaveNotify,      HandleLeaveNotify);		//8
	set_XEventHandle(FocusIn,          HandleFocusIn);		//9
	set_XEventHandle(FocusOut,         HandleFocusOut);		//10
//	set_XEventHandle(Expose,           HandleExpose);		//12
//	set_XEventHandle(VisibilityNotify] := HandleVisibilityNotify;    //15
	set_XEventHandle(DestroyNotify,    HandleDestroyNotify);	//17
	set_XEventHandle(MapNotify,        HandleMapNotify);		//19
	set_XEventHandle(UnmapNotify,      HandleUnmapNotify);		//18
	set_XEventHandle(MapRequest,       HandleMapRequest);		//20




	set_XEventHandle(ConfigureRequest, HandleConfigureRequest);	//23

	set_XEventHandle(PropertyNotify,   HandlePropertyNotify);	//28

//	set_XEventHandle(SelectionRequest, wmHandleSelectionRequest);	//30
//	set_XEventHandle(SelectionNotify,  wmHandleSelectionNotify);	//31

	set_XEventHandle(ClientMessage,    HandleClientMessage);	//33

#ifdef	USE_ev_Not_uses
	set_XEventHandle(CreateNotify,     HandleCreateNotify);		//16  XCreateWindowEvent
	set_XEventHandle(ConfigureNotify,  HandleConfigureNotify);	//22
	set_XEventHandle(SelectionClear,   wmHandleSelectionClear);	//29
#endif
} 
