/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldesk.h"
#include <time.h>

#define tb_DefaultWidth 160

void INTERNAL_FUNC TTaskbar_buWin_SetUpDown(TWins *self, bool down) {
	if (!self) return;
	self->BorderFu &= ~mBorderL;
	if (down) self->BorderFu |= mBorderL;
	Refresh(self);
}

void TTaskbar_buWin_OnMouseDown(TWins *self, XButtonEvent *ev) {
  TwmFrame *wmf = (TwmFrame*)self->tag;
	if (ev->button == Button1) {
	    SetVisible((TWins*)wmf, True);
	    if (wmRoot->Focus_wmWin != wmf->wmWin) {
		SetXInputFocus(wmf->wmWinID);
	    }
	} else if (ev->button == Button3) {
	    if (wmRoot->WinOptMenu) {
		wmRoot->WinOptMenu->owner = (TWins*)wmf->wmWin;
		TMenu_Show((TMenu*)wmRoot->WinOptMenu, self, ev->x_root, ev->y_root);
	    }
	}
}

void TTaskbar_buWin_OnPaint(TWins *self) {
  TwmFrame *wmf = (TwmFrame*)self->tag;
	if (wmf && wmf->wmTitle) {
	    DrawPic(self->WinID, wmf->wmTitle->pic, 3, 3);
	}
}

static int TTaskbar_buWin_GetMax_w(TWins *self) {
  int max_w;
	max_w = self->width - self->CaptionOffX - 4;
	if (max_w < 60) max_w = 60;
	return max_w;
}

void TTaskbar_buWin_OnResize(TWins *self) {
  TwmFrame *wmf = (TwmFrame*)self->tag;
  char *s;
  int max_w;
	if (!wmf || !wmf->wmTitle) return;
	max_w = TTaskbar_buWin_GetMax_w(self);
	s = vds_path_cut(&wmf->vds, wmf->vds.s, max_w);
	if (!pccmp(s, self->vds.s)) {
	    vds_SetStr(&self->vds, s);
	}
	xfree(s);
}

void TTaskbar_buWin_SetCaption(TWins *self, const char *s) {
  char *tmps;
  int max_w;

	max_w = TTaskbar_buWin_GetMax_w(self);
	tmps = vds_path_cut(&self->vds, s, max_w);
	TWins_SetCaption(self, tmps);
	xfree(tmps);
}

static void TTaskbar_SetPosSizeWin(TTaskbar *self, int Desk) {
  TWins *tbw;
  int count = 0;
  int w, x;

	if (Desk != wmRoot->CurrentDesk) return;
	tbw = self->buTaskbarWin[Desk];
	while (tbw) {count++; tbw = tbw->owner;}
	if (!count) return;

	x = self->buStart->left + self->buStart->width + 2;
	w = self->width - x - self->tbMEM->width - self->tbTime->width -
		self->tbSysTray->width - self->tbDesktop->width - 6;
	w = w / count;
	if (w > tb_DefaultWidth) w = tb_DefaultWidth;
	if (w < 4) w = 4;
	tbw = self->buTaskbarWin[Desk];
	while (tbw) {
	    SetRC(tbw, x, tbw->top, w, tbw->height);
	    x += w;
	    tbw = tbw->owner;
	}
}

void TTaskbar_AddWin(TTaskbar *self, TwmFrame *wmf) {
  TWins **pptbw, *tbw;

	if (!self || !wmf || !wmf->wmTitle) return;
	tbw = TWins_CreateLabel((TWins*)self, 60, 2, tb_DefaultWidth, 22, "New", 
				mCaptionLeft, mBorder | mBorderBu | mBorderW2);
	tbw->tag = (int)wmf;
	tbw->Visible = (wmf->wmWin->Desk == wmRoot->CurrentDesk);
	tbw->CaptionOffX = 3;
	if (wmf->wmTitle->pic) tbw->CaptionOffX += wmf->wmTitle->pic->width + 1;

	tbw->OnPaint = TTaskbar_buWin_OnPaint;
	tbw->OnMouseDown = TTaskbar_buWin_OnMouseDown;
	tbw->OnResize = TTaskbar_buWin_OnResize;
	tbw->SetCaption = TTaskbar_buWin_SetCaption;

#ifdef USE_DECOR
	tbw->Decor = TDecor_dup(&self->tbbDecor);
#endif

	Set_Parent((TWins*)tbw, (TWins*)self);

	pptbw = &self->buTaskbarWin[wmf->wmWin->Desk];
	while (*pptbw) pptbw = &(*pptbw)->owner;
	*pptbw = tbw;

	wmf->buTaskbarWin = tbw;
	TTaskbar_SetPosSizeWin(self, wmf->wmWin->Desk);
}

static TWins **get_pptbw(TTaskbar *self, TwmFrame *wmf) {
  TWins **pptbw;
	pptbw = &self->buTaskbarWin[wmf->wmWin->Desk];
	while (*pptbw && ((*pptbw)->tag != (int)wmf)) pptbw = &(*pptbw)->owner;
	return pptbw;
}

void TTaskbar_DelWin(TTaskbar *self, TwmFrame *wmf) {
  TWins **pptbw, *tbw;

	if (!self || !wmf || !wmf->buTaskbarWin) return;
	wmf->buTaskbarWin = NULL;
	pptbw = get_pptbw(self, wmf);
	if (*pptbw) {
	    tbw = *pptbw;
	    *pptbw = tbw->owner;
	    Tfree(tbw);
	}
	TTaskbar_SetPosSizeWin(self, wmf->wmWin->Desk);
}

void TTaskbar_DeskMoveWin(TTaskbar *self, TwmFrame *wmf, int ToNewDesk) {
  TWins **pptbw, *tbw;

	if (!self || !wmf || !wmf->buTaskbarWin) return;
	pptbw = get_pptbw(self, wmf);
	if (!*pptbw) return;

	tbw = *pptbw;
	*pptbw = tbw->owner;
	tbw->owner = NULL;

	pptbw = &self->buTaskbarWin[ToNewDesk];
	while (*pptbw) pptbw = &(*pptbw)->owner;
	*pptbw = tbw;

	SetVisible(tbw, False);
	TTaskbar_SetPosSizeWin(self, wmf->wmWin->Desk);

	MapWinID(wmf->WinID, 0);
	wmf->wmWin->Desk = ToNewDesk;
}



void buStart_OnMouseDown(TWins *self, XButtonEvent *ev) {
	exec_Buildin(self->apph, "menu_start", NULL, cond_0);
}

void buStart_OnPaint(TWins *self) {
	DrawPic(self->WinID, ((TTaskbar*)self->Parent)->buStart_pic, 1, 1);
}

static void TTaskbar_ChangeDeskWin(TTaskbar *self, int Desk) {
  TWins *tbw;

	tbw = self->buTaskbarWin[wmRoot->CurrentDesk];
	while (tbw) {SetVisible(tbw, False); tbw = tbw->owner;}

	tbw = self->buTaskbarWin[Desk];
	while (tbw) {SetVisible(tbw, True); tbw = tbw->owner;}
}

static void SetCurrentDesk(int Desk) {
  Tapph *ap = IappXDat.apph;
  TwmFrame *wmf;
  TwmFrame *topF = NULL;

	while (ap) {
	    if (ap->wins->wtyp == wtyp_TwmFrame) {
		wmf = (TwmFrame*)ap->wins;
		if (wmf->wmWin->Desk == Desk) {
		    if (wmf->Visible) {
			if (!topF) topF = wmf;
			MapWinID(wmf->WinID, 1);
		    }
		} else {
		    MapWinID(wmf->WinID, 0);
		}
	    }
	    ap = ap->next;
	}
	if (topF) SetXInputFocus(topF->wmWinID);
}

void tbDeskButton_DoMouseDown(TWins *self, XButtonEvent *ev) {
  TTaskbar *tb = (TTaskbar*)self->owner;
  int i;

	if (wmRoot->CurrentDesk == self->tag) return;
	for (i=0;i<4;i++) TTaskbar_buWin_SetUpDown(tb->tbDeskButton[i], False);
	tb->tbDeskButton[self->tag]->BorderFu |= mBorderL;
	SetCurrentDesk(self->tag);
	TTaskbar_ChangeDeskWin(tb, self->tag);
	wmRoot->CurrentDesk = self->tag;
	TTaskbar_SetPosSizeWin(tb, self->tag);
}

typedef struct TtbMemInfo {
  dword Total, Free, Cached, SwapTotal, SwapFree;
} TtbMemInfo;



static bool mStrScan(char *keyw, char key, char *pc, char **val) {
  bool res;

	while (*keyw && *pc && (*keyw == *pc)) {keyw++; pc++;}
	res = (*pc == key) && !*keyw;
	if (res) {
	    *val = NULL;
	    pc++;
	    while (*pc && (*pc == ' ')) pc++;
	    if (*pc) {
		*val = pc;
	    } else res = False;
	}
	return res;
}

int Get_tbMemInfo(TtbMemInfo *MemInfo) {
  int resmk = 0;
  FILE *f;
  char *tline, *val;
  char line[256];
  
	f = fopen("/proc/meminfo", "r");
	if (f) {
loop:
	    tline = fgets(line, sizeof(line)-1, f);
	    if (tline) {
		if (mStrScan("MemTotal", ':', tline, &val)) {
		    MemInfo->Total = ppctoint(&val);
		    resmk |= 0x01;
		} else if (mStrScan("MemFree", ':', tline, &val)) {
		    MemInfo->Free = ppctoint(&val);
		    resmk |= 0x02;
		} else if (mStrScan("Cached", ':', tline, &val)) {
		    MemInfo->Cached = ppctoint(&val);
		    resmk |= 0x04;
		} else if (mStrScan("SwapTotal", ':', tline, &val)) {
		    MemInfo->SwapTotal = ppctoint(&val);
		    resmk |= 0x10;
		} else if (mStrScan("SwapFree", ':', tline, &val)) {
		    MemInfo->SwapFree = ppctoint(&val);
		    resmk |= 0x20;
		}
		if (resmk != 0x37) goto loop;
	    }
	    fclose(f);
	}
	return resmk;
}

void TTaskbar_tbMEM_OnPaint(TWins *self) {
	StdBorderPaint(self->WinID, mBorder, 0, 0, 5, self->height);
	StdBorderPaint(self->WinID, mBorder, 5, 0, 5, self->height);
}

static char *get_fmtDate() {
  char *pc = "%a %d-%m-%Y %H:%M:%S";
	if (wmRoot->DateFormat == 1) {
	    pc = "%a %m-%d-%Y %H:%M:%S";
	} else if (wmRoot->DateFormat == 2) {
	    pc = "%a %Y-%m-%d %H:%M:%S";
	}
	return pc;
}

void TTaskbar_OnTimer(TgTimer *ti) {
  TTaskbar *tb = ti->obj;
  Window win;
  int m, mg, cpuh, scpuh, azheight;
  int Total, Free, Cached;
  FILE *f;
  char *tline;
  char *val;
  char line[256];
  TtbMemInfo MemInfo;

	tb->TimerZ++;

	win = tb->tbCPU->WinID;
	azheight = tb->tbCPU->width - 2;
	cpuh = 0;
	scpuh = 0;
	f = fopen("/proc/stat", "r");
	if (f) {
	    tline = fgets(line, sizeof(line)-1, f);
	    if (tline) {
		if (mStrScan("cpu", ' ', tline, &val)) {
  dword dw;
  int syst;
		    dw = tb->oCPU_USER;
		    tb->oCPU_USER = ppctoint(&val);
		    m = tb->oCPU_USER - dw;

		    dw = tb->oCPU_NICE;
		    tb->oCPU_NICE = ppctoint(&val);
		    m += tb->oCPU_NICE - dw;

		    dw = tb->oCPU_SYS;
		    tb->oCPU_SYS = ppctoint(&val);
		    syst = tb->oCPU_SYS - dw;
		    m += syst;

		    dw = tb->oCPU_IDLE;
		    tb->oCPU_IDLE = ppctoint(&val);
		    mg = m + (tb->oCPU_IDLE - dw);
		    if (mg > 0) {
			cpuh = (m * azheight) / mg;
			scpuh = (syst * azheight) / mg;
		    }
		}
	    }
	    fclose(f);
	}

	if (cpuh < azheight) 
		FillRectRGB(win, tb->tbCPU->Color, cpuh+1, 1, azheight-cpuh, 4);
	if (cpuh > 0) FillRectRGB(win, 0xc000ff, 1, 1, cpuh, 4);
	if (scpuh > 0) FillRectRGB(win, 0xff0000, 1, 1, scpuh, 4);

	if (tb->TimerZ & 0x01) {
  int MemInfoMk;
	    azheight = tb->tbMEM->height - 2;
	    win = tb->tbMEM->WinID;
	    MemInfoMk = Get_tbMemInfo(&MemInfo);
	    if ((MemInfoMk & 0x07) == 0x07) {
		Total = MemInfo.Total >> 4; //Div16
		if (Total) {
		  Free = MemInfo.Free >> 4;
		  cpuh = ((Total - Free) * azheight) / Total;
		  if (cpuh < azheight) 
		    FillRectRGB(win, tb->tbCPU->Color, 1, 1, 3, azheight-cpuh);
		  if (cpuh > 0)
		    FillRectRGB(win, 0x808000, 1, azheight-cpuh+1, 3, cpuh);
		  Cached = MemInfo.Cached >> 4;
		  cpuh = ((Total - (Free + Cached)) * azheight) / Total;
		  if (cpuh < azheight) 
		    FillRectRGB(win, 0x4040e0, 1, azheight-cpuh+1, 3, cpuh);
		}
	    }
	    if ((MemInfoMk & 0x30) == 0x30) {
		Total = MemInfo.SwapTotal >> 4; //Div16
		if (Total) {
		  Free = MemInfo.SwapFree >> 4;
		  cpuh = ((Total - Free) * azheight) / Total;
		  if (cpuh < azheight) 
		    FillRectRGB(win, tb->tbCPU->Color, 6, 1, 3, azheight-cpuh);
		  if (cpuh > 0)
		    FillRectRGB(win, 0x808000, 6, azheight-cpuh+1, 3, cpuh);
		}
	    }
	}

  time_t vti;
  struct tm *locti;
	vti = time(NULL);
	locti = localtime(&vti);
	if (tb->Oldtm_sec != locti->tm_sec) {
	    tb->Oldtm_sec = locti->tm_sec;
#ifdef MEM_CHK
  int lMCt;
	    MEM_CHK_Count(&lMCt, NULL);

	    snprintf(line, sizeof(line)-1, "%d", lMCt);
#else
	    strftime(line, sizeof(line)-1, "%H:%M:%S", locti);
#endif
	    SetCaption(tb->tbTime, line);
	    if (tb->tbTime->InfoWin->Visible) {
		strftime(line, sizeof(line)-1, get_fmtDate(), locti);
		SetCaption(tb->tbTime->InfoWin, line);
	    }
	} 

}

void DateInfo_OnVisible(TWins *self, bool Visible) {
  time_t vti;
  struct tm *locti;
  char line[80];

	if (Visible) {
	    XRaiseWindow(IappXDat.dpy, self->WinID);
	    vti = time(NULL);
	    locti = localtime(&vti);
	    strftime(line, sizeof(line)-1, get_fmtDate(), locti);
	    SetCaption(self, line);
	} else {
	    Tlgs_Clr((Tlgs*)&self->vds);
	}
}

void TTaskbar_OnResize(TWins *self) {
  TTaskbar *tb = (TTaskbar*)self;
  int x = self->width - tb->tbMEM->width - 2;
  int h = self->height - 2; 
 
	SetRC(tb->tbMEM, x, 1, tb->tbMEM->width, h);

	x -= tb->tbTime->width + 1;
	SetLeft(tb->tbCPU, x);
	SetLeft(tb->tbTime, x);
	x -= tb->tbSysTray->width + 1;
	SetLeft(tb->tbSysTray, x);
	x -= tb->tbDesktop->width + 1;
	SetLeft(tb->tbDesktop, x);

	x = self->width - tb->tbTime->InfoWin->width - 1;
	SetPosition(tb->tbTime->InfoWin, x, self->top - tb->tbTime->InfoWin->height - 1);
}

static int add_x_InfoCalcX(int width, int x) {
	if (width > 32) return x + 32;
	if (width < 16) return x + 16;
	return x + width + 1;
}

static int TTaskbar_InfoCalcX(TTaskbar *tb) {
  int x;
  Tapph *ap;

	x = 0;
	ap = tb->tbSysTray->apph->next_child;
	while (ap) {
	    x = add_x_InfoCalcX(ap->wins->width, x);
	    ap = ap->next;
	}
	return x;
}

static void TTaskbar_InfoSetSize(TTaskbar *self) {
  int w;
  int x;
  Tapph *ap;
  TWins *ws;

	w = TTaskbar_InfoCalcX(self) + 2;
	SetWidth(self->tbSysTray, w);
	TTaskbar_OnResize((TWins*)self);

	x = 1;
	ap = self->tbSysTray->apph->next_child;
	while (ap) {
	    ws = ap->wins;
	    if (ws->wtyp == wtyp_TTaskbarApplet) {
		if (ws->left != x) {
		    ws->left = x;
		    Window pWinID = ws->WinID;
		    if (((TwmWin*)ws)->SaveParentWinID) pWinID = ((TwmWin*)ws)->SaveParentWinID;
		    XMoveWindow(IappXDat.dpy, pWinID, ws->left, ws->top);
		}
		x = add_x_InfoCalcX(ws->width, x);
	    }
	    ap = ap->next;
	}
}

static int TTaskbar_AppletWin_EvProc(Tobjmsg *msg) {
  long supplied = 0;
  TWins *self;
  XEvent *ev;
  XSizeHints SizeHints;

	self = msg->ap->wins;
	if (msg->msg == msgXEv) {
	    ev = (XEvent *) msg->p;
	    switch(ev->type) {
	    case UnmapNotify:
		Tfree(self);
		return 1;

	    case ConfigureRequest:
		return 1;

	    case PropertyNotify:
		if (ev->xproperty.atom == XA_WM_NORMAL_HINTS) {
		    if (XGetWMNormalHints (IappXDat.dpy, self->WinID, &SizeHints, &supplied) == 0) {
			    SizeHints.flags = 0;
		    }
		    if (SizeHints.flags & (USSize | PSize)) {
		      self->width = SizeHints.width;
		      self->height = SizeHints.height;
		      TTaskbar_InfoSetSize((TTaskbar*)self->owner);
		    }
		}
		return 1;
	    }
	}
	return TWins_EvProc(msg);
}

bool TTaskbar_Chk_AppletWin_EvProc(Tobjmsg *msg, Window win) {
  Tapph *ap;

	ap = wmRoot->Taskbar->tbSysTray->apph->next_child;
	while (ap) {
	    if ((ap->win == win) && (ap->wins->wtyp == wtyp_TTaskbarApplet)) {
		Send_MSG(msg, ap);
		return True;
	    }
	    ap = ap->next;
	}
	return False;
}

/* _____________________ AppletWin ___________________________ */
bool TwmWin__Destroy(TwmWin *self);

bool TwmWin_AppletWin__Destroy(TwmWin *self) {
	TwmWin__Destroy(self);
	TTaskbar_InfoSetSize((TTaskbar*)self->owner);
	return False;
}

void TTaskbar_Add_AppletWin(TTaskbar *tb, TwmWin *wmw) {
  int x;
  Tapph *ap;
  Window pWinID;

	ap = AddAppWin(tb->tbSysTray->apph, (TWins*)wmw);
	ap->EvProc = TTaskbar_AppletWin_EvProc;
	ap->win = wmw->WinID;
	wmw->wtyp = wtyp_TTaskbarApplet;
	wmw->Parent = tb->tbSysTray;
	wmw->owner = (TWins*)tb;
	wmw->destroy = (TDESTROY)TwmWin_AppletWin__Destroy;

	wmw->top = 1;
	XSelectInput(IappXDat.dpy, wmw->WinID, 
				StructureNotifyMask | ResizeRedirectMask |
				PointerMotionMask | PointerMotionHintMask);

	pWinID = tb->tbSysTray->WinID;
	x = 1;
	if (wmw->SaveParentWinID) {
	    pWinID = wmw->SaveParentWinID;
	    x = 0;
	}

	XReparentWindow(IappXDat.dpy, wmw->WinID, pWinID, x, x);
	MapWinID(wmw->WinID, 2);

	TTaskbar_InfoSetSize(tb);
}

#define SYSTEM_TRAY_REQUEST_DOCK  0
#define SYSTEM_TRAY_BEGIN_MESSAGE   1
#define SYSTEM_TRAY_CANCEL_MESSAGE  2

void TTaskbar_Handle_SysTray_ev(TTaskbar *self, XClientMessageEvent *ev) {
  Window WinID;
  Tapph *ap;
  TwmWin *wmw;
  XWindowAttributes attr;
  XSetWindowAttributes XAttr;

	if (ev->data.l[1] == SYSTEM_TRAY_REQUEST_DOCK) {
	    WinID = ev->data.l[2];
	    if (WinID == None) return;
	    TwmRoot_ClearStartNotify();

/* Chk Win exist */
	    ap = self->tbSysTray->apph->next_child;
	    while (ap) {
		if (ap->wins->WinID == WinID) return;
		ap = ap->next;
	    }

	    wmw = TwmWin_Create_base(WinID, &attr);
	    if (!wmw) return;
	    if (wmw->width > 32) wmw->width = 32;
	    if (wmw->width < 4)  wmw->width = 4;
	    wmw->height = 20;
	    wmw->left = 0;
	    wmw->top =  0;

	    Fill0(&XAttr, sizeof(XSetWindowAttributes));
	    XAttr.cursor = IappXDat.XCursors[wmw->cursor_nr];
	    XAttr.background_pixel = GetColor_RGB(self->tbSysTray->Color);

	    XAttr.event_mask = (SubstructureNotifyMask | ResizeRedirectMask |
			   PointerMotionMask);
	    wmw->SaveParentWinID = XCreateWindow(IappXDat.dpy, self->tbSysTray->WinID, 
			1, 1, wmw->width, wmw->height,
			0, CopyFromParent, InputOutput, CopyFromParent,
			CWBackPixel | CWEventMask | CWCursor, &XAttr);
	    XMapWindow(IappXDat.dpy, wmw->SaveParentWinID);
	    UngrabXServer();

	    TTaskbar_Add_AppletWin(self, wmw);
	}; //???
}

/* _____________________ Create DateInfo ______________________ */
void TTaskbar_DateInfo_CreateParams(TWins_Create_struct *wcs) {
  TWins *self = wcs->wins;

	self->Visible = False;
	self->Color = 0xe0e0a0;
	self->BorderFu = mBorder|mBorderL;
	self->CaptionFu = mCaptionPaint;

	vds_SetStr(&self->vds, "Sun 19-09-2010 00:00:00");
	self->width = TextWidthColorString(&self->vds) + 8;
	self->height = self->vds.Font->height;
	self->CaptionOffY = 1;
	Tlgs_Clr((Tlgs*)&self->vds);

	wcs->XMask |= CWOverrideRedirect;
	wcs->XAttr.override_redirect = 1;
	wcs->ParentWinID = IappXDat.RootW;
}

TWins *TTaskbar_DateInfo_Create(TWins *Parent) {
  TWins *self;
  TWins_Create_struct wcs;

	TWins_Create_struct_init(&wcs, sizeof(TWins), wtyp_Wins);
	wcs.Parent = Parent;
	wcs.CreateParams = TTaskbar_DateInfo_CreateParams;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;

	TWins_SetParent(self, NULL);
	self->OnVisible = DateInfo_OnVisible;
	return self;
}

/*____________________________ Destroy __________________________________*/
bool TTaskbar__Destroy(TTaskbar *self) {
	XSetSelectionOwner(IappXDat.dpy, self->XA_NET_SYSTEM_TRAY_S,
			    None, CurrentTime);

	return TWins__Destroy((TWins*)self);
}
/* _____________________ Create _______________________________ */
void TTaskbar_CreateParams(TWins_Create_struct *wcs) {
  TTaskbar *self = wcs->wins;

	self->destroy = (TDESTROY)TTaskbar__Destroy;
	self->top = IappXDat.Root_height - 26;
	self->width = IappXDat.Root_width;
	self->height = 26;

	self->OnResize = TTaskbar_OnResize;

	self->BorderFu = mBorder;
}

TTaskbar *TTaskbar_Create(TWins *Parent) {
  TTaskbar *self;
  TWins_Create_struct wcs;
  char *pc;
  TpcIni *ini;
  TDesktopHead ddkh;
  char buf[512];

	TWins_Create_struct_init(&wcs, sizeof(TTaskbar), wtyp_TTaskbar);
	wcs.CreateParams = TTaskbar_CreateParams;
	wcs.WCSfkt = WCSfkt_OverrideRedirect;
	TWins_DefaultConf(&wcs);
	self = wcs.wins;
	Set_Parent((TWins*)self, NULL);

	pc = buf;
	ppcmove(&pc, wmRoot->menuStart->path);
	ppcmove(&pc, ".directory");
	ini = TpcIni_Create(buf);
	ddkh.name = "Start";
	if (dotdesktop_get_head(ini, &ddkh)) {
	    self->buStart_pic = GetDeskIcon(ddkh.ddkt, 16, ddkh.icon, 0, NULL);
	}
	self->buStart = TWins_CreateLabel((TWins*)self, 2, 2, 60, 22, ddkh.name, 0, mBorder);
	Tfree(ini);

	self->buStart->OnMouseDown = buStart_OnMouseDown;
	self->buStart->OnPaint = buStart_OnPaint;
	SetFont(self->buStart, IappXDat.Default_bold_Font->name);
	if (self->buStart_pic) {
	    self->buStart->CaptionFu |= mCaptionLeft;
	    self->buStart->CaptionOffX = self->buStart_pic->width;
	}

	self->tbMEM = TWins_CreateB((TWins*)self, 66, 1, 10, 22, 0);
	self->tbMEM->OnPaint = TTaskbar_tbMEM_OnPaint;

	self->tbCPU = TWins_CreateB((TWins*)self, 66, 18, 60, 6, mBorder | mBorderL);

	self->tbTime = TWins_CreateLabel((TWins*)self, 66, 2, 60, 16, "00:00:00", 0, mBorder | mBorderL);
	SetFont(self->tbTime, IappXDat.Default_bold_Font->name);
	self->tbTime->InfoWin = TTaskbar_DateInfo_Create(self->tbTime);

	self->tbSysTray = TWins_CreateB((TWins*)self, 66, 2, 2, 22, mBorder | mBorderL);

	sprintf(buf, "_NET_SYSTEM_TRAY_S%u", DefaultScreen(IappXDat.dpy));
	self->XA_NET_SYSTEM_TRAY_S = XInternAtom(IappXDat.dpy, buf, 0);
	self->XA_NET_SYSTEM_TRAY_OPCODE = XInternAtom(IappXDat.dpy, 
					"_NET_SYSTEM_TRAY_OPCODE", 0);
	XSetSelectionOwner(IappXDat.dpy, self->XA_NET_SYSTEM_TRAY_S,
			    self->tbSysTray->WinID, CurrentTime);

	if (XGetSelectionOwner(IappXDat.dpy, self->XA_NET_SYSTEM_TRAY_S) == self->tbSysTray->WinID) {
	    XClientMessageEvent ev;
	    Fill0(&ev, sizeof(ev));
	    ev.type = ClientMessage;
	    ev.window = IappXDat.RootW;
	    ev.message_type = XInternAtom(IappXDat.dpy, "MANAGER", False);
	    ev.format = 32;
	    ev.data.l[0] = CurrentTime;
	    ev.data.l[1] = self->XA_NET_SYSTEM_TRAY_S;
	    ev.data.l[2] = self->tbSysTray->WinID;
	    XSendEvent(IappXDat.dpy, IappXDat.RootW, 0, StructureNotifyMask, (XEvent*)&ev);
	}

	self->tbDesktop = TWins_CreateB((TWins*)self, 66, 2, 31, 22, 0);

  int i;
  int x = 0;
  int y = 0;
  int m = mBorder | mBorderL;
	for (i=0;i<4;i++) {
	    self->tbDeskButton[i] = TWins_CreateB(self->tbDesktop, x, y, 15, 11, m);
	    self->tbDeskButton[i]->DoMouseDown = tbDeskButton_DoMouseDown;
	    self->tbDeskButton[i]->tag = i;
	    self->tbDeskButton[i]->owner = (TWins*)self;
	    x += 16;
	    m = mBorder;
	    if (i == 1) {x = 0; y = 11;}
	}

#ifdef USE_DECOR
	self->tbbDecor = TDecor_w3b3_Create(&TDecoBuR3B3_up_Tab, &TDecoBuR3B3_down_Tab);
#endif
	self->tbTimer = gTimerX_Add(Timer_Enabled, 250, TTaskbar_OnTimer);
	self->tbTimer->obj = self;

	return self;
}


