/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"
#include "dlg.h"

static int LastPosX, LastPosY;

typedef struct {
  AlgTWins
  AlgTForm

  TWins *pnlTop;
  TWins *pnlMiddle;
  TWins *pnlBottom;

  TEdit *edFrom;

  TEdit *edFile;
  TWins *lblStk;
  TWins *lblMB;

  TButton *buCancle;

  TLines *tls;
  TglobPath *gp_src;
  Txpy_wd xpy_wd;
  dword close_us;
  TgTimer *gTimer;
  char *timer_fn;
  char *timer_path_src;
  bool stop;
  bool SkipDir;
} TFormDelRun;

static int msg_CopyEv(Txpy_wd wd, int xcyMsg) {
  int res = 0;
  TFormDelRun *cf;

	cf = XCopy_GetOwner(wd);
	SetEnabled((TWins*)cf->buCancle, False);

	res = xcy_ErrorMsgShow((TWins*)cf, wd, xcyMsg);

	SetEnabled((TWins*)cf->buCancle, True);
	return res;
}

static void set_txt(TFormDelRun *self, char *fn) {
	ppcchange(&self->timer_fn, fn);
	xfree(self->timer_path_src);
	self->timer_path_src = XCopy_GetSrc_gp_fn(self->xpy_wd, gptfu_noPasswd | gptfu_noName);
}

static bool Copy_StartWait(TFormDelRun *self, Tstat64 *st_src) {
	self->SkipDir = False;
	XCopy_Start(self->xpy_wd, st_src, xcySFU_Del);
	while (AppHandleMessage() && XCopy_ChkWait(self->xpy_wd));
	return (XCopy_ChkErr(self->xpy_wd) == 0) && !self->stop;
}

static TDirList *Copy_TDirList(TglobPath *gp) {
  TDirList *dl = TDirList_Create_gp_name(gp);
	if (dl) {
	    if (TDirList_Open(dl, dlobef_level_0)) {
		TDirList_Open_Wait(dl, dlrunstat_ScanEnd);
		TDirFileList_Sort(&dl->dfl, dlsortFu_Name);
		return dl;
	    }
	    Tfree(dl);
	}
	return NULL;
}

bool TFormDelRun_CopyDir(TFormDelRun *self, TDirList *dl, TDirFileList *dfl) {
  char *fn;
  TDirList *ndl;
  TDirFileList *ndfl;
  bool Ok = True;

	if (self->stop) return False;
	fn = dfl->fn;
	if (!fn || ((fn[0] == '.')&&(fn[1] == 0)) ||
	    ((fn[0] == '.')&&(fn[1] == '.')&&(fn[2] == 0))) return True;

// printf("%s(%d): %s %s\n", FL, tmp_gp_dest.path, tmp_gp_dest.name);
	XCopy_NameToPath(self->xpy_wd);
	set_txt(self, fn);
	XCopy_ChangeName(self->xpy_wd, fn);

	Ok = Copy_StartWait(self, dfl->st);
	if (!Ok || self->SkipDir) goto ende;
	if ((dfl->st->st & st_Link) || !(dfl->st->st & st_Dir)) goto ende;

	dl->gp.name = fn;

	ndl = Copy_TDirList(&dl->gp);
	if (ndl) {
	    ndfl = ndl->dfl;
	    while (ndfl && Ok) {
		Ok = TFormDelRun_CopyDir(self, ndl, ndfl) && !self->stop;
		ndfl = ndfl->next;
	    }
	    Tfree(ndl);
	}
	if (Ok) {
	    dword skip = XCopy_GetSkip(self->xpy_wd);
	    XCopy_SetSkip(self->xpy_wd, skip & ~xcySkip_Dir);
	    Ok = Copy_StartWait(self, dfl->st);
	    skip = XCopy_GetSkip(self->xpy_wd) | xcySkip_Dir;
	    XCopy_SetSkip(self->xpy_wd, skip);
	}

ende:
	dl->gp.name = NULL;
	XCopy_PathToName(self->xpy_wd);
	set_txt(self, NULL);
	return Ok;
}

bool TFormDelRun_CopyStart(TFormDelRun *self, char *fn) {
  Tstat64 *src_st;
  TglobPath *src_gp;
  bool Ok;

	if (self->stop) return False;
	if (((fn[0] == '.')&&(fn[1] == 0)) ||
	    ((fn[0] == '.')&&(fn[1] == '.')&&(fn[2] == 0))) return True;
	
	set_txt(self, fn);
	XCopy_ChangeName(self->xpy_wd, fn);

	Ok = Copy_StartWait(self, NULL);
	if (!Ok || self->SkipDir) return Ok;

	src_st = XCopy_GetSrcStat(self->xpy_wd);
	if (!src_st) {
 printf("%s(%d): no src stat ???\n", FL);
	    return True;
	}
	if ((src_st->st & st_Link) || !(src_st->st & st_Dir)) goto ende;

	src_gp = XCopy_GetSrc_gp(self->xpy_wd);

  TDirList *ndl;
	ndl = Copy_TDirList(src_gp);
	if (ndl) {
	    TDirFileList *ndfl = ndl->dfl;
	    while (ndfl && Ok) {
		Ok = TFormDelRun_CopyDir(self, ndl, ndfl) && !self->stop;
		ndfl = ndfl->next;
	    }
	    Tfree(ndl);
	}
	gpFree(src_gp);

	if (Ok) {
	    dword skip = XCopy_GetSkip(self->xpy_wd);
	    XCopy_SetSkip(self->xpy_wd, skip & ~xcySkip_Dir);
	    Ok = Copy_StartWait(self, src_st);
	    skip = XCopy_GetSkip(self->xpy_wd) | xcySkip_Dir;
	    XCopy_SetSkip(self->xpy_wd, skip);
	}

ende:
	xfree(src_st);
	return Ok;
}


/* _____________________________ Ev __________________________________ */
static void buCancle_OnClick(TWins *self, long KeyState, bool DblClick) {
  TFormDelRun *cf = (TFormDelRun*)GetParentForm((TWins*)self);
	XCopy_Stop(cf->xpy_wd);
	cf->stop = True;
}

/* ____________________________ Resize _______________________________ */
static void pnlTop_OnResize(TWins *self) {
  TFormDelRun *cf = (TFormDelRun*)GetParentForm(self);
  int w;
	w = self->width - 43;
	if (w < 40) w = 40;
	SetWidth((TWins*)cf->edFrom, w);
}

static void pnlMiddle_OnResize(TWins *self) {
  TFormDelRun *cf = (TFormDelRun*)GetParentForm(self);
  int w;
	w = self->width - 43;
	if (w < 40) w = 40;
	SetWidth((TWins*)cf->edFile, w);
}


static void pnlBottom_OnResize(TWins *self) {
  TFormDelRun *cf = (TFormDelRun*)GetParentForm(self);
  int w;

	w = self->width - 2 - cf->buCancle->width;
	SetLeft((TWins*)cf->buCancle, w);
}

/* ____________________________ Timer Ev _____________________________ */
static void TFormDelRun_OnTimer(TgTimer* gTimer) {
  TFormDelRun *cf = gTimer->obj;
  float flg;
  TXCopyStats xcyStats;
  char *pc;
  char buf[80];

	SetCaption((TWins*)cf->edFile, cf->timer_fn);
	SetCaption((TWins*)cf->edFrom, cf->timer_path_src);

	if (XCopy_GetStats(cf->xpy_wd, &xcyStats)) {

	    inttopc(buf, xcyStats.FileCount);
	    SetCaption((TWins*)cf->lblStk, buf);

	    flg = xcyStats.GesDataSize + xcyStats.DataSize;
	    flg /= 0x100000;
	    pc = buf;
	    floattoppc(&pc, flg, 0x03);
	    ppcmove(&pc, " MB");
	    SetCaption((TWins*)cf->lblMB, buf);
	}

}

/* ____________________________ Destroy ______________________________ */
bool TFormDelRun__Destroy(TFormDelRun *self) {
	Tfree(self->gTimer);
	Tfree(self->tls);
	gpFree(self->gp_src);
	xfree(self->timer_fn);
	xfree(self->timer_path_src);
	return TWins__Destroy((TWins*)self);
}

static void mf_OnCanClose(TForm *self, int *CanClose) {
	if (((TFormDelRun*)self)->buCancle->Enabled) {
	    *CanClose = FormCloseAndFree;
	} else {
	    *CanClose = False;
	}
}

/* ____________________________ Create _______________________________ */
void TFormDelRun_Show(TWins *Parent, TLines *tls_src, TglobPath *gp_src, int skip) {
  TMainForm *mf;
  TFormDelRun *self;
  char *pc;
  int count;

	if (!tls_src || !gp_src) return;
	mf = IappDat.obj;
	self = TForm_CreatParamAlg(Parent, sizeof(TFormDelRun));
	self->destroy = (TDESTROY)TFormDelRun__Destroy;
	self->FormZStyle = fszTransient;
	if (LastPosX || LastPosY) {
	    self->left = LastPosX;
	    self->top = LastPosY;
	    self->FormPosition = poDesigned;
	} else {
	    self->FormPosition = poCenter;
	}
	self->width = 480;
	self->height = 140;
	self->min_width = 360;
	self->min_height = 140;
	SetCaption((TWins*)self, "Del");
	self->OnCanClose = mf_OnCanClose;
	self->tls = tls_src;
	self->gp_src = gp_src;

	self->pnlTop = TWins_CreateBA((TWins*)self, 0, 0, 200, 41, mBorder, alTop);
	self->pnlTop->OnResize = pnlTop_OnResize;
	self->pnlMiddle = TWins_CreateBA((TWins*)self, 0, 0, 200, 21, mBorder, alClient);
	self->pnlMiddle->OnResize = pnlMiddle_OnResize;
	self->pnlBottom = TWins_CreateBA((TWins*)self, 0, 0, 200, 28, mBorder, alBottom);
	self->pnlBottom->OnResize = pnlBottom_OnResize;

	TWins_CreateLabel(self->pnlTop, 2,  3, 28, 16, "from:", mCaptionLeft, 0);
	TWins_CreateLabel(self->pnlTop, 2, 22, 28, 16, "to :", mCaptionLeft, 0);

	pc = get_DisplayPathStr(NULL, self->gp_src);
	self->edFrom = TEdit_CreateS(self->pnlTop, 40, 2, 45, 18, pc);
	xfree(pc);
	self->edFrom->KeyFu |= mNoKeyFocus;
	self->edFrom->ReadOnly = True;
	SetColor((TWins*)self->edFrom, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edFrom, IappXDat.FileSystemUTF8on);


	TWins_CreateLabel(self->pnlMiddle, 2,  3, 28, 16, "file:", mCaptionLeft, 0);

	self->edFile = TEdit_CreateS(self->pnlMiddle, 40, 2, 45, 18, NULL);
	self->edFile->KeyFu |= mNoKeyFocus;
	self->edFile->ReadOnly = True;
	SetColor((TWins*)self->edFile, 0xe0e0e0);
	TWins_SetUTF8((TWins*)self->edFile, IappXDat.FileSystemUTF8on);

	self->lblStk  = TWins_CreateLabel(self->pnlMiddle,  40, 22, 38, 18, "0", mCaptionRigth, mBorder|mBorderL);
	self->lblMB   = TWins_CreateLabel(self->pnlMiddle,  80, 22, 78, 18, "0.000 MB", mCaptionRigth, mBorder|mBorderL);

	self->buCancle = TButton_CreateS(self->pnlBottom, 2, 2, 64, 24, "CANCEL");
	self->buCancle->OnClick = buCancle_OnClick;

	Set_Focus((TWins*)self->buCancle);

	TForm_Show((TForm*)self);

/* Del Start */

	count = TLines_GetLineCount(self->tls);
	if (count > 0) {
	    char *fn = TLines_Get_pc(self->tls, 0);
	    if (fn) SetCaption((TWins*)self->edFile, fn);
	}

	skip |= xcySkip_Dir;
	self->xpy_wd = XCopy_Create(self, msg_CopyEv, self->gp_src, NULL, skip);

	if (!self->xpy_wd) {
	    MsgBox_Show_Error((TWins*)mf, NULL, msgCANCEL, 
			msgCANCEL, EMSG_ERR, 
			"libaldccp can not handle del");

	} else {
	    self->gTimer = gTimerX_Add(Timer_Enabled, 50, TFormDelRun_OnTimer);
	    self->gTimer->obj = self;

	    for (int i=0;i<count;i++) {
		char *fn = TLines_Get_pc(self->tls, i);
		if (fn) {
		    SetCaption((TWins*)self->edFile, fn);
		    if (!TFormDelRun_CopyStart(self, fn)) break;
		}
	    }
	    Tfree(self->xpy_wd);
	    self->xpy_wd = NULL;
	}

	self->close_us = IappDat.app_us;
	while (((IappDat.app_us - self->close_us) < 100000) &&
		AppHandleMessage());

	LastPosX = self->left;
	LastPosY = self->top;
	TForm_Close((TForm*)self);
	return;
}
