/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldfile.h"
#include "dlg.h"

#define en_all		1
#define en_dfl		(1<<1)
#define en_gp		(1<<2)
#define en_paste	(1<<3)

#define vi_fw		(1<<4)
#define vi_tv		(1<<5)
#define vi_tvnet	(1<<6)
#define vi_tvftp	(1<<7)

#define fkt_cut		(1<<16)

/* ______________________ misc __________________________________________ */
bool get_MenuFileItem_xdfl(TMenuFile *self, Txdfl *xdfl) {
  TMainForm *mf = IappDat.obj;
  int wtyp;

	wtyp = self->owner->wtyp;
	if (wtyp == wtyp_MenuButton) {
	    if (mf->LastNoMenuFocusWin) wtyp = mf->LastNoMenuFocusWin->wtyp;
	}

	xdfl->gpList = NULL;
	if (wtyp == wtyp_TTreeVFile) {
	    xdfl->istv = True;
	    xdfl->o.tv = mf->tv;
	    xdfl->dfl = TTreeVFile_GetSelect_dfl(mf->tv);
	    xdfl->gp = TTreeVFile_GetSelect_gp(mf->tv, False);
	} else {
	    xdfl->istv = False;
	    xdfl->o.fw = mf->fw;
	    xdfl->dfl = TFileWin_Get_Select_dfl(mf->fw, NULL);
	    xdfl->gp = TFileWin_Get_gp(mf->fw);
	}
	xdfl->isLocalF = gp_test_local_drive(xdfl->gp);
	return xdfl->dfl != 0;
}

/* ______________________ menuNew _______________________________________ */
static void menuNew_OnMenuOpen(TMenu *self) {
  TMainForm *mf = (TMainForm*)self->ParentForm;
  TglobPath *gp;
  TMenuItem *mi;
  bool mk;

	gp = TFileWin_Get_gp(mf->fw);
	mi = self->MenuItem;
	while (mi) {
	    mi->fu &= ~miNotEnabled;
	    if (mi->tag == ffnNewDirectory) {
		mk = net_mkdir_gp_test(gp);
	    } else if (mi->tag == ffnNewTxtFile) {
		mk = net_file_new_va_gp_test(gp);
	    } else {
		mk = gp_test_local_drive(gp);
	    }

	    if (!mk) mi->fu |= miNotEnabled;
	    mi = mi->next;
	}
}

static void miNewFile_OnClick(TMenu *self, long KeyState, TMenuItem *mi) {
  TNewFile_Dlg *nf_dlg;
  TMainForm *mf = (TMainForm*)self->ParentForm;

	nf_dlg = Create_NewFile_Dlg((TWins*)mf, mi->tag);
	nf_dlg->gp = gp_dup(TFileWin_Get_gp(mf->fw));
	NewFile_Dlg_Execute(nf_dlg);
	Tfree(nf_dlg);
}

static void miNewURL_OnClick(TMenuFile *self, long KeyState, TMenuItem *mi) {
  TMainForm *mf = IappDat.obj;
  Txdfl xdfl;

	if (!get_MenuFileItem_xdfl(self, &xdfl)) {
	    if (xdfl.istv) {
		TFormPropNet_Show(mf, &xdfl);
	    }
	}
}

/* ______________________ menu OpenWith ______________________________ */
static bool set_next_xdfl(Txdfl *xdfl) {
	if (xdfl->istv) {
	    xdfl->dfl = NULL;
	} else {
	    xdfl->dfl = TFileWin_Get_Select_dfl(xdfl->o.fw, xdfl->dfl);
	}
	return xdfl->dfl != 0;
}

static TDesktop_gpList *get_gpList_xdfl(Txdfl *xdfl) {
  TDesktop_gpList *gpList;

	gpList = xcalloc(sizeof(TDesktop_gpList));
	gp_dup_gpdata_path(&gpList->gp, xdfl->gp);
	if (xdfl->dfl) gpList->gp.name = pcdup(xdfl->dfl->fn);
	return gpList;
}

static void set_xdfl_gpList(Txdfl *xdfl) {
  TDesktop_gpList *gpList;

	xdfl->gpList = get_gpList_xdfl(xdfl);
	gpList = xdfl->gpList;

	while (set_next_xdfl(xdfl)) {
	    gpList->next = get_gpList_xdfl(xdfl);
	    gpList = gpList->next;
	}
}

static void free_xdfl(Txdfl *xdfl) {
  TDesktop_gpList *gpList, *ngpList;

	gpList = xdfl->gpList;
	xdfl->gpList = NULL;
	while (gpList) {
	    ngpList = gpList->next;
	    gp_pc_free(&gpList->gp);
	    xfree(gpList);
	    gpList = ngpList;
	}
}

void OpenWith_Click(TMenuFile *self, int fu) {
#ifndef _OS_WIN32
  int ddkt, appFlag;
  TMenuFileItem *mfi = (TMenuFileItem*)self->SelItem;
  char *path, *fn, *token, *pc_appFlag;
  TpcIni *ini, *appIni;
  Txdfl xdfl;
  TDesktopCMD dskcmd;

	if (!mfi || !mfi->filename) return;
	fn = path_add(self->path, mfi->filename);
	ddkt = ddkt_desktop(fn);
	if (ddkt != ddkt_Application) goto ende_10;

	get_MenuFileItem_xdfl(self, &xdfl);

	path = "/";
	ini = TpcIni_Create(fn);
	dskcmd.Exec = TpcIni_RD(ini, Desktop_Entry, "Exec");

	GetToken(GetFileName_pc(dskcmd.Exec), &token);

	if (!token) goto ende;
	appIni = get_appIni();
	pc_appFlag = TpcIni_RD(appIni, "AppExecFlag", token);
	free_appIni(appIni);
	xfree(token);

	appFlag = 0;
	while (pc_appFlag) {
	    pc_appFlag = GetToken(pc_appFlag, &token);
	    if (!pccasecmpASCII(token, "chdir")) {
	        path = xdfl.gp->path;

	    } else if (!pccasecmpASCII(token, "onlyLocal")) {
		appFlag |= 0x01;

	    } else if (!pccasecmpASCII(token, "onlydir")) {
		appFlag |= 0x02;
	    }
	    xfree(token);
	}

	if ((appFlag & 0x01) && !xdfl.isLocalF) goto ende;

	if (appFlag & 0x02) {
	    xdfl.gpList = xcalloc(sizeof(TDesktop_gpList));
	    gp_dup_gpdata_path(&xdfl.gpList->gp, xdfl.gp);
	} else {
	    if (!xdfl.dfl) goto ende;
	    set_xdfl_gpList(&xdfl);
	}
	dskcmd.gpList = xdfl.gpList;

	get_desktop_dskcmd(&dskcmd);
	if (dskcmd.ret_cmd) exec_pc_co_path(dskcmd.ret_cmd, NULL, path);

	xfree(dskcmd.ret_cmd);
	free_xdfl(&xdfl);
ende:
	Tfree(ini);
ende_10:
	xfree(fn);

#endif /* _OS_WIN32 */
}

static void menuOpenWith_OnClick(TMenuFile *self, long KeyState, bool DblClick) {
	OpenWith_Click(self, 0);
}

static void menuOpenWithFind_OnClick(TMenuFile *self, long KeyState, TMenuItem *mi) {
  TMainForm *mf = (TMainForm*)self->ParentForm;
  TMimeDat md;
  Txdfl xdfl;
  TDesktopCMD dskcmd;

	if (!get_MenuFileItem_xdfl(self, &xdfl)) return;

	if (TFormExecWith_Show((TWins*)mf->fw, xdfl.dfl->fn, xdfl.dfl->st, &md)) {

	    set_xdfl_gpList(&xdfl);
	    dskcmd.gpList = xdfl.gpList;
	    dskcmd.Exec = md.exec;
	    get_desktop_dskcmd(&dskcmd);
	    if (dskcmd.ret_cmd) {
#ifndef _OS_WIN32
		exec_pc(dskcmd.ret_cmd);
#endif
		xfree(dskcmd.ret_cmd);
	    }
	    free_xdfl(&xdfl);
	    xfree(md.exec);
	}
}

/* ______________________ menu Properties ____________________________ */
static void miProperties_OnClick(TMenuFile *self, long KeyState, TMenuItem* mi) {
  TMainForm *mf = IappDat.obj;
  TDirFileList *dfl0, **ppdfl;
  Txdfl xdfl;

	dfl0 = NULL;
	if (!get_MenuFileItem_xdfl(self, &xdfl)) {
	    if (xdfl.istv && xdfl.gp) {
		if (gp_test_local_drive(xdfl.gp)) {
		    goto prop_show;
		} else {
		    TFormPropNet_Show(mf, &xdfl);
		}
	    }
	    return;
	}

	ppdfl = &dfl0;
	*ppdfl = dfl_dup(xdfl.dfl);
	
	while (set_next_xdfl(&xdfl)) {
	    ppdfl = &(*ppdfl)->next;
	    *ppdfl = dfl_dup(xdfl.dfl);
	}
prop_show:
	//dfl0 free in TFormProperty
	TFormProperty_Show((TWins*)mf, xdfl.gp, dfl0, app_mime, xdfl.istv);
}

/* ______________________ menu Copy __________________________________ */
static char *get_sel_str(TMenuFile *self, bool passwd) {
  Txdfl xdfl;
  char *result;
  int gptfu;

	if (!get_MenuFileItem_xdfl(self, &xdfl)) return NULL;
	gptfu = gptfu_noPasswd;
	if (passwd) gptfu = gptfu_PasswdTxt;
	result = NULL;

	if (xdfl.istv) {
	    return gptopc(xdfl.gp, gptfu | gptfu_del_slash);
	}

loop:
	if (!pccmp(xdfl.dfl->fn, "..")) {
	    char *si = xdfl.gp->name;
	    xdfl.gp->name = xdfl.dfl->fn;
	    char *tmp = gptopc(xdfl.gp, gptfu);
	    xdfl.gp->name = si;
	    ppcadd2(&result, tmp, "\n");
	    xfree(tmp);
	}
	if (set_next_xdfl(&xdfl)) goto loop;
	return result;
}

static void miCopy_OnClick(TMenuFile *self, long KeyState, TMenuItem* mi) {
  TMainForm *mf = IappDat.obj;
  char *pcsel;
  bool utf8;

	IappXDat.SelCut = (mi->tag & fkt_cut) != 0;
	Tlgs_Clr(&IappXDat.SelTxt);
	Tlgs_Clr(&IappXDat.SelTxt_URL);

	pcsel = get_sel_str(self, False);
	if (!pcsel) return;

	utf8 = mf->fw->vds.UTF8on;
	if (utf8) {
	    int fuEOL = 0;
	    Chk_fio_UTF8(pcsel, pclen(pcsel), &fuEOL);
	    if ((fuEOL & fio_has_0x80_char) && 
		!(fuEOL & fio_has_UTF8_char)) utf8 = False;
	}
	pctoUTF8esclgs(pcsel, &IappXDat.SelTxt);
	xfree(pcsel);

	pcsel = get_sel_str(self, True);
	utf8 = mf->fw->vds.UTF8on;
	if (utf8) {
	    int fuEOL = 0;
	    Chk_fio_UTF8(pcsel, pclen(pcsel), &fuEOL);
	    if ((fuEOL & fio_has_0x80_char) && 
		!(fuEOL & fio_has_UTF8_char)) utf8 = False;
	}
	pctoUTF8esclgs(pcsel, &IappXDat.SelTxt_URL);
	xfree(pcsel);

	if (utf8) {
#ifdef _OS_WIN32
//	    SetSelection(2, CF_UNICODETEXT, XA_STRING);
#else
	    SetSelection(4, IappXDat.XA_URL_STRING, IappXDat.XA_UTF8_STRING, XA_STRING, IappXDat.XA_KDE_CUT);
#endif
	} else {
#ifdef _OS_WIN32
#else
	    SetSelection(3, IappXDat.XA_URL_STRING, XA_STRING, IappXDat.XA_KDE_CUT);
#endif
	}
}

/* ______________________ menu Del ___________________________________ */
static void miDelete_OnClick(TMenuFile *self, long KeyState, TMenuItem* mi) {
#ifndef _OS_WIN32
  TMainForm *mf = IappDat.obj;
  Txdfl xdfl;
  TTreeVNode *tvn;
  char *pcsel;

	IappXDat.SelCut = 0;

	if (!get_MenuFileItem_xdfl(self, &xdfl)) {
	    if (!xdfl.istv) return;
	    tvn = TTreeVFile_GetSelectNode(xdfl.o.tv);
	    if (!tvn || ((tvn->art & tfvfart_MASK) != tfvfart_ROOT)) return;
	    if (!tvn->tagp) {
		if (TTreeVFileNode_TestIsDummyNode(tvn)) {
		    Tfree(tvn);
		}
		return;
	    }

	    pcsel = pcdup(tvn->tagp);
	} else {
	    pcsel = get_sel_str(self, False);
	}

	TFormDel_Show((TWins*)mf->fw, pcsel);
	xfree(pcsel);
#endif
}
/* ______________________ menu Paste _________________________________ */
static void miPaste_OnClick(TMenu *self, long KeyState, TMenuItem* mi) {
#ifndef _OS_WIN32
  TMainForm *mf = IappDat.obj;
  char *pc;
  Atom SelAtom, CutAtom;
  TglobPath *gp;

	gp = TFileWin_Get_gp(mf->fw);
	if (!gp) return;

	SelAtom = CutAtom = 0;
	SelRetClr();
	if (Get_XA_TARGETS()) {
	    SelAtom = get_nSelAtom(IappXDat.XA_URL_STRING);
	    CutAtom = get_nSelAtom(IappXDat.XA_KDE_CUT);
	}
	SelRetClr();
	if (!SelAtom) return;

	if (CutAtom) {
	    if (GetPasteDaten(IappXDat.CLIPBOARD_ATOM, CutAtom)) {
		IappXDat.SelCut = IappXDat.SelRet_lgs.s[0] & 0x01;
	    }
	}

	if (GetPasteDaten(IappXDat.CLIPBOARD_ATOM, SelAtom)) {
	    if (IappXDat.SelRetFormat != 8) {
		SelRetClr();
		return;
	    }
	}

	pc = UTF8esctopc(IappXDat.SelRet_lgs.s);
	TFormCopy_Show((TWins*)mf->fw, pc, gp);
	xfree(pc);
#endif
}

TglobPath *pcsel_to_tls(char *pcsel, TLines *tls) {
  Tlgs lgs;
  TLines *tmp_tls;
  TColS *cs;
  TglobPath *result;

	tmp_tls = TLines_Create();
	lgs.s = pcsel;
	lgs.lg = pclen(pcsel);
	TLines_Paste_lgs(tmp_tls, NULL, NULL, &lgs, 4096);

	for (int i=0; i<tmp_tls->Count;i++) {
	    cs = TLines_Get_pcs(tmp_tls, i);
	    TLines_Add(tls, GetFileName_pc(cs->s));
	}

	cs = TLines_Get_pcs(tmp_tls, 0);
	result = pctogp(cs->s);
	ppcfree(&result->name);
	Tfree(tmp_tls);

	return result;
}

/* ______________________ OnMenu _____________________________________ */
static void menuEdit_OnMenuOpen(TMenuFile *self) {
  TMenuItem *mi;
  Txdfl xdfl;
  int VisibleMk, EnabledMk;
  bool dflok;

	VisibleMk = 0;
	dflok = get_MenuFileItem_xdfl(self, &xdfl);
	if (xdfl.istv) {
	    if (dflok) {
		VisibleMk = vi_tv;
	    } else {
		if (xdfl.gp) {
		    VisibleMk = vi_tvnet;
		    if (gp_test_local_drive(xdfl.gp)) VisibleMk = vi_tv;
		} else {
		    TMainForm *mf = IappDat.obj;
		    TTreeVNode *tvn = TTreeVFile_GetSelectNode(xdfl.o.tv);
		    while (tvn) {
			if (tvn == mf->tvftp) {
			    VisibleMk = vi_tvftp;
			    break;
			}
			tvn = TTreeVNode_get_prev_child(tvn);
		    }
		}    
	    }

	} else {
	    if (xdfl.dfl && pccmp(xdfl.dfl->fn, "..")) {
		set_next_xdfl(&xdfl);
	    }
	    VisibleMk = vi_fw;
	}

	EnabledMk = 0;
	if (dflok) EnabledMk = en_dfl;
	if (xdfl.gp) EnabledMk |= en_gp;
#ifndef _OS_WIN32
	if (XCopy_GetHande(xdfl.gp, True)) {
	    if (Chk_GetSelection(1, IappXDat.XA_URL_STRING)) EnabledMk |= en_paste;
	}
#endif

	mi = self->MenuItem;
	while (mi) {
	    mi->fu |= miNotVisible|miNotEnabled;
	    if (mi->tag & VisibleMk) mi->fu &= ~miNotVisible;
	    if ((mi->tag & EnabledMk) || (mi->tag & en_all)) mi->fu &= ~miNotEnabled;
	    mi = mi->next;
	}
}

/* ______________________ Create _____________________________________ */
void MenuEdit_Create(TMainForm *self) {
  TMenuItem *mi;

	self->mbEdit = TMenuButton_CreateS(self->pnlTop, self->mbFile, "Edit");

	self->menuEdit = TMenu_Create((TWins*)self);
	self->mbEdit->menu = self->menuEdit;
	self->menuEdit->OnMenuOpen = (TMenuEv)menuEdit_OnMenuOpen;
	mi = TMenu_AddItem(self->menuEdit, "Create New", 0, NULL);
	mi->tag = en_gp|vi_fw;
	  self->menuNew = TMenu_Create((TWins*)self);
	  self->menuNew->OnMenuOpen = menuNew_OnMenuOpen;
	  mi->next_menu = self->menuNew;
	  mi = TMenu_AddItem(self->menuNew, "Directory", 0, miNewFile_OnClick);
	  mi->tag = ffnNewDirectory;
	  mi->pic = get_fileIcon(stft_Folder);
	  mi = TMenu_AddItem(self->menuNew, "Text File", 0, miNewFile_OnClick);
	  mi->tag = ffnNewTxtFile;
	  mi->pic = get_fileIcon(stft_Txt);

	TMenu_AddItem(self->menuEdit, "Create New URL", 0, 
		(TMenuClickEv)miNewURL_OnClick)->tag = en_all|vi_tvftp;


	TMenu_AddItem(self->menuEdit, "-", 0, NULL)->tag = vi_fw;
	self->miOpenWith = TMenu_AddItem(self->menuEdit, "Open With", 0, NULL);
	self->miOpenWith->tag = vi_fw|en_dfl;
	  self->menuOpenWith = TMenuFile_Create((TWins*)self);
	  self->miOpenWith->next_menu = (TMenu*)self->menuOpenWith;
	  self->menuOpenWith->path = pcadd(IappDat.ALDini_dir, "/Menu/OpenWith/");
	  self->menuOpenWith->OnClick = (TClickEv)menuOpenWith_OnClick;
	  self->menuOpenWith->mfk = mfk_AutoFree;

	self->miTools = TMenu_AddItem(self->menuEdit, "Tools", 0, NULL);
	self->miTools->tag = vi_fw|vi_tv|en_gp;
	TMenu_AddItem(self->menuEdit, "Win Exec With", 0, 
		(TMenuClickEv)menuOpenWithFind_OnClick)->tag = vi_fw|en_dfl;

	TMenu_AddItem(self->menuEdit, "-", 0, NULL)->tag = vi_fw|vi_tv;
	TMenu_AddItem(self->menuEdit, "Copy", 0, 
		(TMenuClickEv)miCopy_OnClick)->tag = vi_fw|vi_tv|en_dfl;
	TMenu_AddItem(self->menuEdit, "Paste", 0, miPaste_OnClick)->tag = vi_fw|en_paste;
	TMenu_AddItem(self->menuEdit, "-", 0, NULL)->tag = vi_fw|vi_tv;
	TMenu_AddItem(self->menuEdit, "Cut", 0, 
		(TMenuClickEv)miCopy_OnClick)->tag = vi_fw|en_dfl|fkt_cut;
	TMenu_AddItem(self->menuEdit, "Delete", 0, 
		(TMenuClickEv)miDelete_OnClick)->tag = vi_fw|vi_tv|vi_tvnet|en_dfl|en_gp;
	TMenu_AddItem(self->menuEdit, "-", 0, NULL)->tag = vi_fw|vi_tv|vi_tvnet|en_dfl|en_gp;
	TMenu_AddItem(self->menuEdit, "Properties", 0, 
		(TMenuClickEv)miProperties_OnClick)->tag = vi_fw|vi_tv|vi_tvnet|en_dfl|en_gp;
}
