/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "aldpic.h"

void dvColorsTo_OnValChange(TDVal *dv, double val) {
  TMainForm *fpic = IappDat.obj;
	fpic->QuantOk = False;
	Refresh((TWins*)fpic->AzGraf);
}

void cbColQuantOn_OnClick(TWins *self, long KeyState, bool DblClick) {
  TMainForm *fpic = IappDat.obj;
	fpic->QuantOk = False;
	Refresh((TWins*)fpic->AzGraf);
}

static void dvRGB2cb(TFormPalette *fpal) {
  dword R = fpal->dvR->dval;
  dword G = fpal->dvG->dval;
  dword B = fpal->dvB->dval;
  dword bix = (R<<16)|(G<<8)|B;
	fpal->PaletteColor[fpal->rgbIdx] = bix;
	SetColor((TWins*)fpal->cbPalette[fpal->rgbIdx], bix);
}

void dvRGB_OnValChange(TDVal *dv, double val) {
  TMainForm *fpic = IappDat.obj;
	dvRGB2cb(fpic->fpal);
	if (fpic->fpal->cbColQuantOn->Checked) {
	    fpic->QuantOk = False;
	    Refresh((TWins*)fpic->AzGraf);
	}
}

void dvRGB_OnMouseMove(TWins *self, XMotionEvent *ev) {
  TMainForm *fpic = IappDat.obj;
  int x, y;

	if (!(ev->state & Button1Mask)) return;
	x = ev->x;
	y = ev->y;
	if (x >= 0) {
	    x -= self->width;
	    if (x < 0) x = 0;
	}
	if (y >= 0) {
	    y -= self->height;
	    if (y < 0) y = 0;
	}
	x -= y;
	TDVal_Set_Value_ev((TDVal*)self, fpic->fpal->rgbmdVal + x);
}

void dvRGB_OnMouseDown(TWins *self, XButtonEvent *ev) {
  TMainForm *fpic = IappDat.obj;
	fpic->fpal->rgbmdVal = ((TDVal*)self)->dval;
}

static void Set_dvRGB(TFormPalette *fpal, dword bix) {
	TDVal_Set_Value(fpal->dvB, bix & 0xff);
	TDVal_Set_Value(fpal->dvG, (bix >> 8) & 0xff);
	TDVal_Set_Value(fpal->dvR, (bix >> 16) & 0xff);
}

static void Set_lblRGB(TFormPalette *fpal) {
  char buf[16];
	inttopc(buf+1, fpal->rgbIdx);
	buf[0] = 'R';
	SetCaption(fpal->lblR, buf);
	buf[0] = 'G';
	SetCaption(fpal->lblG, buf);
	buf[0] = 'B';
	SetCaption(fpal->lblB, buf);
}

void cbPalette_OnClick(TWins *self, long KeyState, bool DblClick) {
  TMainForm *fpic = IappDat.obj;
  TFormPalette *fpal = fpic->fpal;

	fpal->rgbIdx = self->tag;
	Set_lblRGB(fpal);
	Set_dvRGB(fpal, fpal->PaletteColor[self->tag]);
	if (!fpal->cbColQuantOn->Checked) return;
	fpic->QuantOk = False;
	Refresh((TWins*)fpic->AzGraf);
}

static void lblColors_Set(TMainForm *fpic, TImage *img) {
  int ColorCount = 0;
  char *pc;
  char buf[32];
	if (img && img->cmap) ColorCount = img->cmap->ColorCount;
	pc = buf;
	inttoppc(&pc, ColorCount);
	ppcmove(&pc, "  Colors");
	SetCaption(fpic->fpal->lblColors, buf);
}

void ColorsQuant(TMainForm *fpic) {
  int QuantColors;
  TFormPalette *fpal = fpic->fpal;
  TImage *img = (TImage*)fpic->vio_pxm;
  int default_quant_count = 0;
  dword default_quant_col_map[20];

	if (!fpal->cbColQuantOn->Checked) {
	    if (!fpic->QuantOk && fpal->Visible) {
		if (!img) img = (TImage*)ALDpic_pxm_get(fpic, ib_base_pxm);
		TImage_GetColorCount(img, pxgcc_RGB_ALL|pxgcc_ColorCount_AddMask);
		lblColors_Set(fpic, img);
	    }
	    return;
	}

	if (fpic->QuantOk || !img) return;
	TImage_GetColorCount(img, pxgcc_RGB_ALL|pxgcc_ColorCount_AddMask);
	lblColors_Set(fpic, img);

	QuantColors = fpal->dvColorsTo->dval;
	if (alpha_akt(img)) QuantColors--;
	if (QuantColors < 1) return;

	for (int i=0;i<20;i++) {
	    if (fpal->cbPalette[i]->Checked) {
		default_quant_col_map[default_quant_count] = fpal->PaletteColor[i];
		default_quant_count++;
	    }
	}
	img->cmap->default_quant_count = default_quant_count;
	img->cmap->default_quant_col_map = (ic32*)default_quant_col_map;

	TImage_GetColorMap(img);
	TImage_CutColorMap(img, QuantColors, 0);
	TImage_ReMapColor(img);

	fpic->RePaint_pxm = True;
	fpic->QuantOk = True;

}

static char *Col2ppc(char *pc, dword val) {
  char *pca = pc;

	ppcmove_v32(&pc, val & 0xff, 0);
	for (int i=5-(pc-pca);i>0;i--) {*pc = ' '; pc++;}
	*pc = 0;
	return pc;
}
void AzRGBA(TMainForm *fpic, int x, int y, TImage *img) {
  TFormPalette *fpal = fpic->fpal;
  dword bix;
  char *pc;
  char buf[80];

	if (!fpic->fpal->Visible) return;
	if (!img || (x < 0) || (x > img->width) ||
	            (y < 0) || (y > img->height)) {
	    SetCaption(fpal->lblPosColor, " --- x ---");
	    return;
	}
	bix = *((dword*)(img->data + (x * 4) + (y * img->bytes_per_line)));
	pc = buf;
	ppcmove_v32(&pc, x, 4);
	*pc = 'x'; pc++;
	ppcmove_v32(&pc, y, 0);
	for (int i=10-(pc-buf);i>0;i--) {*pc = ' '; pc++;}
	*pc = 'R'; pc++;
	pc = Col2ppc(pc, bix >> 16);
	*pc = 'G'; pc++;
	pc = Col2ppc(pc, bix >> 8);
	*pc = 'B'; pc++;
	pc = Col2ppc(pc, bix);
	if (img->stat & pmxStat_has_alpha) {
	    *pc = 'A'; pc++;
	    pc = Col2ppc(pc, bix >> 24);
	}
	SetCaption(fpal->lblPosColor, buf);
	SetColor(fpal->wPosColor, bix);
}

void SetQuant_RGB(TMainForm *fpic, int x, int y, TImage *img) {
  dword bix;

	if (!fpic->fpal->Visible) return;
	if (!img || (x < 0) || (x > img->width) ||
	            (y < 0) || (y > img->height)) {
	    return;
	}
	bix = *((dword*)(img->data + (x * 4) + (y * img->bytes_per_line)));
	Set_dvRGB(fpic->fpal, bix);
	dvRGB2cb(fpic->fpal);
}



int pal16[20] = {
	0xFFFFFF, 0xFF0000, 0x00FF00, 0x0000FF,
	0xC0C0C0, 0x800000, 0x008000, 0x000080,
	0x808080, 0xFF00FF, 0xFFFF00, 0x00FFFF,
	0x000000, 0x800080, 0x808000, 0x008080,
	0xC0C0C0, 0xC0C0C0, 0xC0C0C0, 0xC0C0C0
};

static void TFormPalette_SetDefault(TFormPalette *self) {

	for (int i=0;i<20;i++) {
  TChkBox *cb;
	    cb = self->cbPalette[i];
	    TChkBox_SetChecked(cb, False);
	    SetColor((TWins*)cb, pal16[i]);
	    self->cbPalette[i] = cb;
	    self->PaletteColor[i] = pal16[i];
	}
	TChkBox_SetChecked(self->cbColQuantOn, False);
	self->rgbIdx = 16;

	Set_lblRGB(self);
	Set_dvRGB(self, self->PaletteColor[self->rgbIdx]);
}

void mSetDefault(TMenu *self, long KeyState, TMenuItem *smi) {
  TMainForm *fpic = IappDat.obj;
	TFormPalette_SetDefault(fpic->fpal);
	fpic->QuantOk = False;
	Refresh((TWins*)fpic->AzGraf);
}

TFormPalette *TFormPalette_Create(TWins *parentForm) {
  TFormPalette *self;
	self = TForm_CreatParamAlg(parentForm, sizeof(TFormPalette));
	self->FormZStyle = fszTransient;
	self->FormPosition = poDesigned;
	self->Visible = False;
	SetCaption((TWins*)self, "ALDpic Palette");
	self->width = 230;
	self->height = 137;
	self->min_width = self->width;
	self->min_height = self->height;
	self->max_width = self->width;
	self->max_height = self->height;

	self->lblColors = TWins_CreateLabel((TWins*)self, 0, 2, 92, 16, "000000 Colors", mCaptionLeft, mBorder|mBorderL);
	self->lblColorsTo = TWins_CreateLabel((TWins*)self, 92, 2, 18, 16, "to", 0, 0);
	self->dvColorsTo = TDVal_CreateSV4((TWins*)self, 110, 1, 42, 18, 256, 2, 4096, 0);
	self->dvColorsTo->OnValChange = dvColorsTo_OnValChange;
	self->dvColorsTo->dv_fu = tdv_fu_LostFocusEnter;
	self->cbColQuantOn = TChkBox_CreateS((TWins*)self, 154, 2, 76, 16, "Quant On");
	self->cbColQuantOn->OnClick = cbColQuantOn_OnClick;

	self->pnlPalette = TWins_CreateB((TWins*)self, 0, 22, 140, 99, mBorder|mBorderL);
  int x = 0;
  int y = -16;
  TChkBox *cb;
  char buf[16];
	for (int i=0;i<20;i++) {
	    if (!(i & 0x03)) {
		x = 4;
		y += 19;
	    }
	    inttopc(buf, i);
	    cb = TChkBox_CreateS((TWins*)self->pnlPalette, x, y, 31, 17, buf);
	    cb->OnClick = cbPalette_OnClick;
	    cb->OnlyInBoxCheck = True;
	    cb->tag = i;
	    self->cbPalette[i] = cb;
	    x += 34;
	}


	self->lblPosColor = TWins_CreateLabelA((TWins*)self, 0, 102, 230, 16, " --- x ---", mCaptionLeft, mBorder|mBorderL, alBottom);
	self->wPosColor = TWins_CreateB(self->lblPosColor, 212, 1, 17, 14, 0);

	self->lblR = TWins_CreateLabel((TWins*)self, 140, 23, 26, 16, "R16", mCaptionRigth, 0);
	self->lblG = TWins_CreateLabel((TWins*)self, 140, 43, 26, 16, "G16", mCaptionRigth, 0);
	self->lblB = TWins_CreateLabel((TWins*)self, 140, 63, 26, 16, "B16", mCaptionRigth, 0);
	self->dvR = TDVal_CreateSV4((TWins*)self, 166, 22, 36, 18, 0, 0, 255, 0);
	self->dvR->OnValChange = dvRGB_OnValChange;
	self->dvR->OnMouseDown = dvRGB_OnMouseDown;
	self->dvR->OnMouseMove = dvRGB_OnMouseMove;
	self->dvG = TDVal_CreateSV4((TWins*)self, 166, 42, 36, 18, 0, 0, 255, 0);
	self->dvG->OnValChange = dvRGB_OnValChange;
	self->dvG->OnMouseDown = dvRGB_OnMouseDown;
	self->dvG->OnMouseMove = dvRGB_OnMouseMove;
	self->dvB = TDVal_CreateSV4((TWins*)self, 166, 62, 36, 18, 0, 0, 255, 0);
	self->dvB->OnValChange = dvRGB_OnValChange;
	self->dvB->OnMouseDown = dvRGB_OnMouseDown;
	self->dvB->OnMouseMove = dvRGB_OnMouseMove;

	self->PopupMenu = TMenu_Create((TWins*)self);
	TMenu_AddItem(self->PopupMenu, "Set Default", 0, mSetDefault);

	self->lblColors->PopupMenu = self->PopupMenu;
	self->lblColorsTo->PopupMenu = self->PopupMenu;
	self->lblPosColor->PopupMenu = self->PopupMenu;
	self->lblR->PopupMenu = self->PopupMenu;
	self->lblG->PopupMenu = self->PopupMenu;
	self->lblB->PopupMenu = self->PopupMenu;

	TFormPalette_SetDefault(self);

	return self;
}
