/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "ElMath.h"
#include "dlg.h"

#ifndef _OS_WIN32
static const char ElMath16_AHico[] = {
 0x06,0x08,0x10,0x00,0x10,0x00,0xE7,0xB7,0x6B,0x00,0x1F,0x00,0x86,0x31,0x10,0x84,
 0x14,0xA5,0x18,0xC6,0xDB,0xDE,0xDF,0xFF,0x00,0xF8,0xBD,0xC0,0x7B,0x41,0x77,0x4B,
 0x21,0x77,0x4B,0xD5,0xC8,0xD1,0x1D,0x1D,0x59,0x99,0x95,0x19,0xC9,0x2D,0xD5,0x63,
 0x65,0x56,0x66,0x24,0x47,0x77,0x74,0x14,0x81,0x58,0x66,0x24,0xB7,0x64,0x3F,0x86,
 0x66,0x68,0x41,0x72,0x44,0x23,0x44,0x58,0x66,0x85,0x46,0x72,0x24,0x32,0x42,0x56,
 0x66,0x85,0x46,0x72,0x44,0x33,0x44,0x56,0x66,0x65,0x48,0x72,0x94,0x62,0x43,0x56,
 0x66,0x65,0x48,0x72,0xB7,0x14,0x42,0xBC,0xD2,0x2F,0x00};
#endif

void Form_ELM_mbMM_MIL_OnClick(TWins *self, long KeyState, bool DblClick) {
  TFormElMath *fem = (TFormElMath*)self->ParentForm;
	fem->mm_mil++;
	if (fem->mm_mil > 1) fem->mm_mil= 0;
	switch (fem->mm_mil) {
	    case 0: 
		SetCaption(self, " mm ");
		break;
	    case 1: 
		SetCaption(self, " mil ");
		break;
	} 
	pageLuftSpule_SetUp(fem->pgLuftSpule, 1, fem->mm_mil);
	pageMicrostrip_SetUp(fem->pgMicrostrip, 1, fem->mm_mil);
}

void Form_ELM_mbHelp_OnClick(TWins *self, long KeyState, bool DblClick) {
  TFormElMath *fem = (TFormElMath*)self->ParentForm;
  char buf[64];
	snprintf(buf, sizeof(buf), "ElMath %s   Copyright", VERSION);
	Copyright_MsgBox_Show((TWins*)fem, buf);
}

void fem_pgc_OnActivate_pg(TPageControl *pagc, TPage *pag) {
  TFormElMath *fem = (TFormElMath*)pagc->ParentForm;
  TPageControl *pgc_F = fem->pgFilter->pgc_Filter;
  TPageControl *pgc_DF = fem->pgDIG_Filter->pgc;

	if (pag == (TPage*)fem->pgFilter) {
	    if (pag != fem->lastAktSubPage) {
		fem->lastAktSubPage = pag;
		if (pgc_F->OnActivate_pg) {
		    pgc_F->OnActivate_pg(pgc_F, pgc_F->AktPage);
		}
		if (pgc_DF->OnActivate_pg) {
		    pgc_DF->OnActivate_pg(pgc_DF, NULL);
		}
	    }

	} else if (pag == (TPage*)fem->pgDIG_Filter) {
	    if (pag != fem->lastAktSubPage) {
		fem->lastAktSubPage = pag;
		if (pgc_F->OnActivate_pg) {
		    pgc_F->OnActivate_pg(pgc_F, NULL);
		}
		if (pgc_DF->OnActivate_pg) {
		    pgc_DF->OnActivate_pg(pgc_DF, pgc_DF->AktPage);
		}
	    }

	} else if (fem->lastAktSubPage == (TPage*)fem->pgDIG_Filter) {
		fem->lastAktSubPage = pag;
		if (pgc_DF->OnActivate_pg) {
		    pgc_DF->OnActivate_pg(pgc_DF, NULL);
		}
	}
}

void Form_ELM_mFile_Exit(TMenu *self, long KeyState, TMenuItem *smi) {
  Close();
}

void Form_ELM_OnMove(TForm *self) {
  TpcIni *ti = TpcIni_Create(IappDat.inifile_app);
	TpcIni_WR_int(ti, APPNAME, "Top", self->top);
	TpcIni_WR_int(ti, APPNAME, "Left", self->left);
	TpcIni_SaveAndFree(ti);
}

void Form_ELM_OnResize(TWins *self) {
	if (!self->ResizeStartMk) return;
  TpcIni *ti = TpcIni_Create(IappDat.inifile_app);
	TpcIni_WR_int(ti, APPNAME, "Width", self->width);
	TpcIni_WR_int(ti, APPNAME, "Height", self->height);
	TpcIni_SaveAndFree(ti);
}

void OnAppStart() {
  TFormElMath *fem = IappDat.obj;
	fem->pgMicrostrip = ELM_pageMicrostripAdd(fem->PageControl, "Microstrip");
	if (!AppProcessMessages()) return;
	fem->pgFilter = ELM_pageFilterAdd(fem->PageControl, "Filter");
	if (!AppProcessMessages()) return;
	fem->pgDIG_Filter = ELM_pageDIG_Filter(fem->PageControl, "Dig.Filter");

	Tfree(IappDat.appIni);
	IappDat.appIni = NULL;
	fem->PageControl->OnActivate_pg = fem_pgc_OnActivate_pg;
}

static const char *Help_Msg = {
  "\n"
  "Usage: ElMath [options] \n"
  "\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: ElMath version "VERSION"  (build %d  %s)\n"
};

int build_v = ALD_BUILD_V;
char *build_date = __DATE__;

int main (int argc, char **argv) {
  TFormElMath *fem;
  int i;
  char *val;
  bool help = False;
  bool err = False;

	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		switch (*val) {
		case 'h':
		    help = True; break;
		case '-':
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
#ifdef Test_Root_WinId
		    } else if (pcncmp(val, "TestRootWinID", 13)) {
			i++;
			val = argv[i];
			if (val && *val != '-') {
				TestRootWinId = pctoint(val);
			} else err = True;
#endif
		    } else err = True;
		    break;
		default: 
			err = True;
		}  // switch end
	    } else {
		err = True;
	    }
	    i++;
	}

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(err);
	}

	initXapp(0, argc, argv);
	fem = TForm_CreatParamAlg(NULL, sizeof(TFormElMath));
	IappDat.obj = fem;

#ifdef USE_DECOR
	IappXDat.DefaultButtonDecor = TDecor_button_w3b3_Create();
#endif
#ifdef _OS_WIN32
	IappDat.inifile_app = pcadd(IappDat.ALDini_dir, "\\ElMath.ini");
#else
	IappDat.inifile_app = pcadd(IappDat.ALDini_dir, "/ElMath.ini");
	fem->Icon = GetDeskIcon_z(ddkt_Application, 16, APPNAME, 
				    gdifu_noZoom, ElMath16_AHico);
#endif
	IappDat.appIni = TpcIni_Create(IappDat.inifile_app);

	SetCaption((TWins*)fem, APPNAME);
	fem->IconName = APPNAME;

	fem->top = TpcIni_RD_int(IappDat.appIni, APPNAME, "Top", 0);
	fem->left = TpcIni_RD_int(IappDat.appIni, APPNAME, "Left", 0);
	fem->width = TpcIni_RD_int(IappDat.appIni, APPNAME, "Width", 680);
	fem->height = TpcIni_RD_int(IappDat.appIni, APPNAME, "Height", 380);
	fem->min_width = 720;
	fem->min_height = 400;
	fem->max_width = 1024;
	fem->max_height = 540;
	fem->FormPosition = poDesigned;
	fem->OnMove = Form_ELM_OnMove;
	fem->OnResize = Form_ELM_OnResize;

	fem->pnlTop = TWins_CreateBA((TWins*)fem, 0, 0, 200, 21, 0, alTop);
	fem->pnlMiddle = TWins_CreateBA((TWins*)fem, 0, 0, 200, 42, mBorder | mBorderL, alClient);

// ******* Menu File *******
	fem->mbFile = TMenuButton_Create(fem->pnlTop, (TMenuButton*)fem);
	fem->mbFile->top = 2;
	fem->mbFile->left = 1;
	SetCaption((TWins*)fem->mbFile, "File");
	Set_Parent((TWins*)fem->mbFile, fem->pnlTop);
	fem->menuFile = TMenu_Create((TWins*)fem);
	fem->mbFile->menu = fem->menuFile;
	TMenu_AddItem(fem->menuFile, "Exit", 0, Form_ELM_mFile_Exit);

// ******* Menu MM/MILL *******
	fem->mbMM_MIL = TMenuButton_Create(fem->pnlTop, fem->mbFile);
	fem->mbMM_MIL->OnClick = Form_ELM_mbMM_MIL_OnClick;
	SetCaption((TWins*)fem->mbMM_MIL, " mm ");
	Set_Parent((TWins*)fem->mbMM_MIL, fem->pnlTop);

// ******* Menu Help *******
	MenuHelp_Create(fem);


	fem->PageControl = TPageControl_Create(fem->pnlMiddle);
	fem->PageControl->Align = alClient;
	Set_Parent((TWins*)fem->PageControl, fem->pnlMiddle);

	fem->pgLuftSpule = ELM_pageLuftSpuleAdd(fem->PageControl, "Luft Spule");

	TLineEdit_AddKeyDownPlugIn(NULL, TLineEdit_keyplugin_edit_Create());
	TLineEdit_AddKeyDownPlugIn(NULL, TLineEdit_keyplugin_cursor_Create());

	IappXDat.OnAppStart = OnAppStart;

	AppRun();
	return 0;
}
