/*
 Copyright (C) 2010-2012 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"

bool bi_ift_Set_pc(char *s, char **ret, bool *err) {
  int lg;

	if (!s || (*s == '-')) {*err = True; return False;}
	lg = pclen(s);
	if (*s == '\"') {
		s++;
		lg--;
		if (lg > 0 && s[lg-1] == '\"') lg--;
	}
	*ret = pcndup(s, lg);
	return True;
}

static const char *Help_Msg = {
  "\n"
  "Usage: ini_file_tool [options] -G \"val\" -I \"val\" \"File\"\n"
  "\n"
  "  -r                   Read Item\n"
  "  -w                   Write/Add Item\n"
  "  -rw                  Read Item is no Item Add Item\n"
  "  -wo                  Add Item only not exists\n"
  "  -G \"val\"             Gruppe\n"
  "  -I \"val\"             Item\n"
  "  -D \"val\"             r=Default w=Daten\n"
  "  -t                   Output whit \" \"dat\"\n"
  "  -x                   Output whit label \"item=dat\"\n"
  "  -x1                  Output whit label \"grupp_item=dat\"\n"
  "  -x2                  Output whit label \"grupp.item=dat\"\n"
  "  -h, --help           Prints out this screen\n"
  "  return Item\n"
  "Example:\n"
  "  ini_file_tool -w -Gg1 -Ii1 -Ditem1 -Ii2 -D2 /tmp/test.ini\n"
  "  ini_file_tool -w -G \"g2\" -I \"i1 item1\" -D \"test 1\" /tmp/test.ini\n"
  "  ini_file_tool -r -Gg1 -Ii1 -Ii2 /tmp/test.ini\n\n"
  "  ALDccp: ini_file_tool version "VERSION"  (build %d  %s)\n"
};

#ifdef STANDALONE_BUILDIN

const int build_v = ALD_BUILD_V;
const char *build_date = __DATE__;

int main (int argc, char **argv) {
//	cmdline_slash_korr(argc, argv);
#else

extern const int build_v;
extern const char *build_date;

int bi_ini_file_tool_main (int argc, char **argv) {
#endif
  int i;
  char *pc;
  char *Gruppe = NULL;
  char *Item = NULL;
  char *Default = NULL;
  char *File = NULL;
  bool help = False;
  bool err = False;
  bool gpOk = False;
  bool rd = False;
  bool wr = False;
  bool wrAddOnly = False;
  bool tMk = False;
  bool xMk = False;
  TpcIni *ini;

	i = 1;
	while ((i < argc) && !help && !err) {
	    pc = argv[i];
	    if (*pc == '-') {
		pc++;
OptLoop:
		if (*pc == 'r') {
		    rd = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 'w') {
		    wr = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 'o') {
		    wrAddOnly = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 't') {
		    tMk = True;
		    pc++; if (*pc) goto OptLoop;
		} else if (*pc == 'x') {
		    xMk = 1;
		    pc++;
		    if ((*pc == '1') || (*pc == '2')) {
			if (*pc == '1') xMk = 2; else  xMk = 3;
			pc++;
		    }
		    if (*pc) goto OptLoop;
		} else if (*pc == 'h') {
		    help = True;
		} else if (*pc == 'G') {
		    gpOk = False;
		    pc++; if (!*pc) i++;
		} else if (*pc == 'I') {
		    gpOk = True;
		    pc++; if (!*pc) i++;
		} else if (*pc == 'D') {
		    pc++; if (!*pc) i++;
		} else if (*pc == '-') {
		    pc++;
		    if (pcncmp(pc, "help", 4)) {
			help = True;
		    } else err = True;
		} else err = True;
	    } else {
		bi_ift_Set_pc(argv[i], &File, &err);
	    }
	    i++;
	}

	if (!help && !err) err = (!gpOk || !File);

	if (help || err) {
	    printf(Help_Msg, build_v, build_date);
	    exit(err);
	}
	if (!rd && !wr) rd = True;

	ini = TpcIni_Create(File);
	i = 1;
Loop:
	gpOk = 0;
	while ((i < argc) && !err) {
	    pc = argv[i];
	    if (*pc == '-') {
		pc++;
		if (*pc == 'G') {
		    if (Gruppe) {
			gpOk = 0x03;
			break;
		    }
		    pc++; if (!*pc) {i++; pc = argv[i];}
		    bi_ift_Set_pc(pc, &Gruppe, &err);
		} else if (*pc == 'I') {
		    if (Item) {
			gpOk = 0x02;
			break;
		    }
		    pc++; if (!*pc) {i++; pc = argv[i];}
		    bi_ift_Set_pc(pc, &Item, &err);
		} else if (*pc == 'D') {
		    pc++; if (!*pc) {i++; pc = argv[i];}
		    bi_ift_Set_pc(pc, &Default, &err);
		}
	    }
	    i++;
	}

	if (Gruppe && Item) {
	    if (rd) {
		pc = TpcIni_RD_pc(ini, Gruppe, Item, NULL);
		if (!pc) {
		    if (wr) TpcIni_WRstr(ini, Gruppe, Item, Default);
		    pc = Default;
		}
		if (xMk == 2) printf("%s_", Gruppe);
		if (xMk == 3) printf("%s.", Gruppe);
		if (xMk != 0) printf("%s=", Item);
		if (tMk) printf("\"");
		if (pc) printf("%s", pc);
		if (tMk) printf("\"");
		printf("\n");
	    } else if (wr) {
		if (wrAddOnly) {
		    pc = TpcIni_RD_pc(ini, Gruppe, Item, NULL);
		    if (!pc) TpcIni_WRstr(ini, Gruppe, Item, Default);
		} else TpcIni_WRstr(ini, Gruppe, Item, Default);
	    }
	    if (gpOk & 0x01) ppcfree(&Gruppe);
	    if (gpOk & 0x02) ppcfree(&Item);
	    ppcfree(&Default);

	} else err = True;

	if ((i < argc) && !err) goto Loop;
	
	if (ini->Changed) TpcIni_Save(ini);
	Tfree(ini);
	return 0;
}
