/*
 Copyright (C) 2010-2011 by Hoefer Alfred

 This program is free software; you can redistribute it and/or
 modify it under the terms of the GNU General Public License
 as published by the Free Software Foundation; either
 version 2 of the License, or (at your option) any later version. 

 This program is distributed in the hope that it will be useful, 
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 See the GNU General Public License for more details. 

 You should have received a copy of the GNU General Public
 License along with this program; if not, write to the Free
 Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110, USA

 ----------------------------------------------------------------
 Licensed under GPL v2, see file COPYING in this tarball
*/

#include "sysbase.h"
#include "sysbaseio.h"
#include "zlib.h"

#define INBUF_SIZE 0x4000
#define OUTBUF_SIZE INBUF_SIZE+256

static const char *Help_Msg = {
  "\n"
  "Usage: txt2zsrc [options] file\n"
  "\n"
  "  -o outfile           output file (default tmp_z.c)\n"
  "  -h, --help           Prints out this screen\n\n"
  "  ALDccp: txt2zsrc version "VERSION"  (build %d  %s)\n"
};

int main (int argc, char **argv) {
  int i, inlg, out_count;
  char *val;
  char *InFile, *OutFile, *fn, *cname, *pc, *wrbuf;
  byte *inBuf, *outBuf;
  z_stream s;
  FILE *inf, *outf;
  bool help = False;
  bool err = False;

	InFile = NULL;
	OutFile = "tmp_z.c";
	i = 1;
	while (!err && !help && (i<argc)) {
	    val = argv[i];
	    if (*val == '-') {
		val++;
		if (*val == 'h') {
		    help = True;

		} else if (*val == 'o') {
		    val++;
		    if (*val == '=') {
			val++;
			OutFile = pcdup(val);
		    } else if (((i+1)<argc) && (*argv[i+1] != '-')) {
			i++;
			OutFile = pcdup(argv[i]);
		    } else err = True;

		} else if (*val == '-') {
		    val++;
		    if (pcncmp(val, "help", 4)) {
			help = True;
		    } else err = True;
		} else err = True;
	    } else {
		InFile = val;
	    }
	    i++;
	}

	err = err || !InFile;
	if (help || err) {
	    printf(Help_Msg, ALD_BUILD_V, __DATE__);
	    exit(1);
	}

	fn = path_rel2abs(InFile, NULL);
	inf = fopen(fn, "rb");
	if (!inf) {
	    appError(EMSG_ERR | EMSG_SYS, " open %s ", fn);
	    exit(1);
	}
	xfree(fn);

	fn = path_rel2abs(OutFile, NULL);
	outf = fopen(fn, "w");
	if (!outf) {
	    fclose(inf);
	    appError(EMSG_ERR | EMSG_SYS, " open %s ", fn);
	    exit(1);
	}
	xfree(fn);

	cname = ChangeFileExt(GetFileName_pc(OutFile), NULL);
	wrbuf = xcalloc(1024);

	pc = wrbuf;
	ppcmove(&pc, "//zlib compressed file \"");
	ppcmove(&pc, GetFileName_pc(InFile));
	ppcmove(&pc, "\"\n\n");

	ppcmove(&pc, "const unsigned char ");
	ppcmove(&pc, cname);
	ppcmove(&pc, "[] = {\n");
	fwrite(wrbuf, 1, pclen(wrbuf), outf);

	out_count = 0;
	Fill0(&s, sizeof(s));
	inBuf = xcalloc(INBUF_SIZE);
	outBuf = xcalloc(OUTBUF_SIZE);

	if (deflateInit2(&s, Z_BEST_COMPRESSION, Z_DEFLATED, -15, 8, Z_FILTERED) != Z_OK) {
	    fclose(inf);
	    appError(EMSG_ERR, " zlib_deflateInit");
	    exit(1);
	}
	pc = wrbuf;
loop:
	inlg = fread(inBuf, 1, INBUF_SIZE, inf);
	s.avail_in = inlg;
	s.avail_out = OUTBUF_SIZE;
	s.next_in = inBuf;
	s.next_out = outBuf;

	int m = Z_SYNC_FLUSH;
	if (!inlg) m = Z_FINISH;
	if (deflate(&s, m) < Z_OK) {
	    fclose(inf);
	    appError(EMSG_ERR, " zlib_deflate");
	    exit(1);
	}
	byte *pb = outBuf;
	while (out_count < s.total_out) {
	    out_count++;
	    m = kCOMMA_ADD | 2;
	    if ((inlg < INBUF_SIZE) && (out_count == s.total_out)) m = 2;
	    hexto0xppc(&pc, *pb, m);
	    pb++;
	    if ((out_count & 0xf) == 0) {
		*pc = '\n'; pc++; *pc = 0;
		fwrite(wrbuf, 1, pclen(wrbuf), outf);
		pc = wrbuf;
	    }
	}
	if (inlg == INBUF_SIZE) goto loop;

	fclose(inf);
	if (pc != wrbuf) fwrite(wrbuf, 1, pclen(wrbuf), outf);

	pc = wrbuf;
	ppcmove(&pc, "};\n\nconst int ");
	ppcmove(&pc, cname);
	ppcmove(&pc, "_size = ");
	ppcmove_v32(&pc, s.total_out, kSEMICOL_ADD|kLF_ADD);
	ppcmove(&pc, "const int ");
	ppcmove(&pc, cname);
	ppcmove(&pc, "_size_org = ");
	ppcmove_v32(&pc, s.total_in, kSEMICOL_ADD|kLF_ADD);

	fwrite(wrbuf, 1, pclen(wrbuf), outf);
	fclose(outf);

  printf("total_in=%lu total_out=%lu %lu%%\n", s.total_in, s.total_out, s.total_out * 100 / s.total_in);
	deflateEnd(&s);

	return 0;
}
